/* programme/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un Graphe d'evenements.			*/

#include "eg.h"
#include "math.h"

#ifdef SUN4SOL2
extern double rint(double);
#endif 

#define INFINI		1e12

#define MAX(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"eg2c"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODELN		CODE("\n")

/* Global declarations */
extern EventGraph	*Read_Network(char*);
extern void		dumpe_network(EventGraph*);
extern boolean		topological_analysis(EventGraph*);
extern void		Free_Event_Graph(EventGraph*);
extern char		*todays_date(void);

int	Buf_Size = 1000;	/* default buffer size */
double	Trace_Start = 0.0;	/* trace control parameters 	*/
double	Trace_End = 100.0;	/* with their initial values	*/
int	Event_Start = 0;	/* maybe for a future trace 	*/
int	Event_End = 100;	/* based on event numbers...	*/
char	*Tmp_Dir;		/* where to put the files 	*/
boolean	Slice;

/* forward declarations */
void write_law(Law_Desc);
int Max_Param_Nb(EventGraph*);
int gen_c_code(EventGraph	*Reseau, int	Sim_Length,
	       boolean	Trace,
	       boolean	Waiting_Time, int		From_Trans,
	       int		To_Trans );

int main(int argc, char	*argv[])
{
  static char		File_Name[128] = "";
  EventGraph 		*Reseau;
  boolean		Trace = FALSE;
  int			i;
  int			Sim_Length = 1000;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  int			From_Trans = -1;
  int			To_Trans = -1;
  boolean		Waiting_Time = FALSE;

  Tmp_Dir = (char*) malloc( sizeof( char ) );
  strcpy( Tmp_Dir, "" );

  i = 1;

  if (argc>1)
    do {

      if GET_INT( "-bs", Buf_Size )
      else if GET_STRING( "-f", File_Name )
      else if GET_INT( "-nevt", Sim_Length )
      else if GET_INT_SWITCH( "-from", From_Trans, Waiting_Time, TRUE )
      else if GET_INT_SWITCH( "-to", To_Trans, Waiting_Time, TRUE  )
      else if GET_REAL_SWITCH( "-tb", Trace_Start, Trace, TRUE )
      else if GET_REAL_SWITCH( "-te", Trace_End, Trace, TRUE )
      else if SWITCH( "-slice", Slice, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-bs <buf_size>] [-f <file>] ", TOOLNAME );
	printf("[-debug] [-tb <time>]\n\t[-te <time>] [-from <int>] ");
	printf("[-to <int>] [-tmp <dir>] [-nevt <int>]\n\t");
	printf("[-slice] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmp", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Tmp_Dir = (char*) malloc( (2+strlen( argv[i] )) * sizeof( char ) );
	  strcpy( Tmp_Dir, argv[i] );
	  if ( ( strlen( Tmp_Dir ) != 0 )
	       && (Tmp_Dir[strlen(Tmp_Dir)-1] != '/' ) )
	    strcat( Tmp_Dir, "/" );
	  i++;
	}
	else {  
	  printf("%s: directory expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    Reseau = Read_Network(File_Name);

    if ( NULL != Reseau ) {

      if ( topological_analysis(Reseau) ) {
	if (Debug)
	  dumpe_network(Reseau);

	/* check if waiting time specifications seem ok */
	if ( Waiting_Time ) {
	  Waiting_Time = ( ( From_Trans > 0 ) 
			   && ( From_Trans <= Reseau->Size ) );
	  if ( !Waiting_Time ) {
	    printf( "%s: error: incorrect parameter for -from (%d)\n",
		    TOOLNAME, From_Trans );
	  }
	  else {
	    Waiting_Time = ( ( To_Trans > 0 ) 
			     && ( To_Trans <= Reseau->Size ) );
	    if ( !Waiting_Time ) {
	      printf( "%s: error: incorrect parameter for -to (%d)\n",
		      TOOLNAME, To_Trans );
	    }
	  }
	  if (!Waiting_Time) {
	    return -4;
	  }
	}
	
	gen_c_code( Reseau, Sim_Length, Trace, 
		    Waiting_Time, From_Trans, To_Trans );

	Free_Event_Graph( Reseau );
	free( Tmp_Dir );

	return 0;
      }
      else {
	fprintf( stderr, "Topological analysis failed: ");
	fprintf( stderr, "look for an empty cycle.\n");
	return -3;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network from file %s.\n", 
	       TOOLNAME, File_Name );
      return -2;
    }
    return 0;
  }
  else if ( !Help )
    return -1;
  else 
    return 0;
}

/****************************************************************/
/* Procedures for code generation...				*/
/****************************************************************/
void	gen_file_header( EventGraph 	*Reseau, 
			 int 		Sim_Length,
			 boolean 	Trace )
{
  char	*The_Date;

  CODE("/* Configuration file for an event graph \t\t\t*/\n");
  CODE("/* simulation code generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  CODE("\t\t*/\n");
  The_Date = todays_date();
  CODE1("/* Date: %s \t\t\t\t*/\n\n", The_Date );
  free( The_Date );

  /* write definitions, includes */
  CODE1("int     TMAX = %d;\n", Sim_Length );
  CODE1("#define NB_TRANS\t%d\n", Reseau->Size );
  CODE1("#define NB_PLACES\t%d\n", Reseau->Nb_Places );
  CODE1("#define BUF_SIZE\t%d\n", Buf_Size );
  CODE1("#define MAX_PARAM_NB\t%d\n", Max_Param_Nb(Reseau) );
  CODE1("#define Infinity\t%d\n", 100000 );
  CODE1("#define Minus_Infinity\t%d\n", -100000 );

  /* trace defines */
  if ( Trace ) {
    CODE("#define Trace\n");

    CODE("#define CODE(x)\t\tfprintf(Tracefile,x)\n");
    CODE("#define CODE1(x,y)\tfprintf(Tracefile,x,y)\n");
    CODE("#define CODE2(x,y,z)\tfprintf(Tracefile,x,y,z)\n");
    CODE("#define CODELN\t\tCODE(\"\\n\")\n");
  }
  CODELN;

}

void	gen_declarations( EventGraph *Reseau, boolean Trace )
{
  int	i;
  int	j;
  int	Min_Pop;
  int	Max_Pop;

  CODE("/* table of law descriptors\t\t\t\t\t*/\n");
  CODE("LAW_DESC\t\tTr_Law[NB_TRANS] = {\n");
  for( i=0; i<Reseau->Size-1; i++ ) {
    write_law( Reseau->T[i]->Trans_Law );
    CODE(",\n");
  }
  write_law( Reseau->T[Reseau->Size-1]->Trans_Law );
  CODE("\n};\n\n");

  Min_Pop = Max_Pop = Reseau->P[Reseau->Nb_Places-1]->Init_Pop;

  CODE("/* table of law descriptors\t\t\t\t\t*/\n");
  CODE("LAW_DESC\t\tPl_Law[NB_PLACES] = {\n");
  for( i=0; i<Reseau->Nb_Places-1; i++ ) {
    write_law( Reseau->P[i]->Place_Law );
    CODE(",\n");
    Min_Pop = min( Min_Pop, Reseau->P[i]->Init_Pop );
    Max_Pop = MAX( Max_Pop, Reseau->P[i]->Init_Pop );
  }
  write_law( Reseau->P[Reseau->Nb_Places-1]->Place_Law );
  CODE("\n};\n\n");

  CODE1("#define MIN_POP %d\n", Min_Pop );
  CODE1("#define MAX_POP %d\n", Max_Pop );

  CODE("/* table of initial availabilities\t\t\t\t\t*/\n");
  CODE("int\t\t\tAvail[NB_PLACES][MAX_POP] = {\n");

  for( i=0; i<Reseau->Nb_Places-1; i++ ) {
    CODE("\t");
    for( j=0; j<Reseau->P[i]->Init_Pop; j++ ) {
      CODE1("%8.4f,", Reseau->P[i]->Lag_Times[j] );
    }
    for( ; j<Max_Pop; j++ ) {
      CODE("  0.0000,");
    }
    CODE("\n");
  }
  CODE("\t");
  i = Reseau->Nb_Places-1;
  for( j=0; j<Reseau->P[i]->Init_Pop; j++ ) {
    CODE1("%8.4f,", Reseau->P[i]->Lag_Times[j] );
  }
  for( ; j<Max_Pop-1; j++ ) {
    CODE("  0.0000,");
  }
  if ( j == Max_Pop-1 )
    CODE("  0.0000");
  CODE("\n};\n\n");

  CODE("/* tables of flags for random tables generation\t\t\t\t*/\n");
  CODE("boolean\t\t\tTr_Draw_It[NB_TRANS] = {\n");
  for( i=0; i<Reseau->Size-1; i++ ) {
    if ( Reseau->T[i]->Trans_Law->Name == 'C' )
      CODE("\tFALSE,\n");
    else
      CODE("\tTRUE,\n");
  }
  if ( Reseau->T[Reseau->Size-1]->Trans_Law->Name == 'C' )
    CODE("\tFALSE\n");
  else
    CODE("\tTRUE\n");
  CODE("};\n\n");

  CODE("boolean\t\t\tPl_Draw_It[NB_PLACES] = {\n");
  for( i=0; i<Reseau->Nb_Places-1; i++ ) {
    if ( Reseau->P[i]->Place_Law->Name == 'C' )
      CODE("\tFALSE,\n");
    else
      CODE("\tTRUE,\n");
  }
  if ( Reseau->P[Reseau->Nb_Places-1]->Place_Law->Name == 'C' )
    CODE("\tFALSE\n");
  else
    CODE("\tTRUE\n");
  CODE("};\n\n");

  if ( Trace ) {

    CODE("/* declarations for trace handling */\n");

    CODE("char  *gss_trans_name[NB_TRANS] = {\n");
    for( i=0; i<Reseau->Size-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->T[i]->Data );
	
    CODE1("\t\"%s\"\n", Reseau->T[Reseau->Size-1]->Data );
    CODE("};\n\n");

    CODE("char  *gss_place_name[NB_PLACES] = {\n");
    for( i=0; i<Reseau->Nb_Places-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->P[i]->Name );

    CODE1("\t\"%s\"\n", Reseau->P[Reseau->Nb_Places-1]->Name );
    CODE("};\n\n");

    CODE("char  *gss_in_link_name[NB_PLACES] = {\n");
    for( i=0; i<Reseau->Nb_Places-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->P[i]->Data1 );

    CODE1("\t\"%s\"\n", Reseau->P[Reseau->Nb_Places-1]->Data1 );
    CODE("};\n\n");

    CODE("char  *gss_out_link_name[NB_PLACES] = {\n");
    for( i=0; i<Reseau->Nb_Places-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->P[i]->Data2 );

    CODE1("\t\"%s\"\n", Reseau->P[Reseau->Nb_Places-1]->Data2 );
    CODE("};\n\n");

    CODE("/* table of initial populations\t\t\t\t\t*/\n");
    CODE("int\t\t\tInitial_Pop[NB_PLACES] = {\n");

    for( i=0; i<Reseau->Nb_Places-1; i++ ) {
      CODE1("\t%d", Reseau->P[i]->Init_Pop );
      CODE(",\n");
    }
    CODE1("\t%d", Reseau->P[Reseau->Nb_Places-1]->Init_Pop );
    CODE("\n};\n\n");

    CODE("int   Cur_X_Evt[NB_TRANS+1];\n");
    CODE("int   Cur_Y_Evt[NB_TRANS+1];\n");
    CODELN;
    CODE("FILE  *Tracefile;\n");
    CODE("int   More_Events;\n");
    CODE("int   Was_X_Evt, Was_Y_Evt;\n");
    CODE("float Min_X_Val, Min_Y_Val;\n");
    CODE("int   Arg_Min_X, Arg_Min_Y;\n");
    CODE("float Old_Date, New_Date;\n");
    CODELN;
    CODE1("float Trace_Start = %f;\n", Trace_Start );
    CODE1("float Trace_End = %f;\n", Trace_End );
  }
}

void	write_law(Law_Desc The_Law)
{
  int	Nb_Params;
  int	i;

  CODE1("{ '%c', { ", The_Law->Name );
  switch( The_Law->Name ) {
  case 'E':
  case 'C':
  case 'B':
  case 'g':
  case 'G':
    CODE1( "%f } }", The_Law->Parameters[0] );
    break;
  case 'U':
  case 'W':
  case 'N':
  case 'L':
  case 'p':
  case 'R':
    CODE2( "%f, %f } }", The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'H':
    CODE1( "/* nb params= */ %4.1f ", The_Law->Parameters[0] );
    Nb_Params = (int)rint( The_Law->Parameters[0] );
    for ( i=1; i<=2*Nb_Params; i++ ) {
      CODE1( "%f", The_Law->Parameters[i] );
      if ( i < 2*Nb_Params )
	CODE(",");
    }
    CODE(" } }");
    break;
  case 'D':
    CODE1( "/* nb params= */ %4.1f, ", The_Law->Parameters[0] );
    Nb_Params = (int)rint( The_Law->Parameters[0] );
    for ( i=1; i<= 2*Nb_Params; i++ ) {
      CODE1( "%8.4f ", The_Law->Parameters[i] );
      if ( i < 2*Nb_Params )
	CODE(",");
    }
    CODE(" } }");
    break;
  default:
    fprintf( stderr, "%s: error: unexpected law id (%c). Constant 0.0 assumed\n",
	     TOOLNAME, The_Law->Name );
    CODE( "0.0 } }");
  }
}

void	write_trans_name(Trans_Desc The_Trans)
{
  if ( NULL != The_Trans->Name )
    printf( "%s", The_Trans->Name );
  else
    printf( "_TR_%d", The_Trans->Trans_Id );
}

void	write_Y_plus_L(EventGraph	*Reseau,
		       Place_Desc	The_Place,
		       boolean		Initial )
{
  int	The_Pop;
  int	Tr_Id;

  Tr_Id = The_Place->Input_Id - 1;
  The_Pop = The_Place->Init_Pop;

  if ( Initial ) {
    CODE1("( Time < %d ? ", The_Pop );
    /* the buffer size is always larger than Max_init_Pop	*/
    /* no modulo necessary for initial values			*/
    CODE1("Avail[%d][Time] : ", The_Place->Place_Id );
    CODE2("Y[%d][Time-%d] ", Tr_Id, The_Pop );
    CODE(")\n\t\t\t");
  }
  else {
    if ( The_Pop != 0 ) {
      if ( Slice )
	/* In this case, The_Pop has to be 1 */
	CODE1("Y[%d][ Time_M_1 ] ", Tr_Id );
      else
	CODE2("Y[%d][ (Time+%d) %% BUF_SIZE ] ", 
	      Tr_Id, Buf_Size - The_Pop );
    }
    else
      CODE1("Y[%d][Time] ", Tr_Id );
  }

  /* optimization: if constant times, no table access. 		*/
  /* If zero time, nothing at all.					*/
  if ( 'C' == The_Place->Place_Law->Name ) {
    if ( 0.0 != The_Place->Place_Law->Parameters[0] )
      CODE1("+ %10.6f", The_Place->Place_Law->Parameters[0] );
    /* else nothing */
  }
  else
    CODE1("+ L[%d][Time]", The_Place->Place_Id );
}

void	write_Lg(EventGraph	*Reseau,
		 Place_Desc	The_Place,
		 int		Origin )
{
  int	Tr_Id;
  int	The_Pop;

  Tr_Id = The_Place->Trans_Id - 1;
  The_Pop = The_Place->Init_Pop;

  switch( The_Pop ) {
  case 0:
    CODE2("Lg_1[%d][%d] ", Origin, Tr_Id );
    break;
  case 1:
    CODE2("Lg_0[%d][%d] ", Origin, Tr_Id );
    break;
  default:
    fprintf( stderr, "%s: warning: network not flat.\n", TOOLNAME );
  }
}

void	write_plus_L(EventGraph	*Reseau,
		     Place_Desc	The_Place)

{
  /* optimization: if constant times, no table access. 		*/
  /* If zero time, nothing at all.				*/
  if ( 'C' == The_Place->Place_Law->Name ) {
    if ( 0.0 != The_Place->Place_Law->Parameters[0] )
      CODE1("+ %10.6f", The_Place->Place_Law->Parameters[0] );
    /* else nothing */
  }
  else
    CODE1("+ L[%d][Time]", The_Place->Place_Id );
}

void	write_plus_S(EventGraph	*Reseau,
		     int	The_Node)
{
  /* optimization: if constant times, no table access. 		*/
  /* If zero time, nothing at all.				*/
  if ( 'C' == Reseau->T[The_Node]->Trans_Law->Name ) {
    if ( 0 != Reseau->T[The_Node]->Trans_Law->Parameters[0] )
      CODE1(" + %10.6f", Reseau->T[The_Node]->Trans_Law->Parameters[0] );
    /* else nothing */
  }
  else 
    CODE1( " + S[%d][Time]", The_Node );

}

/* The_Node: 0 to NB_TRANS-1 */
void	gen_code_node(EventGraph	*Reseau,
		      int		The_Node,
		      boolean		Initial )
{
  Place_Desc	Pred_Pl;
  Liste_Place	P_Scan;
  int		Indegree;

  CODE("/* evolution of transition ");
  write_trans_name( Reseau->T[The_Node] );
  CODE(" */\n"); 

  CODE1("X[%d][Time] = ", The_Node);

  Indegree =  Reseau->T[The_Node]->Indegree;

  if( Indegree == 0 ) {	
    /* this is a source but it should be recycled...	*/
    if ( Slice )
      CODE1("Y[%d][ Time_M_1 ];\n", The_Node );
    else
      CODE1("Y[%d][(Time+BUF_SIZE-1) %% BUF_SIZE];\n", The_Node );
    fprintf( stderr, "%s: warning: non recycled source. Assumed recycled.\n",
	     TOOLNAME );
  }
  else if ( Indegree == 1 ) { 
    /* only one input */
    Pred_Pl = Reseau->T[The_Node]->Inputs->Val;
    write_Y_plus_L( Reseau, Pred_Pl, Initial );
    CODE(";\n");
  }
  else {
    CODE("MAX( ");
    Pred_Pl = Reseau->T[The_Node]->Inputs->Val;
    write_Y_plus_L( Reseau, Pred_Pl, Initial );
    CODE(",\n\t\t  ");
    Pred_Pl = Reseau->T[The_Node]->Inputs->Next->Val;
    write_Y_plus_L( Reseau, Pred_Pl, Initial );
    CODE(");\n");

    SCAN( P_Scan, Reseau->T[The_Node]->Inputs->Next->Next ) {
      CODE2("X[%d][Time] = MAX( X[%d][Time] ,\n\t\t  ", 
	    The_Node, The_Node);
      Pred_Pl = P_Scan->Val;
      write_Y_plus_L( Reseau, Pred_Pl, Initial );
      CODE(" );\n");
    }
  }

  CODE2("Y[%d][Time] = X[%d][Time]", The_Node, The_Node );
  /* optimization: test if constant times */
  if ( 'C' == Reseau->T[The_Node]->Trans_Law->Name ) {
    if ( 0 != Reseau->T[The_Node]->Trans_Law->Parameters[0] )
      CODE1(" + %10.6f", Reseau->T[The_Node]->Trans_Law->Parameters[0] );
    /* else nothing */
  }
  else 
    CODE1( " + S[%d][Time]", The_Node );

  CODE(";\n");
}

void 	gen_loop_code(EventGraph	*Reseau,
		      boolean		Initial)
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;

  if ( Slice && !Initial ) {
    /* dans ce cas, comme la profondeur est 1, on est sr d'utiliser	*/
    /* "Time-1". Je le mets donc dans une variable...			*/
    /* Note: dans le cas normal, ca doit tre utile galement.		*/
    CODE("Time_M_1 = ( Time + BUF_SIZE - 1 ) %% BUF_SIZE;\n");
  }
  
  for( Num_Layer = 0; Num_Layer < Reseau->Nb_Layers; Num_Layer++)
    for ( Num_Node = 0; 
	  Num_Node < Reseau->Level[Num_Layer]->Nb_Nodes;
	  Num_Node++ ) {
	    
      The_Node = Reseau->Level[Num_Layer]->Node_Id[Num_Node];
      gen_code_node( Reseau, The_Node, Initial );
    }
}

void	gen_code_node_slice(EventGraph	*Reseau,
			    int		The_Node )
{
  Liste_Place	P_Scan;
  Place_Desc	Pred_Pl;
  int		Indegree;
  int		Origin;

  Indegree =  Reseau->T[The_Node]->Indegree;

  CODE("/* evolution of paths ending at transition ");
  write_trans_name( Reseau->T[The_Node] );
  CODE(" */\n"); 

  for( Origin=0; Origin<Reseau->Size; Origin++ ) {

    CODE("Tmp = EPSILON;\n");
    
    if( Indegree == 0 ) {	
      /* this is a source but should be recycled... skipping */
    }
    else {
      SCAN( P_Scan, Reseau->T[The_Node]->Inputs ) {
	Pred_Pl = P_Scan->Val;
	CODE("Date = ");
	write_Lg( Reseau, Pred_Pl, Origin );
	CODE(";\n");
	CODE("Tmp = ( Date > EPSILON ? MAX( Tmp, Date ");
	write_plus_L( Reseau, Pred_Pl );
	CODE(") : Tmp );\n");
      }
      
      CODE2("Lg_1[%d][%d] = ( Tmp > EPSILON ?\n\t\t", Origin, The_Node );
      CODE("Tmp ");
      write_plus_S( Reseau, The_Node );
      CODE(":\n\t\tEPSILON );\n\n");
    }
  }

}

void 	gen_loop_code_slice(EventGraph	*Reseau)
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;

  for( Num_Layer = 0; Num_Layer < Reseau->Nb_Layers; Num_Layer++)
    for ( Num_Node = 0; 
	  Num_Node < Reseau->Level[Num_Layer]->Nb_Nodes;
	  Num_Node++ ) {
	    
      The_Node = Reseau->Level[Num_Layer]->Node_Id[Num_Node];
      gen_code_node_slice( Reseau, The_Node );
    }
}

void	gen_waiting_time_code( Reseau, From_Trans, To_Trans )
EventGraph	*Reseau;
int		From_Trans;
int		To_Trans;
{
  CODELN;
  printf("/* waiting time between transitions ");
  write_trans_name( Reseau->T[From_Trans-1] );
  printf(" and ");
  write_trans_name( Reseau->T[To_Trans-1] );
  printf(" */\n");
  CODE2("printf( \"%%f\\n\", X[%d][Time] - Y[%d][Time] );\n",
	To_Trans-1, From_Trans-1 );
  CODELN;
  printf("Standard = FALSE;\n");
  CODELN;

}

boolean Is_Dummy( The_Place )
Place_Desc	The_Place;
{
  if ( NULL != The_Place->Name )
    return FALSE;
  else
    return ( 0 == strncmp( The_Place->Name, "_Dummy_Place", 12 ) );
}

void gen_check_firing( EventGraph *Reseau )
{
  int		Num;
  int		Pred_Id;
  int		I_Pop;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  CODE("switch( Arg_Min_Y ) {\n");
  for ( Num = 0; Num < Reseau->Size; Num++ ) {

    CODE1("case %d:\n", Num );

    /* check that removing tokens from (non dummy) input places is ok */
    SCAN( P_Scan, Reseau->T[Num]->Inputs ) {
      The_Place = P_Scan->Val;
      Pred_Id = The_Place->Input_Id - 1;
      if ( !Is_Dummy( The_Place ) ) {
	I_Pop = The_Place->Init_Pop;

	CODE("\tFiring_Ok = Firing_Ok && ( 0 <=\n");
	CODE2("\t\t\t\t%d + Cur_Y_Evt[%d] - 1 -", I_Pop, Pred_Id );
	CODE1("Cur_Y_Evt[%d] );\n", Num );
      }
    }

    CODE("\tbreak;\n");
  }
  CODE("}\n\n");
}

void gen_end_firing( EventGraph	*Reseau )
{
  int		Num;
  int		Pred_Id;
  int		Succ_Id;
  int		Pl_Nb;
  int		I_Pop;
  int		Indegree;
  int		Outdegree;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  CODE("if ( Firing_Ok )\n  switch( Arg_Min_Y ) {\n");
  for ( Num = 0; Num < Reseau->Size; Num++ ) {

    CODE1("    case %d:\n", Num );

    Indegree = Reseau->T[Num]->Indegree;
    Outdegree = Reseau->T[Num]->Outdegree;

    if ( Indegree > 1 )
      CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
    /* first remove tokens from input places, except dummy ones */
    SCAN( P_Scan, Reseau->T[Num]->Inputs ) {
      The_Place = P_Scan->Val;
      Pred_Id = The_Place->Input_Id - 1;
      if ( !Is_Dummy( The_Place ) ) {
	I_Pop = The_Place->Init_Pop;
	Pl_Nb = The_Place->Place_Id;

	CODE("\tCODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
	CODE1("\t\t\t\tgss_place_name[%d],\n", Pl_Nb );
	CODE2("\t\t\t\t%d + Cur_Y_Evt[%d] - 1 -", I_Pop, Pred_Id );
	CODE1("Cur_Y_Evt[%d] );\n", Num );
      }
    }
    if ( Indegree > 1 )
      CODE("\tCODE(\"END_MULTIPLE\\n\");\n");

    /* then move the tokens */
    if ( Indegree > 1 )
      CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
    SCAN( P_Scan, Reseau->T[Num]->Inputs ) {
      The_Place = P_Scan->Val;
      if ( !Is_Dummy( The_Place ) ) {
	Pl_Nb = The_Place->Place_Id;

	CODE("\tCODE2(\"MOVE_NEW_OBJECT OF_CLASS token ");
	CODE("FROM %%s TO %%s \",\n");
	CODE1("\t\t\t\tgss_place_name[%d], ", Pl_Nb );
	CODE1("gss_trans_name[%d] );\n", Num );
	CODE1("\tCODE1(\"ON_PATH %%s \", gss_out_link_name[%d] );\n", Pl_Nb );
	CODE("\tCODE(\"WITH_STEPS 20\\n\");\n");
      }
    }
    if ( Indegree > 1 )
      CODE("\tCODE(\"END_MULTIPLE\\n\");\n");

    /* then switch transition status */
    CODE1("\tCODE(\"MODIFY %s \\\"status\\\" = 0\\n\");\n",
	  Reseau->T[Num]->Name );

    /* move tokens to (non dummy) output places */
    if ( Reseau->T[Num]->Outdegree > 1 )
      CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
    SCAN( P_Scan, Reseau->T[Num]->Outputs ) {
      if ( !Is_Dummy( P_Scan->Val ) ) {
	Pl_Nb = P_Scan->Val->Place_Id;

	CODE("\tCODE2(\"MOVE_NEW_OBJECT OF_CLASS token ");
	CODE("FROM %%s TO %%s \",\n");
	CODE1("\t\t\t\tgss_trans_name[%d],\n", Num );
	CODE1("\t\t\t\tgss_place_name[%d] );\n", Pl_Nb );
	CODE1("\tCODE1(\"ON_PATH %%s \", gss_in_link_name[%d] );\n", Pl_Nb );
	CODE("\tCODE(\"WITH_STEPS 20\\n\");\n");
      }
    }
    if ( Outdegree > 1 )
      CODE("\tCODE(\"END_MULTIPLE\\n\");\n");
					
    /* finally change marking of (non dummy) output places */
    if ( Outdegree > 1 )
      CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
    SCAN( P_Scan, Reseau->T[Num]->Outputs ) {
      if ( !Is_Dummy( P_Scan->Val ) ) {
	Pl_Nb = P_Scan->Val->Place_Id;
	I_Pop = P_Scan->Val->Init_Pop;
	Succ_Id = P_Scan->Val->Trans_Id-1;

	CODE("\tCODE2(\"MODIFY %%s \\\"token\\\" = ");
	CODE("%%d\\n\",\n");
	CODE1("\t\t\t\tgss_place_name[%d],\n", Pl_Nb ); 
	CODE2("\t\t\t\t%d + Cur_Y_Evt[%d] + 1 - ", I_Pop, Num );
	CODE1("Cur_Y_Evt[%d] ", Succ_Id );
	if ( Succ_Id == Num )
	  /* dans ce cas, Cur_Y_Evt a avance aussi pour lui */
	  CODE("- 1 );\n");
	else
	  CODE(");\n");
      }
    }
    if ( Outdegree > 1 )
      CODE("\tCODE(\"END_MULTIPLE\\n\");\n");

    CODE("\tbreak;\n");
  }
  CODE("}\n");
}

int	gen_c_code(
		   EventGraph	*Reseau,
		   int		Sim_Length,
		   boolean	Trace,
		   boolean	Waiting_Time,
		   int		From_Trans,
		   int		To_Trans )
{
  char		File_Name[512];

  /* adjusting buffer size and trace parameters		*/
  Buf_Size = min( Buf_Size, Sim_Length );
  Buf_Size = MAX( Buf_Size, Reseau->Max_Pop_Init );

  if ( Trace ) {
    Event_Start = MAX( Event_Start, 0 );
    Event_End = min( Event_End, Sim_Length );
    Trace_Start = MAX( Trace_Start, 0 );
    Buf_Size = MAX( Buf_Size, Trace_Start );
  }
		
  /* preparing names of files and writing them */
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "egc_graph_parameters.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )
    gen_file_header( Reseau, Sim_Length, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "egc_declarations.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )
    gen_declarations(Reseau, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the initialization loop.		*/
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "egc_initial_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) )
    gen_loop_code( Reseau, TRUE );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the main loop. 			*/
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "egc_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) ) {
    gen_loop_code( Reseau, FALSE );
    if ( Waiting_Time )
      gen_waiting_time_code( Reseau, From_Trans, To_Trans );
  }
  else
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  if ( Trace ) {
    /* code for the trace generation 			*/
    strcpy( File_Name, Tmp_Dir );
    strcat( File_Name, "egc_end_firing_trace.c" );
    if ( NULL != freopen( File_Name, "w", stdout ) ) {
      gen_check_firing( Reseau );
      gen_end_firing( Reseau );
    }
    else
      fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );
  }

  if ( Slice ) {
    /* code for the matrix generation 			*/
    strcpy( File_Name, Tmp_Dir );
    strcat( File_Name, "egc_slice_evolution.c" );
    if ( NULL != freopen( File_Name, "w", stdout ) ) {
      gen_loop_code_slice( Reseau );
    }
    else	
      fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );
  }

  return TRUE;
}

/************************************************************************/
/* Functions for computing the (maximal) number of parameters for the	*/
/* stochastic laws.							*/
/************************************************************************/
int Param_Number( Law_Desc The_Law ) 
{
  int	Res;

  switch( The_Law->Name ) {
  case 'E':
  case 'C':
  case 'B':
  case 'g':
  case 'G':
    Res = 1;
    break;
  case 'U':
  case 'W':
  case 'N':
  case 'L':
  case 'p':
  case 'R':
    Res = 2;
    break;
  case 'H':
  case 'D':
    Res = 1 + 2 * (int)rint( The_Law->Parameters[0] );
    break;
  default:
    fprintf( stderr, "%s: error: unexpected law id (%c).\n",
	     TOOLNAME, The_Law->Name );
    Res = 1;
  }

  return Res;
}

int Max_Param_Nb( EventGraph *Reseau )
{
  int	i;
  int	New;
  int	Res = 1;

  for( i=0; i<Reseau->Size-1; i++ ) {
    New = Param_Number( Reseau->T[i]->Trans_Law );
    Res = MAX( New, Res );
  }

  for( i=0; i<Reseau->Nb_Places-1; i++ ) {
    New = Param_Number( Reseau->P[i]->Place_Law );
    Res = MAX( New, Res );
  }

  return Res;
}
