/************************************************************************/
/* module to compile an EG into an executable performing its equational	*/
/* simulation.								*/
/************************************************************************/
#include <unistd.h>
#include "eg.h"
#define RESOURCES
#include "resources.h"

#define TOOLNAME	"eg_sim"

/* from the module transcript.c */
extern boolean  Write_EG_As_EGL(EventGraph *Reseau, char *FileName,
				char *ToolName);
/* from the resource system */
extern void 	get_resources(boolean);
extern char     **Resource_Value;
extern boolean	*Is_Available;
extern int  	execute(char*,...);

#define STOP    (char *)0
#define RV(x)   Resource_Value[ x ]

boolean Resources_Are_Present(void)
{
  return ( Is_Available[ C_COMPILER ] && Is_Available[ EG2C ] );
}

boolean	Compile_EG(
     EventGraph	*Reseau,
     int	Sim_Length,
     boolean 	Checking,
     boolean 	Keep,
     boolean	Waiting_Time,
     int	From_Trans,
     int	To_Trans )
{
  char	Sim_Name[1024];
  char	Src_Name[1024];
  char	Lib_Name[1024];
  char	From_Name[8];
  char	To_Name[8];
  char  SimL[16];

  boolean	All_Ok;

  get_resources( Checking );

  All_Ok = Resources_Are_Present();

  sprintf( SimL, "%d", Sim_Length );

  if ( (!Checking) && All_Ok ) {

    /* Dumping the network for immediate re-reading... 	*/
    /* Sounds stupid, but not simple to avoid...        	*/
    if ( !Write_EG_As_EGL( Reseau, RV( TMP_EGL_FILE ), TOOLNAME ) ) {
      return FALSE;
    }

    strcpy( Sim_Name, RV( TMP_DIR ) );
    strcat( Sim_Name, "/" );
    strcat( Sim_Name, RV( SIM_NAME ) );

    strcpy( Lib_Name, "-I" );
    strcat( Lib_Name, RV( TMP_DIR ) );

    if ( !Waiting_Time ) {
      execute( TOOLNAME,
	       RV( EG2C ), "-f", RV( TMP_EGL_FILE ), "-tmp", 
	       RV( TMP_DIR ), 
	       "-nevt", SimL, 
	       STOP );
    }
    else {
      sprintf( From_Name, "%d", From_Trans );
      sprintf( To_Name, "%d", To_Trans );
      execute( TOOLNAME,
	       RV( EG2C ), "-f", RV( TMP_EGL_FILE ), "-tmp", 
	       RV( TMP_DIR ), "-from", From_Name,
	       "-to", To_Name,
	       "-nevt", SimL,  
	       STOP );
    }
    execute( TOOLNAME,
	     RV( C_COMPILER ), RV( C_SOURCE_FILE ), "-o", Sim_Name,
	     Lib_Name, "-lm", "-O",
	     STOP );

    execute( TOOLNAME, Sim_Name, STOP );

    /* cleaning up */
    if ( !Keep ) {
      unlink( RV( TMP_EGL_FILE ) );
      unlink( Sim_Name );
      strcpy( Src_Name, RV( TMP_DIR ) );
      strcat( Src_Name, "/egc_initial_evolution.c" );
      unlink( Src_Name );
      strcpy( Src_Name, RV( TMP_DIR ) );
      strcat( Src_Name, "/egc_evolution.c" );
      unlink( Src_Name );
      strcpy( Src_Name, RV( TMP_DIR ) );
      strcat( Src_Name, "/egc_graph_parameters.h" );
      unlink( Src_Name );
      strcpy( Src_Name, RV( TMP_DIR ) );
      strcat( Src_Name, "/egc_declarations.h" );
      unlink( Src_Name );
    }
  }
  else if (!Checking) {
    fprintf( stderr, "%s: error: some resources are missing ", TOOLNAME );
    fprintf( stderr, "(check with -c)\n");

    return FALSE;
  }

  /* execution presumed successful */
  return TRUE;
}
