/* Un programme d'analyse de la FIFOite d'un graphe d'evenements	*/

#include <stdio.h>
#include "eg.h"

#define	TOOLNAME	"fifo_eg"

#define FIFO_PN		"fifo_pn"
#define PNL_FILE_NAME	"fifo_eg_tmp.pnl"
#define RES_FILE_NAME	"fifo_eg_tmp.res"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

#define CODE(x)		fprintf(Code_File,x)
#define CODE1(x,y)	fprintf(Code_File,x,y)
#define CODE2(x,y,z)	fprintf(Code_File,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Code_File,x,y,z,t)
#define CODELN		CODE("\n")

#define INFINITY	-2

extern EventGraph	*Read_Network(char*);
extern void		Free_Event_Graph(EventGraph*);
extern boolean		topological_analysis(EventGraph*);

/* from the module transcript.c */
extern boolean Write_EG_As_PNL(EventGraph *Reseau, char *File_Name, 
			       char *Tool_Name);

/* forward declarations */
boolean	Trans_Are_FIFO(EventGraph *Reseau);
boolean Places_Are_FIFO(EventGraph *Reseau);

/* global variables */
FILE	*Code_File;

int main(int argc, char *argv[])
{
  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";
  static char		command[512];
  EventGraph 		*Reseau;
  int			i;
  boolean		Dump = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Lazy = FALSE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-lazy", Lazy, TRUE )
      else if SWITCH( "-d", Dump, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-?", argv[i] ) )
		|| ( 0 == strcmp( "-h", argv[i] ) ) ) {
	printf("Usage: %s [-d] [-lazy] [-f <file>] [-o <file>] [-h] [-?] \n", 
	       TOOLNAME);
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network(In_File_Name);

    if ( Reseau != NULL ) {

      topological_analysis(Reseau);

      /* Checking simple sufficient conditions for FIFOness */
      if ( Trans_Are_FIFO( Reseau ) && Places_Are_FIFO( Reseau ) ) {
	/* The network is FIFO */
	Free_Event_Graph( Reseau );

	return 0;
      }
      else {
	/* The network may not be FIFO. Have to check harder */

	/* First generate PNL code */
	if ( !Write_EG_As_PNL( Reseau, PNL_FILE_NAME, TOOLNAME ) ) {
	  return -1;
	}

	/* construct command to execute */
	sprintf( command, "%s -f %s", FIFO_PN, PNL_FILE_NAME );
	if ( Lazy )
	  strcat( command, " -lazy" );
	if ( Dump ) 
	  strcat( command, " -d" );
	if ( Debug )
	  strcat( command, " -debug" );
	if ( 0 != strlen( Out_File_Name ) ) {
	  strcat( command, " -o " );
	  strcat( command, Out_File_Name );
	}
	   
	/* execute the command */
	if ( Debug )
	  printf( "%s: executing %s\n", TOOLNAME, command );
	 
	system( command );

	return 0;	/* but should be the error code of the command */
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );
      return -2;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* teste si une transition est recyclee (sans utiliser Next) */
boolean	Is_Recycled_Transition(EventGraph      *Reseau,
			       int		Tr_Num )
{
  boolean	Ok = FALSE;
  Liste_Place	P_Scan;

  SCAN( P_Scan, Reseau->T[Tr_Num]->Outputs )
    Ok = ( Ok || ( ( P_Scan->Val->Trans_Id-1 == Tr_Num )
		   && ( P_Scan->Val->Init_Pop == 1 ) ) );

  return	Ok;
}

boolean Places_Are_FIFO( EventGraph* Reseau )
{
  int		i;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;
  boolean	Is_FIFO = TRUE;

  for ( i=0; Is_FIFO && (i<Reseau->Size); i++ )
    for( P_Scan = Reseau->T[i]->Outputs; Is_FIFO && ( P_Scan != NULL );
	 P_Scan = P_Scan->Next ) {

      The_Place = P_Scan->Val;

      Is_FIFO = ( The_Place->Place_Law->Name == 'C' );
  }

  return Is_FIFO;
}

boolean Trans_Are_FIFO(EventGraph* Reseau)
{
  int		i;
  boolean	Is_FIFO = TRUE;

  for ( i=0; Is_FIFO && (i<Reseau->Size); i++ )

    Is_FIFO = ( ( Reseau->T[i]->Trans_Law->Name == 'C' )
		|| Is_Recycled_Transition( Reseau, i ) );

  return Is_FIFO;
}
