/* a function to "flatten" an EG so than its maximum marking		*/
/* becomes one. This involves the creation of "dummy" transitions and	*/
/* places.								*/
/* The function returns a pointer on a new network. The previous one	*/
/* is not touched.							*/

#include "eg.h"

#define MAX(x,y)    	( x > y ? x : y )
#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

#define	MINFINITY	-1e20
#define	AMONT		0
#define	AVAL		1
#define	ECLATEMENT	2

/* external declaration */
extern EventGraph* New_Event_Graph( int Size );
extern Trans_Desc  New_Transition( void );
extern Place_Desc  New_Place( void );

/* Function for copying a string, without using strdup */
char	*Copy_Name( char *Orig_Name )
{
  char	*Dest_Name;

  if ( NULL == Orig_Name ) {
    return (char*) NULL;
  }
  else {
    Dest_Name = (char*) calloc( (1+strlen( Orig_Name )), sizeof(char) );
    strcpy( Dest_Name, Orig_Name );
    return Dest_Name;
  }
}

void	Attach_To_Outputs(Trans_Desc	The_Trans,
			    Place_Desc	The_Place )
{
  Liste_Place	The_List;

  The_List = (Liste_Place) malloc( sizeof( struct PLACE_LIST ) );
  The_List->Val = The_Place;
  The_List->Next = The_Trans->Outputs;
  The_Trans->Outputs = The_List;
  The_Trans->Outdegree++;
  The_Place->Input_Id = The_Trans->Trans_Id;
}

void	Store_Place(EventGraph	*RdP,
		    Place_Desc	The_Place )
{
  RdP->Nb_Places++;
  RdP->P 
    = (Place_Desc*) realloc( RdP->P,
			     (size_t)(RdP->Nb_Places)*sizeof(Place_Desc) );
  RdP->P[RdP->Nb_Places-1] = The_Place;
  The_Place->Place_Id = RdP->Nb_Places-1;
}

Place_Desc Create_Dummy_Place(Trans_Desc	Orig_Trans,
			      int		Dest_Id,
			      double		The_Lag )
{
  Place_Desc	The_Place;

  The_Place = New_Place();
  The_Place->Trans_Id = Dest_Id;
  The_Place->Name = Copy_Name( "Dummy_Place" );
  The_Place->Data1 = NULL;
  The_Place->Data2 = NULL;
  The_Place->Init_Pop = 1;
  The_Place->Lag_Times = (double*) calloc( 1, sizeof(double) );
  The_Place->Lag_Times[0] = The_Lag;

  The_Place->Place_Law = New_Law();
  The_Place->Place_Law->Name = 'C';
  /* Law member: Parameters comes with 1 value pre-allocated. No realloc */
  The_Place->Place_Law->Parameters[0] = 0.0;

  Attach_To_Outputs( Orig_Trans, The_Place );

  return The_Place;
}

/************************************************************************/
/* A procedure to evaluate the number of transitions in the final	*/
/* while computing the index of the first "dummy" transition		*/
/* corresponding to each real transition.				*/
/*									*/
/* Needs:	Reseau		a pointer on the EG to expand		*/
/*		Index, Mu_Star	two already allocated arrays of integers*/
/*				of length N, N being the number of	*/
/*				transitions in Reseau.			*/
/*		Algorithm	a descriptor of the algo used.		*/
/*									*/
/* Returns:	The number of transitions T in the expanded network	*/
/* 		The array Index filled. The value of Index[N] is T	*/
/************************************************************************/
int	Compute_Index_Table(EventGraph	*Reseau,
			    int		*Index,
			    int		*Mu_Star,
			    int		Algorithm )
{
  int		Nb_Trans = 0;
  int		i;
  int		New_Nb;
  Liste_Place	P_Scan;

  for (i=0; i<Reseau->Size; i++) {

    Index[i] = Reseau->Size + Nb_Trans - i;
    New_Nb = 1;	/* there is at least one new transition	*/
    /* for each old one!			*/

    switch( Algorithm ) {
    case AVAL:
      SCAN( P_Scan, Reseau->T[i]->Outputs )
	New_Nb = MAX( New_Nb, P_Scan->Val->Init_Pop );
      break;
    case AMONT:
      SCAN( P_Scan, Reseau->T[i]->Inputs )
	New_Nb = MAX( New_Nb, P_Scan->Val->Init_Pop );
      break;
    case ECLATEMENT:
      SCAN( P_Scan, Reseau->T[i]->Outputs )
	New_Nb += MAX( 0, P_Scan->Val->Init_Pop-1 );
    }

    Mu_Star[i] = New_Nb;
    Nb_Trans += New_Nb;
  }

  return Nb_Trans;
}

EventGraph*	Copy_Invariant_Part(EventGraph*	Reseau,
				    int		New_Size )
{
  EventGraph*	RdP;
  int		i;
  Trans_Desc	The_Trans;
  Trans_Desc	New_Trans;

  RdP = New_Event_Graph( New_Size );

  if ( NULL != RdP ) {
    /* copy old transitions into the new network */
    for ( i=0; i<Reseau->Size; i++ ) {
      The_Trans = Reseau->T[i];
      New_Trans = New_Transition();
      RdP->T[i] = New_Trans;
      New_Trans->Trans_Id = i+1;
      New_Trans->Trans_Law = Copy_Law( The_Trans->Trans_Law );
      New_Trans->Name = Copy_Name( The_Trans->Name );
    }
    /* initialize new transitions */
    for ( i=Reseau->Size; i<RdP->Size; i++ ) {
      New_Trans = New_Transition();
      RdP->T[i] = New_Trans;
      New_Trans->Trans_Id = i+1;
      New_Trans->Trans_Law = New_Law();
      New_Trans->Trans_Law->Name = 'C';
  /* Law member: Parameters comes with 1 value pre-allocated. No realloc */
      New_Trans->Trans_Law->Parameters[0] = 0.0;
      New_Trans->Name = Copy_Name( "Dummy_Transition" );
    }
  }

  return RdP;
}

void Copy_Place( 
Place_Desc	Old_Place,
EventGraph	*N_Reseau,
int		Trans_Nb,      /* the id of the upstream transition */
int		*Index,
int		*Mu_Star,
int		Algorithme )
{
  int		Mu;
  Place_Desc	The_Place;

  Mu = Old_Place->Init_Pop;
  The_Place = New_Place();

  The_Place->Name = Copy_Name( Old_Place->Name );
  The_Place->Place_Law = Copy_Law( Old_Place->Place_Law );
  The_Place->Data1 = Copy_Name( Old_Place->Data1 );
  The_Place->Data2 = Copy_Name( Old_Place->Data2 );

  switch( Mu ) {
  case 0:
    The_Place->Init_Pop = 0;
    The_Place->Lag_Times = NULL;
    break;
  default:
    The_Place->Init_Pop = 1;
    The_Place->Lag_Times = (double*) calloc( 1, sizeof(double) );
    switch( Algorithme ) {
    case AMONT:
      The_Place->Lag_Times[0] = Old_Place->Lag_Times[Mu-1];
      break;
    case AVAL:
    case ECLATEMENT:
      The_Place->Lag_Times[0] = Old_Place->Lag_Times[0];
    }
  }

  switch( Mu ) {
  case 0:
  case 1:
    The_Place->Trans_Id = Old_Place->Trans_Id;
    break;
  default:
    switch( Algorithme ) {
    case AVAL:
    case ECLATEMENT:
      The_Place->Trans_Id = Old_Place->Trans_Id;
      break;
    case AMONT:
      The_Place->Trans_Id = Index[Old_Place->Trans_Id-1]
	+Mu_Star[Old_Place->Trans_Id-1]+1-Mu;
      break;
    }
  }

  Store_Place( N_Reseau, The_Place );
  Attach_To_Outputs( N_Reseau->T[Trans_Nb], The_Place );

}

void Copy_Real_Places(EventGraph	*O_Reseau,
		      EventGraph	*N_Reseau,
		      int		*Index,
		      int		*Mu_Star,
		      int		Algorithme )
{
  int		i;
  Trans_Desc	The_Trans;
  Liste_Place	P_Scan;
  int		Sub_Index;
  int		Mu;

  for ( i=0; i<O_Reseau->Size; i++ ) {
    The_Trans = O_Reseau->T[i];

    switch( Algorithme ) {
    case AVAL:
      SCAN( P_Scan, The_Trans->Outputs ) {
	Mu = P_Scan->Val->Init_Pop;
	if ( Mu <= 1 )
	  Copy_Place( P_Scan->Val, N_Reseau, i, Index,
		      Mu_Star, Algorithme );
	else
	  Copy_Place( P_Scan->Val, N_Reseau,
		      Index[i]+Mu-2,
		      Index, Mu_Star, Algorithme );
	}
      break;
    case ECLATEMENT:
      Sub_Index = Index[i];
      SCAN( P_Scan, The_Trans->Outputs ) {
	Mu = P_Scan->Val->Init_Pop;
	Sub_Index += MAX( Mu-1, 0 );
	if ( Mu <= 1 )
	  Copy_Place( P_Scan->Val, N_Reseau, i, Index,
		      Mu_Star, Algorithme );
	else
	  Copy_Place( P_Scan->Val, N_Reseau, Sub_Index-1, 
		      Index, Mu_Star, Algorithme );
      }
      break;
    case AMONT:
      SCAN( P_Scan, The_Trans->Outputs )
	Copy_Place( P_Scan->Val, N_Reseau, i, Index, 
		    Mu_Star, Algorithme ); 
      break;
    }
  }
}

double	Max_Lags_Of_Inputs(EventGraph    *Reseau,
			  int		Trans_Id,
			  int		The_Pop )
{
  double	The_Lag;
  Place_Desc	The_Place;
  Liste_Place	P_Scan;

  if ( Reseau->T[Trans_Id]->Indegree == 0 )
    /* en principe, pas tellement possible, mais bon.	*/
    return 0.0;
  else {
    The_Lag = MINFINITY;

    SCAN( P_Scan, Reseau->T[Trans_Id]->Inputs ) {
      The_Place = P_Scan->Val;
      if ( The_Place->Init_Pop >= The_Pop )
	The_Lag = MAX( The_Lag, 
		       The_Place->Lag_Times[The_Pop-2] );
    }

    return The_Lag;
  }
}
							
/* Flatten_EG will return a NEW Event graph, and not touch to the old one. */
EventGraph *Flatten_EG(EventGraph *Reseau,
		       int	  Algorithme )
{
  Liste_Place	P_Scan;
  Place_Desc	The_Place;
  int		i;
  int		j;
  int		New_Size;
  int		*Index;
  int		*Mu_Star;
  EventGraph	*RdP;
  int		Sub_Index;
  int		Mu;

  Index = (int*) calloc( 1+Reseau->Size, sizeof(int) );
  Mu_Star = (int*) calloc( 1+Reseau->Size, sizeof(int) );
  New_Size = Compute_Index_Table( Reseau, Index, Mu_Star, Algorithme );

  RdP = Copy_Invariant_Part( Reseau, New_Size );
  if ( NULL != RdP ) {

    /* je vais construire d'abord la liste des voisins,		*/
    /* avec les identites correctes et creer les places.	*/
    Copy_Real_Places( Reseau, RdP, Index, Mu_Star, Algorithme );

    if ( NULL != RdP ) {
      /* maintenant, creation des places et transitions fictives qui	*/
      /* correspondent a cette transition reele (valable pour tout	*/
      /* algorithme...							*/
      for ( i=0; i<Reseau->Size; i++ ) {
	if ( Mu_Star[i] != 1 ) {
	  switch( Algorithme ) {
	    case AVAL:
	      The_Place = Create_Dummy_Place( RdP->T[i], Index[i]+1, 0.0 );
	      Store_Place( RdP, The_Place );
	      for ( j=0; j<Mu_Star[i]-2; j++ ) {
		The_Place = Create_Dummy_Place( RdP->T[j+Index[i]],
						j+Index[i]+2, 0.0 );
		Store_Place( RdP, The_Place );
	      }
	      break;
	    case AMONT:
	      The_Place = 
		Create_Dummy_Place( RdP->T[Index[i]+Mu_Star[i]-2],
				    i+1, 
				    Max_Lags_Of_Inputs( Reseau, i, 2 ) );
	      Store_Place( RdP, The_Place );
	      for ( j=0; j<Mu_Star[i]-2; j++ ) {
		The_Place = 
		  Create_Dummy_Place( RdP->T[j+Index[i]], 
				      Index[i]+j+2,
				      Max_Lags_Of_Inputs( Reseau, i, 
							 Mu_Star[i]-j ) );
		Store_Place( RdP, The_Place );
	      }
	      break;
	    case ECLATEMENT:
	      Sub_Index = Index[i];
	      SCAN( P_Scan, Reseau->T[i]->Outputs ) {

		Mu = P_Scan->Val->Init_Pop;
		if ( Mu > 1 ) {
		  The_Place = 
		    Create_Dummy_Place( RdP->T[i], Sub_Index+1, 
					P_Scan->Val->Lag_Times[Mu-1] );
		  Store_Place( RdP, The_Place );
		}
		for (j=1; j<Mu-1; j++) {
		  The_Place =
		    Create_Dummy_Place( RdP->T[Sub_Index+j-1],
					Sub_Index+j+1,
					P_Scan->Val->Lag_Times[Mu-j-1] );
		  Store_Place( RdP, The_Place );
		}
		Sub_Index += MAX( Mu-1, 0 );
	      }
	      break;
	  }
	}
      }
    }
  }

  free( Index );
  free( Mu_Star );

  return RdP;

}

