/* Program to generate Event Graphs with given structure in EGL.	*/

#include <string.h>
#include "eg.h"

#define TOOLNAME	"gen_eg"

#define	Empty		0
#define	Full		1
#define First_Place	2
#define Even_Places	3
#define Diagonal	4
#define Checker		5
#define Random		6

typedef int Init_Type;

#define	Ser_Open	0
#define Ser_Closed	1
#define	Job_Shop	2
#define	Double_Loop	3
#define Random_Edge	4

typedef int Net_Type;

#define Illimited	-1		/* for capacity purposes	*/
#define MAX_RANDOM_VALUE          0x7FFFFFFF

#ifndef ALPHA
extern void srandom( unsigned seed );
extern long random( void );
#endif

double U_0_1(void)
{
  return (double)random()/MAX_RANDOM_VALUE;
}

void fprint_init_mark(FILE *OutFile, int Number)
{	
  int j;

  if (Number>0) {
    fprintf(OutFile, "%d ",Number);
    for (j=0; j<Number; j++)
      fprintf(OutFile, "0.0 ");
  }
}

void gen_series_open(FILE	*OutFile,
		     int	Net_Size,
		     Law_Desc	The_Trans_Law,
		     Law_Desc	The_Place_Law,
		     Init_Type	Init_Mode,
		     int	Buf_Capa,
		     double	Proba )
{
  int	i;

  fprintf(OutFile, "Size <- %d\n", Net_Size);

  for (i=0; i<Net_Size; i++) {
    fprintf(OutFile, "%d ", i+1);
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );

    /* recycling of the transition */
    fprintf(OutFile, ", %d C 0.0 1 0.0", i+1);

    /* place for limited buffer size */
    if ( ( Buf_Capa != Illimited ) && ( i != 0 ) ) {
      fprintf(OutFile, ", %d ", i); /* i.e. i-1+1 */
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
      switch(Init_Mode) {
      case Empty:
	fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Full:
	fprint_init_mark(OutFile, Buf_Capa);;
	break;
      case First_Place:
      case Diagonal:
	if (i==1)
	  fprint_init_mark(OutFile, Buf_Capa);
	else	
	  fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Even_Places:
      case Checker:
	if (i % 2 == 1)
	  fprint_init_mark(OutFile, Buf_Capa);
	else
	  fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Random:
	fprint_init_mark(OutFile, Buf_Capa+(U_0_1()<Proba?1:0) );
	break;
      }
    }

    /* the next place, if any */
    if ( i < Net_Size-1 ) {
      fprintf(OutFile, ", %d ", i+2);
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
      switch(Init_Mode) {
      case Empty:
	fprintf(OutFile, "0;\n");
	break;
      case Full:
	fprintf(OutFile, "1 0.0;\n");
	break;
      case First_Place:
      case Diagonal:
	if (i==0)
	  fprintf(OutFile, "1 0.0;\n");
	else
	  fprintf(OutFile, "0;\n");
	break;
      case Even_Places:
      case Checker:
	if (i % 2 == 0)
	  fprintf(OutFile, "1 0.0;\n");
	else
	  fprintf(OutFile, "0;\n");
	break;
      case Random:
	if ( U_0_1() < Proba ) 
	  fprintf(OutFile, "1 0.0;\n");
	else
	  fprintf(OutFile, "0;\n");
	break;
      }
    }
    else	fprintf(OutFile, ";\n");
  }
}
			
void gen_series_closed(FILE		*OutFile,
		       int		Net_Size,
		       Law_Desc		The_Trans_Law,
		       Law_Desc		The_Place_Law,
		       Init_Type	Init_Mode,
		       int		Buf_Capa,
		       double		Proba )
{
  int	i;

  fprintf(OutFile, "Size <- %d\n", Net_Size);

  for (i=0; i<Net_Size; i++) {
    fprintf(OutFile, "%d ", i+1);
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );

    /* recycling of the transition */
    fprintf(OutFile, ", %d C 0.0 1 0.0", i+1);

    if ( ( Buf_Capa != Illimited ) && ( i != 0 ) ) {
      fprintf(OutFile, ", %d ", i); /* i.e. i-1+1 */
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
      switch(Init_Mode) {
      case Empty:
	fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Full:
	fprint_init_mark(OutFile, Buf_Capa);;
	break;
      case First_Place:
      case Diagonal:
	if (i==1)
	  fprint_init_mark(OutFile, Buf_Capa);
	else
	fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Even_Places:
      case Checker:
	if (i % 2 == 1)
	  fprint_init_mark(OutFile, Buf_Capa);
	else
	fprint_init_mark(OutFile, Buf_Capa+1);
	break;
      case Random:
	if ( U_0_1() < Proba ) 
	  fprintf(OutFile, "1 0.0;\n");
	else
	  fprintf(OutFile, "0;\n");
	break;
      }
    }

    /* the next place */
    fprintf(OutFile, ", %d ", ( (i+1) % Net_Size ) + 1);
    Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
    switch(Init_Mode) {
    case Empty:
      fprintf(OutFile, "0;\n");
      break;
    case Full:
      fprintf(OutFile, "1 0.0;\n");
      break;
    case First_Place:
    case Diagonal:
      if (i==0)
	fprintf(OutFile, "1 0.0;\n");
      else
	fprintf(OutFile, "0;\n");
      break;
    case Even_Places:
    case Checker:
      if (i % 2 == 0)
	fprintf(OutFile, "1 0.0;\n");
      else
	fprintf(OutFile, "0;\n");
      break;
    case Random:
      if ( U_0_1() < Proba ) 
	fprintf(OutFile, "1 0.0;\n");
      else
	fprintf(OutFile, "0;\n");
      break;
    }
  }
}

void gen_random(FILE		*OutFile,
		int		Net_Size,
		Law_Desc	The_Trans_Law,
		Law_Desc	The_Place_Law,
		Init_Type	Init_Mode,
		double		Proba,
		double		Proba2 )
{
  int	i;
  int	j;

  fprintf(OutFile, "Size <- %d\n", Net_Size);

  for (i=0; i<Net_Size; i++) {
    fprintf(OutFile, "%d ", i+1);
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );

    /* recycling of the transition */
    fprintf(OutFile, ", %d C 0.0 1 0.0", i+1);

    for (j=0; j<Net_Size; j++) {
      if ( ( i != j ) && ( U_0_1() < Proba ) ) {
	fprintf(OutFile, ", %d ", j+1); 
	Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
	switch(Init_Mode) {
	case Empty:
	  fprintf(OutFile, "0;\n");
	  break;
	case First_Place:
	case Diagonal:
	  if (i==0)
	    fprintf(OutFile, "1 0.0;\n");
	  else
	    fprintf(OutFile, "0;\n");
	  break;
	case Even_Places:
	case Checker:
	  if (i % 2 == 0)
	    fprintf(OutFile, "1 0.0;\n");
	  else
	    fprintf(OutFile, "0;\n");
	  break;
	case Full:
	  fprint_init_mark(OutFile, 1);
	  break;
	case Random:
	  if ( U_0_1() < Proba ) 
	    fprintf(OutFile, "1 0.0;\n");
	  else
	    fprintf(OutFile, "0;\n");
	  break;
	}
      }
    }
    fprintf(OutFile, ";\n");
  }
}


void fprint_place_js(FILE*	OutFile,
		     int	Destination,
		     Law_Desc	The_Law,
		     Init_Type	Init_Mode,
		     int	Line, 
		     int	Column,
		     double	Proba )
{	
  fprintf(OutFile, ", %d ", Destination );
  Write_Law(OutFile, The_Law, NORMAL_PRINT_MODE );
  switch(Init_Mode) {
  case Empty:
    fprintf(OutFile, "0");
    break;
  case Full:
    fprintf(OutFile, "1 0.0");
    break;
  case First_Place:
    if ( (Line==0) && (Column==0) )
      fprintf(OutFile, "1 0.0");
    else
      fprintf(OutFile, "0");
    break;
  case Diagonal:
    if (Line==Column)
      fprintf(OutFile, "1 0.0");
    else
      fprintf(OutFile, "0");
    break;
  case Even_Places:
  case Checker:
    if ( (Line+Column) % 2 == 0 )
      fprintf(OutFile, "1 0.0");
    else
      fprintf(OutFile, "0");
    break;
  case Random:
    if ( U_0_1() < Proba ) 
      fprintf(OutFile, "1 0.0\n");
    else
      fprintf(OutFile, "0\n");
    break;
  }
}
			
void gen_job_shop(FILE		*OutFile,
		  int		Size_Machine, 
		  int		Size_Piece,
		  Law_Desc	The_Trans_Law, 
		  Law_Desc	The_Place_Law,
		  Init_Type	Init_Mode,
		  double	Proba )
{
  int	i;
  int	j;
  int	Trans_Id;

  fprintf(OutFile, "Size <- %d\n", Size_Machine*Size_Piece);

  for (i=0; i<Size_Machine; i++) {
    for (j=0; j<Size_Piece; j++) {
      Trans_Id = i*Size_Piece+j+1;
      fprintf(OutFile, "%d ", Trans_Id );
      fprintf(OutFile, "\"Trans_%d_%d\" ", i, j );
      Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );
      /* recycling of the transition */
      fprintf(OutFile, ", %d C 0.0 1 0.0", Trans_Id );
      /* the next machine place */
      fprint_place_js(OutFile, 
		      ( (i+1) % Size_Machine ) * Size_Piece + j + 1,
		      The_Place_Law, Init_Mode, i, j, Proba );
      /* the next piece place */
      fprint_place_js(OutFile, 
		      i * Size_Piece + ( j + 1 ) % Size_Piece + 1,
		      The_Place_Law, Init_Mode, i, j, Proba );
      fprintf(OutFile, ";\n");

    }
  }
}
			
void fprint_place_dl(FILE*	OutFile,
		     int	Destination,
		     Law_Desc	The_Law,
		     Init_Type	Init_Mode,
		     int	Size1,
		     int	Size2,
		     double	Proba )
{	
  fprintf(OutFile, ", %d ", Destination+1 );
  Write_Law(OutFile, The_Law, NORMAL_PRINT_MODE );
  switch(Init_Mode) {
  case Empty:
    fprintf(OutFile, "0");
    break;
  case Diagonal:
  case First_Place:
    if ( (Destination==0) || (Destination==Size1) )
      fprintf(OutFile, "1 0.0");
    else	fprintf(OutFile, "0");
    break;
  case Even_Places:
  case Checker:
    if ( Destination % 2 == 0 )
      fprintf(OutFile, "1 0.0");
    else	fprintf(OutFile, "0");
    break;
  case Full:
    fprintf(OutFile, "1 0.0");
    break;
  case Random:
    if ( U_0_1() < Proba ) 
      fprintf(OutFile, "1 0.0\n");
    else
      fprintf(OutFile, "0\n");
    break;
    fprint_init_mark(OutFile, (U_0_1()<Proba?1:0) );
  }
}
			
void gen_double_loop(FILE	*OutFile,
		     int	Net_Size,
		     int	Net_Size1,
		     Law_Desc	The_Trans_Law, 
		     Law_Desc	The_Place_Law,
		     Init_Type	Init_Mode,
		     double	Proba )
{
  int	i;
  int	dest;

  fprintf(OutFile, "Size <- %d\n", Net_Size+Net_Size1);

  for (i=0; i<Net_Size+Net_Size1; i++) {
    fprintf(OutFile, "%d ", i+1);
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );
    /* recycling of the transition */
    fprintf(OutFile, ", %d C 0.0 1 0.0", i+1);
    /* transition 0 has an extra output, with zero initial marking */
    if (i==0) {
      fprintf(OutFile, ", %d ", Net_Size+1);
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
      fprintf(OutFile, " 0");
    }
    /* the next place */
    if ( i < Net_Size-1 )
      dest = i+1;
    else if ( i == Net_Size-1 )
      dest = 0;
    else if ( i == Net_Size+Net_Size1 - 1 )
      dest = Net_Size;
    else
      dest = i+1;

    fprint_place_dl(OutFile, dest, The_Place_Law, Init_Mode,
		    Net_Size, Net_Size1, Proba );
    fprintf(OutFile, ";\n");

    }
}

void display_help( char *Prog_Name )
{
  char	Format[100];
	
  sprintf(Format, "%%%dc [-tl <law>] [-pl <law>] [-o <file>] [-h]\n",
	  (int)(strlen(Prog_Name)+7) );

  printf("Usage: %s [-t <type>] [-s <size(s)>] [-i <init>]",Prog_Name);
  printf(" [-p <proba(s)>] [-seed <int>]\n");
  printf(Format,' ');
  printf("  <type> : ser_closed (def) ser_open double job_shop random\n");
  printf("  <init> : full (def) empty first even diagonal checker random\n");
  printf("  <law>  : E (def) C B G R U H + parameters (def: 1.0)\n");
}

int Parse_Law_From_Args( char **argv, int index, int argc,
			   Law_Desc The_Law )
{
  int		i;
  int		j;
  int		Nb_Params;
  boolean 	Error;
  boolean	Short;
  boolean 	Ok;
  double	Foo;
  
  i = index;
  Error = FALSE;
  Short = FALSE;

  The_Law->Name = argv[i][0];
  i++;

  if ( i < argc ) {
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
    case 'B':
    case 'g':
    case 'G':
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else
	i++;
      break;
    case 'R':
    case 'U':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 2*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      } 
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  }
	  else 
	    i++;
	} 
	else {
	  Short = TRUE;
	}
      }
      break;
    case 'O':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 3*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  } 
	  else {
	    i++;
	    if ( i < argc ) {
	      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[2] ) ) ) {
		Error = TRUE;
	      } 
	    }
	  }
	}
	else 
	  Error = TRUE;
      }
      break;
    case 'H':
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first.							*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 1 == sscanf( argv[j], "%lf", &Foo ) );
      }
      /* counting the number of floats read */
      if ( j < argc )
	j--;
      Nb_Params = (j-i);
      if ( Nb_Params %2 != 0 ) {
	Error = TRUE;
      } else {
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=i; j<i+Nb_Params; j++ ) {
	  sscanf( argv[j], "%lf", &(The_Law->Parameters[1+j-i]) );
	}
	i = i+Nb_Params;
      }
      break;
    case 'D':
      /* skip the (a priori) "[" argument */
      i++;
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first, looking for the next "]".			*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 0 != strcmp( argv[j], "]" ) );
      }
      /* counting the number of floats read */
      if ( j >= argc ) {
	Short = TRUE;
      }
      else {
	j--;
	Nb_Params = (j-i);
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+2*Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=0; j<Nb_Params; j++ ) {
	  sscanf( argv[i+j], "%lf", &(The_Law->Parameters[1+j]) );
	}
	/* skip "]" and "[" */
	i += Nb_Params+2;
	/* continue reading. Have to protect against short arg list */
	Ok = TRUE;
	for( j=0; Ok && (j<Nb_Params) && (i+j<argc); j++ ) {
	  Ok = ( 1 == sscanf( argv[i+j], "%lf", 
			      &(The_Law->Parameters[1+Nb_Params+j]) ) );
	}
	if ( !Ok ) {
	  Error = TRUE;
	  i = i+j-1;
	} 
	else if ( i+j >= argc ) {
	  Short = TRUE;
	}
	else {
	  /* skip final "]", hoping it is actually there */
	  i = i + Nb_Params + 1;
	}
      }
      break;
    case 'P':
      /* read first parameter */
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	/* recursively read two laws */
	The_Law->Law_Parameters = (Law_Desc*) calloc( 2, sizeof(Law_Desc) );
	The_Law->Law_Parameters[0] = New_Law();
	The_Law->Law_Parameters[1] = New_Law();
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[0] );
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[1] );
      }
    default:
      printf("%s: unknown law id (%c)\n", TOOLNAME, The_Law->Name );
    }
  }
  else {
    Short = TRUE;
  }

  if ( Short ) {
    printf("%s: error parsing law %c: not enough arguments.\n", 
	   TOOLNAME, The_Law->Name );
    i = index;
  }
  else if ( Error ) {
    printf("%s: error parsing law %c at argument %s.\n", 
	   TOOLNAME, The_Law->Name, argv[i] );
    i = index;
  }

  return i;
}

int main(int argc, char	*argv[])
{
  Net_Type	RdP_Type = Ser_Closed;
  Init_Type	Init_Mode = Full;
  static char	File_Name[128] = "";
  int		Net_Size = 2;
  int		Net_Size2 = 2;
  int		Buf_Capa = Illimited;
  Law_Desc	The_Trans_Law;
  Law_Desc	The_Place_Law;
  int		i;
  double	Proba = 0.5;
  double	Proba2 = 0.5;
  int		Seed = -1;
  FILE*		OutFile;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;

  The_Trans_Law = New_Law();
  The_Place_Law = New_Law();
  The_Trans_Law->Name = The_Trans_Law->Name = 'E';
  The_Place_Law->Name = The_Trans_Law->Name = 'E';
  The_Place_Law->Parameters[0] = 1.0;
  The_Trans_Law->Parameters[0] = 1.0;

  i = 1;

  if (argc>1) {
    do {
      if ( 0 == strcmp( "-t", argv[i] ) ) {
	i++;
	if ( 0 == strcmp( "ser_open", argv[i] ) )
	  RdP_Type = Ser_Open;
	else if ( 0 == strcmp( "ser_closed", argv[i] ) )
	  RdP_Type = Ser_Closed;
	else if ( 0 == strcmp( "job_shop", argv[i] ) )
	  RdP_Type = Job_Shop;
	else if ( 0 == strcmp( "double", argv[i] ) )
	  RdP_Type = Double_Loop;
	else if ( 0 == strcmp( "random", argv[i] ) )
	  RdP_Type = Random_Edge;
	else {
	  printf("%s: unrecognized network type\n", TOOLNAME);
	  Args_Ok = FALSE;
	}
	i++;
      }
      else if ( 0 == strcmp( "-s", argv[i] ) ) {
	i++;
	switch( RdP_Type ) {
	case Ser_Open:
	case Ser_Closed:
	case Random_Edge:
	  if ( 1 != sscanf( argv[i], "%d", &Net_Size ) ) {
	    printf("%s: size expected\n", TOOLNAME ); 
	    Args_Ok = FALSE;
	  }
	  else
	    i++;
	  break;
	case Job_Shop:
	case Double_Loop:
	  if ( ( 1 != sscanf( argv[i], "%d", &Net_Size ) )
	       || ( 1 != sscanf( argv[i+1], "%d", &Net_Size2 ) ) ) {
	    printf("%s: two sizes expected\n", TOOLNAME );
	    Args_Ok = FALSE;
	  }
	  else {
	    i++; i++;
	  }
	  break;
	}
      }

      else if GET_INT( "-c", Buf_Capa )
      else if GET_INT( "-seed", Seed )

      else if ( 0 == strcmp( "-p", argv[i] ) ) {
	i++;
	switch( RdP_Type ) {
	case Ser_Open:
	case Ser_Closed:
	case Job_Shop:
	case Double_Loop:
	  if ( 1 != sscanf( argv[i], "%lf", &Proba ) ) {
	    printf("%s: probability expected after -p\n", TOOLNAME );
	    Args_Ok = FALSE;
	  }
	  else
	    i++;
	  break;
	case Random_Edge:
	  if ( ( 1 != sscanf( argv[i], "%lf", &Proba ) )
	       || ( 1 != sscanf( argv[i+1], "%lf", &Proba2 ) ) ) {
	    printf("%s: two probabilities expected after -p\n", TOOLNAME );
	    Args_Ok = FALSE;
	  }
	  else {
	    i++; i++;
	  }
	  break;
	}
      }

      else if ( 0 == strcmp( "-i", argv[i] ) ) {
	i++;
	if ( 0 == strcmp( "empty", argv[i] ) )
	  Init_Mode = Empty;
	else if ( 0 == strcmp( "full", argv[i] ) )
	  Init_Mode = Full;
	else if ( 0 == strcmp( "first", argv[i] ) )
	  Init_Mode = First_Place;
	else if ( 0 == strcmp( "even", argv[i] ) )
	  Init_Mode = Even_Places;
	else if ( 0 == strcmp( "diagonal", argv[i] ) )
	  Init_Mode = Diagonal;
	else if ( 0 == strcmp( "checker", argv[i] ) )
	  Init_Mode = Checker;
	else if ( 0 == strcmp( "random", argv[i] ) )
	  Init_Mode = Random;
	else {
	  printf("%s: unrecognized init mode\n", TOOLNAME );
	  Args_Ok = FALSE;
	}
	i++;
      }
      else if ( 0 == strcmp( "-tl", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, The_Trans_Law );
      }
      else if ( 0 == strcmp( "-pl", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, The_Place_Law );
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc )
	  strcpy( File_Name, argv[i] );
	else	
	  printf("%s: filename expected after %s\n", TOOLNAME, argv[i] );
	i++;
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		||  0 == strcmp( "-?", argv[i] ) ) {
	i++;
	display_help(argv[0]);
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else if IS_VERSION

      else 
	i++;	/* skip argument */
    } while ( ( i < argc ) && Args_Ok );
  } 

  if (Args_Ok) {

    if ( strlen(File_Name) != 0 )
      OutFile = fopen( File_Name, "w" );
    else
      OutFile = stdout;

    if ( Seed >= 0 )
      srandom((unsigned)Seed);

    if ( OutFile != NULL ) {
      switch( RdP_Type ) {
      case Ser_Open:
	gen_series_open( OutFile, Net_Size, 
			 The_Trans_Law, The_Place_Law, Init_Mode,
			 Buf_Capa, Proba );
	break;
      case Ser_Closed:
	gen_series_closed( OutFile, Net_Size, 
			   The_Trans_Law, The_Place_Law, Init_Mode,
			   Buf_Capa, Proba );
	break;
      case Job_Shop:
	gen_job_shop( OutFile, Net_Size, Net_Size2,
		      The_Trans_Law, The_Place_Law, Init_Mode,
		      Proba );
	break;
      case Double_Loop:
	gen_double_loop( OutFile, Net_Size, Net_Size2,
			 The_Trans_Law, The_Place_Law, Init_Mode,
			 Proba );
	break;
      case Random_Edge:
	gen_random( OutFile, Net_Size, 
		    The_Trans_Law, The_Place_Law, Init_Mode,
		    Proba, Proba2 );
	break;
      }

      /* successful! */
      Free_Law( The_Trans_Law );
      Free_Law( The_Place_Law );

      return 0;

    }
    else {
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, File_Name );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}
