/************************************************************************/
/* module where calls to external programs are performed in the task	*/
/* of computing distributions for M/EG networks.			*/
/************************************************************************/

#include <unistd.h>
#include "eg.h"
#define RESOURCES
#include "resources.h"

#define TOOLNAME	"m_eg"

/* externs from parse_EG */
extern EventGraph	*Read_Network(char*);
extern boolean		topological_analysis(EventGraph*);

/* extern from module transcript.c */
extern boolean 		Write_EG_As_EGL(EventGraph*,char*,char*);

/* externs from the resource system */
extern void 	get_resources(boolean);
extern char     **Resource_Value;
extern boolean	*Is_Available;
extern int  	execute(char*,...);

#define STOP    (char *)0
#define RV(x)   Resource_Value[ x ]

boolean Resources_Are_Present()
{
  return ( Is_Available[ MK1 ] 
	   && Is_Available[ EG2EG ] 
	   && Is_Available[ EG2MATRIX ]
	   && Is_Available[ SCSCYC ] );
}

boolean Initialize_Resources(boolean	Checking)
{
  get_resources( Checking );

  return Resources_Are_Present();
}


EventGraph *Flatten_Network(
EventGraph	*Reseau,
boolean		Keep )
{
  /* Dumping the network for immediate re-reading... 	*/
  /* Sounds stupid, but not simple to avoid...        	*/
  Write_EG_As_EGL( Reseau, RV( TMP_EGL_FILE ), TOOLNAME );
  
  execute( TOOLNAME,
	   RV( EG2EG ), "-f", RV( TMP_EGL_FILE ), "-flatten", 
	   "-o", RV( TMP_EGL_FILE2 ),
	   STOP );

  Reseau = Read_Network( RV( TMP_EGL_FILE2 ) );

  if ( NULL != Reseau ) {
    topological_analysis( Reseau );
  }
  else {
    fprintf( stderr, "%s: Error reading flattened network.\n", TOOLNAME );
  }

  if ( !Keep ) {
    unlink( RV( TMP_EGL_FILE ) );
    unlink( RV( TMP_EGL_FILE2 ) );
  }

  return Reseau;

}

void	Compute_Matrix( 
EventGraph	*Reseau,
int		Trans_Nb,
boolean 	Keep )
{
  char	The_Trans[8];

  sprintf( The_Trans, "%d", Trans_Nb );
  /* Dumping the network for immediate re-reading... 	*/
  /* Sounds stupid, but not simple to avoid...        	*/
  Write_EG_As_EGL( Reseau, RV( TMP_EGL_FILE ), TOOLNAME );

  execute( TOOLNAME,
	   RV( EG2MATRIX ), "-f", RV( TMP_EGL_FILE ), 
	   "-o", RV( TMP_MAT_FILE ),
	   "-int", "-bruno",
	   "-flag_trans", The_Trans,
	   STOP );

  if ( !Keep ) {
    unlink( RV( TMP_EGL_FILE ) );
  }

}

void	Compute_Transients( 
EventGraph	*Reseau,
boolean 	Keep )
{
  /* Dumping the network for immediate re-reading... 	*/
  /* Sounds stupid, but not simple to avoid...        	*/
  Write_EG_As_EGL( Reseau, RV( TMP_EGL_FILE ), TOOLNAME );

  execute( TOOLNAME,
	   RV( SCSCYC ), "-f", RV( TMP_MAT_FILE ), 
	   "-o", RV( TMP_RES_FILE ),
	   "-transient",
	   STOP );

  if ( !Keep ) {
    unlink( RV( TMP_MAT_FILE ) );
  }
}

void	Compute_Distribution(
double	Lambda,
boolean	Value,
double	x_Start, 
double	x_Stop,
double	x_Val,
double  x_Step,
boolean	Moment,
int	Order,
boolean	Log,
boolean	Queue,
boolean	Keep )
{
  char	Lambda_Val[20];
  char	X_Val1[20], X_Val2[20], X_Val3[20];
  char	Mom_Val[4];

  sprintf( Lambda_Val, "%f", Lambda );
  sprintf( X_Val1, "%f", x_Start );
  sprintf( X_Val2, "%f", x_Stop );
  sprintf( X_Val3, "%f", x_Val );

  if ( Value ) {

    execute( TOOLNAME,
	     RV( MK1 ), "-f", RV( TMP_MK1_FILE ), 
	     "-lambda", Lambda_Val,
	     "-xval", X_Val3,
	     STOP );
  }
  else if ( !Moment ) {

    if ( Queue ) {
      if ( Log ) {
	execute( TOOLNAME,
		 RV( MK1 ), "-f", RV( TMP_MK1_FILE ), 
		 "-lambda", Lambda_Val,
		 "-xfrom", X_Val1,
		 "-xto", X_Val2,
		 "-queue", "-log", 
	     STOP );
      }
      else {
	execute( TOOLNAME,
		 RV( MK1 ), "-f", RV( TMP_MK1_FILE ), 
		 "-lambda", Lambda_Val,
		 "-xfrom", X_Val1,
		 "-xto", X_Val2,
		 "-queue",
		 STOP );
      }
    }
    else {
      execute( TOOLNAME,
	       RV( MK1 ), "-f", RV( TMP_MK1_FILE ), 
	       "-lambda", Lambda_Val,
	       "-xfrom", X_Val1,
	       "-xto", X_Val2,
	       STOP );
    }
  }
  else {
    sprintf( Mom_Val, "%d", Order );
    execute( TOOLNAME,
	     RV( MK1 ), "-f", RV( TMP_MK1_FILE ), 
	     "-lambda", Lambda_Val,
	     "-moment", Mom_Val,
	     STOP );
  }
    

  if ( !Keep ) {
    unlink( RV( TMP_MK1_FILE ) );
  }
}

