#include <stdio.h>
#include <math.h>

#include "eg.h"
#define RESOURCES
#include "resources.h"

#define	TOOLNAME	"m_eg"

#define	MAX(x,y)	( (x>y) ? (x) : (y) )

boolean	Debug;
boolean	Queue = FALSE;
boolean Log = FALSE;

/* externs from parse_EG */
extern EventGraph	*Read_Network(char*);
extern void		dumpe_network(EventGraph*);
extern boolean		topological_analysis(EventGraph*);
/* externs from module calls.c */
extern boolean		Initialize_Resources(boolean);
extern EventGraph 	*Flatten_Network(EventGraph*,boolean);
extern void		Compute_Matrix(EventGraph*,int,boolean);
extern void		Compute_Transients(EventGraph*,boolean);
extern void		Extract_Parameters(int,int,int*,int*,
					   double**,double**);
extern void		Compute_Distribution(double,boolean,double,double,
					     double,double,boolean,int,
					     boolean,boolean,boolean);

/* externs from the resource system */
extern void 	get_resources(boolean);
extern char     **Resource_Value;
extern boolean	*Is_Available;

#define STOP    (char *)0
#define RV(x)   Resource_Value[ x ]

/* forward declarations */
boolean Dump_Data_File(double Lambda, int T, int D, double *B,
		       double *Inc, int N_Max );

/* main program */
int main(int argc, char *argv[])
{
  static char		Result_File_Name[256] = "";
  static char		EG_File_Name[256] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Value = FALSE;
  boolean		Moment = FALSE;
  boolean		Stationary = FALSE;
  boolean		Keep = FALSE;
  boolean		Checking = FALSE;
  double		x_Start = 0.0;
  double		x_Stop = 10.0;
  double		x_Step = 0.1;
  double		x_Val;
  double		Lambda = 0.0;
  int			Order = 1;	/* for the moments	*/
  int			T;		/* lg of transients 	*/
  int			D;		/* lg of period 	*/
  double		*B;		/* B of transients 	*/
  double		*Inc;		/* Increments of period */
  int			Trans_Nb = -1;
  int			Input_Nb = -1;
  int			Event_Nb = -1;
  EventGraph		*Reseau;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", EG_File_Name )
      else if GET_STRING( "-o", Result_File_Name )

      else if SWITCH( "-keep", Keep, TRUE )
      else if SWITCH( "-c", Checking, TRUE )
      else if SWITCH( "-stationary", Stationary, TRUE )
      else if SWITCH( "-queue", Queue, TRUE )
      else if SWITCH( "-log", Log, TRUE )

      else if GET_INT_SWITCH( "-moment", Order, Moment, TRUE )

      else if GET_INT( "-trans_nb", Trans_Nb )
      else if GET_INT( "-input_nb", Input_Nb )
      else if GET_INT( "-event_nb", Event_Nb )

      else if GET_REAL( "-lambda", Lambda )
      else if GET_REAL( "-xfrom", x_Start )
      else if GET_REAL( "-xto", x_Stop )
      else if GET_REAL_SWITCH( "-xval", x_Val, Value, TRUE )
      else if GET_REAL( "-xstep", x_Step )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {

	printf("Usage: %s [-f <file>] [-o <file>]\n\t-lambda <real> ", 
	       TOOLNAME );
	printf("-trans_nb <int> -input_nb <int> -event_nb <int>\n\t");
	printf("[-xfrom <real>] [-xto <real>] [-xval <real>] ");
	printf("[-xstep <real>] [-queue]\n\t[-log] [-moment <int>] ");
	printf("[-stationary] [-c] [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    if ( ! Initialize_Resources( Checking ) ) {
      printf("%s: error: some resources are missing", TOOLNAME );
      printf(" (check with -c). No analysis.\n" );
      
      return -2;
    }

    if ( Checking )
      return 0;

    /* checking that required arguments have been supplied */
    if ( Trans_Nb < 0 ) {
      printf("%s: error: argument to -trans_nb required.\n",
	     TOOLNAME );
      return -4;
    }

    if ( Input_Nb < 0 ) {
      printf("%s: error: argument to -input_nb required.\n",
	     TOOLNAME );
      return -4;
    }

    if ( !Stationary && ( Event_Nb < 0 ) ) {
      printf("%s: error: argument to -event_nb required.\n",
	     TOOLNAME );
      return -4;
    }

    if ( Lambda <= 0.0 ) {
      printf("%s: error: positive argument to -lambda required.\n",
	     TOOLNAME );
      return -4;
    }
      
    /* reading the network */

    Reseau = Read_Network( EG_File_Name );

    if ( Reseau != NULL ) {

      if (Debug)
	dumpe_network(Reseau);

      if ( Trans_Nb >= Reseau->Size ) {
	printf("%s: error: argument to -trans_nb (%d)", TOOLNAME, Trans_Nb );
	printf("exceeds network size (%d).\n", Reseau->Size );
	return -4;
      }

      if ( topological_analysis(Reseau) ) {

	/* now we can begin to work... */
	if ( Reseau->Max_Pop_Init > 1 ) {
	  /* network is not flat... need to flatten it */
	  Reseau = Flatten_Network( Reseau, Keep );
	}

	if ( Reseau != NULL ) {
	  Compute_Matrix( Reseau, Input_Nb, Keep );
	  Compute_Transients( Reseau, Keep );
	  Extract_Parameters( Reseau->Size, Trans_Nb, &T, &D, &B, &Inc );

	  /* decide what exactly is to be computed */
	  if ( Value ) {
	    x_Start = x_Val;
	    x_Stop = x_Val;
	  }
	  /* overrides stationarity */
	  if ( Stationary )
	    Event_Nb = -1;

	  /* compute it ! */
	  if ( Dump_Data_File( Lambda, T, D, B, Inc, Event_Nb ) ) {

	    if ( strlen(Result_File_Name) != 0 )
	      freopen( Result_File_Name, "w", stdout );

	    Compute_Distribution( Lambda,
				  Value, x_Start, x_Stop, x_Val, x_Step, 
				  Moment, Order,
				  Log, Queue, Keep );

	    return 0;
	  }
	  else {
	    return -2;
	  }
	}
	else {
	  fprintf( stderr, "%s: Error: flattening failed. Abort.\n",
		   TOOLNAME );
	  return -1;
	}
      }
      else {
	fprintf( stderr, "%s: Error: topological analysis failed. Abort.\n",
		 TOOLNAME );
	return -1;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No analysis.\n",
	       TOOLNAME );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* Writing down the data for the mk1 program */
boolean Dump_Data_File(
double	Lambda,
int	T,
int	D,
double	*B,
double	*Inc,
int	N_Max )
{
  FILE		*Data_File;
  boolean	Ok = TRUE;
  int		i;

  Data_File = fopen( RV( TMP_MK1_FILE ), "w" );
  Ok = ( NULL != Data_File );

  if ( Ok ) {

    fprintf( Data_File, "%12.8f\n", Lambda );
    fprintf( Data_File, "%d\n", T );
    fprintf( Data_File, "%d\n", D );

    for ( i=0; i<=T; i++ )
      fprintf( Data_File, " %12.8f", B[i] );
    fprintf( Data_File, "\n");

    for ( i=0; i<D; i++ )
      fprintf( Data_File, " %12.8f", Inc[i] );
    fprintf( Data_File, "\n");

    fprintf( Data_File, "%d\n", N_Max );

    fclose( Data_File );

    return TRUE;
  }
  else {
    fprintf( stderr, "%s: Error: could not open file %s for writing.\n",
	     TOOLNAME, RV( TMP_MK1_FILE ) );
    return FALSE;
  }

}
