%{

#include "eg.h"

#define YYDEBUG 1

extern boolean		Debug;

extern EventGraph 	*Parse_EG(int, Liste_Transition);
extern Trans_Desc 	Parse_Transition(int, Law_Desc, char*, char* );
extern Trans_Desc 	Add_Sons(Trans_Desc, Liste_Place);
extern Population	*Parse_Init_Pop(int, Liste_Reel);
extern Place_Desc	Parse_Place(int, Law_Desc, Population *, 
				    char*, char*, char*);
extern Liste_Transition Append_Transition(Liste_Transition, Trans_Desc);
extern Liste_Place	Append_Place(Liste_Place, Place_Desc);

/* The result of the parsing is to be put into the (global!) variable */
/* The_Event_Graph. It is defined in "parse_eg.c".                     */
/* I used to use "yyval" but this does not work with flex/bison.      */
extern EventGraph	*The_Event_Graph;

%}

%start reseau

%union{
	int		ival;
	double		rval;
	char		*sval;
	EventGraph	*un_reseau;
	Law_Desc	loi;
	Place_Desc	place;
	Liste_Place	plist;
	Trans_Desc	transition;
	Liste_Transition 
			tlist;
	Parametres	params;
	Population	*pop;
	Liste_Reel	rlist;
	Liste_Loi	Llist;
}

%token	<ival>		SIZE TMAX AFFECT INTGR
%token	<ival>		SEP ENDINSTR 	/* pas de valeur, en fait */
%token	<ival>		PARB PARE
%token	<rval>		REAL
%token	<sval>		STRING
%token	<loi>		EXPO ERLG BERN GEOM0 GEOM1 UNIF CST DISC ONOFF IPP HEXP
%token  <loi>		LOGNO PARETO WEIBULL

%type	<un_reseau>	reseau
%type	<transition>	transition_dcl transition_sons
%type	<tlist>		liste_transitions
%type	<params>	declarations
%type	<ival>		trans_id
%type	<loi>		loi_dcl
%type	<place>		place_dcl
%type	<plist>		place_list
%type	<pop>		pop_dcl
%type	<ival>		taille_dcl duree_dcl
%type	<rlist>		real_list
%type   <Llist>		law_list
%type	<sval>		name

%%

reseau	:	declarations liste_transitions
		{
#ifdef USE_YYLVAL
                        $$ = Parse_EG( $1.T_Size, $2 );
#else
                        The_Event_Graph = Parse_EG( $1.T_Size, $2 );
#endif

			if ( Debug )
			  printf("Reseau ok\n");
		}
		;

declarations :	taille_dcl duree_dcl
		{	$$.T_Size = $1;
		}
		;

taille_dcl :	SIZE AFFECT INTGR
		{	$$ = $3;
		}
		;
		
duree_dcl :	TMAX AFFECT INTGR
		{	$$ = $3;
			if ( Debug ) {
		  	  printf("Warning: obsolete duration declaration.\n");
			}
		}
		|
		{       $$ = 1000;
		}
 		;

liste_transitions : 
		transition_sons
		{	$$ = Append_Transition( NULL, $1 );
		}
		| liste_transitions transition_sons
		{	$$ = Append_Transition( $1, $2 );
		}
		;


transition_sons:
		transition_dcl SEP place_list ENDINSTR
		{ 	$$ = Add_Sons( $1, $3 );
		}
		| transition_dcl ENDINSTR
		{ 	$$ = $1;
		}
		;

transition_dcl:
		trans_id loi_dcl
		{ 	$$ = Parse_Transition( $1, $2, (char*) NULL,
						     (char*) NULL );
			if ( Debug )
				printf("Transition %d recognized\n",$1);
		}
		| trans_id name loi_dcl
		{ 	$$ = Parse_Transition( $1, $3, $2,
						     (char*) NULL );
			if ( Debug )
				printf("Transition %s [%d] recognized\n",$2, $1);
		}
		| trans_id name name loi_dcl
		{ 	$$ = Parse_Transition( $1, $4, $2, $3 );
			if ( Debug )
				printf("Transition %s [%d][%s] recognized\n",
					$2, $1, $3);
		}

place_list :	place_dcl
		{	$$ = Append_Place( NULL, $1 );
		}
		| place_list SEP place_dcl 	
		{	$$ = Append_Place( $1, $3 );
		}
		;

place_dcl :	trans_id loi_dcl pop_dcl
		{	$$ = Parse_Place( $1, $2, $3, (char*) NULL,
					(char*) NULL, (char*) NULL );
		}
		| trans_id name loi_dcl pop_dcl
		{	$$ = Parse_Place( $1, $3, $4, $2,
					(char*) NULL, (char*) NULL );
		}
		| trans_id name name loi_dcl pop_dcl
		{	$$ = Parse_Place( $1, $4, $5, $2,
					$3, (char*) NULL );
		}
		| trans_id name name name loi_dcl pop_dcl
		{	$$ = Parse_Place( $1, $5, $6, $2, $3, $4 );
		}
		;

trans_id :	INTGR
		;

name :		STRING
		{	$$ = $1;
		}

loi_dcl : 	EXPO real_list
		{	$$ = Parse_Law( 'E', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| CST real_list
		{	$$ = Parse_Law( 'C', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| BERN real_list
		{	$$ = Parse_Law( 'B', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| GEOM0 real_list
		{	$$ = Parse_Law( 'g', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| GEOM1 real_list
		{	$$ = Parse_Law( 'G', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| ERLG INTGR real_list
		{	$$ = Parse_Law( 'R', Append_Real( $3, (double)$2 ),
					(Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| UNIF real_list
		{	$$ = Parse_Law( 'U', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| PARETO real_list
		{	$$ = Parse_Law( 'p', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| WEIBULL real_list
		{	$$ = Parse_Law( 'W', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| LOGNO real_list
		{	$$ = Parse_Law( 'L', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| DISC PARB real_list PARE PARB real_list PARE
		{	$$ = Parse_Law( 'D', $3, $6,
					(Liste_Loi)NULL );
		}
		| HEXP real_list
		{	$$ = Parse_Law( 'H', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| ONOFF real_list
		{	$$ = Parse_Law( 'O', $2, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| IPP PARB real_list SEP law_list PARE
		{	$$ = Parse_Law( 'P', $3, (Liste_Reel)NULL, $5 );
		}
		;

pop_dcl :	INTGR real_list
		{	$$ = Parse_Init_Pop( $1, $2 );
		}
		| INTGR 
		{	$$ = Parse_Init_Pop( $1, (Liste_Reel)NULL );
		}
		;

real_list :	REAL
		{	$$ = Append_Real( (Liste_Reel)NULL, $1 );
		}
		| real_list REAL
		{	$$ = Append_Real( $1, $2 );
		}
		;

law_list :	loi_dcl
		{	$$ = Append_Law( (Liste_Loi)NULL, $1 );
		}
		| law_list SEP loi_dcl
		{	$$ = Append_Law( $1, $3 );
		}
		;

%%
#include "egl_lex.yy.c"
