/****************************************************************/
/* Partie relative a lex/yacc					*/
/****************************************************************/

#include "eg.h"
#include "egl.tab.h"

#define MAX(x,y) ( x > y ? x : y )

#define	E_MEM(x) fprintf(stderr,"Warning: not enough memory: %s not created\n",x);
#define W_EXT(x) fprintf(stderr,"Warning: extra parameters for %c (ignored)\n", x );
#define W_MIS(x) fprintf(stderr,"Warning: missing parameters for %c. Default 0.0.\n", x );

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

extern FILE *yyin;
extern int yydebug;
extern char yytext[];
#ifdef YY_FLEX_LEX_COMPAT
extern int yylineno;
#endif
extern int yyparse();
extern int yylex();
extern int EGL_lineno;

/* forced external declaration */
extern char *strdup(const char *);

/* forward declaration */
void dumpe_network(EventGraph *);
void Free_Transition_List(Liste_Transition);

boolean	Debug = FALSE;
#ifdef USE_YYLVAL
extern  YYSTYPE yylval;
#define PARSE_RESULT yylval
#else 
EventGraph *The_Event_Graph;
#define PARSE_RESULT The_Event_Graph
#endif

EventGraph *Read_Network(char *File_Name)
{
  FILE	*Input_File;
  FILE	*Current_Input;
  int	Res;

  if ( strlen( File_Name ) > 0 ) {
    Input_File = fopen( File_Name, "r" );
  }
  else {
    Input_File = stdin;
  }

  if ( NULL != Input_File ) {

    Current_Input = yyin;
    yyin = Input_File;

    yydebug = 0;
    Res = yyparse();

    yyin = Current_Input;

    fclose( Input_File );

    if ( Res == 0 )
      return PARSE_RESULT;
    else
      return (EventGraph*)NULL;
  }
  else {
    printf("Error: could not open file %s for reading.\n", File_Name );

    return (EventGraph*)NULL;
  }
}

void yyerror(char *s)
{
#ifdef YY_LEX_FLEX_COMPAT
  printf("---%s---[%s] at line %d\n",s,yytext,yylineno);
#else
  printf("---%s---[%s] at line %d\n",s,yytext,EGL_lineno);
#endif
}

/****************************************************************/
/* Procedures de construction du reseau				*/
/****************************************************************/
EventGraph *New_Event_Graph( int The_Size )
{
  EventGraph	*RdP;
  
  RdP = (EventGraph*) malloc( sizeof ( EventGraph ) );
  if (RdP != NULL ) {

    RdP->Size = The_Size;
    RdP->Nb_Places = 0;
    RdP->P = (Place_Desc*) NULL;
    RdP->Nb_Layers = 0;
    RdP->Level = (Layer_Desc*) NULL;
    RdP->Max_Pop_Init = 0;

    RdP->T = (Trans_Desc*) calloc( (size_t)RdP->Size, sizeof( Trans_Desc ) );
    if ( (Trans_Desc*)NULL == RdP->T ) {
      free( RdP );
      RdP = (EventGraph*) NULL;
    }
  }
  else {
    E_MEM("Event Graph");
  }

  return RdP;
}

Trans_Desc New_Transition(void)
{
  Trans_Desc	The_Trans;

  The_Trans = (Trans_Desc) malloc( sizeof( struct TRANS_DESC ) );

  if ( The_Trans != NULL ) {
    The_Trans->Trans_Id = -1;
    The_Trans->Trans_Law = (Law_Desc)NULL;
    The_Trans->Name = (char*)NULL;
    The_Trans->Data = (char*)NULL;
    The_Trans->Outdegree = 0;
    The_Trans->Outputs = (Liste_Place)NULL;
    The_Trans->Inputs = (Liste_Place)NULL;
  }
  else {
    E_MEM("Transition");
  }

  return The_Trans;
}

Place_Desc New_Place(void)
{
  Place_Desc	The_Place;

  The_Place = (Place_Desc) malloc( sizeof ( struct PLACE_DESC ) );
  The_Place->Place_Id = -1; /* unknown at this stage */
  The_Place->Trans_Id = -1; /* unknown at this stage */
  The_Place->Input_Id = -1; /* unknown at this stage */
  The_Place->Name = (char*)NULL;
  The_Place->Data1 = (char*)NULL;
  The_Place->Data2 = (char*)NULL;
  The_Place->Place_Law = (Law_Desc)NULL;
  The_Place->Init_Pop = 0;
  The_Place->Lag_Times = (double*)NULL;

  return The_Place;

}

EventGraph *Parse_EG(int The_Size, 
		     Liste_Transition The_Transitions )
{
  EventGraph		*RdP;
  int			idx;
  Liste_Transition 	Scan;
  Liste_Transition 	T_Scan;
  Liste_Place	 	P_Scan;
  Trans_Desc	 	The_Trans;

  RdP = New_Event_Graph( The_Size );

  /* scanner la liste des trans. et la mettre dans le tableau	*/
  /* pour chacune, convertir la liste des voisins en tableau	*/
  for (	Scan = The_Transitions; Scan != NULL ; Scan = Scan->Next ) {
    if ( Scan->Val->Trans_Id <= The_Size ) {
      The_Trans = Scan->Val;

      /* Scan the list of neighbors and compute max init pop. */
      SCAN( P_Scan, The_Trans->Outputs ) {
	if ( P_Scan->Val->Trans_Id <= The_Size ) {
	  RdP->Max_Pop_Init = MAX( P_Scan->Val->Init_Pop, RdP->Max_Pop_Init );
	  /* set also the Origin field in place structure */
	  P_Scan->Val->Input_Id = Scan->Val->Trans_Id;
	}
	else {
	  fprintf(stderr,"Number of Transition (%d) > Network Size (%d).",
		  P_Scan->Val->Trans_Id, 
		  The_Size);
	  fprintf(stderr," Place '%d->%d' ignored.\n",
		  Scan->Val->Trans_Id,
		  P_Scan->Val->Trans_Id );
	}
      }

      /* Movind the transition into the T table. */
      RdP->T[ Scan->Val->Trans_Id-1] = The_Trans;
      RdP->Nb_Places += RdP->T[ Scan->Val->Trans_Id-1 ]->Outdegree;

    }
    else {
      fprintf(stderr,"Number of Transition (%d) > Network Size (%d).",
	      Scan->Val->Trans_Id, The_Size );
      fprintf(stderr," Transition '%d' ignored.\n",
	      Scan->Val->Trans_Id );
    }
  }

  /* Now make a second pass to store the adresses of places in the	*/
  /* "P" table.							*/
  RdP->P = (Place_Desc*) calloc( (size_t)RdP->Nb_Places, 
				 sizeof( Place_Desc ) );

  idx = 0;
  SCAN( T_Scan, The_Transitions ) {
    if ( T_Scan->Val->Trans_Id <= The_Size ) {
      SCAN( P_Scan, T_Scan->Val->Outputs ) {
	if ( P_Scan->Val->Trans_Id <= The_Size ) {
	  RdP->P[idx] = P_Scan->Val;
	  P_Scan->Val->Place_Id = idx;
	  idx++;
	}
      }
    }
  }

  Free_Transition_List( The_Transitions );

  return RdP;

}


Trans_Desc Parse_Transition(int		Trans_Num,
			    Law_Desc	The_Law,
			    char	*The_Name,
			    char	*The_Data )
{
  Trans_Desc	The_Trans;

  The_Trans = New_Transition();

  if ( The_Trans != NULL ) {
    The_Trans->Trans_Id = Trans_Num;
    The_Trans->Trans_Law = The_Law;
    The_Trans->Name = The_Name;
    The_Trans->Data = The_Data;
  }

  return The_Trans;
}

Trans_Desc Add_Sons(Trans_Desc	The_Trans,
		    Liste_Place	The_Outputs )
{	
  Liste_Place	Scan;

  if ( The_Trans != NULL ) {

    The_Trans->Outputs = The_Outputs;

    for ( Scan = The_Outputs; Scan != NULL; Scan=Scan->Next )
      The_Trans->Outdegree++;

  }
  else {
    fprintf(stderr, "Cannot add place list to null transition.\n");
  }

  return The_Trans;
}

Place_Desc Parse_Place( 
int		Num_Transition,
Law_Desc	Loi,
Population	*The_Pop,
char		*The_Name, 
char		*The_Data1, 
char		*The_Data2 )
{
  Place_Desc	Resultat;

  Resultat = New_Place();
  Resultat->Trans_Id = Num_Transition;
  Resultat->Name = The_Name;
  Resultat->Data1 = The_Data1;
  Resultat->Data2 = The_Data2;
  Resultat->Place_Law = Loi;
  Resultat->Init_Pop = The_Pop->Pop;
  Resultat->Lag_Times = The_Pop->Delais;

  /* Note: At this stage, Place_Id and Input_Id are still unknown */

  free( The_Pop );

  return Resultat;
}

Liste_Transition Append_Transition(Liste_Transition Trans_List,
				   Trans_Desc	New_Trans )
{
  Liste_Transition The_List;

  The_List = (Liste_Transition) malloc( sizeof( struct TRANS_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Trans;
    The_List->Next = Trans_List;
  }
  else {
    E_MEM("Transition");
  }

  return( The_List );
}

Liste_Place Append_Place( Liste_Place Place_List,
			  Place_Desc  New_Place )
{
  Liste_Place The_List;

  The_List = (Liste_Place) malloc( sizeof( struct PLACE_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Place;
    The_List->Next = Place_List;
  }
  else {
    E_MEM("Place");
  }

  return The_List;
}

Population *Parse_Init_Pop( int		Nb_Tokens,
			    Liste_Reel	Lags )
{
  Population  *The_Pop;
  int		Nb;
  int		Nb_Reals;
  Liste_Reel  Scan;

  The_Pop = (Population *) malloc(sizeof(struct POPULATION));

  if (The_Pop != NULL) {
    The_Pop->Pop = Nb_Tokens;
    if ( Nb_Tokens > 0 ) {
      The_Pop->Delais = (double*) calloc(Nb_Tokens, sizeof(double));
    }
    else {
      The_Pop->Delais = (double*) NULL;
    }

    if ( ( Nb_Tokens != 0 ) && (The_Pop->Delais != NULL) ) {
      Nb = 0;
      Nb_Reals = 0;
      /* count the number of reals in the list */
      for ( Scan = Lags; Scan != NULL; Scan=Scan->Next )
	Nb_Reals++;
      for (Scan = Lags;
	   (Scan != NULL) && (Nb < Nb_Tokens); Scan = Scan->Next) {
	if ( Nb_Reals-1-Nb < Nb_Tokens )
	  The_Pop->Delais[Nb_Reals-1-Nb] = Scan->Val;
	Nb++;
      }
      if (Nb < Nb_Tokens)
	for (; Nb < Nb_Tokens; Nb++) {
	  The_Pop->Delais[Nb] = 0.0;
	  fprintf(stderr,
		  "Warning: not enough initial lags (0.0 assumed)\n");
	}
      else if (Scan != NULL) {
	fprintf(stderr,
		"Warning: too many initial lags (ignored)\n");
      }
    }
    else if ( Nb_Tokens != 0 ) {
      E_MEM("real");
    }
  }
  else {
    E_MEM("population");
  }

  Free_Real_List( Lags );

  return The_Pop;
}

/****************************************************************/
/* Procedure de Trace						*/
/****************************************************************/
void dumpe_network( EventGraph *RdP )
{
  int		i;
  int		j;
  Liste_Place	P_Scan;
  char		*Info;
  char		Pop_S[32];

  Info = (char*) calloc( (size_t)RdP->Size, sizeof(char) );

  printf("Network of size %d transitions and %d places\n",
	 RdP->Size, RdP->Nb_Places );

  printf("      ");
  for (i=0; i<2*RdP->Size+3; i++)
    printf(".");
  printf("\n");
	
  for (i=0; i<RdP->Size; i++) {

    printf("%3d[%c].",i+1, RdP->T[i]->Trans_Law->Name );
    
    for (j=0; j<RdP->Size; j++) 
      Info[i] = ' ';
    SCAN( P_Scan, RdP->T[i]->Outputs ) 
      Info[P_Scan->Val->Trans_Id-1] = P_Scan->Val->Place_Law->Name;

    for (j=0; j<RdP->Size; j++)
      printf("%c", Info[i] );
    printf(".");

    SCAN( P_Scan, RdP->T[i]->Outputs ) {
      if ( P_Scan->Val->Init_Pop > 9 ) {
	Info[P_Scan->Val->Trans_Id-1] = '+';
      }
      else {
	sprintf( Pop_S, "%d1", P_Scan->Val->Init_Pop );
	Info[P_Scan->Val->Trans_Id-1] = Pop_S[0];
      }
    }
    
    for (j=0; j<RdP->Size; j++)
      printf("%c", Info[i] );
    printf(".\n");
  }

  printf("      ");
  for (i=0; i<2*RdP->Size+3; i++)
    printf(".");
  printf("\n\n");

  printf("Maximal initial population: %d\n\n",RdP->Max_Pop_Init);

  for (i=0; i<RdP->Nb_Layers; i++) {
    printf("Level %d : ",i+1);
    for (j=0; j<RdP->Level[i]->Nb_Nodes; j++)
      printf(" %d",RdP->Level[i]->Node_Id[j]+1);
    printf("\n");
  }
  printf("\n");

  for (i=0; i<RdP->Size; i++) {
    printf("Transitions input of %d: ",i+1);
    SCAN( P_Scan, RdP->T[i]->Inputs ) {
      printf(" %d", P_Scan->Val->Input_Id );
    }
    printf("\n");
  }

  free( Info );
}

/****************************************************************/
/* Cleanup procedures						*/
/****************************************************************/
void Free_Place_List( Liste_Place L )
{
  Liste_Place	M;
  Liste_Place	N;

  for ( M = L; M != NULL; ) {
    N = M;
    M = N->Next;
    free( N );
  }
  
}

void Free_Transition_List( Liste_Transition L )
{
  Liste_Transition	M;
  Liste_Transition	N;

  for ( M = L; M != NULL; ) {
    N = M;
    M = N->Next;
    free( N );
  }
  
}

void Free_Transition( Trans_Desc T ) 
{
    Free_Law( T->Trans_Law );
    Free_Place_List( T->Outputs );
    if ( (Liste_Place)NULL != T->Inputs )
      Free_Place_List( T->Inputs );
    if ( (char*)NULL != T->Name )
      free( T->Name );
    if ( (char*)NULL != T->Data )
      free( T->Data );
    free( T );
}

void Free_Event_Graph( EventGraph* RdP ) 
{
  int		i;
  
  for ( i=0; i<RdP->Size; i++ ) {
    Free_Transition( RdP->T[i] );
  }
  for ( i=0; i<RdP->Nb_Places; i++ ) {
    Free_Law( RdP->P[i]->Place_Law );
    if ( RdP->P[i]->Init_Pop > 0 )
      free( RdP->P[i]->Lag_Times );
    if ( (char*)NULL != RdP->P[i]->Name )
      free( RdP->P[i]->Name );
    if ( (char*)NULL != RdP->P[i]->Data1 )
      free( RdP->P[i]->Data1 );
    if ( (char*)NULL != RdP->P[i]->Data2 )
      free( RdP->P[i]->Data2 );
    free( RdP->P[i] );
  }

  free( RdP->T );
  free( RdP->P );
  for ( i=0; i<RdP->Nb_Layers; i++ ) {
    free( RdP->Level[i]->Node_Id );
    free( RdP->Level[i] );
  }
  if ( (Layer_Desc*)NULL != RdP->Level )
    free( RdP->Level );

  free( RdP );
}
