/* modules pour generer des RdP reguliers dans le format PNL.	*/

#include "pn.h"

#define TOOLNAME	"gen_fc"

#define	Empty		0
#define	Full		1
#define First_Place	2
#define Even_Places	3
#define Diagonal	4
#define Checker		5
#define Column		6

typedef int Init_Type;

#define	Ser_Open	0
#define Ser_Closed	1
#define	Job_Shop	2
#define	Double_Loop	3

typedef int Net_Type;

#define Illimited	-1		/* for capacity purposes	*/

void fprint_init_mark(FILE *OutFile, int Number)
{	
  int j;

  fprintf(OutFile, "%d ",Number);
  for (j=0; j<Number; j++)
    fprintf(OutFile, "0.0 ");
}

void gen_series_open(
FILE		*OutFile,
int		Net_Size,
Law_Desc	The_Trans_Law, 
Law_Desc	The_Place_Law,
Init_Type	Init_Mode,
int		Marking,
int		Sim_Length )
{
  int	i;

  fprintf(OutFile, "Transitions <- %d\n", Net_Size );
  fprintf(OutFile, "Places <- %d\n", Net_Size-1 );

  for (i=0; i<Net_Size; i++) {

    /* generation de la transition */
    fprintf(OutFile, "T %d ", i+1);
    fprintf(OutFile, "\"Trans%d\" ", i );
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );

    if ( i < Net_Size-1 )
      fprintf(OutFile, ", %d ;\n", i+1); /* i.e. i-1+1 */
    else
      fprintf( OutFile, ";\n");

    /* generation de la place */
    if ( i < Net_Size-1 ) {
      fprintf( OutFile, "P %d ", i+1 );
      fprintf( OutFile, "\"Place%d\" ", i );
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );

      switch(Init_Mode) {
      case Empty:
	fprint_init_mark( OutFile, 0 );
	break;
      case Full:
	fprint_init_mark( OutFile, Marking );
	break;
      case First_Place:
      case Diagonal:
      case Column:
	if (i==0)
	  fprint_init_mark( OutFile, Marking);
	else	
	  fprint_init_mark( OutFile, 0 );
	break;
      case Even_Places:
      case Checker:
	if (i % 2 == 0) 
	  fprint_init_mark( OutFile, Marking);
	else	
	  fprint_init_mark( OutFile, 0 );
	break;
      }

      /* transition en sortie */
      fprintf( OutFile, ", %d;\n", ( i == Net_Size-1 ? 1 : i+2 ) );
    }
  }
}

void gen_series_closed(
FILE		*OutFile,
int		Net_Size,
Law_Desc	The_Trans_Law, 
Law_Desc	The_Place_Law,
Init_Type	Init_Mode,
int		Marking,
int		Sim_Length)
{
  int	i;

  fprintf(OutFile, "Transitions <- %d\n", Net_Size );
  fprintf(OutFile, "Places <- %d\n", Net_Size );

  for (i=0; i<Net_Size; i++) {

    /* generation de la transition */
    fprintf(OutFile, "T %d ", i+1);
    fprintf(OutFile, "\"Trans%d\" ", i );
    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );
    fprintf(OutFile, ", %d ;\n", i+1); /* i.e. i-1+1 */

    /* generation de la place */
    fprintf( OutFile, "P %d ", i+1 );
    fprintf(OutFile, "\"Place%d\" ", i );
    Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );
	
    switch(Init_Mode) {
    case Empty:
      fprint_init_mark( OutFile, 0 );
      break;
    case Full:
      fprint_init_mark( OutFile, Marking );
      break;
    case First_Place:
    case Diagonal:
    case Column:
      if (i==0)
	fprint_init_mark( OutFile, Marking );
      else
	fprint_init_mark( OutFile, 0 );
      break;
    case Even_Places:
    case Checker:
      if (i % 2 == 0)
	fprint_init_mark( OutFile, Marking );
      else
	fprint_init_mark( OutFile, 0 );
      break;
    }

    /* transition en sortie */
    fprintf( OutFile, ", %d;\n", ( i == Net_Size-1 ? 1 : i+2 ) );

  }
}

void fprint_trans_js(
FILE*		OutFile,
int		Destination,
Law_Desc	The_Law,
char		*Dir,
int		Orig,
int		Dest )
{	
  fprintf(OutFile, "T %d \"Trans%s_%d_%d\" ", Destination, 
	  Dir, Orig, Dest );
  Write_Law(OutFile, The_Law, NORMAL_PRINT_MODE );

  /* place en sortie */
  fprintf( OutFile, ", %d ;\n", Dest );
}

void gen_job_shop(
FILE		*OutFile,
int		Size_Machine, 
int		Size_Piece,
Law_Desc	The_Trans_Law, 
Law_Desc	The_Place_Law,
Init_Type	Init_Mode,
int		Marking,
int		Sim_Length )
{	
  int	i;
  int	j;
  int	Place_Id;

  fprintf(OutFile, "Transitions <- %d\n", 2*Size_Machine*Size_Piece);
  fprintf(OutFile, "Places <- %d\n", Size_Machine*Size_Piece);

  for (i=0; i<Size_Machine; i++) {
    for (j=0; j<Size_Piece; j++) {
      Place_Id = i*Size_Piece+j+1;
      fprintf(OutFile, "P %d \"Place_%d_%d\" ", Place_Id, i+1, j+1 );
      Write_Law(OutFile, The_Place_Law, NORMAL_PRINT_MODE );

      switch(Init_Mode) {
      case Empty:
	fprintf(OutFile, "0");
	break;
      case Full:
	fprint_init_mark( OutFile, Marking );
	break;
      case First_Place:
	if ( (i==0) && (j==0) )
	  fprint_init_mark( OutFile, Marking );
	else	
	  fprintf(OutFile, "0");
	break;
      case Column:
	if ( j==0 )
	  fprint_init_mark( OutFile, Marking );
	else	
	  fprintf(OutFile, "0");
	break;
      case Diagonal:
	if ( i == j )
	  fprint_init_mark( OutFile, Marking );
	else
	  fprintf(OutFile, "0");
	break;
      case Even_Places:
      case Checker:
	if ( (i+j) % 2 == 0 )
	  fprint_init_mark( OutFile, Marking );
	else	
	  fprintf(OutFile, "0");
	break;
      }

      /* routage */
      fprintf( OutFile, " Random[ 0.5 0.5 ]" );

      /* transitions en sortie */
      fprintf(OutFile, ", %d, %d ;\n",
	      2*i*Size_Piece + j + 1,
	      (2*i+1)*Size_Piece + j + 1 );

      /* the next machine place */
      fprint_trans_js(OutFile,
		      2*i*Size_Piece + j + 1, 
		      The_Trans_Law, "H", 
		      Place_Id, 1 + ( j < Size_Piece-1 ? Place_Id :
				      i*Size_Piece ) );

      /* the next piece place */
      fprint_trans_js(OutFile, 
		      (2*i+1)*Size_Piece + j + 1,
		      The_Trans_Law, "V", 
		      Place_Id, ( i < Size_Machine-1 ? 
				  Place_Id + Size_Piece : 
				  j + 1 ) );
    }
  }
}
			
void fprint_place_dl(
FILE*		OutFile,
int		Destination,
Law_Desc	The_Law,
Init_Type	Init_Mode,
int		Marking,
int		Size1, 
int		Size2 )
{
  int dest;

  /* generation des transitions de la premiere boucle */
  fprintf(OutFile, "P %d ", Destination+1 );

  if ( Destination < Size1 )
    fprintf( OutFile, "\"Pl_1_%d\" ", Destination );
  else
    fprintf( OutFile, "\"Pl_2_%d\" ", Destination-Size1 );

  Write_Law(OutFile, The_Law, NORMAL_PRINT_MODE );

  switch(Init_Mode) {
  case Empty:
    fprint_init_mark( OutFile, 0 );
    break;
  case Diagonal:
  case First_Place:
  case Column:
    if ( (Destination==0) || (Destination==Size1) )
      fprint_init_mark( OutFile, Marking );
    else
      fprint_init_mark( OutFile, 0 );
    break;
  case Even_Places:
  case Checker:
    if ( Destination % 2 == 0 )
      fprint_init_mark( OutFile, Marking );
    else
      fprint_init_mark( OutFile, 0 );
    break;
  case Full:
  default:
    fprint_init_mark( OutFile, Marking );
  }

  /* routage (quand necessaire) */
  if ( Destination == 0 )
    fprintf( OutFile, "Random[ 0.5 0.5 ], %d", Size1 + Size2 + 1 );

  /* transition destination */
  if ( Destination < Size1 - 1 )
    dest = Destination+1;
  else if ( Destination == Size1 - 1 )
    dest = 0;
  else if ( Destination == Size1 + Size2 - 1 )
    dest = Size1;
  else	dest = Destination+1;

  fprintf( OutFile, ", %d;\n", dest+1 );
	
}
			
void gen_double_loop(
FILE		*OutFile,
int		Size_Loop1, 
int		Size_Loop2,
Law_Desc	The_Trans_Law, 	
Law_Desc	The_Place_Law,
Init_Type	Init_Mode,
int		Marking,
int		Sim_Length )
{
  int	i;

  fprintf(OutFile, "Transitions <- %d\n", Size_Loop1 + Size_Loop2 + 1);
  fprintf(OutFile, "Places <- %d\n", Size_Loop1 + Size_Loop2 );

  for (i=0; i< Size_Loop1 + Size_Loop2; i++) {

    /* generation de la transition */
    fprintf(OutFile, "T %d ", i+1 );

    if ( i < Size_Loop1 )
      fprintf( OutFile, "\"Tr_1_%d\" ", i);
    else
      fprintf( OutFile, "\"Tr_2_%d\" ", i-Size_Loop1 );

    Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );
    fprintf( OutFile, ", %d;\n", i+1 );

    /* generation de la place */
    fprint_place_dl(OutFile, i, The_Place_Law, Init_Mode, Marking,
		    Size_Loop1, Size_Loop2 );

  }
  /* transition supplementaire */
  fprintf(OutFile, "T %d ", Size_Loop1 + Size_Loop2 + 1 );
  fprintf( OutFile, "\"Tr_1_to_2\" ");
  Write_Law(OutFile, The_Trans_Law, NORMAL_PRINT_MODE );
  fprintf( OutFile, ", %d;\n", Size_Loop1 + 1 );

}

void display_help( )
{
  char	Format[100];
	
  sprintf(Format, "%%%dc [-tl <law>] [-pl <law>] [-o <file>] [-h]\n",
	  (int)strlen(TOOLNAME)+7 );

  printf("Usage: %s [-t <type>] [-s <size(s)>] [-i <init>]", TOOLNAME);
  printf(" [-l <sim length>] [-m <int>]\n");
  printf(Format,' ');
  printf("  <type> : ser_closed (def) ser_open double job_shop\n");
  printf("  <init> : full (def) empty first even diagonal checker");
  printf(" column\n");
  printf("  <law>  : E (def) C B G R U + parameters (def: 1.0)\n");
}

int main(int argc, char *argv[] )
{	
  Net_Type	RdP_Type = Ser_Closed;
  Init_Type	Init_Mode = Full;
  static char	File_Name[128] = "";
  int		Net_Size = 2;
  int		Net_Size2 = 2;
  int		Marking = 1;
  int		Sim_Length = 1000;
  Law_Desc	The_Trans_Law, The_Place_Law;
  int		i;
  FILE*		OutFile;
  boolean	Args_Ok = TRUE;
  boolean	Help = FALSE;

  The_Trans_Law = (Law_Desc) malloc( sizeof( struct LAW_DESC ) );
  The_Place_Law = (Law_Desc) malloc( sizeof( struct LAW_DESC ) );
  The_Place_Law->Name = The_Trans_Law->Name = 'E';
  The_Place_Law->Parameters = (double*) malloc( sizeof( double ) );
  The_Trans_Law->Parameters = (double*) malloc( sizeof( double ) );
  The_Place_Law->Parameters[0] = 1.0;
  The_Trans_Law->Parameters[0] = 1.0;

  i = 1;
  if (argc>1)
    do if ( 0 == strcmp( "-t", argv[i] ) ) {
      i++;
      if ( 0 == strcmp( "ser_open", argv[i] ) )
	RdP_Type = Ser_Open;
      else if ( 0 == strcmp( "ser_closed", argv[i] ) )
	RdP_Type = Ser_Closed;
      else if ( 0 == strcmp( "job_shop", argv[i] ) )
	RdP_Type = Job_Shop;
      else if ( 0 == strcmp( "double", argv[i] ) )
	RdP_Type = Double_Loop;
      else printf("%s: unrecognized network RdP_Type\n",argv[0]);
      i++;
    }
    else if ( 0 == strcmp( "-s", argv[i] ) )
      {	i++;
      switch( RdP_Type ) {
      case Ser_Open:
      case Ser_Closed:
	if ( 1 != sscanf( argv[i], "%d", &Net_Size ) )
	  printf("%s: size expected\n",argv[0]);
	else	i++;
	break;
      case Job_Shop:
      case Double_Loop:
	if ( 
	    ( 1 != sscanf( argv[i], "%d", &Net_Size ) )
	    || ( 1 != sscanf( argv[i+1], "%d", &Net_Size2 ) ) )
	  printf("%s: two sizes expected\n",argv[0]);
	else { i++; i++;};
	break;
      }
      }
    else if ( 0 == strcmp( "-m", argv[i] ) )
      {	i++;
      if ( 1 != sscanf( argv[i], "%d", &Marking ) )
	printf("%s: integer marking expected\n",argv[0]);
      else	i++;
      }
    else if ( 0 == strcmp( "-i", argv[i] ) )
      {	i++;
      if ( 0 == strcmp( "empty", argv[i] ) )
	Init_Mode = Empty;
      else if ( 0 == strcmp( "full", argv[i] ) )
	Init_Mode = Full;
      else if ( 0 == strcmp( "first", argv[i] ) )
	Init_Mode = First_Place;
      else if ( 0 == strcmp( "even", argv[i] ) )
	Init_Mode = Even_Places;
      else if ( 0 == strcmp( "diagonal", argv[i] ) )
	Init_Mode = Diagonal;
      else if ( 0 == strcmp( "checker", argv[i] ) )
	Init_Mode = Checker;
      else if ( 0 == strcmp( "column", argv[i] ) )
	Init_Mode = Column;
      else printf("%s: unrecognized init mode\n", TOOLNAME);
      }
    else if ( 0 == strcmp( "-l", argv[i] ) )
      {	i++;
      if ( i<argc )
	sscanf( argv[i], "%d", &Sim_Length );
      else	printf("%s: simulation length expected\n", TOOLNAME);
      }
    else if ( 0 == strcmp( "-tl", argv[i] ) )
      {	i++;
      switch( The_Trans_Law->Name = argv[i][0] ) {
      case 'C':
      case 'E':
      case 'B':
      case 'G':
	i++;
	sscanf( argv[i], "%lf", 
		&(The_Trans_Law->Parameters[0] ) );
	break;
      case 'R':
      case 'U':
	i++;
	The_Trans_Law->Parameters = (double*) 
	  realloc ( 
		   The_Trans_Law->Parameters, 
		   2*sizeof(double) );
	sscanf( argv[i], "%lf", 
		&(The_Trans_Law->Parameters[0] ) );
	i++;
	sscanf( argv[i], "%lf", 
		&(The_Trans_Law->Parameters[1] ) );
	break;
      default:
	printf("%s: unknown law\n", TOOLNAME );
      }
      i++;
      }
    else if ( 0 == strcmp( "-pl", argv[i] ) )
      {	i++;
      switch( The_Place_Law->Name = argv[i][0] ) {
      case 'C':
      case 'E':
      case 'B':
      case 'G':
	i++;
	sscanf( argv[i], "%lf", 
		&(The_Place_Law->Parameters[0] ) );
	break;
      case 'R':
      case 'U':
	i++;
	The_Place_Law->Parameters = (double*) 
	  realloc ( 
		   The_Place_Law->Parameters, 
		   2*sizeof(double) );
	sscanf( argv[i], "%lf", 
		&(The_Place_Law->Parameters[0] ) );
	i++;
	sscanf( argv[i], "%lf", 
		&(The_Place_Law->Parameters[1] ) );
	break;
      default:
	printf("%s: unknown law\n", TOOLNAME);
      }
      i++;
      }
    else if ( 0 == strcmp( "-o", argv[i] ) ) {
      i++;
      if ( i < argc )
	strcpy( File_Name, argv[i] );
      else	printf("%s: filename expected\n", TOOLNAME);
      i++;
    }
    else if ( ( 0 == strcmp( "-h", argv[i] ) )
	      || ( 0 == strcmp( "-?", argv[i] ) ) ) {
      i++;
      display_help(argv[0]);
      Args_Ok = FALSE;
      Help = TRUE;
    }
    else if ( 0 == strcmp( "-v", argv[i] ) ) {
      i++;
      printf("%s (Version %s)\n",argv[0],VERSION);
      Args_Ok = FALSE;
    }
    else 
      i++;	/* skip argument */
    while ( ( i < argc-1 ) && Args_Ok );

  if (Args_Ok) {

    if ( strlen(File_Name) != 0 )
      OutFile = fopen( File_Name, "w" );
    else	OutFile = stdout;

    switch( RdP_Type ) {
    case Ser_Open:
      gen_series_open( OutFile, Net_Size, 
		       The_Trans_Law, The_Place_Law, Init_Mode,
		       Marking, Sim_Length );
      break;
    case Ser_Closed:
      gen_series_closed( OutFile, Net_Size, 
			 The_Trans_Law, The_Place_Law, Init_Mode,
			 Marking, Sim_Length );
      break;
    case Job_Shop:
      gen_job_shop( OutFile, Net_Size, Net_Size2,
		    The_Trans_Law, The_Place_Law, Init_Mode,
		    Marking, Sim_Length );
      break;
    case Double_Loop:
      gen_double_loop( OutFile, Net_Size, Net_Size2,
		       The_Trans_Law, The_Place_Law, Init_Mode,
		       Marking, Sim_Length );
      break;
    }
    return 0;
  }
  else {
    if ( Help )
      return 0;
    else
      return -1;
  }
}
