/* programme/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un graphe d'interaction.		*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "boolean.h"
#include "parseargs.h"
#include "maxplusmat.h"
#include "bernoulli.h"

#define	TOOLNAME	"mat2c"
#ifdef USE_FULL_MATRICES
#define VERSION		"1.1 (full matrices)"
#else
#define VERSION		"1.1 (sparse matrices)"
#endif

#define CODE(x)		printf(x);
#define CODE1(x,y)	printf(x,y);
#define CODE2(x,y,z)	printf(x,y,z);
#define CODE3(x,y,z,t)	printf(x,y,z,t);
#define CODELN		CODE("\n")

#define MAX_NAME_LENGTH	256

/* Global declarations */
extern char	*todays_date(void);

void sort_probas( Bernoulli_Product Problem );
void gen_c_code( Bernoulli_Product Problem,
		 char File_Name[MAX_NAME_LENGTH], int Sim_Length,
		 int Nb_Replicates, boolean Gradient );

boolean	Debug = FALSE;

int main(int argc, char *argv[] )
{
  static char		In_File_Name[MAX_NAME_LENGTH] = "";
  static char		Out_File_Name[MAX_NAME_LENGTH] = "";
  int			Sim_Length = 100000;
  int			Nb_Replicates = 1;
  int			i;
  boolean		Gradient = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  Bernoulli_Product	Problem;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )

      else if GET_STRING( "-o", Out_File_Name )

      else if GET_INT( "-nevt", Sim_Length )
      else if GET_INT( "-nrep", Nb_Replicates )
      else if SWITCH( "-gradient", Gradient, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", TOOLNAME );
	printf(" [-nevt <int>] [-nrep <int>] [-gradient]\n\t[-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    if ( strlen(In_File_Name) != 0 )
      freopen( In_File_Name, "r", stdin );

    Problem = Read_Bernoulli_Product( stdin );

    if ( (Bernoulli_Product)NULL != Problem ) {
      sort_probas( Problem );
      gen_c_code( Problem, Out_File_Name, 
		  Sim_Length, Nb_Replicates, Gradient );
    }
    else
      fprintf( stderr, "%s: bad input data. No code generated.\n", TOOLNAME );

    /* cleanup... */
    Free_Bernoulli_Product( Problem );

    return 0;
  }
  else if ( Help )
    return 0;
  else
    return -1;
}

void	sort_probas( Bernoulli_Product Problem )
{
  int	i;

  /* to be implemented... */
  for ( i=0; i<Problem->Nb_Matrices; i++ )
    Problem->Permut[i] = i;
}

void	gen_file_header( Bernoulli_Product Problem,
			 int Sim_Length,
			 int Nb_Replicates )
{
  CODE("/* Simulation code for bernoulli products of (max,+) matrices */\n");
  CODE("/* generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  CODE("\t*/\n");
  CODE1("/* Date: %s \t\t\t\t\t*/\n\n", todays_date() );

  /* write definitions, includes */
  CODE("#include <stdio.h>\n");
  CODE("#include <string.h>\n");
  CODE("#include <malloc.h>\n");
  CODE("#include <math.h>\n");
  CODELN;
  CODE1("#define TMAX\t\t%d\n", Sim_Length );
  CODE1("#define NB_REPLICATES\t\t%d\n", Nb_Replicates );
  CODE1("#define NB_NODES\t%d\n", Problem->Size );
  CODE1("#define NB_MATRICES\t%d\n", Problem->Nb_Matrices );
  CODE("#define MAX_RANDOM_VALUE\t\t0x7FFFFFFF\n" );
  CODE("#define EPSILON\t\t-1e38\n" );
  CODE("#define MAX(x,y)\t( x > y ? x : y )\n");
  CODELN;
  CODE("double U_0_1(void)\n");
  CODE("{\n  return (double)random()/MAX_RANDOM_VALUE;\n");
  CODE("}\n\n");
  CODELN;
}

void	gen_declarations( Bernoulli_Product Problem,
			  boolean Gradient )
{
  int		i;

  CODE("int\t\tTime;\n");
  CODE("int\t\tk;\n");
  CODE("double\t\tProba;\n");
  CODE("/* table of X values */\n");
  CODE("double\t\t*X;\n");
  CODE("double\t\t*Y;\n");
  if ( Gradient ) {
    CODE("double\t\t*X_Plus;\n");
    CODE("double\t\t*X_Minus;\n");
  }
  CODE("double\t\t*Tmp; /* for swapping array pointers */\n");
  CODE("/* table of cumulated probabilities */\n");
  CODE("double\t\tPr_Cum[ NB_MATRICES ] = {\n");

  for ( i=0; i<Problem->Nb_Matrices-1; i++ ) {
    CODE1("\t%f,\n", Problem->Pr_Cum[i] );
  }
  /* Normally, Problem->Pr_Cum[Problem->Nb_Matrices-1] is 1.0, but I 	*/
  /* take no chances...							*/
  CODE1("\t%f };\n", 1.0 );

  CODE("/* tables for statistical estimators */\n");
  CODE("double\t\t*Gamma;\n");
  CODE("double\t\t*Gamma2;\n");
  if ( Gradient ) {
    CODE("double\t\t*DGamma;\n");
    CODE("double\t\t*DGamma2;\n");
  }
  CODELN;
}

void	write_sum( int Idx, double Value )
{
  CODE1( "X[%d]", Idx );
  if ( Value != 0 )
    CODE1(" + %f", Value );
}

void	gen_matrix_code( Bernoulli_Product Problem,
			 int The_Node )
{
  int		i;
  int		j;
  int		Nb_Nodes;
  int		Nb_In;
  int		*Neighbors;
  double	*Values;
  double	Val;

  Nb_Nodes = Problem->Size;

  Neighbors = (int*) calloc( Nb_Nodes, sizeof(int) );
  Values = (double*) calloc( Nb_Nodes, sizeof(double) );

  CODE1("\t  /* multiplication by matrix %d */\n", 
	Problem->Permut[The_Node] ); 

  for ( i=0; i<Nb_Nodes; i++ ) {
    /* counting the number of input arcs. If matrices are sparse,	*/
    /* this could be done in a much cheaper way! Later...		*/
    Nb_In = 0;
    for ( j=0; j<Nb_Nodes; j++ ) {
      Val = Get_Matrix_Entry( Problem->A[The_Node], i, j );
      if ( Val != M_INFINITY ) {
	Neighbors[Nb_In] = j;
	Values[Nb_In] = Val;
	Nb_In++;
      }
    }

    if ( Nb_In == 0 ) {
      CODE1("\t  Y[%d] = EPSILON;\n", i );
    }
    else if ( Nb_In == 1 ) {
      CODE1("\t  Y[%d] = ", i );
      write_sum( Neighbors[0], Values[0] );
      CODE(";\n");
    }
    else {
      CODE1("\t  Y[%d] = MAX( ", i );
      write_sum( Neighbors[0], Values[0] );
      CODE(", ");
      write_sum( Neighbors[1], Values[1] );
      CODE(" );\n");
      for ( j=2; j < Nb_In; j++ ) {
	CODE2("\t  Y[%d] = MAX( Y[%d], ", i, i );
	write_sum( Neighbors[j], Values[j] );
	CODE(" );\n");
      }
    }
  }

  free( Neighbors );
  free( Values );
}

void	gen_loop_code( boolean 		Gradient )
{
  CODELN;
  CODE("int main( int argc, char *argv[] )\n{\n");

  CODE("  int i = 1;\n");
  CODE("  int Num_Rep;\n");
  CODE("  int Nb_Rep = NB_REPLICATES;\n");
  CODE("  int Sim_Length = TMAX;\n");
  CODE("  double G;\n");
  CODE("  double Sigma;\n");
  CODE("  int Stats = 0;\n");
  CODE("  int Alloc_Ok;\n");
  CODELN;
  CODE("  /* parsing of arguments */\n");
  CODE("  if (argc>1)\n");
  CODE("    do {\n");
  CODE("      if ( !strcmp( \"-nevt\", argv[i] ) ) {\n");
  CODE("        i++;\n");
  CODE("        if ( i < argc ) {\n");
  CODE("          sscanf( argv[i], \"%%d\", &Sim_Length );\n");
  CODE("          i++;\n");
  CODE("        }\n");
  CODE("        else {\n");
  CODE("          printf( \"number expected after -nevt\\n\" );\n");
  CODE("          return -2;\n");
  CODE("        }\n");
  CODE("      }\n");
  CODE("      else if ( !strcmp( \"-nrep\", argv[i] ) ) {\n");
  CODE("        i++;\n");
  CODE("        if ( i < argc ) {\n");
  CODE("          sscanf( argv[i], \"%%d\", &Nb_Rep );\n");
  CODE("          i++;\n");
  CODE("        }\n");
  CODE("        else {\n");
  CODE("          printf( \"number expected after -nrep\\n\" );\n");
  CODE("          return -2;\n");
  CODE("        }\n");
  CODE("      }\n");
  CODE("      else if ( !strcmp ( argv[i], \"-stats\" ) ) {\n");
  CODE("        i++;\n");
  CODE("        Stats = 1;\n");
  CODE("      }\n");
  CODE("      else if ( !strcmp ( argv[i], \"-h\" ) ) {\n");
  CODE("        printf( \"Usage: %%s [-nevt <int>] [-nrep <int>] [-stats] [-h]\\n\", argv[0] );\n");
  CODE("        return -2;\n");
  CODE("      }\n");
  CODE("      else\n");
  CODE("        i++;\n");
  CODE("    } while ( i < argc );\n");
  CODELN;
  CODE("  /* Initializations. zeroing of arrays is done by calloc. */\n");
  CODE("  X = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
  CODE("  Y = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
  CODE("  Gamma = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
  CODE("  Gamma2 = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
  CODE("  Alloc_Ok =  ( ( (double*)NULL != X )\n");
  CODE("             && ( (double*)NULL != Y )\n");
  CODE("             && ( (double*)NULL != Gamma )\n");
  CODE("             && ( (double*)NULL != Gamma2 ) );\n");
  if ( Gradient ) {
    CODE("  X_Plus = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
    CODE("  X_Minus = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
    CODE("  DGamma = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
    CODE("  DGamma2 = (double*) calloc( (size_t)NB_NODES, sizeof(double) );\n");
    CODE("  Alloc_Ok =  ( Alloc_Ok\n");
    CODE("             && ( (double*)NULL != X_Plus )\n");
    CODE("             && ( (double*)NULL != X_Minus )\n");
    CODE("             && ( (double*)NULL != DGamma )\n");
    CODE("             && ( (double*)NULL != DGamma2 ) );\n");
  }
  CODELN;
  CODE("  if ( !Alloc_Ok ) {\n");
  CODE("    fprintf( stderr, \"Not enough memory. Simulation not launched.\\n\" );\n");
  CODE("    return -1;\n");
  CODE("  }\n");
  CODELN;
  CODE("  for ( Num_Rep = 0; Num_Rep < Nb_Rep; Num_Rep++ ) {\n");
  CODE("    for ( i=0; i<NB_NODES; i++ )\n");
  CODE("      X[i] = 0.0;\n");
  CODELN;
  CODE("    for ( Time = 0; Time < Sim_Length; Time ++ ) {\n\n");
  CODE("      Proba = U_0_1();\n");
  CODE("      for ( k=0; ( k < NB_MATRICES ) && ( Proba > Pr_Cum[k] ); k++ );\n");
  
  CODE("      Product( k, X, Y );");
  CODELN;
}

void    gen_product_code( Bernoulli_Product Problem )
{
  int i;

  CODE("void Product( int k, double *X, double *Y )\n");
  CODE("{\n");
  CODE("  switch( k ) {\n");
  for ( i=0; i<Problem->Nb_Matrices; i++ ) {
    CODE1("  case %d:\n", i );
    gen_matrix_code( Problem, i );
    CODE ("  break;\n\n");
  }
  CODE("  } /* end of switch(k) */\n");
  CODE("}\n");
  CODELN;
}

void	gen_file_trailer( boolean Gradient )
{
  CODE("        /* swapping Y and X */\n");
  CODE("        Tmp = X; X = Y; Y = Tmp;\n");
  CODELN;
  CODE("    } /* end of main simulation loop */\n\n");
  CODELN;
  CODE("    /* accumulating stats */\n");
  CODE("    for ( i = 0; i < NB_NODES; i++ ) {\n");
  CODE("      if ( X[i] != EPSILON ) {\n");
  CODE("  	G = X[i]/(double)Sim_Length;\n");
  CODE("  	Gamma[i] = ( Gamma[i] == EPSILON ? EPSILON : Gamma[i] + G );\n");
  CODE("  	Gamma2[i] += G*G;\n");
  CODE("      }\n");
  CODE("      else {\n");
  CODE("  	Gamma[i] = EPSILON;\n");
  CODE("      }\n");
  if ( Gradient ) {
    CODE("      if ( ( X_Minus[i] != EPSILON ) && ( X_Plus[i] != EPSILON ) ) {\n");
    CODE("        G = X_Plus[i] - X_Minus[i];\n");
    CODE("  	  DGamma[i] = ( DGamma[i] == EPSILON ? EPSILON : DGamma[i] + G );\n");
    CODE("        DGamma2[i] += G*G;\n");
    CODE("      }\n");

    CODE("      else {\n");
    CODE("        DGamma[i] = EPSILON;\n");
    CODE("      }\n");
  }
  CODE("    }\n");
  CODE("  } /* end of main replication loop */\n\n");

  CODE("  /* computation of statistics */\n");
  CODE("  if ( ( Nb_Rep > 0 ) && ( Sim_Length > 0 ) ) {\n");
  CODE("    for( i=0; i<NB_NODES; i++ ) {\n");
  CODE("      printf( \"gamma[%%d] = \", i );\n");
  CODE("      if ( Gamma[i] == EPSILON )\n");
  CODE("        printf(\"eps\");\n");
  CODE("      else {\n");
  CODE("        printf(\"%%f\", Gamma[i] / (double)Nb_Rep );\n");
  CODE("        if ( Stats && ( Nb_Rep > 1 ) ) {\n");
  CODE("	  Sigma = sqrt( ( Gamma2[i] - Gamma[i]*Gamma[i]/Nb_Rep ) / (Nb_Rep-1) );\n");
  CODE("          printf(\"  +/- %%f \", Sigma * 1.96 / sqrt(Nb_Rep) );\n");
  if ( !Gradient ) {
    CODE("          printf(\"(conf. 95%%%%)\");\n");
  }
  CODE("        }\n");
  CODE("      }\n");
  if ( Gradient ) {
    CODE("      printf(\"  grad: \");\n");
    CODE("      if ( DGamma[i] == EPSILON )\n");
    CODE("        printf(\"eps\");\n");
    CODE("      else {\n");
    CODE("        printf(\"%%f\", DGamma[i] / (double)Nb_Rep );\n");
    CODE("        if ( Stats && ( Nb_Rep > 1 ) ) {\n");
    CODE("	    Sigma = sqrt( ( DGamma2[i] - DGamma[i]*DGamma[i]/Nb_Rep ) / (Nb_Rep-1) );\n");
    CODE("          printf(\"  +/- %%f (conf. 95%%%%)\", Sigma * 1.96 / sqrt(Nb_Rep) );\n");
    CODE("        }\n");
    CODE("      }\n");
  }
  CODE("      printf(\"\\n\");\n");
  CODE("    }\n");
  CODE("  }\n");
  CODELN;

  CODE("  return 0;\n");
  CODE("}\n");
}

void	gen_c_code( Bernoulli_Product Problem,
		    char 	Name[MAX_NAME_LENGTH], 
		    int  	Sim_Length,
		    int  	Nb_Replicates,
		    boolean 	Gradient )
{
  boolean	File_Ok = TRUE;

  if ( strlen( Name ) > 0 )
    File_Ok = ( NULL != freopen( Name, "w", stdout ) );

  if ( File_Ok ) {
    gen_file_header( Problem, Sim_Length, Nb_Replicates );
    gen_declarations( Problem, Gradient );
    gen_product_code( Problem );
    gen_loop_code( Gradient );
    gen_file_trailer( Gradient );
  }
  else
    fprintf( stderr, "%s: error opening file %s\n", TOOLNAME, Name );
  
}
