/************************************************************************/
/* Definitions related to (maxplus) matrix manipulations. 		*/
/* The internal representation of matrices (full or sparse) is chosen	*/
/* at compile time with the USE_FULL_MP_MATRICES flag.			*/
/************************************************************************/
/* The package defines two types:					*/
/*	MaxPlus_Matrix							*/
/*	MaxPlus_Vector							*/
/* and methods:								*/
/*      New_MaxPlus_Matrix						*/
/*      Free_MaxPlus_Matrix						*/
/*      New_MaxPlus_Vector						*/
/*      Free_MaxPlus_Vector						*/
/************************************************************************/

#ifndef _MAXPLUSMAT_H
#define _MAXPLUSMAT_H

#include <stdio.h>
#include "boolean.h"
#include "maxpluscore.h"
#include "fullmat.h"
#include "spmat.h"

/* formats for printing out */
#define	FORMAT_FULL	0
#define FORMAT_SPARSE	1
#define FORMAT_SCILAB	2
#define FORMAT_DEBUG	3

#ifdef USE_FULL_MP_MATRICES
/* Interface to type-independent procedures */
/* For full matrices, a "vector" is a column vector by default. */
typedef Full_MaxPlus_Matrix MaxPlus_Matrix;
typedef Full_MaxPlus_Matrix MaxPlus_Vector;
#define New_MaxPlus_Matrix(R,C,I) New_Full_MaxPlus_Matrix(R,C,I)
#define Free_MaxPlus_Matrix(A) Free_Full_MaxPlus_Matrix(A)
#define New_MaxPlus_Vector(S,M) New_Full_MaxPlus_Matrix(S,1,M)
#define Free_MaxPlus_Vector(V) Free_Full_MaxPlus_Matrix(V)
#define Init_MaxPlus_Matrix(A,M) Init_Full_MaxPlus_Matrix(A,M)
#define Make_MaxPlus_Matrix(R,C,V) Make_Full_MaxPlus_Matrix(R,C,V)
#define Convert_MaxPlus_Matrix_To_Full(A) Make_Full_MaxPlus_Matrix(A->Rows,A->Cols,A->Vals)
#define Copy_MaxPlus_Matrix(A) Copy_Full_MaxPlus_Matrix(A)
#define Get_Matrix_Entry(A,i,j) Get_Full_Matrix_Entry(A,i,j)
#define Set_Matrix_Entry(A,i,j,x) Set_Full_Matrix_Entry(A,i,j,x)
#define Read_Full_Matrix_Entry(F,A,i,j) Full_Read_Full_Matrix_Entry(F,A,i,j)
#define Read_Full_Matrix(F,A) Full_Read_Full_Matrix(F,A)
#define Read_MaxPlus_Matrix(F,A) Full_Read_MaxPlus_Matrix(F,A)
#define Write_MaxPlus_Matrix(A,F,M,N) Write_Full_MaxPlus_Matrix(A,F,M,N)
#define Write_MaxPlus_Vector(A,F,M) Write_Full_MaxPlus_Matrix(A,F,M)
#define MM_Mult(A,B,C) Full_MM_Mult(A,B,C)
#define MV_Mult(A,B,C) Full_MM_Mult(A,B,C)
#define M_Norm(A) Full_Matrix_Norm(A)
#define V_Norm(V) Full_Matrix_Norm(V)
#define V_Min(V) Full_Matrix_Min(V)

#else
/* interface to type-independent procedures */
typedef SP_MaxPlus_Matrix MaxPlus_Matrix;
typedef SP_MaxPlus_Vector MaxPlus_Vector;
#define New_MaxPlus_Matrix(R,C,M) New_SP_MaxPlus_Matrix(R,C,M)
#define Free_MaxPlus_Matrix(A) Free_SP_MaxPlus_Matrix(A)
#define New_MaxPlus_Vector(S,M) New_SP_MaxPlus_Vector(S,M)
#define Free_MaxPlus_Vector(V) Free_SP_MaxPlus_Vector(V)
#define Init_MaxPlus_Matrix(A,M) Init_SP_MaxPlus_Matrix(A,M)
#define Convert_MaxPlus_Matrix_To_Full(A) Convert_SP_Matrix_To_Full(A)
#define Get_Matrix_Entry(A,i,j) Get_SP_Matrix_Entry(A,i,j)
#define Set_Matrix_Entry(A,i,j,x) Set_SP_Matrix_Entry(A,i,j,x)
#define Read_Full_Matrix_Entry(F,A,i,j) SP_Read_Full_Matrix_Entry(F,A,i,j)
#define Read_Full_Matrix(F,A) SP_Read_Full_Matrix(F,A)
#define Read_MaxPlus_Matrix(F,A) SP_Read_MaxPlus_Matrix(F,A)
#define Write_MaxPlus_Matrix(A,F,M,N) SP_Write_Matrix(A,F,M,N)
#define Write_MaxPlus_Vector(V,F,M) SP_Write_Vector(V,F,M)
#define MV_Mult(A,X,Y) SP_MV_Mult(A,X,Y)
#define VM_Mult(X,A,Y) SP_VM_Mult(X,A,Y)
#define M_Norm(A) SP_Matrix_Norm(A)
#define V_Norm(A) SP_Vector_Norm(A)
#define V_Min(A) SP_Vector_Min(A)

#endif

#endif
/* of #ifdef _MAXPLUSMAT_H */
