%{

#include "tetris.h"

#define YYDEBUG 1

extern boolean		Debug;

extern Tetris	 	Parse_Tetris(int, Piece_List);
extern Piece_Desc	Parse_Piece(int,double,char *,Sparse_Int_Array,
				    Sparse_Int_Array);
extern Piece_List	Append_Piece(Piece_List,Piece_Desc);
extern Sparse_Int_Array	Append_Pair(Sparse_Int_Array,BiPoint);

/* The result of the parsing is to be put into the (global!) variable	*/
/* The_Tetris. It is defined in "parse_tetris.c".			*/
/* Before I used "yyval" but this does not work with flex/bison.	*/
extern Tetris	The_Tetris;
extern int	Nb_Of_Columns;

extern boolean	Parse_Debug;
%}

%start tetris

%union{
  int			ival;
  double		rval;
  char			*sval;
  BiPoint		bpval;
  Tetris		un_reseau;
  Piece_Desc		pval;
  Piece_List 		plist;
  Sparse_Int_Array 	prval;	/* for storing profiles */
}

%token	<sval>		STRING
%token	<ival>		INTGR
%token	<ival>		COLUMNS COLOR
%token	<ival>		PIECE WEIGHT EPS HIPROF LOPROF /* tokens w/o value */
%token	<rval>		REAL

%type	<un_reseau>	tetris
%type   <pval>		a_piece
%type   <plist>         liste_pieces
%type   <ival>		col_dcl
%type	<rval>		weight_dcl
%type	<sval>		color_dcl
%type   <prval>         profile
%type	<bpval>		a_pair
%%

tetris	:	col_dcl liste_pieces
		{
		  The_Tetris = Parse_Tetris( $1, $2 );
		  /* Free_Piece_List( $3 ); */
		  if ( Debug )
		    printf("Input ok\n");
		}
		;

col_dcl :	COLUMNS INTGR
		{
		  $$ = $2;
		  /* set global variable needed for parsing pieces before
		     parsing the tetris game itself.			*/
		  Nb_Of_Columns = $2;
		}
		;

liste_pieces :  a_piece
		{
		  $$ = Append_Piece( (Piece_List)NULL, $1 );
		}
		| liste_pieces a_piece
		{
		  $$ = Append_Piece( $1, $2 );
		}
		;

a_piece :	PIECE INTGR weight_dcl color_dcl LOPROF profile HIPROF profile
		{ 	
		  $$ = Parse_Piece( $2, $3, $4, $6, $8 );
		}
		| 
		PIECE INTGR weight_dcl color_dcl HIPROF profile LOPROF profile
		{ 	
		  $$ = Parse_Piece( $2, $3, $4, $8, $6 );
		}
		;

color_dcl :	COLOR STRING
		{	
		  $$ = $2;
		}
		;

weight_dcl :	WEIGHT REAL
		{	
		  $$ = $2;
		}
		| WEIGHT INTGR
		{	
		  $$ = (double)$2;
		}
		| /* nothing */
		{
		  $$ = 1.0;
		}
		;

profile :	a_pair
		{	
		  $$ = Append_Pair( (Sparse_Int_Array)NULL, $1 );
		}
		| profile a_pair
		{	
		  $$ = Append_Pair( $1, $2 );
		}
		;

a_pair :	INTGR INTGR
		{	
		  $$.Col = $1;
		  $$.Height = $2;
		}
		|
		INTGR EPS
		{	
		  $$.Col = $1;
		  $$.Height = EPSILON;
		}
		;
%%
#include "tetris_lex.yy.c"
