/************************************************************************/
/*  Ce progamme calcule a partir de la matrice standard (donnee         */
/*  avec une precision fixee a l'avance) d'un                           */
/*  graphe d'evenement (un jeton par place) sa classe scs-cyc,          */
/*  son sous-graphe critique avec sa decomp. en compo. fort. connexes   */
/*  ainsi que sa valeur propre et la longueur du regime transient       */
/*  en partant  du vecteur de son choix.                                */
/*  Ce programme utilise les algorithmes les plus performants           */
/*  connus (de moi) a ce jour (30/12/96).                               */  
/************************************************************************/


#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <values.h>
#include <string.h>

#define TOOLNAME	"scscyc"
#define VERSION		"2.2 (13/03/1999)"
/*#define BIG 10000 */  /* arbitraire, doit pouvoir etre ameliore */

#define TRUE 1
#define FALSE 0
/* #define DEBUG */

#define EPSILON 1e-7 	/* precision for testing integrality of entries */

#ifdef SUN4SOL2
extern double rint(double);
#endif 

typedef struct chaine {
    int *X;
    struct chaine *next;
} chain;

typedef struct RATIONAL {
  int Num;
  int Den;
} *rational;

/**************************************************************************/
/*     definitions des variables globales                                 */
/**************************************************************************/

char str[128];        /* lecture fichier  entree  */
int trans= FALSE ;    /* booleen de calcul de la periode transiente */
int Cycles= FALSE ;   /* booleen de calcul des cycles critiques */

int Precision = 1;    /* int. represents the precision of the computation*/

int infty;            /* integer that represent - infinity */
int plusinfty;        /* integer that represent + infinity */
int N;                /* nb de sommets */
int K;                /* nb sommets critiques */
int scs,cyc;          /* nb de Compo connexes de Critic et sa cyclicite */

int lambdaD;          /* denominateur de la val propre */
int lambdaN;          /* numerateur de la val propre */

int Tr ;              /* periode transiente */
int *X;               /* vecteur itere */

int
**A,                   /* matrice initiale */
**An,                  /* matrice normalise: A - lambda */
**Col;                 /* matrice des colonnes 0 pour le calcul de Karp */

int
**Astar,               /* mat des chemins les plus longs dans An */
**Aplus,               /* Aplus = A otimes Astar */
**MatVectProp,         /* Mat diag pour un vect. propre */
**InvMatVectProp,      /* Inverse de MatVectProp */
**CriticMat,           /* mat intermediaire */
**Critic2;             /* matrice critique */

int
**Critic,   /* matrice d'adjacence du graphe critique agrege */
**Comp,     /* Composantes connexes */
**Cstar;    /* Critic to the star */



int *R,     /* passage de  Critic a A */
*U;         /* passage de A a Critic  */

int *nbr;        /* nbr[i] : nb de sommets dans la Compo i du graph. crit. */


int *circuit;    /* circuit[i] = 1 <=> i \in circuit */
int *cyclicite;  /* cyclicite[i] = cyc de la Compo i du graphe critique */


int now ;        /* ces trois variables servent pour le                     */
                 /* calcul des composantes fortement connexe en depth first */
int *val,*stack;

int p ;

int Debug = FALSE;

/*************************************************************************/
/*           Procedures elementaires                                     */
/*************************************************************************/

/****************************/
int IABS(int a)
{
  return((a>0)?a:-a);
}


/****************************/
int max(int a, int b)
{
  if  ((a == plusinfty) || (b == plusinfty))  {return(plusinfty);}
  else if (a == infty) return(b);
  else if (b == infty) return(a);
  else return((a>b)?a:b);
}


/*************************/
int plus(int a, int b)
{
  if ((a == infty) || (b == infty))  {return(infty);}
  else if  ((a == plusinfty) || (b == plusinfty))  {return(plusinfty);}
  else return(a+b);
}

/*************************/
int moins(int a, int b)
{
  if (a == infty)  {return(infty);}
  else if (b == infty) {return(plusinfty);}
  else return(a-b);
}

/*************************/
int mult(int a,int b)          /* b n'est jamais infini (denominateur) */
{
  if  ((a == infty) && (b >= 0))  {return(infty);}
  else if  ((a == infty) && (b < 0))  {return(infty);}
  else if  ((a == plusinfty) && (b >= 0))  {return(plusinfty);}
  else if  ((a == plusinfty) && (b < 0))  {return(infty);}
  else return(a*b);
}


int compa(int a, int b,int c,int d) /*  a/b < c/d ? , b and d not infinite*/
{
  if  ((a == infty) || (c == plusinfty)) {return(TRUE);}
  else if  ((a == plusinfty) || (c == infty)) {return(FALSE);}
  else return (((double)a/b <(double)c/d )?TRUE:FALSE);
}

/************** Procedure to compute C = A \oplus B , size S ***********/
/* void Add(int **A, int **B, int **C, int S ) */
/* { */
/*   int i,j; */

/*   for (i = 0 ;i<S;i++) */
/*     { */
/*       for (j = 0 ;j<S;j++) */
/* 	{ */
/* 	  C[i][j] = max(A[i][j],B[i][j]); */
/* 	} */
/*     } */
/* } */


/************** Procedure to compute C = A \otimes B , size S **********/
void Otimes(int **A, int **B, int **C, int S )
{
  int i,j,k;

  for (i = 0 ;i<S;i++)
    {
      for (j = 0 ;j<S;j++)
	{
	  C[i][j] = infty;
	}
    }

  for (i = 0 ;i<S;i++)
    {
      for (j = 0 ;j<S;j++)
	{
	  for (k = 0 ;k<S;k++)
	    {
	      C[i][j] = max(plus(A[i][k],B[k][j]),C[i][j]);
	    }
	}
    }
}

/*********************************/
int gcd(int a, int b)          /*a and b are never infinite and positive */
{


  if (a == 0)         { return(b); }
  else if (b == 0)    { return(a); }
  else if (a > b)    { return(gcd(a%b,b));}
  else               { return(gcd(a,b%a)); }
}
    
			   
/*********************************/
int lcm(int a, int b)          /*a and b are never infinite */
{
  int c;

  c = gcd(a,b);
  return(a*b/c);
}

/************************************************/
/* A procedure to reduce a fraction __in situ__ */
/************************************************/
void Reduce_Fraction( rational F ) {
  int d;

  if ( F->Den != 0 ) {
    d = gcd( F->Num, F->Den );
    F->Num = F->Num / d;
    F->Den = F->Den / d;
  }

}

/**********************************/
/* int calcul_pgcd(int *circuit, int c) */
/* { */

/*   int lgr,i; */

/*   lgr = 0; */


/*   for(i=0;i<K;i++) */
/*     { */
/*       if (circuit[i] == 1) lgr++; */
/*     } */

/*   return(gcd(lgr,c)); */

/* } */

/************** Procedure to compute Y = A \otimes X , size S **********/
void Product(int **A, int *X , int *Y, int S)
{
  int i,k;

  for (i = 0 ;i<S;i++)
    {
      Y[i] = infty;
    }

  for (i = 0 ;i<S;i++)
    {
    
      for (k = 0 ;k<S;k++)
	{
	  Y[i] = max(plus(A[i][k],X[k]),Y[i]);
	}
    }

}


/*************************************************************************/
/*                      Procedures de calcul                             */
/*************************************************************************/

/*********************************/
void cycle_time()                         /* algo de Karp avec tous les j */
{
  int n,k,c,i,j;
  int lD,lN,miniN,miniD,maxiD,maxiN;

  lambdaN = infty;
  lambdaD = 1;                      /* lambda est mis a -infinity*/


  for (j=0;j<N;j++)              /* on passe toutes les colonnes */
    {
      miniN = plusinfty ;  /* init  de maxi et de mini */
      miniD = 1;
      maxiN = infty;
      maxiD = 1;

      for (i=0;i<N+1;i++)          /* init Col a infty */
	{
	  for (c = 0; c<N;c++)
	    {
	      Col[i][c] = infty;
	    }
	}

      /* Col[0]: vecteur identite */
      Col[0][j] = 0;  





      /* Col[1]: colonne j de la matrice A */
      for (i=0;i<N;i++)
	{
	  Col[1][i] = A[i][j];
	}
 

			
      /* Col[n]: colonne j de la matrice A^n */
      for (n = 2; n <= N;n++)
	{
	  for (i = 0 ; i<N; i++)
	    {
	      for (c = 0; c<N;c++)
		{
		  Col[n][i] = max(plus(A[i][c],Col[n-1][c]),Col[n][i]);
		}
	    }
	}


      for (i=0;i<N;i++)
	{
	  miniN = plusinfty;
	  miniD = 1;
	  for (k=0;k<N;k++)
	    {
	      lN = moins(Col[N][i] ,Col[k][i]);
	      lD = (N-k);
	      if (compa(lN,lD,miniN,miniD))
		{miniN = lN;miniD = lD;}
	    }
	  if (compa(maxiN,maxiD,miniN,miniD) )
	    {maxiN = miniN;maxiD = miniD;}
	}


      if (compa(lambdaN,lambdaD,maxiN,maxiD) )
	{lambdaN = maxiN;lambdaD = maxiD;}

    }

}
		

/******************************/
void calcul_chemins()                    /* algorithme de Floyd */
{
  int i,j,k;
  int aux;

  for(i = 0;i<N;i++)
    {
      for (j= 0;j<N;j++)
	{
	  Astar[i][j] = An[i][j];
	}
    }


  for (k=0;k<N;k++)
    {
      for(i = 0;i<N;i++)
	{
	  for (j= 0;j<N;j++)
	    {
	      aux = plus(Astar[i][k],Astar[k][j]);
	      if (aux > Astar[i][j]) Astar[i][j] = aux;
	    }
	}
    }


  Otimes(An,Astar,Aplus,N);

}

/****************************/
void calcul_vec_propre()
{
  int i,j;
  for (i= 0 ;i < N;i++)
    {
      for (j = 0;j<N;j++)
	{
	  MatVectProp[i][j] = infty;
	  InvMatVectProp[i][j] = infty;
	}
    }

  for (i= 0 ;i < N;i++)
    {
      if (Aplus[i][i] == 0)       
	{
	  for (j = 0;j<N;j++)
	    {
	      MatVectProp[j][j] = max(MatVectProp[j][j], Aplus[j][i]);
	      InvMatVectProp[j][j] = -MatVectProp[j][j];
		
	    }
		
	  /* break; */
	}


    }
}

/********************/
void calcul_mat_critic(int Silent)      /* Methode de Jean Mairesse */
{
  int i,j;

  K=0;
  for (i = 0;i< N;i++)
    {
      U[i] = infty;          /* not defined  au debut */
      R[i] = infty;          /* not defined  au debut */
      if (Aplus[i][i] == 0)
	{
	  U[i] = K;
	  R[K++] = i;
	}
    }


  Otimes(InvMatVectProp,An,CriticMat,N);
  Otimes(CriticMat,MatVectProp,Critic2,N);



  Critic = (int **) calloc(K, sizeof(int*));
  for(i=0;i<K;i++)
    {
      Critic[i] = (int *) calloc(K, sizeof(int));
    }





  for (i = 0;i< K;i++)
    {
      for (j = 0;j< K;j++)
	{
	  if (Critic2[R[i]][R[j]] == 0) {Critic[i][j] = 0;}
	  else Critic[i][j] = infty;

	}
    }


  if ( !Silent ) {


  printf("Critical sub-graph:\n");
  for (i = 0;i< N;i++)
    {
      for (j = 0;j< N;j++)
	{
	
	  if (Critic2[i][j] != 0) {printf("   eps");}
	  else printf("%6d",Critic2[i][j]);
	}
      printf("\n");
    }

  printf("Agregated critical sub-graph:\n");
  for (i = 0;i< K;i++)
    {
      for (j = 0;j< K;j++)
	{
	
	  if (Critic[i][j] != 0) {printf("   eps");}
	  else printf("%6d",Critic[i][j]);
	}
      printf("\n");

    }

  }
}

/******************************/
int visit(int k,int S)                  /* proc. utilisee par calcul_Comp2 */
{
  int m,min,j,i;


  now++;
  val[k] = now;
  min = now;

  stack[p] = k;
  p++;

  for (i = 0;i < S;i++)
    {
      if (Critic[k][i] != infty)
	{
	  if (val[i] == 0)
	    {
	      m = visit(i,S);
	      if (m < min) min = m;
	    }
	  else if (val[i] < min) min = val[i];
	}
    }
  if (min == val[k])
    {
      j=0;
      scs++;
      do
	{
		
	  p--;
	  Comp[scs][j++] = stack[p];
	}
      while (stack[p] != k);
      nbr[scs] = j;
	
    }
  val[k]= S+1;
  return(min);
}

/******************************/
void calcul_Comp2()                  /* appelle la fonction visite */
{
  int i;
  int temp;

  scs = -1;
  now = 0;
  p = 0;


  val =  (int *) calloc(K, sizeof(int));
  stack =  (int *) calloc(K, sizeof(int));

  for (i = 0;i<K;i++)
    {
      val[i] =0;                  /* initialisations a "rien" */
      stack[i] =-1;
    }


  for (i = 0;i<K;i++)
    {
      if (val[i] == 0) temp = visit(i,K);
    }
}

/******************************/
void calcul_Comp3()                  /* calcule les composantes fort. con. */
{
  int i,j,k;
  
  int aux;
  
  stack =  (int *) calloc(K, sizeof(int));  /* to recall treated nodes */

  Cstar = (int **) calloc(K, sizeof(int*)); /* Cstar = Critic ^ *      */
    for(i=0;i<N;i++)
      {
	Cstar[i] = (int *) calloc(K, sizeof(int));
      }


  for(i = 0;i<K;i++)
    {
      for (j= 0;j<K;j++)
	{
	  Cstar[i][j] = Critic[i][j];
	}
    }


  for (k=0;k<K;k++)
    {
      for(i = 0;i<K;i++)
	{
	  for (j= 0;j<K;j++)
	    {
	      aux = plus(Cstar[i][k],Cstar[k][j]);
	      if (aux > Cstar[i][j]) Cstar[i][j] = aux;
	    }
	}
    }



  scs = -1;  
  k = 0;  



#ifdef DEBUG
  printf("C star: \n");
  for (i = 0;i< K;i++)
    {
      for (j = 0;j< K;j++)
	{
	  printf(" %d ",Cstar[i][j]);
	}
      printf("\n");
    }
  printf("\n"); 
#endif


  for (i = 0;i<K;i++)
    {
      stack[i] = 0;
    }


  for (i = 0;i<K;i++)
    {
      if (stack[i] == 0)
	{
	  scs++;
	  k = 0;
	  stack[i] = 1;
	  Comp[scs][k] = i;
	  nbr[scs] = k+1;
	  k++;
	  for (j = 0;j<K;j++)
	    {
	      if ((Cstar[i][j] != infty) && (Cstar[j][i] != infty) && (j != i))
		{
		  stack[j] = 1;
		  Comp[scs][k] = j;
		  nbr[scs] = k+1;
		  k++;
		}
	    }
	}
    }
}



/****************** Calcul de cyc sur la compo Node (Denardo) ***********/
int calcul_cyc(int *Node, int nb)      
{
  int i,j,sum,cycle;
  int *label;
  int *scan;

  label = (int *) calloc(nb, sizeof(int));
  scan = (int *) calloc(nb, sizeof(int));

  for (i = 0 ;i<nb;i++)
    {
      label[i] = -1;
      scan[i] = FALSE;



    }

  sum = 0;
  cycle = 0;
  label[0] = 0;

  while ((cycle != 1) && (sum < nb)) {

    for (i = 0 ; (label[i] == -1) || (scan[i] == TRUE) ; i++);


    scan[i] = TRUE;
    sum++;


    for (j = 0; j<nb;j++)
      {
	if (Critic[Node[i]][Node[j]] != infty)
	  {
	    if (label[j] != -1) 
	      {
		cycle = gcd(cycle,IABS(label[i]+1-label[j]));

	      }

	    else label[j] = label[i]+1;
	  }
      }

  }

  return(cycle);
}


/*****************************/
int transient()
{
  int i,fin,T,equal;

  chain *ch,*start,*current,*list;

  int *Y,*Yn,*Z;

  start = (chain *) malloc(sizeof(chain));

  Y = (int *) calloc(N, sizeof(int));



  T = 1;



  printf("\n");

  printf("T = %d: ",0);
  for (i = 0;i< N;i++)
    {
      if (X[i] == infty) {
	printf("       eps");
      }
      else 
	printf(" %9.4f",(double)X[i]/Precision);
    }
  printf("\n");


  for (i = 0;i< N;i++)
    {Y[i] = X[i];}


  for (i = 0;i< N;i++)
    {X[i] = mult(X[i],lambdaD);}

  start -> X = X;
  current = start;



  fin  = FALSE;
  while (fin == FALSE)
    {
      Z = (int *) calloc(N, sizeof(int));
      Yn = (int *) calloc(N, sizeof(int));
    
    
      Product(A,Y,Z,N);
    
      for (i = 0;i< N;i++) {Y[i] = Z[i];}



      printf("T = %d: ",T);
      for (i = 0;i< N;i++) {
	if (Y[i] == infty) {
	  printf("       eps");
	}
	else 
	  printf(" %9.4f",(double)Y[i]/Precision);
      }
      printf("\n");

    
      for (i = 0;i< N;i++)
	{
	  Yn[i] = moins(mult(Y[i],lambdaD),T*lambdaN);
	}


#ifdef DEBUG
      printf("(normalise)T = %4d lambda = %d/%d \n ",T,lambdaN,lambdaD);
      for (i = 0;i< N;i++)
	{
	  if (Yn[i] == infty) {printf("   eps");} else printf("%6d",Yn[i]);
	}
      printf("\n");
#endif

    

      ch = (chain *) malloc(sizeof(chain));
      ch->X = Yn;
      current->next = ch;
      ch->next = NULL;
      current = ch;

      list = start;
      equal = FALSE;
      while ((list != ch) && (equal == FALSE))
	{
	  equal = TRUE;
	  for (i = 0;i<N;i++)
	    {
	      if ( list->X[i] !=  ch->X[i] ) equal = FALSE;
	    }
	  list = list->next;
	}
      fin = (equal) && (T >=cyc);
      T++;
    }

  return(T);
}

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[])
{
  static char	In_File_Name[128] = "";
  static char	Out_File_Name[128] = "";

  int 		i;
  int		j;
  int		k;
  double 	brut;
  int  		Args_OK = TRUE;
  int  		Help = FALSE;
  int	  	InOut = TRUE;

  rational	F;

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( In_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"%s: usage:filename expected after -f\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      else if ( ( 0 == strcmp( "-h", argv[i] )) || 
		( 0 == strcmp( "-?", argv[i] )) ||
		( 0 == strcmp( "-help", argv[i] ) ) ) {
	  i++;
	  fprintf(stdout,"usage: %s [-transient] [-f inputfile] ",
		  TOOLNAME );
	  fprintf(stdout,"[-o outputfile] [-precision <int>]\n");
	  fprintf(stdout,"[-cycles]\n");
	  fprintf(stdout,"The precision is normally a power of 10 (ex: 100 for two digit precision).\nThe inputfile must contain:\n\t- the size of the matrix, N, \n\t- an NxN integer matrix and an N vector (with option -transient),\n  with the following convention:\n\tthe string 'eps' represents -infinity,\n\tnumbers out of range [-%d/precision,%d/precision]\n\t are not accepted and\n\tall the insignificant digits (over the precision) are dropped.\n",MAXINT,MAXINT);
	  Args_OK = FALSE;
	  Help = TRUE;
      }
      else if ( 0 == strcmp( "-transient", argv[i] )) {
	i++;
	trans = TRUE;
      }
      else if ( 0 == strcmp( "-cycles", argv[i] )) {
	i++;
	Cycles = TRUE;
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {	
	i++;
	if ( i < argc) {
	  strcpy( Out_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"usage: %s: filename expected after -o\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      else if ( 0 == strcmp( "-precision", argv[i] ) ) {
	i++;
	if ( (i < argc) && ( 1 == sscanf( argv[i], "%d", &Precision ) ) ) {
	  i++;
	}
	else {
	  fprintf(stderr,"usage: %s: integer expected after %s\n",
		  argv[0], argv[i-1]);
	  Args_OK = FALSE;
	}
      }
      else if ( 0 == strcmp( "-debug", argv[i] )) {
	i++;
	Debug = TRUE;
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n", TOOLNAME, VERSION);
	Args_OK = FALSE;
	Help = TRUE;
	i++;
      }
      else 
	i++;
    } while( Args_OK && ( i < argc ) );

  if (Args_OK) {
    if  ( ( strlen(In_File_Name) != 0 ) &&
	  (NULL == freopen( In_File_Name, "r", stdin ) )) {
      fprintf(stderr,"%s: could not open file %s\n",argv[0], In_File_Name );
      InOut = FALSE;
    }

    if ( ( strlen(Out_File_Name) != 0 ) &&
	 (NULL == freopen( Out_File_Name, "w", stdout ))) {
      
      fprintf(stderr,"%s: cannot open file %s\n",argv[0],Out_File_Name);
      InOut = FALSE;
    }
  }

  /************* start of the program if everything is OK ****************/

  if (Args_OK && InOut ) {

    /********************* read the input matrix *******************/

    infty = -MAXINT;
    plusinfty = MAXINT;

    scanf("%s",str);
    sscanf(str,"%d",&N);



    A = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	A[i] = (int *) calloc(N, sizeof(int));
      }


    R = (int *) calloc(N, sizeof(int));
    U = (int *) calloc(N, sizeof(int));


    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {

	scanf("%s",str);
	
	if (strcmp(str,"eps")!=0) {

	  sscanf(str,"%lf",&brut);
	  if (fabs(brut*Precision) >= plusinfty) {

	    fprintf(stderr,"%s: Arithmetic overflow. ", TOOLNAME );
	    fprintf(stderr,"Decreasing the precision may help.\n");

	    return(2);
	  }
	  else {
	    /* rounding off at requested precision */
	    if ( Debug ) {
		printf("Value*Precision = %f, cast = %d, trunc = %d\n",
		       brut*Precision, 
		       (int)(brut*Precision),
		       (int)rint(brut*Precision) );
	    }
	    A[i][j] = (int) rint(brut*Precision);
	    if ( fabs( A[i][j] - brut*Precision ) > EPSILON ) {
	      fprintf(stderr,"%s: Warning: Matrix entry rounded off:", 
		      TOOLNAME );
	      fprintf(stderr," %f -> %f.\n", 
		      brut, 
		      (double)A[i][j]/(double) Precision );
	    }
	  }
	}
	else { 
	  A[i][j] = infty ;
	}
      }
    }

    X = (int *) calloc(N, sizeof(int));

    if (trans)
      for (i=0;i<N;i++)
	{
	  scanf("%s",str);
	  if (strcmp(str,"eps")==0) 
	    { X[i] = infty ;}
	  else { 
	    sscanf(str,"%lf",&brut);
	    X[i] = (int) rint(brut*Precision);
	    /*sscanf(str,"%d",&X[i]);*/
	  }
	}



    /**************** calcul de la valeur propre  ***************/

    /** memory allocation **/

    Col = (int **) calloc(N+1, sizeof(int*));
    for(i=0;i<=N;i++)
      {
	Col[i] = (int *) calloc(N, sizeof(int));
      }



    cycle_time();


    printf("\n");
    if (lambdaN==infty) {

      printf("cycle mean: eps\n");
      printf("No cycles. \n");

      return(0);
    }

/*     else if (lambdaN==plusinfty)  */
/*       { */
/* 	printf("cycle mean: infinite\n"); */
/* 	printf("This case is not studied.\n"); */
/* 	return(1); */
/*       } */
    else {
      F = (rational) malloc( sizeof(struct RATIONAL) );
      F->Num = lambdaN;
      F->Den = lambdaD*Precision;
      Reduce_Fraction( F );

      if ( !Cycles ) {
	/* printf("cycle mean: %d / %d  \n",lambdaN,lambdaD*Precision); */
	printf("cycle mean: %d / %d  \n", F->Num, F->Den );
      }

      free(F);
    }
    
    printf("\n");

    /******************* calcul de la matrice normalise ****************/

    /** memory allocation **/

    An = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	An[i] = (int *) calloc(N, sizeof(int));
      }



    for(i=0;i<N;i++)
      {
	for(j=0;j<N;j++)
	  {
	    An[i][j] = moins(mult(A[i][j],lambdaD),lambdaN);
	  }
      }

    /***************** calcul de la matrice critique  **********/

    /** memory allocation **/

    InvMatVectProp = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	InvMatVectProp[i] = (int *) calloc(N, sizeof(int));
      }

    MatVectProp = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	MatVectProp[i] = (int *) calloc(N, sizeof(int));
      }

    Astar = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Astar[i] = (int *) calloc(N, sizeof(int));
      }

    Aplus = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Aplus[i] = (int *) calloc(N, sizeof(int));
      }

    CriticMat = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	CriticMat[i] = (int *) calloc(N, sizeof(int));
      }

    Critic2 = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Critic2[i] = (int *) calloc(N, sizeof(int));
      }



    calcul_chemins();
    calcul_vec_propre();

    if ( Cycles ) {
      /* compute the critical graph, but silently */
      calcul_mat_critic(1);
    }
    else {
      /* compute and write it down */
      calcul_mat_critic(0);
    }

    /******************* calcul de scs et Comp ******************/


    /** allocation memoire **/

    Comp = (int **) calloc(K, sizeof(int*));
    for(i=0;i<K;i++)
      {
	Comp[i] = (int *) calloc(K, sizeof(int));
      }

    nbr = (int *) calloc(K, sizeof(int));



    calcul_Comp3();

    if ( Cycles ) {
      /* write down critical arcs, only if they belong to some Comp */
      for (i = 0; i < scs+1; i++) {
	/* component #i */
	for (j = 0; j < nbr[i]; j++) {
	  /* node #j in component #i */
	  for ( k = 0; k < nbr[i]; k++ ) {
	    if ( Critic2[ R[Comp[i][j]] ][ R[Comp[i][k]] ] == 0 ) {
	      /* Print arc. WARNING beware of the transposition of A!!! */
	      printf("%6d %6d %14f\n", 
		     R[Comp[i][k]], R[Comp[i][j]], 
		     A[ R[Comp[i][j]] ][ R[Comp[i][k]] ]
		     / (double)Precision );
	    }
	  }
	}
      }

      /* old exhaustive way: catches non-critical stuff also
      for ( i=0; i<N; i++ ) {
	for (j = 0;j< N;j++) {
	  if (Critic2[i][j] == 0) {
	    printf("%6d %6d %14f\n", j, i, A[i][j] / (double)Precision );
	  }
	}
      }
      */
      /* get out to skip rest */
      return 0;
    }

    /** display Comp **/

    printf("Number of critical strongly connected components: scs = %d \n ",scs+1);
    for (i = 0;i< scs+1;i++)
      {
	printf("component %d:\n",i);
	for (j = 0;j< nbr[i];j++)
	  {
	    printf("%5d",R[Comp[i][j]]+1);
	  }
	printf("\n");
      }


    /**************** calcul de cyc  *************************/
    cyclicite = (int *) calloc(scs+1, sizeof(int));
    cyc =1;

    for (i = 0; i<scs+1; i++)           /* pour chaque Compo connexe */
      {
	cyclicite[i] = calcul_cyc(Comp[i],nbr[i]);
	cyc = lcm(cyc,cyclicite[i]);
      }

    printf("\n");
    printf("Cyclicity: cyc = %d \n",cyc);

    /********************* calcul periode transiente ***************/

    /* On part de X et on teste quand on retombe sur un point deja visite */

    if (trans)
      {
	Tr = transient();

	printf("\n");
	printf("length of transient period:  %d \n",Tr - cyc);
	printf("\n");
      }

    return 0;
  }
  else if ( Help )
    return 0;
  else
    return -1;
}


