/* Procedures applicable to (Max,+) spectral analysis manipulations */

#include <stdio.h>
#include <stdlib.h>
#include "boolean.h"
#include "maxplusmat.h"
#include "spectral.h"

int	infty;            /* integer that represent - infinity */
int	plusinfty;        /* integer that represent + infinity */

/************************************************************************/
/* Procedures useful in Karp's Algo.					*/
/************************************************************************/
int moins(int a, int b)
{
  if (a == infty)  {return(infty);}
  else if (b == infty) {return(plusinfty);}
  else return(a-b);
}

int compa(int a, int b,int c,int d) /*  a/b < c/d ? , b and d not infinite*/
{
  if  ((a == infty) || (c == plusinfty)) {return(TRUE);}
  else if  ((a == plusinfty) || (c == infty)) {return(FALSE);}
  else return (((double)a/b <(double)c/d )?TRUE:FALSE);
}

/************************************************************************/
/* Spectral Analysis.							*/
/************************************************************************/

/* to be done... */

/* Computation of the cycle time of the matrix, using Karp's algorithm	*/
/* with all columns computed.						*/
void Compute_Cycle_Time( Max_Plus_Matrix The_Mat )
{
  int n,k,c,i,j;
  int	N;
  int lD,lN,miniN,miniD,maxiD,maxiN;
  int		lambdaN;
  int		lambdaD;
  int	*A;		/* the matrix A, stored linearly */
  int	**Col;
  boolean	Alloc_Ok = TRUE;

  if ( The_Mat->Rows != The_Mat->Cols ) {
    printf("(maxplus): cannot compute cycle times for non-square matrices.\n");
  }
  else {
    N = The_Mat->Rows;

    /* allocation of temporary matrices */
    Col = (int **) calloc( (unsigned)(N+1), sizeof(int*));
    A = (int *) calloc( (unsigned)(N*N), sizeof(int*));
    Alloc_Ok = ( NULL != Col ) && ( NULL != A );
    for( i=0; Alloc_Ok && (i<=N); i++ ) {
      Col[i] = (int *) calloc( (unsigned)N, sizeof(int));
      Alloc_Ok = ( NULL != Col[i] );
    }
    /* filling the A matrix in */

    if ( Alloc_Ok ) {
      for ( i=0; i<N; i++ )
	for ( j=0; j<N; j++ ) 
	  A[i*N+j] = The_Mat->Val[i][j];

      lambdaN = infty;
      lambdaD = 1;                      /* lambda est mis a -infinity*/

      for (j=0;j<N;j++) {              /* on passe toutes les colonnes */
	
	miniN = plusinfty ;  /* init  de maxi et de mini */
	miniD = 1;
	maxiN = infty;
	maxiD = 1;

	for (i=0;i<N+1;i++) {          /* init Col a infty */
	  for (c = 0; c<N;c++) {
	    Col[i][c] = infty;
	  }
	}

	/* Col[0]: vecteur identite */
	Col[0][j] = 0;  

	/* Col[1]: colonne j de la matrice A */
	for (i=0;i<N;i++) {
	  Col[1][i] = A[i*N+j];
	}
 
	/* Col[n]: colonne j de la matrice A^n */
	for (n = 2; n <= N; n++) {
	  for (i = 0; i < N; i++) {
	    for (c = 0; c < N; c++) {

	      Col[n][i] = max(plus(A[i*N+c],Col[n-1][c]),Col[n][i]);
	    }
	  }
	}


	for (i=0;i<N;i++) {
	  miniN = plusinfty;
	  miniD = 1;
	  for (k=0;k<N;k++) {

	    lN = moins(Col[N][i] ,Col[k][i]);
	    lD = (N-k);
	    if (compa(lN,lD,miniN,miniD)) {
	      miniN = lN;
	      miniD = lD;
	    }
	  }
	  if ( compa(maxiN,maxiD,miniN,miniD) ) {
	    maxiN = miniN;
	    maxiD = miniD;
	  }
	}

	if (compa(lambdaN,lambdaD,maxiN,maxiD) ) {
	  lambdaN = maxiN;
	  lambdaD = maxiD;
	}
      }
      /* Computation finished. Storing the cycle time */
      if ( NULL != The_Mat->Sp )
	free( The_Mat->Sp );
      The_Mat->Sp = (Spectral_Elements) 
	malloc( sizeof(struct SPECTRAL_ELEMENTS ) );
      The_Mat->Sp->Cycle_Mean_Rational.Numerator = lambdaN;
      The_Mat->Sp->Cycle_Mean_Rational.Denominator = lambdaD;
      The_Mat->Sp->Cycle_Mean_Double = (double)lambdaN / (double)lambdaD;
    }

    /* cleanup */
    for ( i=0; i<=N; i++ )
      free( Col[i] );
    free( Col );
    free( A );
  }
}
