/* A program for converting a trajectory (eg of some Markov chain)	*/
/* into a nicely colored fig file...					*/

#define X_SCALE		25
#define Y_SCALE		500

#define FIG_BLACK	0
#define FIG_BLUE	1
#define FIG_GREEN	2
#define FIG_CYAN	3
#define FIG_RED		4
#define FIG_MAGENTA	5
#define FIG_YELLOW	6
#define FIG_WHITE	7
#define FIG_BROWN	24
#define FIG_GOLD	31

#include <stdio.h>
#include <math.h>

#include "boolean.h"
#include "parseargs.h"

#define TOOLNAME	"traj2fig"
#define VERSION		"1.0"

/* Declaration of Debug, normally in the parser... */
int Debug = FALSE;

void Write_Fig_Rectangle( double T1, double T2, int State,
			  double X_Scale, double Y_Scale );
int  Map_Color( int State );

main(
int	argc,
char	*argv[] )
{
  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";
  static char		Title[128] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  double		T0;
  double		Old_Time;
  double		Time;
  int			State;
  int			New_State;
  int			Nb_Read;
  int			Nb_Max_Samples = 1000;
  double		T_Max = 1000.0;
  double		X_Scale = 1.0;
  double		Y_Scale = 1.0;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )

      else if GET_STRING( "-o", Out_File_Name )

      else if GET_STRING( "-title", Title )

      else if GET_INT( "-nmax", Nb_Max_Samples )

      else if GET_REAL( "-tmax", T_Max )

      else if GET_REAL( "-xscale", X_Scale )

      else if GET_REAL( "-yscale", Y_Scale )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>] ", TOOLNAME );
	printf("[-nmax <int>] [-tmax <real>]\n\t[-title <string>] [-?] [-h]\n" );
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Debug = TRUE;

    if ( strlen( In_File_Name ) != 0 )
      File_Ok = ( NULL != freopen( In_File_Name, "r", stdin ) );

    if ( File_Ok ) {

      if ( strlen( Out_File_Name ) != 0 )
	File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( File_Ok ) {

	printf("#FIG 3.1\n");
	printf("Landscape\n");
	printf("Center\n");
	printf("Metric\n");
	printf("1200 2\n");

	File_Ok = ( 2 == scanf( "%lf%d", &Time, &State ) );
	Nb_Read = 1;
	if ( File_Ok ) {
	  T0 = Time;
	  Old_Time = Time;
	}
	for( ; File_Ok && ( Nb_Read <= Nb_Max_Samples )
	       && ( Time <= T_Max ); ) {
	  File_Ok = ( 2 == scanf( "%lf%d", &Time, &New_State ) );
	  if ( File_Ok && ( Time <= T_Max ) ) {
	    Write_Fig_Rectangle( Old_Time-T0, Time-T0, State,
				 X_Scale, Y_Scale );
	  }
	  Old_Time = Time;
	  State = New_State;
	  Nb_Read++;
	}

	/* write comment, if any */
	if ( strlen( Title ) > 0 ) {
	  printf("4 0 -1 0 0 0 16 0.0000 6 150 400 0 %d %s\001\n",
		 (int)floor(1.5 * Y_SCALE * Y_Scale),
		 Title );
	}
      }

      return 0;
    }
    else {
      fprintf( stderr, "%s: Error reading Markov chain.\n", 
	       TOOLNAME );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

void Write_Fig_Rectangle( double T1, double T2, int State,
			  double X_Scale, double Y_Scale )
{
  int		Color;
  int		x1;
  int		x2;
  int		y1;
  int		y2;

  Color = Map_Color( State );

  x1 = (int)floor( T1 * X_SCALE * X_Scale );
  x2 = (int)floor( T2 * X_SCALE * X_Scale );
  y1 = 0;
  y2 = (int)floor( Y_SCALE * Y_Scale );

  printf("2 2 0 1 %d %d 0 0 20 0.0 0 0 -1 0 0 5\n", Color, Color );
  printf("\t\t");
  printf("%d %d ", x1, y1 );
  printf("%d %d ", x1, y2 );
  printf("%d %d ", x2, y2 );
  printf("%d %d ", x2, y1 );
  printf("%d %d ", x1, y1 );
  printf("\n");

}

int Map_Color( int State )
{
  int	Res;

  switch( State ) {
  case 0: Res = FIG_GREEN;
    break;
  case 1: Res = FIG_BLUE;
    break;
  case 2: Res = FIG_RED;
    break;
  case 3: Res = FIG_YELLOW;
    break;
  case 4: Res = FIG_CYAN;
    break;
  case 5: Res = FIG_MAGENTA;
    break;
  case 6: Res = FIG_BROWN;
    break;
  case 7: Res = FIG_GOLD;
    break;
  case 8: Res = FIG_BLACK;
    break;
  default: Res = FIG_WHITE;
  }

  return Res;
}
