/************************************************************************/
/* Module providing the function:					*/
/* 	PetriNet *Flatten_PN( PetriNet* Reseau, 			*/
/*		 	      boolean	Expand_Names );			*/
/************************************************************************/

#include "pn.h"

#define TOOLNAME	"Flatten_PN"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

extern char	Find_Type(PetriNet *RdP, Address The_Address );
extern void 	Write_Address( Address The_Add );
extern Address	Expand_Address( PetriNet* Reseau, Address Addr );

void Set_Offsets( 
PetriNet	*RdP,
int		*Cur_Place_Nb,
int		*Cur_Trans_Nb )
{
  int	i;

  for ( i=0; i<RdP->Nb_Submodels; i++ ) {
    RdP->S[i].First_Place = (*Cur_Place_Nb);
    RdP->S[i].First_Trans = (*Cur_Trans_Nb);
    (*Cur_Place_Nb) += RdP->S[i].Submodel->Nb_Places;
    (*Cur_Trans_Nb) += RdP->S[i].Submodel->Nb_Trans;

    Set_Offsets( RdP->S[i].Submodel, Cur_Place_Nb, Cur_Trans_Nb );
  }

}

void Reset_Link_Addresses( 
Subm_Des	*The_Submodel )
{
  int		i;
  int		Place_Offset;
  int		Trans_Offset;
  Pl_Des	The_Place;
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  PetriNet	*RdP;

  Place_Offset = The_Submodel->First_Place;
  Trans_Offset = The_Submodel->First_Trans;

  RdP = The_Submodel->Submodel;

  /* do it for transitions... */
  for ( i=0; i<RdP->Nb_Trans; i++ ) {
    The_Trans = RdP->T[i];
    SCAN( L_Scan, The_Trans.Inputs )
      L_Scan->Dest += Place_Offset;
    SCAN( L_Scan, The_Trans.Outputs ) 
      L_Scan->Dest += Place_Offset;
  }
  /* now for places... */
  for ( i=0; i<RdP->Nb_Places; i++ ) {
    The_Place = RdP->P[i];
    SCAN( L_Scan, The_Place.Inputs ) 
      L_Scan->Dest += Trans_Offset;
    SCAN( L_Scan, The_Place.Outputs ) 
      L_Scan->Dest += Trans_Offset;
  }
  /* and explore submodels */
  for ( i=0; i<RdP->Nb_Submodels; i++ ) {
    Reset_Link_Addresses( &(RdP->S[i]) );
  }
}

void Reset_Routing_Addresses( 
Subm_Des	*The_Submodel )
{
  int		i;
  int		j;
  int		Trans_Offset;
  Routing_Desc	The_Routing;
  PetriNet	*RdP;

  Trans_Offset = The_Submodel->First_Trans;

  RdP = The_Submodel->Submodel;

  for ( i=0; i<RdP->Nb_Places; i++ ) {
    if ( NULL != ( The_Routing = RdP->P[i].Routing ) ) {

      switch( The_Routing->Type ) {
      case 'P':
      case 'D':
	for ( j=0; j<The_Routing->Period; j++ ) {
	  The_Routing->Cycle[j] = Expand_Address( RdP, The_Routing->Cycle[j] );
	  /* Je pense que faire le changement de numero ici est une	*/
	  /* erreur: c'est le travail de Link_Routings(). En tout etat	*/
	  /* de cause, changer le champ "Number" quand l'adresse est 	*/
	  /* de type 'S' est un bug. Je corrige donc l'instruction 	*/
	  /* en attendant une action plus en profondeur.		*/
	  if ( The_Routing->Cycle[j]->Next == NULL ) {
	    /* this is a direct address */
	    The_Routing->Cycle[j]->Number += Trans_Offset;
	  }
	}
	break;
	/* default: do nothing... */
      }
    }
  }
  /* Explore submodels */
  for ( i=0; i<RdP->Nb_Submodels; i++ ) {
    Reset_Routing_Addresses( &(RdP->S[i]) );
  }
}

void Reset_Names( 
Subm_Des	*The_Submodel,
char		*Prefix )
{
  int		i;
  PetriNet	*RdP;
  char		*Buffer;
  char		*Old_Name;

  /* If name is null, put nomething else instead */
  if ( strlen( The_Submodel->Name ) == 0 ) {
    Buffer = (char*) malloc( (unsigned)(3+strlen(Prefix) ) );
    strcpy( Buffer, Prefix );
    strcat( Buffer, "*/");
  }
  else {
    Buffer = (char*) malloc( (unsigned)(2+strlen(Prefix)+strlen(The_Submodel->Name) ) );
    strcpy( Buffer, Prefix );
    strcat( Buffer, The_Submodel->Name );
    strcat( Buffer, "/" );
  }

  RdP = The_Submodel->Submodel;

  for ( i=0; i<RdP->Nb_Places; i++ ) {
    Old_Name = RdP->P[i].Name;
    if ( strlen( Old_Name ) > 0 ) {
      /* Expansion only if name exists */
      RdP->P[i].Name = 
	(char*) malloc( (unsigned)(1+strlen(Old_Name)+strlen(Buffer) ) );
      strcpy( RdP->P[i].Name, Buffer );
      strcat( RdP->P[i].Name, Old_Name );
      free( Old_Name );
    }
  }
  for ( i=0; i<RdP->Nb_Trans; i++ ) {
    Old_Name = RdP->T[i].Name;
    if ( strlen( Old_Name ) > 0 ) {
      RdP->T[i].Name = 
	(char*) malloc( (unsigned)(1+strlen(Old_Name)+strlen(Buffer) ) );
      strcpy( RdP->T[i].Name, Buffer );
      strcat( RdP->T[i].Name, Old_Name );
      free( Old_Name );
    }
  }

  /* Explore submodels */
  for ( i=0; i<RdP->Nb_Submodels; i++ ) {
    Reset_Names( &(RdP->S[i]), Buffer );
  }

  free( Buffer );

}

void Copy_Objects( 
PetriNet	*Top_PN,
Subm_Des	*The_Submodel )
{
  int		i;
  int		Place_Offset;
  int		Trans_Offset;
  PetriNet	*RdP;

  Place_Offset = The_Submodel->First_Place;
  Trans_Offset = The_Submodel->First_Trans;

  RdP = The_Submodel->Submodel;

  /* do it for transitions... */
  for ( i=0; i<RdP->Nb_Trans; i++ ) {
    Top_PN->T[i+Trans_Offset] = RdP->T[i];
  }
  /* now for transitions... */
  for ( i=0; i<RdP->Nb_Places; i++ ) {
    Top_PN->P[i+Place_Offset] = RdP->P[i];
  }
  /* and explore submodels */
  for ( i=0; i<RdP->Nb_Submodels; i++ ) {
    Copy_Objects( Top_PN, &(RdP->S[i]) );
  }
}

/************************************************************************/
/* A function that computes the absolute object number in the expanded	*/
/* Petri net.								*/
/* Care is taken of the fact that the object numbers in linked addresses*/
/* are NOT corrected so as to begin at 0.				*/
/************************************************************************/
int  Find_Number_In_Submodel( 
Subm_Des	*The_Submodel,
Address		The_Address )
{
  Subm_Des	*Next_Subm;
  PetriNet	*The_PN;
  int		i;

  /* Assertion: submodel is never NULL.					 */
  if ( The_Address->Type == 'P' )
    return The_Address->Number + The_Submodel->First_Place;
  else if ( The_Address->Type == 'T' )
    return The_Address->Number + The_Submodel->First_Trans;
  else if ( The_Address->Type == 'S' ) {
    /* find the number of the submodel */
    The_PN = The_Submodel->Submodel;
    for ( i=0; (i < The_PN->Nb_Submodels) 
	    && (The_Address->Number != The_PN->S[i].Subm_Id );
	  i++ );
    if ( The_Address->Number == The_PN->S[i].Subm_Id ) {
      Next_Subm = &(The_Submodel->Submodel->S[i]);
      return Find_Number_In_Submodel( Next_Subm, The_Address->Next );
    }
    else {
      printf("%s: error: illegal submodel id found (%d) when linking",
	     TOOLNAME, The_Address->Number );
      printf(" bindings.\n");
      return -1;
    }
  }
  else {
    printf("%s: error: illegal address type found (%c) when linking",
	   TOOLNAME, The_Address->Type );
    printf(" bindings.\n");
    return -1;
  }
}

void Link_Bindings( 
PetriNet	*Top_PN,
Subm_Des	*The_Submodel )
{
  int		i;
  Binding	B_Scan;
  char		From_Type;
  char		To_Type;
  Liste_Link	New_Link;
  int		From_Number;
  int		To_Number;
  PetriNet	*The_PN;

  The_PN = The_Submodel->Submodel;

  /* edit own bindings */
  SCAN( B_Scan, The_PN->Binding_Table ) {
    if ( Debug ) {
      printf("Adding I/O link ");
      Write_Address( B_Scan->From );
      printf(" -> ");
      Write_Address( B_Scan->To );
    }
    /* Have to find type and number. In two passes for simplicity */
    From_Type = Find_Type( The_PN, B_Scan->From );
    From_Number = Find_Number_In_Submodel( The_Submodel, B_Scan->From );

    /* To_Type probably useless. Computed for debugging purposes.	*/
    To_Type = Find_Type( The_PN, B_Scan->To );
    To_Number = Find_Number_In_Submodel( The_Submodel, B_Scan->To );

    if ( From_Number >= 0 ) {
      /* Now add in output list */
      if ( From_Type == 'T' ) {
	B_Scan->Link->Dest = To_Number;
	B_Scan->Link->Next = Top_PN->T[From_Number].Outputs;
	Top_PN->T[From_Number].Outputs = B_Scan->Link;
	/* Do NOT increase Outdegree. Normally, it has been kept to	*/
	/* its real value since the parsing of the place links.		*/
	/* Top_PN->T[From_Number].Outdegree ++; */
	if ( Debug )
	  printf(": now -> P %d\n", B_Scan->Link->Dest );
      }
      else if ( From_Type == 'P' ) {
	B_Scan->Link->Dest = To_Number;
	B_Scan->Link->Next = Top_PN->P[From_Number].Outputs;
	Top_PN->P[From_Number].Outputs = B_Scan->Link;
	/* Do NOT increase Outdegree. Normally, it has been kept to	*/
	/* its real value since the parsing of the place links.		*/
	/* Top_PN->P[From_Number].Outdegree ++; */
	if ( Debug )
	  printf(": now -> T %d\n", B_Scan->Link->Dest );
      }

      /* and duplicate for input list */
      New_Link = (Liste_Link) malloc( sizeof( struct LINK_LIST) );
      New_Link->Dest = From_Number;
      New_Link->Weight = B_Scan->Link->Weight;
      /* duplicating the data to make things cleaner */
      New_Link->Data = 
	(char*) malloc( (unsigned)(1+strlen( B_Scan->Link->Data )) );
      New_Link->Data = strcpy( New_Link->Data, B_Scan->Link->Data );
      New_Link->Is_Inhibitor = B_Scan->Link->Is_Inhibitor;
      
      if ( From_Type == 'T' ) {
	New_Link->Next = Top_PN->P[B_Scan->Link->Dest].Inputs;
	Top_PN->P[B_Scan->Link->Dest].Inputs = New_Link;
	Top_PN->P[B_Scan->Link->Dest].Indegree ++;
      } else if ( From_Type == 'P' ) {
	New_Link->Next = Top_PN->T[B_Scan->Link->Dest].Inputs;
	Top_PN->T[B_Scan->Link->Dest].Inputs = New_Link;
	Top_PN->T[B_Scan->Link->Dest].Indegree ++;
      }

      if ( New_Link->Weight > 1 )
	Top_PN->Has_Weights = TRUE;
      if ( New_Link->Is_Inhibitor )
	Top_PN->Has_Inhibitors = TRUE;

      if ( Debug )
	printf("%s: bound %c%d to %c%d.\n", TOOLNAME,
	       From_Type, From_Number+1, To_Type, To_Number+1 );
	
    }
  }      

  /* edit bindings in submodels */
  for ( i=0; i<The_PN->Nb_Submodels; i++ )
    Link_Bindings( Top_PN, &(The_PN->S[i]) );
    
}

void Link_Routings( 
PetriNet	*Top_PN,
Subm_Des	*The_Submodel )
{
  int		i;
  int	        j;
  char		Add_Type;
  int		Add_Number;
  Routing_Desc	The_Routing;
  PetriNet	*The_PN;

  The_PN = The_Submodel->Submodel;

  /* edit own routings */
  for ( i=0; i<The_PN->Nb_Places; i++ ) {
    The_Routing = The_PN->P[i].Routing;
    if ( The_Routing != NULL ) {
      switch( The_Routing->Type ) {
      case 'P':
      case 'D':
	for ( j=0; j<The_Routing->Period; j++ ) {
	  Add_Type = Find_Type( The_PN, The_Routing->Cycle[j] );
	  if ( The_Routing->Cycle[j]->Next != NULL ) {
	    /* meaning: this is an address of type 'S' */
	    if ( Debug ) {
	      printf("Routing #%d (", j );
	      Write_Address( The_Routing->Cycle[j] );
	      printf(")");
	    }
	    Add_Number = Find_Number_In_Submodel( The_Submodel, 
						  The_Routing->Cycle[j] );
	    The_Routing->Cycle[j]->Type = Add_Type;
	    The_Routing->Cycle[j]->Number = Add_Number;
	    /* this is a dirty leak, but oh well. */
	    The_Routing->Cycle[j]->Next = (Address)NULL;
	    if ( Debug ) {
		printf(" edited to T%d in ", Add_Number );
		printf(" in place %d.\n", i );
	    }
	  }
	}
	break;
      default:
	/* nothing to do */
	break;
      }
    }
  }
      
  /* edit routings in submodels */
  for ( i=0; i<The_PN->Nb_Submodels; i++ )
    Link_Routings( Top_PN, &(The_PN->S[i]) );

}

void Link_Addresses( 
Address		The_Address,
Subm_Des	*The_Submodel )
{
  char		The_Type;
  int		The_Number;

  /* First expand the address */
  Expand_Address( The_Submodel->Submodel, The_Address );

  /* Then find its type and number */
  The_Type = Find_Type( The_Submodel->Submodel, The_Address );
  The_Number = Find_Number_In_Submodel( The_Submodel, The_Address );
  The_Address->Type = The_Type;
  The_Address->Number = The_Number;
  The_Address->Next = (Address)NULL;

}

PetriNet	*Flatten_PN(PetriNet	*Reseau,
			    boolean	Expand_Names )
{
  int		i;
  int		Tot_Nb_Places;
  int		Tot_Nb_Trans;
  PetriNet	*Last_PN;
  Subm_Des	Last_Submodel;
  Subm_Des	*Top_Submodel;
  int		Nb_Places;
  int		Nb_Trans;

  if ( Reseau->Nb_Submodels == 0 )
    return Reseau;
  else {

    /* Creation of a dummy top-level submodel to uniformize things.	*/
    Top_Submodel = (Subm_Des*) malloc( sizeof( struct SUBM_DES ) );
    Top_Submodel->Subm_Id = 0;
    Top_Submodel->Father = (PetriNet*)NULL;
    Top_Submodel->Submodel = Reseau;
    Top_Submodel->Name = "Top_Model";
    Top_Submodel->File_Name = Reseau->File_Name;
    Top_Submodel->First_Place = 0;
    Top_Submodel->First_Trans = 0;

    /* First go through the submodels, setting the offset numbers for	*/
    /* places and transitions.						*/
    Nb_Places = Reseau->Nb_Places;
    Nb_Trans = Reseau->Nb_Trans;
    Set_Offsets( Reseau, &Nb_Places, &Nb_Trans );

    /* Now go through submodels, renumbering the ids of internal links	*/
    for ( i=0; i<Reseau->Nb_Submodels; i++ )
      Reset_Link_Addresses( &(Reseau->S[i]) );

    /* Now go through submodels, renumbering ids in routing tables	*/
    Reset_Routing_Addresses( Top_Submodel );

    /* Now go through submodels, renaming objects			*/
    if ( Expand_Names )
      for ( i=0; i<Reseau->Nb_Submodels; i++ )
	Reset_Names( &(Reseau->S[i]), "" );

    /* Copy all new transitions into the root network.			*/
    /* Find first total number of places/trans: in the "last" subnet.	*/
    Last_PN = Reseau;
    Last_Submodel = Last_PN->S[Last_PN->Nb_Submodels-1];
    Last_PN = Last_Submodel.Submodel;
    if ( Last_PN->Nb_Submodels > 0 ) {
      do {
	Last_Submodel = Last_PN->S[Last_PN->Nb_Submodels-1];
	Last_PN = Last_Submodel.Submodel;
      }
      while ( Last_PN->Nb_Submodels > 0 );
    }
    Tot_Nb_Places = Last_Submodel.First_Place + Last_PN->Nb_Places;
    Tot_Nb_Trans = Last_Submodel.First_Trans + Last_PN->Nb_Trans;

    /* Recursively transform routings into hard addresses. */
    Link_Routings( Reseau, Top_Submodel );

    Reseau->Nb_Trans = Tot_Nb_Trans;
    Reseau->Nb_Places =Tot_Nb_Places;
  
    Reseau->T = (Tr_Des*) realloc( Reseau->T, 
				   (size_t)(Tot_Nb_Trans*sizeof( Tr_Des ) ) );
    Reseau->P = (Pl_Des*) realloc( Reseau->P, 
				   (size_t)(Tot_Nb_Places*sizeof( Pl_Des ) ) );

    /* recursively move places and transitions into top level net     	*/
    for ( i=0; i<Reseau->Nb_Submodels; i++ )
      Copy_Objects( Reseau, &(Reseau->S[i]) );

    /* Recursively transform bindings into hard addresses. */
    Link_Bindings( Reseau, Top_Submodel );

    /* flatten I/O also */
    for ( i=0; i<Reseau->Nb_Inputs; i++ )
      Link_Addresses( Reseau->Inputs[i], Top_Submodel  );
    for ( i=0; i<Reseau->Nb_Outputs; i++ )
      Link_Addresses( Reseau->Outputs[i], Top_Submodel );

    /* Finally, remove references to submodels */
    Reseau->Nb_Submodels = 0;

  }

  return Reseau;
}

