#include "pn.h"

#define	TOOLNAME	"gen_c++"

#define FILENAME_LG	256

#define CODE(x)		fprintf(OutFile,x)
#define CODE1(x,y)	fprintf(OutFile,x,y)
#define CODE2(x,y,z)	fprintf(OutFile,x,y,z)
#define CODE3(x,y,z,t)	fprintf(OutFile,x,y,z,t)
#define COMLINE		CODE("//============================================================\n");
#define CODELN		CODE("\n")

extern int		Buf_Size;
extern boolean		*Is_Source;
extern boolean		*Is_Compet_Pl;
extern char		*Prefix;

/* Externs from the module parse_pn.o */
extern char		*todays_date();
extern char		Find_Type(PetriNet *RdP, Address The_Address );

/* Externs from the module pn_to_c.c */
extern void	Gen_Defines( FILE	*OutFile,
			     PetriNet	*Reseau,
			     int	Max_Delay,
			     char	*Class_Name,
			     boolean	C_Plus_Plus,
			     boolean	Prosit );
void	Gen_Includes(FILE	*OutFile,
		     PetriNet	*Reseau,
		     boolean	C_Plus_Plus,
		     boolean	Prosit,
		     boolean	Distributed );
void	Gen_Variables(
FILE		*OutFile, 
PetriNet	*Reseau,
int		Max_Delay,
int		Scale,
char		*Class_Name,
boolean		C_Plus_Plus,
boolean		Prosit );

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

void	Gen_Class_File_Header( FILE	*OutFile,
			       PetriNet	*Reseau, 
			       char	*Path,
			       boolean	Proxy )
{
  char	*The_Date;

  The_Date = todays_date();

  CODE ("//=========================\n");
  CODE ("//  COPYRIGHT INRIA 1997 ==\n");
  CODE ("//=========================\n");
  COMLINE;
  CODE1("// File        : %s\n", Path );
  CODE2("// Author      : %s %s\n", TOOLNAME, VERSION );
  if ( strlen( Reseau->File_Name ) > 0 )
    CODE1("// Origin      : %s\n", Reseau->File_Name );
  CODE1("// Date        : %s\n", The_Date);
  CODE ("// Description : A class encapsulating the equational simulation\n");
  CODE ("//               of a Petri Net.\n");
  if ( Proxy )
    CODE("//               Proxy for the distributed PROSIT.\n");
  COMLINE;
  CODELN;

  free( The_Date );

}

void Gen_Class_Methods(
FILE		*OutFile,
PetriNet	*Reseau,
int		Max_Delay,
char		*Class_Name,
boolean		Prosit,
boolean		Distributed,
boolean		Single_Marking_Function )
{
  int		i;
  int		Nb_Sources;
  int		Nb_Compets;

  /* initializers */
  CODE ("public:\n");
  if ( Prosit ) {
    CODE1("  %s(Time activation=0, Time step=1.0, int stacksize = 0x8000 );\n",
	  Class_Name );
  }
  CODE1("  ~%s();\n", Class_Name );
  CODELN;

  if ( Distributed ) { /* implies Prosit! */
    CODE("  static Comm_Object *init_cstr(){\n");
    CODE1("    return new %s(0,1,0x8000);\n", Class_Name );
    CODE("  }\n");
    CODELN;
  }

  /* private methods not inherited from EQ_Thing */
  CODE("protected:\n");
  CODE("    virtual void Init();\n");
  if ( Prosit )
    CODE("    virtual void Arrival(int Num_Class, double Arr_Time);\n");

  CODE("    virtual void Increment();\n");
  CODE("    virtual void Increment_Initial();\n");
  CODE("    virtual void Increment_Regular();\n");

  if ( Prosit )
    CODE1("    virtual %sRouting_Server *Next_Departure();\n",
	  ( Distributed ? "CM_" : "" ) );

  CODE("    virtual void Terminate();\n");
  CODE("    virtual void Compute_Stat();\n");
  CODE("    virtual void Report_Stat(ostream &sortie);\n");  
  CODE("    virtual void Report_Trans_Stat(ostream&,int,char*,int,int);\n");  
  CODELN;

  /* private methods related to sources */
  Nb_Sources = 0;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Is_Source[i] )
      Nb_Sources++;

  if ( Nb_Sources > 0 ) {
    CODE("private:\n");
    CODE("    void Re_Init_Sources( int From, int To );\n");
  }

  /* private methods related to competition */
  Nb_Compets = 0;
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Is_Compet_Pl[i] )
      Nb_Compets++;

  if ( Nb_Compets > 0 ) {
    CODE("private:\n");
    if ( Single_Marking_Function ) {
      CODE1("    int I_Marking( int Place_Id, int The_Time  );\n", i );
      CODE1("    int Marking( int Place_Id );\n", i );
    }
    else {
      for ( i=0; i<Reseau->Nb_Places; i++ )
	if ( Is_Compet_Pl[i] ) {
	  CODE1("    int I_Marking_%d( int The_Time  );\n", i );
	  CODE1("    int Marking_%d( );\n", i );
	}
    }
  }


  /* public methods for input/output management */
  CODE("public:\n");
  if ( Prosit ) {
    CODE("    int Register_Input(int);\n");
    CODE1("    bool Register_Output(int,%sRouting_Server*);\n",
	  ( Distributed ? "CM_" : "" ) );
  }
  CODELN;
}

void	Gen_Class_Members(
FILE		*OutFile,
PetriNet	*Reseau,
int		Max_Delay,
int		Scale,
char		*Class_Name,
boolean		Prosit,
boolean		Distributed )
{
  /* variables that do not depend (or little!) on network structure */
  if ( Prosit &&
       ( ( Reseau->Nb_Inputs > 0 ) || ( Reseau->Nb_Outputs > 0 ) ) ) {
    CODE("protected:\n");
    if ( Reseau->Nb_Inputs > 0 ) {
      CODE ("    int\tNb_Input_Pl;\n");
      CODE ("    int\tNb_Input_Tr;\n");
      CODE1("    int\tOrig_Nb[%d];\n", Reseau->Nb_Inputs );
      CODE1("    int\t*Orig_Table[%d];\n", Reseau->Nb_Inputs );
      CODE ("    int\t**Ext_H;\n");
      CODE ("    int\t**Ext_Y;\n");
      CODE1("    bool\tIs_Input_Trans[%d];\n", Reseau->Nb_Trans);
      CODE1("    bool\tIs_Input_Place[%d];\n", Reseau->Nb_Places);
    }
    if ( Reseau->Nb_Outputs > 0 ) {
      CODE1("    int Dest_Nb[%d];\n", Reseau->Nb_Outputs );
      CODE2("    %sRouting_Server **Dest_Table[%d];\n", 
	    ( Distributed ? "CM_" : "" ),
	    Reseau->Nb_Outputs );
    }
  }
  CODELN;
  /* variables that do depend on network structure */
  CODE("protected:\n");
  CODELN;
  Gen_Variables( OutFile, Reseau, Max_Delay, Scale, Class_Name,
		 /* C_Plus_Plus= */ TRUE, /* prosit= */ TRUE );

}

int	Gen_Class_Prototype(
FILE		*OutFile,
PetriNet	*Reseau,
int		Max_Dependency,
int		Scale,
char		*Class_Name,
boolean		Prosit,
boolean		Distributed,
boolean		Single_Marking_Function )
{
  Gen_Includes( OutFile, Reseau, TRUE, Prosit, Distributed );
  CODELN;

  CODE1("#include \"%sEQ_Thing.hh\"\n",	( Distributed ? "CM_" : "" ) );
  CODELN;

  Gen_Defines( OutFile, Reseau, Max_Dependency, Class_Name, TRUE, Prosit );
  CODELN;

  CODE2("class %s : public %sEQ_Thing {\n", 
	Class_Name,
	( Distributed ? "CM_" : "" ) );

  CODELN;

  Gen_Class_Methods( OutFile, Reseau, Max_Dependency, 
		     Class_Name, Prosit, Distributed, 
		     Single_Marking_Function );

  Gen_Class_Members( OutFile, Reseau, Max_Dependency, Scale,
		     Class_Name, Prosit, Distributed );

  CODE("};\n");

  return TRUE;
}

int	Gen_Class_Proxy_Prototype(
FILE	*OutFile,
char	*Class_Name )
{
  CODE1("#ifndef P_%s_H\n", Class_Name );
  CODE1("#define P_%s_H\n", Class_Name );
  CODELN;

  CODE("#include \"Proxy/PCM_EQ_Thing.hh\"\n");
  CODELN;
  CODE("class Cluster;\n");
  CODE("class DSim_Obj;\n");
  CODE("class Proxy_Comm_Object;\n");
  CODE("class PCM_Routing_Server;\n");
  CODELN;

  CODE2("class P_%s : public PCM_EQ_Thing {\n", Class_Name, Class_Name );
  CODELN;
  CODE ("public:\n");
  CODE2("  P_%s(Cluster *destination):PCM_EQ_Thing(destination,\"%s\"){};\n",
	Class_Name, Class_Name );
  CODE1("  P_%s(DSim_Obj *local_obj):PCM_EQ_Thing(local_obj){};\n",
	Class_Name );
  CODE1("  P_%s(Entity_Id &id,Cluster *c,Proxy_Comm_Object *pco):",
	Class_Name );
  CODE("PCM_EQ_Thing(id,c,pco){};\n");
  CODELN;

  CODE("};\n");

  CODE1("#endif /* ifndef P_%s_H */\n", Class_Name );

  return TRUE;
}


void Gen_Class_Cons_Des(
FILE		*OutFile,
PetriNet	*Reseau,
char		*Class_Name,
boolean		Distributed )
{
  CODE("// Constructor\n");
  CODE2("%s::%s( Time activation, Time step, int stacksize  ) :\n", 
	Class_Name, Class_Name );
  CODE1("\t%sEQ_Thing( activation, step, stacksize ) {\n",
	( Distributed ? "CM_" : "" ) );
  CODE("  Init( );\n");
  CODE("}\n");
  CODELN;
  CODE("// Destructor\n");
  CODE2("%s::~%s( ) {\n", Class_Name, Class_Name );
  CODE("  Terminate( );\n");
  CODE("}\n");
  CODELN;
}

void Gen_Input_Handling(
FILE		*OutFile,
PetriNet	*Reseau,
int		Max_Dependency,
char		*Class_Name )
{
  int	Nb_Obj;
  int	i;

  CODE1("int  %s::Register_Input( int Input_Id )\n", Class_Name );
  CODE("{\n");
  CODELN;
  CODE("  int Res;\n");
  CODELN;
  CODE1("  if ( ( Input_Id <= 0 ) || ( Input_Id > %d ) ) {\n",
	Reseau->Nb_Inputs );
  CODE("    printf(\"Error. Illegal Input number (%%d).\\n\", Input_Id );\n");
  CODE("    return -1;\n");
  CODE("  }\n");
  if ( Reseau->Nb_Inputs > 0 ) {
    CODE("  else {\n");
    CODE("    Input_Id--;\n");
    CODE("    switch( Input_Id ) {\n");
  
    /* counting the number of transitions that are inputs */
    Nb_Obj = 0;
    for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
      if ( 'T' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	Nb_Obj++;
    }
    if ( Nb_Obj > 0 ) {
      CODE("    // transitions\n");
      for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
	if ( 'T' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	  CODE1("    case %2d:\n", i );
      }
      CODE("      Orig_Nb[ Input_Id ]++;\n");
      CODE("      Orig_Table[ Input_Id ] = \n");
      CODE("        (int*) realloc( Orig_Table[ Input_Id ], \n");
      CODE("        (unsigned) Orig_Nb[ Input_Id ] ");
      CODE("* sizeof( int ) );\n");
      CODE("      Orig_Table[ Input_Id] [ Orig_Nb[ Input_Id ]-1 ] ");
      CODE("= Nb_Input_Pl;\n");
      CODE("      // Creation and initialization of the virtual ");
      CODE("place counter\n");
      CODE("      Ext_H = (int**)realloc( Ext_H,");
      CODE("(unsigned)(1+Nb_Input_Pl)*sizeof(int*) );\n");
      CODE("      Ext_H[Nb_Input_Pl] = ");
      CODE1("(int*)calloc( %sBUF_SIZE, sizeof(int) );\n", Prefix );
      CODE1("      for( int i=0; i<%sBUF_SIZE; i++ )\n", Prefix );
      CODE("        Ext_H[Nb_Input_Pl][i] = 0;\n");
      CODE("      Res = Nb_Input_Pl;\n");
      CODE("      Nb_Input_Pl ++;\n");
      CODE("    break;\n");
    }
    /* counting the number of places that are inputs */
    Nb_Obj = 0;
    for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
      if ( 'P' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	Nb_Obj++;
    }
    if ( Nb_Obj > 0 ) {
      CODE("    // places\n");
      for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
	if ( 'P' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	  CODE1("    case %2d:\n", i );
      }
      CODE("      if ( Orig_Nb[ Input_Id ] > 0 ) {\n");
      CODE("        // already connected. Just return channel number\n");
      CODE("        Res = Orig_Table[ Input_Id ][0];\n");
      CODE("      }\n");
      CODE("      else {\n");
      CODE("        Orig_Nb[ Input_Id ]++;\n");
      CODE("        Orig_Table[ Input_Id ] = \n");
      CODE("          (int*) realloc( Orig_Table[ Input_Id ], \n");
      CODE("          (unsigned) Orig_Nb[ Input_Id ] * sizeof( int ) );\n");
      CODE("        Orig_Table[ Input_Id] [ Orig_Nb[ Input_Id ]-1 ] ");
      CODE("= Nb_Input_Tr;\n");
      CODE("        // Creation and initialization of the virtual ");
      CODE("transition counter\n");
      CODE("        Ext_Y = (int**)realloc( Ext_Y,");
      CODE("(unsigned)(1+Nb_Input_Tr)*sizeof(int*) );\n");
      CODE("        Ext_Y[Nb_Input_Tr] = ");
      CODE("(int*)calloc( BUF_SIZE, sizeof(int) );\n");
      CODE("        for( int i=0; i<BUF_SIZE; i++ )\n");
      CODE("          Ext_Y[Nb_Input_Tr][i] = 0;\n");
      CODE("        Res = Nb_Input_Tr;\n");
      CODE("        Nb_Input_Tr ++;\n");
      CODE("      }\n");
      CODE("    break;\n");
    }
    CODE("    } // switch\n");
    CODELN;
    /* I am not proud of this way of doing it, but well... */
    for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
      if ( 'T' == Find_Type( Reseau, Reseau->Inputs[i] ) ) {
	CODE1("    if ( Input_Id == %d )\n", i );
	CODE1("      Is_Input_Trans[%d] = TRUE;\n", Reseau->Inputs[i]->Number );
      }
      else if ( 'P' == Find_Type( Reseau, Reseau->Inputs[i] ) ) {
	CODE1("    if ( Input_Id == %d )\n", i );
	CODE1("      Is_Input_Place[%d] = TRUE;\n", Reseau->Inputs[i]->Number );
      }
    }
    CODELN;
    CODE("    return Res;\n");
    CODE("  }\n");
  }
  CODE("}\n");
  CODELN;
}
  
void Gen_Output_Handling(
FILE		*OutFile,
PetriNet	*Reseau,
int		Max_Dependency,
char		*Class_Name,
boolean		Distributed )
{
  int	Nb_Obj;
  int	i;

  CODE2("bool %s::Register_Output( int Output_Id, %sRouting_Server *Dest )\n", 
	Class_Name,
	( Distributed ? "CM_" : "" ) );
  CODE("{\n");
  CODELN;
  CODE("  bool Res;\n");
  CODELN;

  CODE1("  if ( ( Output_Id <= 0 ) || ( Output_Id > %d ) ) {\n",
	Reseau->Nb_Outputs );
  CODE("    printf(\"Error. Illegal Output number (%%d).\\n\", Output_Id );\n");
  CODE("    Res = FALSE;\n");
  CODE("  }\n");
  CODE("  else {\n");
  CODE("    Output_Id--;\n");
  CODE("    switch( Output_Id ) {\n");
  /* counting the number of transitions that are outputs */
  Nb_Obj = 0;
  for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
    if ( 'T' == Find_Type( Reseau, Reseau->Outputs[i] ) )
      Nb_Obj++;
  }
  if ( Nb_Obj > 0 ) {
    CODE("    /* transitions */\n");
    for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
      if ( 'T' == Find_Type( Reseau, Reseau->Outputs[i] ) )
	CODE1("    case %2d:\n", i );
    }
  
    CODE("      Dest_Table[ Output_Id ] = \n");
    CODE1("        (%sRouting_Server**) realloc( Dest_Table[ Output_Id ], \n",
	  ( Distributed ? "CM_" : "" ) );
    CODE ("        (unsigned) ( 1 + Dest_Nb[ Output_Id ] ) * ");
    CODE1("sizeof( %sRouting_Server* ) );\n",
	  ( Distributed ? "CM_" : "" ) );
    CODE("      Dest_Table[ Output_Id ] [ Dest_Nb[ Output_Id ] ] = Dest;\n");
    CODE("      Dest_Nb[ Output_Id ]++;\n");
    CODE("      Res = TRUE;\n");
    CODE("    break;\n");
  }
  /* counting the number of places that are outputs */
  Nb_Obj = 0;
  for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
    if ( 'P' == Find_Type( Reseau, Reseau->Outputs[i] ) )
      Nb_Obj++;
  }
  if ( Nb_Obj > 0 ) {
    CODE("    /* places */\n");
    for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
      if ( 'P' == Find_Type( Reseau, Reseau->Outputs[i] ) )
	CODE1("    case %2d:\n", i );
    }
    CODE("      if ( Dest_Nb[ Output_Id ] > 0 ) {\n");
    CODE("      // already connected. Just return channel number\n");
    CODE("      printf(\"Error. Cannot have more than one output \");\n");
    CODE("      printf(\"for place output (#%%d).\\n\", Output_Id );\n");
    CODE("      Res = FALSE;\n");
    CODE("      }\n");
    CODE("      else {\n");
    CODE("      Dest_Table[ Output_Id ] = \n");
    CODE("        (Routing_Server**) realloc( Dest_Table[ Output_Id ], \n");
    CODE("        (unsigned) Dest_Nb[ Output_Id ] * sizeof( int ) );\n");
    CODE("      Dest_Table[ Output_Id] [ Dest_Nb[ Output_Id ] ] = Dest;\n");
    CODE("      Dest_Nb[ Output_Id ]++;\n");
    CODE("      Res = TRUE;\n");
    CODE("      }\n");
    CODE("    break;\n");
  }
  CODE("    }\n");
  CODE("  }\n");
  CODELN;
  CODE("  return Res;\n");
  CODE("}\n");
}
