#include <stdio.h>

#define CODE(x)			fprintf(OutFile,x)
#define CODE1(x,y)		fprintf(OutFile,x,y)
#define CODE2(x,y,z)		fprintf(OutFile,x,y,z)
#define CODE3(x,y,z,t)		fprintf(OutFile,x,y,z,t)
#define CODE4(x,y,z,t,u)	fprintf(OutFile,x,y,z,t,u)
#define CODELN		CODE("\n")
#define COMLINE		CODE("//=========================================\n");

extern int	*Delay_Table;
extern boolean	Y_Static;

void Gen_Loop_Switch_Code( FILE		*OutFile,
			   PetriNet	*Reseau,
			   int		Max_Dependency,
			   boolean	C_Plus_Plus,
			   char		*Prefix )
{
  int		i;
  Liste_Link	L_Scan;
  boolean	Some;

  if ( C_Plus_Plus ) {
    COMLINE;
    CODE("// Procedure for incrementing the firing count vector\n");
    CODE("// Switch according to what time it is...\n");
    COMLINE;
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t\t\n");
    CODE("/* Switch according to what time it is...\t\t\t*/\n");
  }
  CODELN;

  if ( C_Plus_Plus )
    CODE1("void %s::", Prefix );
  else 
    CODE("void ");
    
  CODE("Increment( )\n{\n");
  CODE1("  if ( Real_EQ_Time >= %d )\n", Max_Dependency );
  CODE("    Increment_Regular();\n" );
  CODE("  else\n");
  CODE("    Increment_Initial();\n");
  CODELN;
  CODE("  /* Increment of the time variables */\n");
  CODE("  Real_EQ_Time ++;\n");
  CODE1("  EQ_Time = ( EQ_Time + 1 ) %% %sBUF_SIZE;\n", Prefix );
  for (i=1; i<=Max_Dependency; i++ )
    if ( Delay_Table[i] )
      CODE3("  EQ_Time_M_%d = ( EQ_Time + %d ) %% %sBUF_SIZE;\n", 
	    i, Buf_Size - i, Prefix );
  if ( Y_Static ) {
    CODE("  Y_0 = Y[ EQ_Time ];\n");
    for (i=1; i<=Max_Dependency; i++ )
      if ( Delay_Table[i] )
	CODE2("  Y_Time_M_%d = Y[ EQ_Time_M_%d ];\n", i, i );
    CODELN;
  }

  CODE("}\n\n");
}

void	Gen_Preparation_Code(FILE	*OutFile,
			     PetriNet	*Reseau,
			     boolean	Initial,
			     boolean	C_Plus_Plus,
			     char	*Prefix )
{
  int		i;
  int		Nb_Sources;
  Liste_Link	L_Scan;

  /* check first if there are some non tagged sources */
  Nb_Sources = 0;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( ( Reseau->T[i].Tag == EQ_OBJ ) && Is_Source[i] )
      Nb_Sources++;

  if ( Nb_Sources > 0 ) {
    CODE ("  /* Draw source arrivals when necessary */\n");
    CODE1("  if ( EQ_Time %% %sHALF_BUF_SIZE == 0 ) {\n", Prefix );
    CODE ("    if ( EQ_Time == 0 )\n");
    CODE1("      Re_Init_Sources( 0, %sHALF_BUF_SIZE );\n", Prefix );
    CODE ("    else\n");
    CODE2("      Re_Init_Sources( %sHALF_BUF_SIZE, %sBUF_SIZE );\n", 
	  Prefix, Prefix );
    CODE ("  }\n");
    CODELN;
  }

  if ( C_Plus_Plus )
    CODE("  // Copy present firing count for everybody\n");
  else
    CODE("  /* Copy present firing count for everybody  */\n");

  if ( C_Plus_Plus ) {
    CODE1("  for ( int i=0; i<%sNB_TRANS; i++ )\n", Prefix );
    if ( Initial )
      CODE("    Y[EQ_Time][ i ] = ( EQ_Time>0 ? Y[EQ_Time-1][ i ] : 0 );\n");
    else if ( Y_Static )
      CODE("    Y_0[ i ] = Y_Time_M_1[ i ];\n");
    else
      CODE("    Y[ EQ_Time ][ i ] = Y[ EQ_Time_M_1 ][ i ];\n");
  }
  else {
    for( i=0; i<Reseau->Nb_Trans; i++ )
      if ( !Is_Source[i] ) {
	if ( Initial )
	  CODE2("  Y[EQ_Time][%d] = ( EQ_Time>0 ? Y[EQ_Time-1][%d] : 0 );\n",
		i, i );
	else if ( Y_Static )
	  CODE2("  Y_0[%d] = Y_Time_M_1[%d];\n", i, i );
	else
	  CODE2("  Y[ EQ_Time ][%d] = Y[ EQ_Time_M_1 ][%d];\n", i, i );
      }
  }

  CODELN;
}

void	Gen_Increment_Code( FILE *OutFile, 
			    int	Max_Delay )
{
  int	i;

  CODE("/* Increment of the time variables */\n");
  CODE("Time = ( Time + 1 ) %% BUF_SIZE;\n");
  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] )
      CODE2("Time_M_%d = ( Time + %d ) %% BUF_SIZE;\n", 
	    i, Buf_Size - i );
  if ( Y_Static ) {
    CODE("Y_0 = Y[ Time ];\n");
    for (i=1; i<=Max_Delay; i++ )
      if ( Delay_Table[i] )
	CODE2("Y_Time_M_%d = Y[ Time_M_%d ];\n", i, i );
    CODELN;
  }
}

void 	Gen_Loop_Code(FILE		*OutFile,
		      PetriNet		*Reseau,
		      Layer_Struct	The_Layer_Struct,
		      boolean		Initial,
		      boolean		C_Plus_Plus,
		      boolean		Single_Marking_Function )
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;
  int	Time_Scale;
  int	i;
  int	j;
  int	Nb_Nodes;

  for( i=1; i<=The_Scales[0]; i++ ) {
    Time_Scale = The_Scales[i];

    /* find the number of nodes within this time scale */
    Nb_Nodes = 0;
    for ( j=0; j<Reseau->Nb_Trans; j++ )
      if ( !Is_Source[j] && ( TST[j] == Time_Scale ) )
	Nb_Nodes++;

    if ( Nb_Nodes != 0 ) {
      CODE3("  /* Computing the %d transition%s with time scale %d */\n", 
	    Nb_Nodes,
	    ( Nb_Nodes > 1 ? "s" : "" ),
	    Time_Scale );

      if ( Time_Scale != 1 ) {
	CODE1("  if ( ( Real_EQ_Time %% %d ) == 0 ) {\n\n", Time_Scale );
	Indent_Nb += 2;
      }

      for( Num_Layer = 0; Num_Layer < The_Layer_Struct->Nb_Levels; 
	   Num_Layer++)
	for ( Num_Node = 0; 
	      Num_Node < The_Layer_Struct->Level[Num_Layer]->Nb_Nodes;
	      Num_Node++ ) {
	    
	  The_Node = The_Layer_Struct->Level[Num_Layer]->Node_Id[Num_Node];
	  if ( ( Reseau->T[The_Node].Tag == EQ_OBJ ) 
	       && ( !Is_Source[The_Node] || C_Plus_Plus )
	       && ( TST[The_Node] == Time_Scale ) )
	    gen_code_node( Reseau, The_Node, Initial, C_Plus_Plus,
			   Single_Marking_Function );
	}

      if ( Time_Scale != 1 ) {
	CODE1("  } /* end of time scale %d */\n\n", Time_Scale );
	Indent_Nb -= 2;
      }
    }
  }

  /* Now generate code for places. This is necessary for EQ places that	*/
  /* have Prosit output(s). At the same time, check if all routing	*/
  /* places have been visited.						*/
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( ( Reseau->P[i].Tag == EQ_OBJ ) 
	 && !Already_Visited[i] 
	 && Has_Prosit_Trans_Succ( Reseau, i ) ) {

      Write_Token_Total( Reseau, i, Initial, C_Plus_Plus );
      /* call Gen_EQ_H_Function with "-1" as trans. id. This is ok as 	*/
      /* long as routing is not 'P', which should be prevented at 	*/
      /* checking time.							*/
      Gen_EQ_H_Function( Reseau, i, -1, Initial,Single_Marking_Function );

    }
  }

}

void Gen_Loop_Code(FILE		*OutFile,
		   PetriNet	*Reseau,
		   boolean	Initial,
		   Layer_Struct	The_Layer_Struct,
		   int		Max_Dependency,
		   boolean	C_Plus_Plus,
		   char		*Prefix,
		   boolean	Single_Marking_Function )
{
  if ( C_Plus_Plus ) {
    CODE("// Procedure for incrementing the firing count vector\n");
    if ( Initial ) {
      CODE("// Version to handle initial firings\n" );
      CODE1("void %s::Increment_Initial()\n{\n", Prefix );
    }
    else {
      CODE("// Version to handle regular firings\n" );
      CODE1("void %s::Increment_Regular()\n{\n", Prefix );
    }
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t*/\n");
    if ( Initial ) {
      CODE("/* Version to handle initial firings\t\t\t*/\n" );
      CODE("void Increment_Initial()\n{\n");
    }
    else {
      CODE("/* Version to handle regular firings\t\t\t*/\n" );
      CODE("void Increment_Regular()\n{\n");
    }
  }
  CODE("  static int\tj;\n");
  CODE("  static int\tTmp_Y;\n");
  CODE("  static double\tU;\n");
  if ( C_Plus_Plus )
    CODE("  static int\tIn_Nb;\n");
  CODELN;

  Gen_Preparation_Code( OutFile, Reseau, Initial, C_Plus_Plus, Prefix );
  Gen_Loop_Code( OutFile, Reseau, The_Layer_Struct, Initial, C_Plus_Plus,
		 Single_Marking_Function );
  Gen_Increment_Code( OutFile, Reseau, Initial );

  CODELN;
  CODE("} /* end of loop code */\n\n");
}
