/* programme/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un r\'eseau de Petri en "temps discret" */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "pn.h"

#define MAXINT          0x7FFFFFFF
#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"pn2c"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

/* Definitions and declarations for code generation */
int	Indent_Nb = 0;
void	Indent();		/* declared below */

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODELN		CODE("\n")

/* Externs from the module parse_pn.o */
extern PetriNet		*Read_Network();
extern void		dumpe_network();
extern char		*todays_date();

/* Externs from the module analyze_pn.o */
extern boolean		topological_analysis();
extern boolean		Has_All_Deterministic_Places();
extern boolean		Has_Integral_Transitions();
extern boolean		Has_All_Routing_Places();
extern int		Find_Time_Scale();
extern void		Dump_Layer_Struct();
extern double		Max_Lag();
extern double		Max_Delay();

/* Forward declarations */
boolean Is_Dummy();
int	gen_c_code();
void	TimeScale();

int	Buf_Size = 1024;	/* default buffer size */
int	Scale = 10;		/* for integrality check */

int	*Hold_Time;
int	*Fire_Time;
boolean	*Is_Compet_Pl;		/* flag for places involved in a conflict */
boolean	*Is_Compet_Tr;		/* flag for transitions involved */
boolean	*Is_Source;
boolean	*Already_Visited;
boolean	*Delay_Table;
boolean	Has_Competition = FALSE;
boolean Y_Static = TRUE;
int	Trace_Start = 0;	/* trace control parameters 	*/
int	Trace_End = 100;	/* with their initial values	*/
char	*Tmp_Dir;		/* where to put the files 	*/
int 	*TSP;          		/* table of place time scales 	*/
int 	*TST;          		/* table of trans. time scales	*/
int	*The_Scales;		/* table of different time sc.	*/

main(argc, argv)
int	argc;
char	*argv[];
{
  static char		File_Name[128] = "";
  static char		Out_File_Name[128] = "";
  PetriNet 		*Reseau;
  boolean		Trace = FALSE;
  boolean		Testing = FALSE;
  boolean		For_GSS = FALSE;
  int			i;
  int			Max_Time = -1;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Network_Ok = TRUE;
  Layer_Struct		The_Layer_Struct;
  FILE			*OutFile;
  int			*Bad_Objects;

  Tmp_Dir = (char*) malloc( sizeof( char ) );
  strcpy( Tmp_Dir, "" );

  i = 1;

  if (argc>1)
    do {

      if GET_INT( "-bs", Buf_Size )
      else if GET_INT( "-scaling", Scale )
      else if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_INT( "-tmax", Max_Time )
      else if GET_INT_SWITCH( "-tb", Trace_Start, Trace, TRUE )
      else if GET_INT_SWITCH( "-te", Trace_End, Trace, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-testing", Testing, TRUE )
      else if SWITCH( "-gss", For_GSS, TRUE )
      else if SWITCH( "-noYstatic", Y_Static, FALSE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-bs <buf_size>] [-f <file>] ", argv[0]);
	printf("[-scaling <int>]\n\t[-tb <time>] [-te <time>] ");
	printf("[-tmp <dir>] [-tmax <int>] [-testing]\n\t[-gss] [-o <file>] ");
	printf("[-noYstatic] [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmp", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Tmp_Dir = (char*) realloc( Tmp_Dir, 
				     2 + (unsigned)strlen( argv[i] ) * 
				     sizeof( char ) );
	  strcpy( Tmp_Dir, argv[i] );
	  if ( ( strlen( Tmp_Dir ) != 0 )
	       && (Tmp_Dir[strlen(Tmp_Dir)-1] != '/' ) )
	    strcat( Tmp_Dir, "/" );
	  i++;
	}
	else {  
	  printf("%s: directory expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {
    if ( strlen(File_Name) != 0 )
      File_Ok = ( NULL != freopen( File_Name, "r", stdin ) );

    if ( File_Ok ) {
      Reseau = Read_Network();

      if ( Reseau != NULL ) {
	invert_network( Reseau );

	if ( For_GSS )
	  Testing = TRUE;

	if ( Testing ) {
	  if ( strlen( Out_File_Name ) != 0 )
	    File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );
	  OutFile = stdout;

	  if ( File_Ok && For_GSS ) {
	    fprintf( OutFile, 
		     "# List of non simulable elements generated by %s\n", 
		     TOOLNAME );
	    fprintf( OutFile, "# Codes: transition = 0, place = 1\n#\n");
	    fprintf( OutFile, 
		     "# Number of non simulable elements:\n#\n" );

	    Bad_Objects = (int*) calloc( (unsigned)(1 + 6*Reseau->Nb_Places +
					 4*Reseau->Nb_Trans), sizeof( int ) );
	    Bad_Objects[0] = 0;
	  }
	}
	else
	  Bad_Objects == NULL;

	if ( File_Ok ) {
	  /* Checking that the network satisfies the conditions */
	  Network_Ok = 
	    Has_Integral_Transitions( Reseau, Testing, Bad_Objects )
	    && Has_All_Deterministic_Places( Reseau, Testing, Bad_Objects )
	    && Has_Simulable_Places( Reseau, Testing, Bad_Objects );

	  if ( Network_Ok ) {
	    Scale = Find_Time_Scale( Reseau, Scale, Testing, Bad_Objects );
	    Network_Ok = ( Scale != 0 );
	  }

	  if ( Testing && For_GSS ) {
	    fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[0] );
	    for( i=1; i<=Bad_Objects[0]; i++ ) {
	      fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i-1] );
	      fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i] + 1 );
	    }
	    fclose( OutFile );
	    return Network_Ok;
	  }
	  else {

	    if ( Debug )
	      printf("%s: time scale found: %d\n", TOOLNAME, Scale );

	    if ( Network_Ok ) {
	      /* Go for structural analysis and code generation */
	      The_Layer_Struct = (Layer_Struct) 
		malloc( sizeof( struct LAYER_STRUCT ) );

	      if ( topological_analysis( Reseau, The_Layer_Struct ) ) {

		if (Debug)
		  Dump_Layer_Struct( The_Layer_Struct );

		if ( Max_Time >= 0 )
		  Reseau->Sim_Params->Nb_Firings = Max_Time;

		gen_c_code( Reseau, The_Layer_Struct, Trace );
	      }
	      else {
		fprintf( stderr, "Topological analysis failed: ");
		fprintf( stderr, "look for an zero-time cycle.\n");
		return -3;
	      }
	    }
	    else {
	      printf("%s: the network doesn't satisfy the assumptions. ", 
		     TOOLNAME );
	      printf("No translation.\n");
	    }
	  }
	}
	else 
	  fprintf( stderr, "%s: Error opening file %s.\n", 
		   TOOLNAME, Out_File_Name );
      }
      else {
	printf("%s: Error reading network. No translation.\n", TOOLNAME );
      }
    }
    else {
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, File_Name );
      return -2;
    }
    return 0;
  }
  else if ( !Help )
    return -1;
  else 
    return 0;
}

/****************************************************************/
/* Procedure to initialize various tables.			*/
/****************************************************************/
void Init_Generation( Reseau, Max_Delay, Scale )
PetriNet	*Reseau;
int		Scale;
{
  int		i;
  int		Delay;
  int		Pl_Num;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  Hold_Time = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof(int) );
  Fire_Time = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );

  Delay_Table = (boolean*) calloc( (unsigned)(1+Max_Delay), 
				   sizeof(boolean) );
  Is_Compet_Pl = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				   sizeof(boolean) );
  Is_Compet_Tr = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Is_Source = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Already_Visited = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				       sizeof( boolean ) );

  for ( i=0; i<Reseau->Nb_Places; i++ ) 
    Hold_Time[i] = rint( Scale * Reseau->P[i].Pl_Law->Parameters[0] );

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    Is_Compet_Tr[i] = FALSE;
    SCAN( L_Scan, Reseau->T[i].Inputs ) {
      The_Routing = Reseau->P[L_Scan->Dest].Routing;
      if ( The_Routing != NULL ) {
	if ( The_Routing->Type == 'P' ) {
	  Is_Compet_Tr[i] = TRUE;
	  Has_Competition = TRUE;
	}
      }
    }
    /* find out if it is a source or not */
    Is_Source[i] = FALSE;
    if ( Reseau->T[i].Indegree == 0 )
      /* I assume this is meant to be a source... */
      Is_Source[i] = TRUE;
    else if ( Reseau->T[i].Indegree == 1 ) {
      Pl_Num = Reseau->T[i].Inputs->Dest;
      if ( ( Reseau->P[Pl_Num].Indegree == 1 ) 
	   && ( Reseau->P[Pl_Num].Inputs->Dest == i ) )
	Is_Source[i] = TRUE;
    }

    if ( !Is_Source[i] )
      Fire_Time[i] = rint( Scale * Reseau->T[i].Tr_Law->Parameters[0] );
    else
      Fire_Time[i] = 0.0;

    if ( Is_Compet_Tr[i] ) {
      /* if it is, generate Routing() code for **EACH** input transition */
      SCAN( L_Scan, Reseau->T[i].Inputs ) {
	Is_Compet_Pl[L_Scan->Dest] = TRUE;
	Has_Competition = TRUE;
      }
    }
  }

  for ( i=0; i<Max_Delay+1; i++ )
    Delay_Table[i] = FALSE;
  /* Always need Time-1 to ensure firing counts will increase.		 */
  Delay_Table[1] = TRUE;

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    Delay = Hold_Time[ i ];
    SCAN( L_Scan, Reseau->P[i].Inputs ) {
      if ( !Is_Source[ L_Scan->Dest ] ) 
	Delay += Fire_Time[ L_Scan->Dest ];
      Delay_Table[ Delay ] = TRUE;
    }
  }

  for ( i=0; i<Reseau->Nb_Places; i++ )
    Already_Visited[i] = FALSE;

  /* structures for time scale analysis */
  TSP = (int *)calloc( (unsigned)Reseau->Nb_Places, sizeof( int ) );
  TST = (int *)calloc( (unsigned)Reseau->Nb_Trans, sizeof( int ) );

}
/****************************************************************/
/* Procedures for code generation...				*/
/****************************************************************/
void	gen_file_header( Reseau, Trace )
PetriNet	*Reseau;
boolean		Trace;
{
  CODE("/* Configuration file for a Petri net \t\t\t*/\n");
  CODE("/* simulation code generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  CODE("\t\t*/\n");
  CODE1("/* Date: %s \t\t\t\t*/\n\n", todays_date() );

  /* write definitions, includes */
  CODE1("int     TMAX = %d;\n", Scale*Reseau->Sim_Params->Nb_Firings );
  CODE1("int     SCALE = %d;\n", Scale );
  CODE1("#define NB_TRANS\t%d\n", Reseau->Nb_Trans );
  CODE1("#define NB_PLACES\t%d\n", Reseau->Nb_Places );
  CODE1("#define BUF_SIZE\t%d\n", Buf_Size );
  CODE1("#define HALF_BUF_SIZE\t%d\n", Buf_Size/2 );
  CODE("#define Infinity 0x7FFFFFFF\n" );

  /* trace defines */
  if ( Trace ) {
    CODE("#define Trace\n");

    CODE("#define CODE(x)\t\tfprintf(Tracefile,x)\n");
    CODE("#define CODE1(x,y)\tfprintf(Tracefile,x,y)\n");
    CODE("#define CODE2(x,y,z)\tfprintf(Tracefile,x,y,z)\n");
    CODE("#define CODELN\t\tCODE(\"\\n\")\n");
  }
  CODELN;

}

void Gen_Source_Code( Reseau )
PetriNet	*Reseau;
{
  int		i;
  int		Delay;
  int		Nb_Tokens;
  double	Proba;
  Law_Desc	The_Law;

  CODE("/* Procedure for computing the number of tokens generated ");
  CODE("by the sources */\n");
  CODE("void Re_Init_Sources( From, To )\n");
  CODE("int\tFrom;\nint\tTo;\n");
  CODE("{\n  int\ti, j;\n");

  CODE("  for( i = From; i < To; i++ ) {\n\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Is_Source[i] ) {
      CODE2("    /* Transition %d (%s) */\n", i, Reseau->T[i].Name );
      The_Law = Reseau->T[i].Tr_Law;

      if ( Reseau->T[i].Indegree == 0 )
	/* assume simple loop */
	Nb_Tokens = 1;
      else 
	Nb_Tokens = Reseau->P[Reseau->T[i].Inputs->Dest].Init_Pop;

      if ( Nb_Tokens > 1 ) {
	CODE1("    for( j = 0; j < %d; j ++ ) {\n", Nb_Tokens );
	Indent_Nb += 2;
      }

      switch( The_Law->Name ) {
      case 'C':
	Delay = IRINT( The_Law->Parameters[0] * Scale );
	
	CODE1("    T_Phase_%d --;\n", i );
	CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	CODE1("      Nb_Tok_%d ++;\n", i );
	CODE2("      T_Phase_%d = %d;\n", i, Delay );
	CODE("    }\n");
  
	break;

      case 'G':
      case 'B':	
	Proba = The_Law->Parameters[0];

	if ( Scale != 1 ) {
	  CODE1("    T_Phase_%d --;\n", i );
	  CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	}
	CODE1("      Nb_Tok_%d += ", i );
	if ( The_Law->Name == 'B' )
	  CODE("1 + ");
	CODE1(" (int)floor( - log(U_0_1())/ %12.6f );\n", -log(1.0-Proba) );
	if ( Scale != 1 ) {
	  CODE2("      T_Phase_%d = %d;\n", i, Scale );
	  CODE("    }\n");
	}

	break;

      case 'g':
	Proba = The_Law->Parameters[0];

	CODE1("    T_Phase_%d --;\n", i );
	CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	CODE1("      Nb_Tok_%d ++;\n", i );
	CODE1("      T_Phase_%d = 1 + ", i );
	CODE1(" (int)floor( - log(U_0_1())/ %12.6f );\n", -log(Proba) );
	CODE( "    }\n");

	break;

      case 'D':
	CODE1("    T_Phase_%d --;\n", i );
	CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	CODE1("      Nb_Tok_%d ++;\n", i );
	CODE1("      if( T_Pr_Zero_%d > 0.0 )\n", i );
	CODE2("        Nb_Tok_%d += (int)floor( log(U_0_1())/ log(T_Pr_Zero_%d ) );\n",
	      i, i );
	CODE( "      U = U_0_1();\n");
	CODE1("      for( j=0; U > T_Proba_%d[j]; j++ );\n", i );
	CODE2("      T_Phase_%d = Interarrival_%d[j];\n", i, i );
	CODE( "    }\n");
	
	break;

      default:
	printf("%s: warning: unknown law type at source %d\n",
	       TOOLNAME, i );

      }
      CODE2("    R_%d[i] = Nb_Tok_%d;\n", i, i );

      if ( Nb_Tokens > 1 ) {
	Indent_Nb -= 2;
	CODE("    }\n");
      }
      CODELN;
    }
  CODE("  }\n");

  CODE("}\n");
}

void Gen_Stats_Code( Reseau )
PetriNet	*Reseau;
{
  int		i;

  CODE("void Write_Source_Stats( Sim_Length, Buf_Size )\n");
  CODE("int\tSim_Length;\n");
  CODE("int\tBuf_Size;\n");
  CODE("{\n");
  CODE("  int\ti;\n");
  CODE("  int\tNb_Enabl;\n");
  CODELN;
  
  for (i=0; i<Reseau->Nb_Trans; i++) {
    if ( Is_Source[i] ) {
      CODE1("  Nb_Enabl = R_%d[Sim_Length %% Buf_Size];\n", i );
      CODE1("  printf(\"R_%d[%%d] = %%6d    Cycle time: \", ", i );
      CODE("Sim_Length / SCALE, Nb_Enabl );\n");
      CODE("  if ( Nb_Enabl > 0 )\n");
      CODE("    printf(\"%%10.4f\", (double)Sim_Length / ");
      CODE(" (double)Nb_Enabl / SCALE );\n");
      CODE("  else\n");
      CODE("    printf(\"  Infinity\");\n");
      CODE1("  if ( strlen( Trans_Name[%d] ) > 0 )\n", i );
      CODE1("    printf(\"  (%%s)\\n\", Trans_Name[%d] );\n", i );
      CODE("  else\n");
      CODE("    printf(\"\\n\");\n");
      CODELN;
    }
  }
  CODE("}\n");

  CODELN;
  CODE("int Nb_Arrived_Tokens( Time )\n");
  CODE("int\tTime;\n");
  CODE("{\n  int\tRes = 0;\n\n");
  for (i=0; i<Reseau->Nb_Trans; i++) {
    if ( Is_Source[i] ) {
      CODE1("  Res += R_%d[Time %% BUF_SIZE];\n", i );
    }
  }
  CODELN;
  CODE("  return Res;\n");
  CODE("}\n");
  
}

void Gen_Marking_Function( Reseau, Place_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  Pl_Des	The_Place;
  int		Delay;

  The_Place = Reseau->P[Place_Id];

  if ( Initial )
    CODE("int I_");
  else
    CODE("int ");

  CODE1("Marking_%d()\n{\n  register int Res;\n\n", Place_Id );
  if ( Initial )
    CODE1("  Res = Avail[%d][ Time ];\n", Place_Id );
  else
    CODE1("  Res = %d;\n", Reseau->P[Place_Id].Init_Pop );

  SCAN( L_Scan, The_Place.Inputs ) {
    CODE("  Res += ");
    Delay = Hold_Time[Place_Id];
    if ( !Is_Source[L_Scan->Dest] )
      Delay += Fire_Time[L_Scan->Dest];

    if ( Delay == 0 ) {
      if ( Is_Source[L_Scan->Dest] )
	CODE1( "R_%d[ Time ]", L_Scan->Dest );
      else if ( Y_Static )
	CODE1( "Y_0[%d]", L_Scan->Dest );
      else
 	CODE1( "Y[ Time ][%d]", L_Scan->Dest );
    }
    else {
      if ( Is_Source[L_Scan->Dest] )
	CODE2( "R_%d[ Time_M_%d ]", L_Scan->Dest, Delay );
      else if ( Y_Static )
	CODE2( "Y_Time_M_%d[%d]", Delay, L_Scan->Dest );
      else
 	CODE2( "Y[ Time_M_%d ][%d]", Delay, L_Scan->Dest );
    }
    if ( L_Scan->Weight > 1 )
      CODE1(" * %d", L_Scan->Weight );
    CODE(";\n");
  }
  SCAN( L_Scan, The_Place.Outputs ) {
    if ( Y_Static )
      CODE1("  Res -= Y_0[%d]", L_Scan->Dest );
    else
      CODE1("  Res -= Y[ Time ][%d]", L_Scan->Dest );
    if ( L_Scan->Weight > 1 )
      CODE1(" * %d;\n", L_Scan->Weight );
    else
      CODE(";\n");
  }
  CODE("  return Res;\n}\n\n");
}

void	gen_declarations(Reseau, Max_Delay, Scale, Trace)
PetriNet	*Reseau;
int		Max_Delay;
int		Scale;
boolean		Trace;
{
  int		i, j, k, l;
  int		Nb_Params;
  int		Nb_Values;
  int		Nb_Inserted;
  int		Max_Degree;
  double	Pr0;
  double	Proba;
  double	Cum_Proba;
  double	*The_Probas;
  double	*The_Values;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;
  Law_Desc	The_Law;

  CODE1("#define MAX_DELAY %d\n", Max_Delay );

  /* Compute the max outdegree of places "involved" in EQ. 		*/
  /* For use with the H_* tables.					*/
  Max_Degree = -1;
  for ( i=0; i<Reseau->Nb_Places; i++ ) 
    Max_Degree = max( Max_Degree, Reseau->P[i].Outdegree );
  CODE1("#define MAX_OUTDEGREE\t%d\n", Max_Degree );

  /* defining link numbers for use with H_* tables 			*/
  CODE("/* defining the link numbers at the output of places */\n");
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    j = 0;
    SCAN( L_Scan, Reseau->P[i].Outputs ) {
      CODE3("#define NB_P%d_TO_T%d\t%d\n", i, L_Scan->Dest, j );
      j++;
    }
  }
  CODELN;

  /* Defining the number of available tokens for the MAX_DELAY first	*/
  /* units of time. Used at the beginning of the simulation.		*/
  if ( Reseau->Nb_Places > 0 ) {
    CODE("/* table of initial availabilities\t\t\t\t\t*/\n");
    CODE("int\t\t\tAvail[NB_PLACES][MAX_DELAY] = {\n");

    for( i=0; i<Reseau->Nb_Places-1; i++ ) {
      CODE("\t");
      for( j=0; j<Max_Delay; j++ ) {
	CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
      }
      CODE("\n");
    }
    CODE("\t");
    i = Reseau->Nb_Places-1;
    for( j=0; j<Max_Delay-1; j++ ) {
      CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
    }
    CODE1("%2d", Nb_Available_Tokens( Reseau->P[i], Max_Delay-1 ) );

    CODE("\n};\n\n");
  }

  if ( Reseau->Nb_Trans > 0 ) {
    CODE("/* table of firing times\t\t\t\t\t*/\n");
    CODE("int\t\t\tFire_Time[NB_TRANS] = {\n");

    for( i=0; i<Reseau->Nb_Trans-1; i++ )
      CODE1("\t%2d,\n", Fire_Time[i] );

    i = Reseau->Nb_Trans-1;
    CODE1("\t%2d", Fire_Time[i] );
    CODE("\n};\n\n");
  }

  if ( Reseau->Nb_Trans > 0 ) {
    CODE("/* table of source flags\t\t\t\t\t*/\n");
    CODE("int\t\t\tIs_Source[NB_TRANS] = {\n");

    for( i=0; i<Reseau->Nb_Trans-1; i++ )
      CODE1("\t%s,\n", ( Is_Source[i] ? "TRUE" : "FALSE" ) );

    i = Reseau->Nb_Trans-1;
    CODE1("\t%s\n", ( Is_Source[i] ? "TRUE" : "FALSE" ) );
    CODE("};\n\n");
  }

  if ( Reseau->Nb_Trans > 0 ) {
    CODE("/* table of transition names\t\t\t\t\t*/\n");
    CODE("char\t\t\t*Trans_Name[NB_TRANS] = {\n");

    for( i=0; i<Reseau->Nb_Trans-1; i++ )
      if ( strlen( Reseau->T[i].Name ) > 0 )
	CODE1("\t\"%s\",\n", Reseau->T[i].Name );
      else
	CODE("\t\"\",\n" );

    i = Reseau->Nb_Trans-1;
    if ( strlen( Reseau->T[i].Name ) > 0 )
      CODE1("\t\"%s\",\n", Reseau->T[i].Name );
    else
      CODE("\t\"\"\n" );
    CODE("\n};\n\n");
  }

  CODE("/* variables for use with the circular buffer\t\t\t*/\n");
  CODE("int\tTime;\t\t/* The time, modulo the buffer size */\n");
  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] )
      CODE2("int\tTime_M_%d;\t\t/* Time - %d */\n", i, i );
  CODELN;
  if ( Y_Static ) {
    CODE("int\t*Y_0;\t\t/* Y[ Time ] */\n");
    for (i=1; i<=Max_Delay; i++ )
      if ( Delay_Table[i] || (i == 1 ) )
	/* Y_Time_M_1 is needed in all cases because of the scaling system */
	CODE2("int\t*Y_Time_M_%d;\t\t/* Y[ Time - %d ] */\n", i, i );
    CODELN;
  }

  CODE("/* tables and counts for routing generation\t\t\t*/\n");
  for( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Routing = Reseau->P[i].Routing;

    if ( The_Routing != NULL ) {
      switch ( The_Routing->Type ) {

      case 'R':
	CODE1("int Count_%d = 0;\n", i );
	CODE2("double P_Proba_%d[%d] = {", i, Reseau->P[i].Outdegree );
      
	Cum_Proba = 0.0;
	for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	  Cum_Proba += The_Routing->Probas[j];
	  CODE1(" %6.4f,", Cum_Proba );
	}
	CODE(" 1.0 };\n");
	CODE2("int Dest_%d[%d] = {", i, Reseau->P[i].Outdegree );
	L_Scan = Reseau->P[i].Outputs;
	for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	  CODE1(" %d,", L_Scan->Dest );
	  L_Scan = L_Scan->Next;
	}
	CODE1(" %d };\n", L_Scan->Dest );

	break;

      case 'D':
	CODE1("int Count_%d = 0;\n", i );
	CODE1("int P_Phase_%d = 0;\n", i );
	CODE2("int Cycle_%d[%d] = {", i, The_Routing->Period );
	for ( j=0; j<The_Routing->Period-1 ; j++ ) {
/* 	  CODE1(" %d,", The_Routing->Cycle[j] ); */
	  CODE2(" NB_P%d_TO_T%d,", i, The_Routing->Cycle[j] );
	}
/* 	CODE1(" %d };\n", */
	CODE2(" NB_P%d_TO_T%d };\n", i,
	      The_Routing->Cycle[The_Routing->Period-1] );

	break;

      default:
	if ( Is_Compet_Pl[i] )
	  CODE1("int Count_%d = 0;\n", i );

	break;
      }
    } /* if ( Routing != NULL ) */
    else if ( Is_Compet_Pl[i] )
	  CODE1("int Count_%d = 0;\n", i );

/*     CODE1("int H_%d[NB_TRANS];\n", i ); */
    CODE1("int H_%d[MAX_OUTDEGREE];\n", i );

  }

  /* Tables and counts for sources */
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Is_Source[i] ) {
      The_Law = Reseau->T[i].Tr_Law;

      CODE2("/* Variables for source %d (%s) */\n", i, Reseau->T[i].Name );
      CODE1("int\tNb_Tok_%d = 0;\n", i );
      CODE1("int\tR_%d[ BUF_SIZE ];\n", i );

      switch( The_Law->Name ) {
      case 'C':
	CODE2("int\tT_Phase_%d = %d;\n", i,
	      IRINT( The_Law->Parameters[0] ) );
	break;

      case 'D':
	Nb_Params = IRINT(The_Law->Parameters[0]);
	/* We have to eliminate the null interarrivals. In the process,	*/
	/* we sort the probabilities by decreasing values		*/
	Pr0 = 0.0;
	Nb_Values = Nb_Params;
	for ( j=0; j<Nb_Params ; j++ ) {
	  if ( The_Law->Parameters[1+j] == 0.0 ) {
	    Pr0 += The_Law->Parameters[Nb_Params+1+j];
	    Nb_Values--;
	  }
	}
	if ( Pr0 >= 1.0 )
	  printf("%s: error: infinite throughput at source %d\n",
		 TOOLNAME, i );
	else {
	  The_Probas = (double*) calloc( Nb_Values, sizeof( double ) );
	  The_Values = (double*) calloc( Nb_Values, sizeof( double ) );
	  Nb_Inserted = 0;
	  for ( j=0; j<Nb_Params; j++ )
	    if ( The_Law->Parameters[1+j] != 0.0 ) {
	      Proba = The_Law->Parameters[Nb_Params+1+j];
	      for( k=0; (k<Nb_Inserted) && (Proba<The_Probas[k]); k++ );
	      for( l=Nb_Inserted; l>k; l-- ) {
		The_Probas[l] = The_Probas[l-1];
		The_Values[l] = The_Values[l-1];
	      }
	      The_Probas[k] = Proba;
	      The_Values[k] = The_Law->Parameters[1+j] * Scale;
	      Nb_Inserted++;
	    }
	    
	  CODE1("int\tT_Phase_%d = 1;\n", i );
	  CODE2("double\tT_Pr_Zero_%d = %6.4f;\n", i, Pr0 );
	  CODE2("double\tT_Proba_%d[%d] = {", i, Nb_Values );
	  Cum_Proba = 0.0;
	  for ( j=0; j<Nb_Values-1 ; j++ ) {
	    Cum_Proba += The_Probas[j];
	      CODE1(" %6.4f,", Cum_Proba/(1.0-Pr0) );
	  }
	  CODE(" 1.0 };\n");
	}

	CODE2("double\tInterarrival_%d[%d] = {", i, Nb_Values );
	for ( j=0; j<Nb_Values-1 ; j++ ) {
	  CODE1(" %6.4f,", The_Values[j] );
	}
	CODE1(" %6.4f };\n", The_Values[Nb_Values-1] );

	break;

      case 'B':
      case 'G':
	if ( Scale != 1 )
	  CODE2("int\tT_Phase_%d = %d;\n", i, Scale );
	
	break;
      }
    }

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( Is_Compet_Pl[i] ) && !Already_Visited[i] ) {
      /* generate Marking code 						*/
      CODE2("/* Function to compute the marking of place %d (%s) */\n",
	    i, Reseau->P[i].Name );
      Gen_Marking_Function( Reseau, i, TRUE );
      Gen_Marking_Function( Reseau, i, FALSE );
      Already_Visited[i] = TRUE;
    }

  CODELN;
  Gen_Source_Code( Reseau );

  CODELN;
  Gen_Stats_Code( Reseau );

  if ( Trace ) {

    CODE("/* declarations for trace handling */\n");

    if ( Reseau->Nb_Trans > 0 ) {
      CODE("char  *gss_trans_name[NB_TRANS] = {\n");
      for( i=0; i<Reseau->Nb_Trans-1; i++ )
	CODE1("\t\"%s\",\n", Reseau->T[i].Data );
	
      CODE1("\t\"%s\"\n", Reseau->T[Reseau->Nb_Trans-1].Data );
      CODE("};\n\n");
    }

    if ( Reseau->Nb_Places > 0 ) {
      CODE("char  *gss_place_name[NB_PLACES] = {\n");
      for( i=0; i<Reseau->Nb_Places-1; i++ )
	CODE1("\t\"%s\",\n", Reseau->P[i].Data );
      
      CODE1("\t\"%s\"\n", Reseau->P[Reseau->Nb_Places-1].Data );
      CODE("};\n\n");
    }

/*     CODE("char  *gss_platra_link_name[NB_PLACES][NB_TRANS];\n"); */
/*     for( i=0; i<Reseau->Nb_Places; i++ ) */
/*       SCAN( L_Scan, Reseau->P[i].Outputs ) */
/* 	CODE3("gss_platra_link_name[%d][%s] = \"%s\",\n",  */
/* 	      i, L_Scan->Dest, L_Scan->Data ); */

/*     CODE("char  *gss_trapla_link_name[NB_TRANS][NB_PLACES];\n"); */
/*     for( i=0; i<Reseau->Nb_Trans; i++ ) */
/*       SCAN( L_Scan, Reseau->T[i].Outputs ) */
/* 	CODE3("gss_trapla_link_name[%d][%s] = \"%s\",\n",  */
/* 	      i, L_Scan->Dest, L_Scan->Data ); */

    if ( Reseau->Nb_Places > 0 ) {
      CODE("/* table of initial populations\t\t\t\t\t*/\n");
      CODE("int\t\t\tInitial_Pop[NB_PLACES] = {\n");

      for( i=0; i<Reseau->Nb_Places-1; i++ ) {
	CODE1("\t%d", Reseau->P[i].Init_Pop );
	CODE(",\n");
      }

      CODE1("\t%d", Reseau->P[Reseau->Nb_Places-1].Init_Pop );
      CODE("\n};\n\n");
    }

    CODE("int   Cur_X_Evt[NB_TRANS+1];\n");
    CODE("int   Cur_Y_Evt[NB_TRANS+1];\n");
    CODELN;
    CODE("FILE  *Tracefile;\n");
    CODE("int   More_Events;\n");
    CODE("int   Was_X_Evt, Was_Y_Evt;\n");
    CODE("int   Min_X_Val, Min_Y_Val;\n");
    CODE("int   Arg_Min_X, Arg_Min_Y;\n");
    CODE("int   Old_Date, New_Date;\n");
    CODELN;
    CODE1("int  Trace_Start = %d;\n", Trace_Start );
    CODE1("int  Trace_End = %d;\n", Trace_End );
  }
}

void gen_trace_before_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  if ( The_Trans.Indegree > 1 )
    CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
  /* first remove tokens from input places, except dummy ones */
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Inputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( Time>%d ? Y[ Time - %d ][%d] : 0 )", 
	      Delay, Delay, M_Scan->Dest );
	else if ( Y_Static )
	  CODE2("  Marking += Y_Time_M_%d[%d]", Delay, M_Scan->Dest );
	else
 	  CODE2("  Marking += Y[ Time_M_%d ][%d]", Delay, M_Scan->Dest );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	if ( Y_Static )
	  CODE1("  Marking -= Y_0[%d]", M_Scan->Dest );
	else 
	  CODE1("  Marking -= Y[ Time ][%d]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then move the tokens */
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_place_name[%d], ", L_Scan->Dest );
      CODE1("gss_trans_name[%d] );\n", Trans_Id );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then switch transition status */
  CODE1("  CODE(\"MODIFY %s \\\"status\\\" = 0\\n\");\n",
	The_Trans.Name );
}

void gen_trace_after_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  /* move tokens to (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {

      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_trans_name[%d],", Trans_Id );
      CODE1("gss_place_name[%d] );\n", L_Scan->Dest );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* finally change marking of (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( Time>%d ? Y[ Time - %d ][%d] : 0 )", 
	      Delay, Delay, M_Scan->Dest );
	else
	  CODE2("  Marking += Y[ Time - %d ][%d]", Delay, M_Scan->Dest );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	CODE1("  Marking -= Y[ Time ][%d]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  CODE("  CODE(\"eol+\\n\");\n");
}

void	write_trans_name( The_Trans )
Tr_Des	The_Trans;
{
  if ( NULL != The_Trans.Name )
    printf( "%s", The_Trans.Name );
  else
    printf( "_TR_%d", The_Trans.Tr_Id );
}

void	Write_Token_Total( Reseau, Place_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  int		Delay;

  if ( Already_Visited[ Place_Id ] ) {
    CODE1("Nb_Tokens = Count_%d;\n", Place_Id );
  }
  else {
    if ( Initial )
      CODE1("Nb_Tokens = Avail[%d][Time];\n", Place_Id );
    else
      CODE1("Nb_Tokens = %d;\n", Reseau->P[Place_Id].Init_Pop );

    SCAN( L_Scan, Reseau->P[Place_Id].Inputs ) {
      CODE( "Nb_Tokens += ");

      if ( Is_Source[L_Scan->Dest] )
	Delay = IRINT( Scale * 
		       Reseau->P[Place_Id].Pl_Law->Parameters[0] );
      else
	Delay = IRINT( Scale * 
		       ( Reseau->P[Place_Id].Pl_Law->Parameters[0]
			 + Reseau->T[L_Scan->Dest].Tr_Law->Parameters[0] ) );

      if ( Initial ) {
	CODE1( "( Time >= %d ? ", Delay );
	if ( Is_Source[L_Scan->Dest] )
	  CODE2( "R_%d[Time-%d] ", L_Scan->Dest, Delay );
	else
	  CODE2( "Y[Time-%d][%d] ", Delay, L_Scan->Dest );
	if ( L_Scan->Weight > 1 )
	  CODE1("* %d", L_Scan->Weight );
	CODE(" : 0 );\n");
      }
      else {
	if ( Delay == 0 ) {
	  if ( Is_Source[L_Scan->Dest] )
	    CODE1( "R_%d[ Time ]", L_Scan->Dest );
	  else if ( Y_Static )
	    CODE1( "Y_0[%d]", L_Scan->Dest );
	  else
 	    CODE1( "Y[ Time ][%d]", L_Scan->Dest );
	}
	else {
	  if ( Is_Source[L_Scan->Dest] )
	    CODE2( "R_%d[ Time_M_%d ]", L_Scan->Dest, Delay );
	  else if ( Y_Static )
	    CODE2( "Y_Time_M_%d[%d]", Delay, L_Scan->Dest );
	  else
 	    CODE2( "Y[ Time_M_%d ][%d]", Delay, L_Scan->Dest );
	}
	if ( L_Scan->Weight > 1 )
	  CODE1(" * %d", L_Scan->Weight );
	CODE(";\n");
      }
    }
  }
}

void	Write_Token_Routing( Reseau, Place_Id, Trans_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
int		Trans_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  The_Routing = Reseau->P[Place_Id].Routing;

  if ( Reseau->P[Place_Id].Outdegree > 1 ) {
    if ( !Already_Visited[Place_Id] ) {
      CODE2("/* Computing number of tokens arrived in place %d (%s) */\n",
	    Place_Id, Reseau->P[Place_Id].Name );

      if ( The_Routing != NULL ) {
	switch( The_Routing->Type ) {

	case 'R':

	  CODE2("for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
		Place_Id, Place_Id );
	  CODE("  U = U_0_1();\n");
	  CODE1("  for( j=0; U > P_Proba_%d[j]; j++ );\n", Place_Id );
/* 	  CODE2("  H_%d[Dest_%d[j]]++;\n", Place_Id, Place_Id ); */
	  CODE2("  H_%d[ j ]++;\n", Place_Id, Place_Id );
	  CODE("}\n");

	  Already_Visited[ Place_Id ] = TRUE;

	  break;

	case 'D':

	  CODE2("for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
		Place_Id, Place_Id );
	  CODE3("  H_%d[ Cycle_%d[ P_Phase_%d ] ]++;\n",
		Place_Id, Place_Id, Place_Id );
	  CODE1("  P_Phase_%d++;\n", Place_Id );
	  CODE3("  P_Phase_%d = ( P_Phase_%d %% %d );\n", 
		Place_Id, Place_Id, The_Routing->Period );
	  CODE("}\n");

	  Already_Visited[ Place_Id ] = TRUE;

	  break;

	case 'P':

	  CODE("/* Start firing as long as  transition ");
	  CODE2("%d (%s) is enabled */\n", 
		Trans_Id, Reseau->T[Trans_Id].Name );
	  CODE("Is_Enabled = TRUE;\n");
	  CODE1("while( Is_Enabled && ( Count_%d < Nb_Tokens ) ) {\n", 
	       Place_Id );
	  CODE("  Is_Enabled = ");
	  L_Scan = Reseau->T[Trans_Id].Inputs;
	  /* Call to the marking function, with all Time parameters */
	  if ( Initial )
	    CODE("( I_");
	  else
	    CODE("( ");
	  CODE2("Marking_%d() >= %d )", L_Scan->Dest, L_Scan->Weight );
	  SCAN( L_Scan, Reseau->T[Trans_Id].Inputs->Next ) {
	    CODE2("\n      && ( Marking_%d () >= %d )",
		 L_Scan->Dest, L_Scan->Weight );
	  }
	  CODE(";\n");
	  CODE("  if ( Is_Enabled ) {\n");
/* 	  CODE2("    H_%d[ %d ]++;\n", Place_Id, Trans_Id ); */
	  CODE3("    H_%d[ NB_P%d_TO_T%d ]++;\n", 
		Place_Id, Place_Id, Trans_Id );
	  CODE1("    Count_%d++;\n", Place_Id );
	  CODE("  }\n");
	  CODE("}\n");

	  break;

	default:

	  printf("%s: error: unexpected routing policy (%c)\n",
		 TOOLNAME, The_Routing->Type );
	} /* switch(...) */
      } /* if ( Routing != NULL ) */
      else {
	  CODE1("Count_%d = Nb_Tokens;\n", Place_Id );
	  Already_Visited[ Place_Id ] = TRUE;
      }
    }

/*     CODE2("Nb_Tokens = H_%d[%d];\n", Place_Id, Trans_Id ); */
    CODE3("Nb_Tokens = H_%d[ NB_P%d_TO_T%d ];\n",
	  Place_Id, Place_Id, Trans_Id );
  }
}

void Write_Token_Reduction( The_Link )
Liste_Link	The_Link;
{
  if ( The_Link->Weight > 1 )
    CODE1("Nb_Tokens /= %d;\n", The_Link->Weight );
}

void	gen_code_node( Reseau, The_Node, Initial, Trace )
PetriNet	*Reseau;
int		The_Node;
boolean		Initial;
boolean		Trace;
{
  Liste_Link	L_Scan;
  Tr_Des	The_Trans;
  int		Pred_Pl_Id;
  int		Delay;

  The_Trans = Reseau->T[ The_Node ];

  CODE("/* Evolution of transition ");
  write_trans_name( Reseau->T[The_Node] );
  CODE(" */\n"); 

  switch( The_Trans.Indegree ) {

  case 0:
    /* this is a source but it should be recycled...	*/
    /* Actually, we shouldn't be here at all...		*/
    CODE2("Nb_Tokens = Y[Time_M_%d][%d] + 1;\n",
	  Fire_Time[The_Node], The_Node );
    printf("%s: warning: non recycled source (%d). ", TOOLNAME, The_Node );
    printf("Assumed recycled.\n" );

    if ( Initial || !Y_Static )
      CODE1("Y[Time][%d] = Nb_Tokens;\n", The_Node );
    else
      CODE1("Y_0[%d] = Nb_Tokens;\n", The_Node );

    break;

  case 1:
    /* single input. No min */
    Pred_Pl_Id = The_Trans.Inputs->Dest;
    Write_Token_Total( Reseau, Pred_Pl_Id, Initial );
    Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial );
    Write_Token_Reduction( The_Trans.Inputs );

    if ( Initial || !Y_Static )
      CODE1("Y[Time][%d] = Nb_Tokens;\n", The_Node );
    else
      CODE1("Y_0[%d] = Nb_Tokens;\n", The_Node );

    break;

  default:

    CODE("Tmp_Y = Infinity;\n" );

    for( L_Scan=The_Trans.Inputs; L_Scan != NULL; L_Scan=L_Scan->Next ) {

      Pred_Pl_Id = L_Scan->Dest;
      Write_Token_Total( Reseau, Pred_Pl_Id, Initial );
      Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial );
      Write_Token_Reduction( L_Scan );
      
      CODE("Tmp_Y = MIN( Tmp_Y, Nb_Tokens );\n" );
    }

    if ( Initial || !Y_Static )
      CODE1("Y[Time][%d] = Tmp_Y;\n", The_Node );
    else
      CODE1("Y_0[%d] = Tmp_Y;\n", The_Node );

  } /* switch () */

#define OLDY(x)	CODE3("( Time>%d ? Y[Time-%d][%d] : 0 )", x, x, The_Node )

  if ( Trace ) {
    Delay = Fire_Time[The_Node];
    CODE("Present_Nb = ");
    if ( Initial )
      OLDY( Delay );
    else if ( Y_Static )
      CODE2("Y_Time_M_%d[%d]", Delay, The_Node );
    else
      CODE2("Y[ Time - %d ][%d]", Delay, The_Node );

    CODE(";\n");
    
    CODE3("Y[(Time>%d?Time-%d:0)][%d]",
	  Delay, Delay, The_Node );
    CODE3(" = Y[(Time>%d?Time-%d:0)][%d];\n", 
	  Delay+1, Delay+1, The_Node );
    CODE("if( ( Time <= Trace_End ) && ( Time >= Trace_Start ) )\n");
    CODE3("while( Y[(Time>%d?Time-%d:0)][%d] < Present_Nb ) {\n",
	  Delay, Delay, The_Node );
    CODELN;
    gen_trace_before_firing( Reseau, The_Node, Initial );
    gen_trace_after_firing( Reseau, The_Node, Initial );
    CODELN;
    CODE3("  Y[(Time>%d?Time-%d:0)][%d]++;\n", Delay, Delay, The_Node );
    CODE("}\n");
  }
  CODELN;
}

void	gen_preparation_code( Reseau, Initial )
PetriNet	*Reseau;
boolean		Initial;
{
  int		i;
  Liste_Link	L_Scan;

  CODE("/* Draw source arrivals when necessary */\n");
  CODE("if ( Time %% HALF_BUF_SIZE == 0 ) {\n");
  CODE("  if ( Time == 0 )\n");
  CODE("    Re_Init_Sources( 0, HALF_BUF_SIZE );\n");
  CODE("  else\n");
  CODE("    Re_Init_Sources( HALF_BUF_SIZE, BUF_SIZE );\n");
  CODE("}\n");
  CODELN;

  CODE("/* Copy present firing count for everybody  */\n");
  for( i=0; i<Reseau->Nb_Trans; i++ )
    if ( !Is_Source[i] ) {
      if ( Initial )
	CODE2("Y[Time][%d] = ( Time>0 ? Y[Time-1][%d] : 0 );\n", i, i );
      else if ( Y_Static )
	CODE2("Y_0[%d] = Y_Time_M_1[%d];\n", i, i );
      else
 	CODE2("Y[ Time ][%d] = Y[ Time_M_1 ][%d];\n", i, i );
    }

  CODELN;

/*   if ( Has_Competition ) { */
/*     CODE("/* Prepare enabling count variables for transitions with "); */
/*     CODE("conflicts * /\n"); */

/*     for( i=0; i<Reseau->Nb_Trans; i++ ) */
/*       if ( Is_Compet_Tr[i] ) { */
/* 	if ( Initial ) */
/* 	  CODE2("Y[Time][%d] = ( Time>0 ? Y[Time-1][%d] : 0 );\n", i, i ); */
/* 	else if ( Y_Static ) */
/* 	  CODE2("Y_0[%d] = Y_Time_M_1[%d];\n", i, i ); */
/* 	else */
/*  	CODE2("Y[ Time ][%d] = Y[ Time_M_1 ][%d];\n", i, i ); */
/*       } */

/*     CODELN; */
/*   } */
}

void	gen_increment_code( Max_Delay )
int	Max_Delay;
{
  int	i;

  CODE("/* Increment of the time variables */\n");
  CODE("Time = ( Time + 1 ) %% BUF_SIZE;\n");
  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] )
      CODE2("Time_M_%d = ( Time + %d ) %% BUF_SIZE;\n", 
	    i, Buf_Size - i );
  if ( Y_Static ) {
    CODE("Y_0 = Y[ Time ];\n");
    for (i=1; i<=Max_Delay; i++ )
      if ( Delay_Table[i] )
	CODE2("Y_Time_M_%d = Y[ Time_M_%d ];\n", i, i );
    CODELN;
  }
}

void 	gen_loop_code( Reseau, The_Layer_Struct, Initial, Trace )
PetriNet	*Reseau;
Layer_Struct	The_Layer_Struct;
boolean		Initial;
boolean		Trace;
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;
  int	Time_Scale;
  int	i;
  int	j;
  int	Nb_Nodes;

  if ( Trace ) {
    CODE("CODE1(\"MODIFY TheClock \\\"ctime\\\" = %%7.3f");
    CODE("\\n\", (double)Time );\n");
  }

  for( i=1; i<=The_Scales[0]; i++ ) {
    Time_Scale = The_Scales[i];

    /* find the number of nodes within this time scale */
    Nb_Nodes = 0;
    for ( j=0; j<Reseau->Nb_Trans; j++ )
      if ( !Is_Source[j] && ( TST[j] == Time_Scale ) )
	Nb_Nodes++;

    if ( Nb_Nodes != 0 ) {
      CODE2("/* Computing the %d transition(s) with time scale %d */\n", 
	    Nb_Nodes, Time_Scale );

      if ( Time_Scale != 1 ) {
	CODE1("if ( ( R_Time %% %d ) == 0 ) {\n\n", Time_Scale );
	Indent_Nb += 2;
      }

      for( Num_Layer = 0; Num_Layer < The_Layer_Struct->Nb_Levels; 
	   Num_Layer++)
	for ( Num_Node = 0; 
	      Num_Node < The_Layer_Struct->Level[Num_Layer]->Nb_Nodes;
	      Num_Node++ ) {
	    
	  The_Node = The_Layer_Struct->Level[Num_Layer]->Node_Id[Num_Node];
	  if ( !Is_Source[The_Node] && ( TST[The_Node] == Time_Scale ) )
	    gen_code_node( Reseau, The_Node, Initial, Trace );
	}

      if ( Time_Scale != 1 ) {
	CODE1("} /* end of time scale %d */\n\n", Time_Scale );
	Indent_Nb -= 2;
      }
    }
  }
}

boolean Is_Dummy( The_Place )
Pl_Des	The_Place;
{
  if ( NULL != The_Place.Name )
    return FALSE;
  else
    return ( 0 == strncmp( The_Place.Name, "_Dummy_Place", 12 ) );
}

int	gen_c_code(Reseau, The_Layer_Struct, Trace )
PetriNet	*Reseau;
Layer_Struct	The_Layer_Struct;
boolean		Trace;
{
  int		Max_Dependency;
  int		Max_Source_Dependency;
  char		File_Name[512];
  int		i;

  Max_Dependency = rint( Scale * max( Max_Delay( Reseau ), 
				      Max_Lag( Reseau ) ) );
  
  /* adjusting buffer size and trace parameters		*/
  Buf_Size = min( Buf_Size, Scale*Reseau->Sim_Params->Nb_Firings + 1 );
  Buf_Size = max( Buf_Size, 2*Max_Dependency );

  if ( Trace ) {
    Trace_Start = max( Scale*Trace_Start, 0 );
    Buf_Size = max( Buf_Size, Trace_Start );
  }

  /* Buf_Size has to be even because of the "source dependency problem" */
  if ( 1 == ( Buf_Size % 2 ) )
    Buf_Size++;

  /* initializing some data structures */
  Init_Generation( Reseau, Max_Dependency, Scale );

  /* finding more precisely the time scale of the various objects */
  TimeScale( Reseau, Scale );

  /* generating global parameters of the graph */
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_graph_parameters.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )
    gen_file_header( Reseau, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_declarations.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )    
    gen_declarations( Reseau, Max_Dependency, Scale, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the initialization loop.		*/
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_initial_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) ) {
    gen_preparation_code( Reseau, TRUE );
    gen_loop_code( Reseau, The_Layer_Struct, TRUE, Trace );
    gen_increment_code( Max_Dependency );
  }
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the main loop. 			*/
  for ( i=0; i<Reseau->Nb_Places; i++ )
    Already_Visited[i] = FALSE;
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) ) {
    gen_preparation_code( Reseau, FALSE );
    gen_loop_code( Reseau, The_Layer_Struct, FALSE, Trace );
    gen_increment_code( Max_Dependency );
  }
  else
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  return TRUE;
}

/************************************************************************/
/* Module for computing the time scale of each transition.		*/
/* Author: Bruno Gaujal, March 1996.					*/
/************************************************************************/
int Original_Trans_Time_Scale( Reseau, Trans_Id, Scale )
PetriNet	*Reseau;
int		Trans_Id;
int		Scale;
{
  Law_Desc	The_Law;
  int		Nb_Params;
  double	Val;
  int		j;
  int		Trans_Scale;

  The_Law = Reseau->T[ Trans_Id ].Tr_Law;

  if ( !Is_Source[ Trans_Id ] ) {
    Trans_Scale = Fire_Time[ Trans_Id ];
  }
  else {
    switch( The_Law->Name ) {
    case 'C':
      Val = Scale * The_Law->Parameters[0];
      Trans_Scale = IRINT( Val );

      break;

    case 'D':
      Nb_Params = IRINT( The_Law->Parameters[0] );
      Trans_Scale = Scale;
      for ( j=1; j<= Nb_Params; j++ ) {
	Val = Scale * The_Law->Parameters[j];
	Trans_Scale = gcd( Trans_Scale, IRINT( Val ) );
      }
      break;

    case 'G':
    case 'B':
      Trans_Scale = Scale;
      break;

    default:
      Trans_Scale = 0;
    }
  }

  return Trans_Scale;
}

void TimeScale( Petri, Scale )
PetriNet	*Petri;
int		Scale;
{
  Liste_Link 	LinkPt;
  int 		pl_name,tr_name;
  int 		fire,hold;
  int 		p,t,i;
  int		Nb_Scales;
  int		Val;
  int		k;
  int		l;

  /******* remplissage initial ******/
  for (p = 0 ; p <  Petri->Nb_Places; p++)
    {
      TSP[p] = IRINT( Scale * Petri->P[p].Pl_Law->Parameters[0] );
    }

  for (t = 0; t <  Petri->Nb_Trans; t++)
    {
      TST[t] =  Original_Trans_Time_Scale( Petri, t, Scale );
    }


  /*** iterations ****/
  for (i=0; i< Petri->Nb_Trans + Petri->Nb_Places; i++)
    {
      /** les places **/
      for (p = 0 ; p <  Petri->Nb_Places; p++)
	{

	  fire = 0;

	  LinkPt =  Petri->P[p].Inputs;
	
	  while (LinkPt != NULL) 
	    {
	      tr_name = LinkPt->Dest;
	      if ((fire != 0) && (TST[tr_name] != 0)) 
		{fire = gcd(fire, TST[tr_name]);}
	      else {fire = max(fire,TST[tr_name]);}
	      LinkPt = LinkPt->Next;
	    }
	
	  if ((fire != 0) && (TSP[p] != 0)) {TSP[p] = gcd(TSP[p],fire);}
	  else {TSP[p] = max(fire,TSP[p]);}

	}

      /** les transitions **/
      for (t = 0; t <  Petri->Nb_Trans; t++)
	{

	  hold =0;
	  LinkPt =  Petri->T[t].Inputs;

	  while (LinkPt != NULL) 
	    {
	      pl_name = LinkPt->Dest;
	      if ((hold != 0) && (TSP[pl_name] != 0)) 
		{hold = gcd(hold, TSP[pl_name]);}
	      else {hold = max(hold,TSP[pl_name]);}
	    
	      LinkPt = LinkPt->Next;
	    }
	
	  if ((TST[t]!= 0)  && (hold != 0)){TST[t] = gcd(TST[t],hold);}
	  else  {TST[t] = max(hold,TST[t]);}
	}
    
    }

  /* Recognition and sort of the different time scales */
  The_Scales = (int*) calloc( Petri->Nb_Trans, sizeof( int ) );
  Nb_Scales = 0;
  for ( i=0; i<Petri->Nb_Trans; i++ )
    if ( !Is_Source[i] ) {
      Val = TST[i];
      for( k=1; (k<1+Nb_Scales) && (Val>The_Scales[k]); k++ );
      if ( ( k == 1+Nb_Scales ) || ( Val < The_Scales[k] ) ) {
	for( l=Nb_Scales+1; l>k; l-- )
	  The_Scales[l] = The_Scales[l-1];
	The_Scales[k] = Val;
	Nb_Scales++;
      }
    }
  The_Scales[0] = Nb_Scales;

  /* Analysis finished. Tracing if needed. */
  if ( Debug ) {
    printf("Found %d different time scales:", Nb_Scales );
    for ( i=1; i<=Nb_Scales; i++ )
      printf(" %d", The_Scales[i] );
    printf("\n");
    printf("Time Scales:\n");
    for ( i = 0 ; i < Petri->Nb_Places; i++) {
      printf("Pl %2d (%8s) = %2d    ",
	     i, Petri->P[i].Name, TSP[i]);
      if ( ( ( i+1 ) % 3 ) == 0 )
	printf("\n");
    }
    printf("\n");
    for ( i = 0 ; i < Petri->Nb_Trans; i++) {
      printf("Tr %2d (%8s) = %2d    ",
	     i, Petri->T[i].Name, TST[i]);
      if ( ( ( i+1 ) % 3 ) == 0 )
	printf("\n");
    }
    printf("\n");
  }

}

/************************************************************************/
/* Procedure for pretty printing. Normally called only by CODE macros	*/
/************************************************************************/
void Indent()
{
  int	i;

  printf("%*c", Indent_Nb, ' ' );
}
