/* programme/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un r\'eseau de Petri en "temps discret" */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "pn.h"

#define MAXINT          0x7FFFFFFF
#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"pn2c"

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODELN		CODE("\n")

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

/* Externs from the module parse_pn.o */
extern PetriNet		*Read_Network();
extern void		dumpe_network();
extern char		*todays_date();

/* Externs from the module analyze_pn.o */
extern boolean		topological_analysis();
extern boolean		Has_All_Deterministic_Times();
extern int		Find_Time_Scale();
extern boolean		Has_All_Routing_Places();
extern void		Dump_Layer_Struct();
extern double		Max_Lag();
extern double		Max_Delay();

/* Forward declarations */
boolean Is_Dummy();

int	Buf_Size = 1024;	/* default buffer size */
int	Scale = 10;		/* for integrality check */

int	*Hold_Time;
int	*Fire_Time;
boolean	*Is_Compet_Pl;		/* flag for places involved in a conflict */
boolean	*Is_Compet_Tr;	/* flag for transitions involved */
boolean	*Already_Visited;
boolean	*Delay_Table;
boolean	Has_Competition = FALSE;
int	Trace_Start = 0;	/* trace control parameters 	*/
int	Trace_End = 100;	/* with their initial values	*/
char	*Tmp_Dir;		/* where to put the files 	*/

main(argc, argv)
int	argc;
char	*argv[];
{
  static char		File_Name[128] = "";
  static char		Out_File_Name[128] = "";
  PetriNet 		*Reseau;
  boolean		Trace = FALSE;
  boolean		Testing = FALSE;
  boolean		For_GSS = FALSE;
  int			i;
  int			Max_Time = -1;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Network_Ok = TRUE;
  Layer_Struct		The_Layer_Struct;
  FILE			*OutFile;
  int			*Bad_Objects;

  Tmp_Dir = (char*) malloc( sizeof( char ) );
  strcpy( Tmp_Dir, "" );

  i = 1;

  if (argc>1)
    do {

      if GET_INT( "-bs", Buf_Size )
      else if GET_INT( "-scaling", Scale )
      else if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_INT( "-tmax", Max_Time )
      else if GET_INT_SWITCH( "-tb", Trace_Start, Trace, TRUE )
      else if GET_INT_SWITCH( "-te", Trace_End, Trace, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-testing", Testing, TRUE )
      else if SWITCH( "-gss", For_GSS, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-bs <buf_size>] [-f <file>] ", argv[0]);
	printf("[-scaling <int>] [-tb <time>] [-te <time>]\n\t");
	printf("[-tmp <dir>] [-tmax <int>] [-testing] [-gss] [-o <file>] ");
	printf("[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmp", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Tmp_Dir = (char*) realloc( Tmp_Dir, 
				     2 + (unsigned)strlen( argv[i] ) * 
				     sizeof( char ) );
	  strcpy( Tmp_Dir, argv[i] );
	  if ( ( strlen( Tmp_Dir ) != 0 )
	       && (Tmp_Dir[strlen(Tmp_Dir)-1] != '/' ) )
	    strcat( Tmp_Dir, "/" );
	  i++;
	}
	else {  
	  printf("%s: directory expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {
    if ( strlen(File_Name) != 0 )
      File_Ok = ( NULL != freopen( File_Name, "r", stdin ) );

    if ( File_Ok ) {
      Reseau = Read_Network();

      if ( Reseau != NULL ) {
	invert_network( Reseau );

	if ( For_GSS )
	  Testing = TRUE;

	if ( Testing ) {
	  if ( strlen( Out_File_Name ) != 0 )
	    File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );
	  OutFile = stdout;

	  if ( File_Ok && For_GSS ) {
	    fprintf( OutFile, 
		     "# List of non simulable elements generated by %s\n", 
		     TOOLNAME );
	    fprintf( OutFile, "# Codes: transition = 0, place = 1\n#\n");
	    fprintf( OutFile, 
		     "# Number of non simulable elements:\n#\n" );

	    Bad_Objects = (int*) calloc( (unsigned)(1 + 6*Reseau->Nb_Places +
					 4*Reseau->Nb_Trans), sizeof( int ) );
	    Bad_Objects[0] = 0;
	  }
	  else {
	    Bad_Objects == NULL;
	  }

	}
	if ( File_Ok ) {
	  /* Checking that the network satisfies the conditions */
	  Network_Ok = 
	    Has_All_Deterministic_Times( Reseau, Testing, Bad_Objects )
	    && Has_Simulable_Places( Reseau, Testing, Bad_Objects );

	  if ( Network_Ok ) {
	    Scale = Find_Time_Scale( Reseau, Scale, Testing, Bad_Objects );
	    Network_Ok = ( Scale != 0 );
	  }

	  if ( Testing && For_GSS ) {
	    fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[0] );
	    for( i=1; i<=Bad_Objects[0]; i++ ) {
	      fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i-1] );
	      fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i] + 1 );
	    }
	    fclose( OutFile );
	    return Network_Ok;
	  }
	  else {

	    if ( Debug )
	      printf("%s: time scale found: %d\n", TOOLNAME, Scale );

	    if ( Network_Ok ) {
	      /* Go for structural analysis and code generation */
	      The_Layer_Struct = (Layer_Struct) 
		malloc( sizeof( struct LAYER_STRUCT ) );

	      if ( topological_analysis( Reseau, The_Layer_Struct ) ) {

		if (Debug)
		  Dump_Layer_Struct( The_Layer_Struct );

		if ( Max_Time >= 0 )
		  Reseau->Sim_Params->Nb_Firings = Max_Time;

		gen_c_code( Reseau, The_Layer_Struct, Trace );
	      }
	      else {
		fprintf( stderr, "Topological analysis failed: ");
		fprintf( stderr, "look for an zero-time cycle.\n");
		return -3;
	      }
	    }
	    else {
	      printf("%s: the network doesn't satisfy the assumptions. ", 
		     TOOLNAME );
	      printf("No translation.\n");
	    }
	  }
	}
	else 
	  fprintf( stderr, "%s: Error opening file %s.\n", 
		   TOOLNAME, Out_File_Name );
      }
      else {
	printf("%s: Error reading network. No translation.\n", TOOLNAME );
      }
    }
    else {
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, File_Name );
      return -2;
    }
    return 0;
  }
  else if ( !Help )
    return -1;
  else 
    return 0;
}

/****************************************************************/
/* Procedure to initialize various tables.			*/
/****************************************************************/
void Init_Generation( Reseau, Max_Delay, Scale )
PetriNet	*Reseau;
int		Scale;
{
  int		i;
  int		Delay;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  Hold_Time = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof(int) );
  Fire_Time = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );

  Delay_Table = (boolean*) calloc( (unsigned)(1+Max_Delay), 
				   sizeof(boolean) );
  Is_Compet_Pl = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				   sizeof(boolean) );
  Is_Compet_Tr = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Already_Visited = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				       sizeof( boolean ) );

  for ( i=0; i<Reseau->Nb_Places; i++ ) 
    Hold_Time[i] = rint( Scale * Reseau->P[i].Pl_Law->Parameters[0] );

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    Fire_Time[i] = rint( Scale * Reseau->T[i].Tr_Law->Parameters[0] );

    Is_Compet_Tr[i] = FALSE;
    SCAN( L_Scan, Reseau->T[i].Inputs ) {
      The_Routing = Reseau->P[L_Scan->Dest].Routing;
      if ( The_Routing != NULL ) {
	if ( The_Routing->Type == 'P' ) {
	  Is_Compet_Tr[i] = TRUE;
	  Has_Competition = TRUE;
	}
      }
    }
    if ( Is_Compet_Tr[i] ) {
      /* if it is, generate Routing() code for **EACH** input transition */
      SCAN( L_Scan, Reseau->T[i].Inputs ) {
	Is_Compet_Pl[L_Scan->Dest] = TRUE;
	Has_Competition = TRUE;
      }
    }
  }

  for ( i=0; i<Max_Delay+1; i++ )
    Delay_Table[i] = FALSE;

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    SCAN( L_Scan, Reseau->P[i].Inputs ) {
      Delay = Hold_Time[ i ] + Fire_Time[ L_Scan->Dest ];
      Delay_Table[ Delay ] = TRUE;
    }
  }
  /* need Time-1 if there is competition */
  if ( Has_Competition )
    Delay_Table[1] = TRUE;

  for ( i=0; i<Reseau->Nb_Places; i++ )
    Already_Visited[i] = FALSE;
  
}
/****************************************************************/
/* Procedures for code generation...				*/
/****************************************************************/
void	gen_file_header( Reseau, Trace )
PetriNet	*Reseau;
boolean		Trace;
{
  CODE("/* Configuration file for a Petri net \t\t\t*/\n");
  CODE("/* simulation code generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  CODE("\t\t*/\n");
  CODE1("/* Date: %s \t\t\t\t*/\n\n", todays_date() );

  /* write definitions, includes */
  CODE1("int     TMAX = %d;\n", Scale*Reseau->Sim_Params->Nb_Firings );
  CODE1("int     SCALE = %d;\n", Scale );
  CODE1("#define NB_TRANS\t%d\n", Reseau->Nb_Trans );
  CODE1("#define NB_PLACES\t%d\n", Reseau->Nb_Places );
  CODE1("#define BUF_SIZE\t%d\n", Buf_Size );
  CODE("#define Infinity 0x7FFFFFFF\n" );

  /* trace defines */
  if ( Trace ) {
    CODE("#define Trace\n");

    CODE("#define CODE(x)\t\tfprintf(Tracefile,x)\n");
    CODE("#define CODE1(x,y)\tfprintf(Tracefile,x,y)\n");
    CODE("#define CODE2(x,y,z)\tfprintf(Tracefile,x,y,z)\n");
    CODE("#define CODELN\t\tCODE(\"\\n\")\n");
  }
  CODELN;

}

void Gen_Marking_Function( Reseau, Place_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  Pl_Des	The_Place;
  int		Delay;

  The_Place = Reseau->P[Place_Id];

  if ( Initial )
    CODE("int I_");
  else
    CODE("int ");

  CODE1("Marking_%d()\n{\n  register int Res;\n\n", Place_Id );
  if ( Initial )
    CODE1("  Res = Avail[%d][ Time ];\n", Place_Id );
  else
    CODE1("  Res = %d;\n", Reseau->P[Place_Id].Init_Pop );

  SCAN( L_Scan, The_Place.Inputs ) {
    Delay = Hold_Time[Place_Id] + Fire_Time[L_Scan->Dest];
    CODE1("  Res += Y[%d][ Time", L_Scan->Dest );
    if ( Delay > 0 )
      CODE1("_M_%d", Delay );
    if ( L_Scan->Weight > 1 )
      CODE1(" ] * %d;\n", L_Scan->Weight );
    else
      CODE(" ];\n");
  }
  SCAN( L_Scan, The_Place.Outputs ) {
    CODE1("  Res -= Y[%d][ Time ]", L_Scan->Dest );
    if ( L_Scan->Weight > 1 )
      CODE1(" * %d;\n", L_Scan->Weight );
    else
      CODE(";\n");
  }
  CODE("  return Res;\n}\n\n");
}

void	gen_declarations(Reseau, Max_Delay, Scale, Trace)
PetriNet	*Reseau;
int		Max_Delay;
int		Scale;
boolean		Trace;
{
  int		i, j;
  double	Cum_Proba;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  CODE1("#define MAX_DELAY %d\n", Max_Delay );

  CODE("/* table of initial availabilities\t\t\t\t\t*/\n");
  CODE("int\t\t\tAvail[NB_PLACES][MAX_DELAY] = {\n");

  for( i=0; i<Reseau->Nb_Places-1; i++ ) {
    CODE("\t");
    for( j=0; j<Max_Delay; j++ ) {
      CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
    }
    CODE("\n");
  }
  CODE("\t");
  i = Reseau->Nb_Places-1;
  for( j=0; j<Max_Delay-1; j++ ) {
    CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
  }
  CODE1("%2d", Nb_Available_Tokens( Reseau->P[i], Max_Delay-1 ) );

  CODE("\n};\n\n");

  CODE("/* table of firing times\t\t\t\t\t*/\n");
  CODE("int\t\t\tFire_Time[NB_TRANS] = {\n");

  for( i=0; i<Reseau->Nb_Trans-1; i++ )
    CODE1("\t%2d,\n", Fire_Time[i] );

  i = Reseau->Nb_Trans-1;
  CODE1("\t%2d", Fire_Time[i] );
  CODE("\n};\n\n");

  CODE("/* table of transition names\t\t\t\t\t*/\n");
  CODE("char\t\t\t*Trans_Name[NB_TRANS] = {\n");

  for( i=0; i<Reseau->Nb_Trans-1; i++ )
    if ( strlen( Reseau->T[i].Name ) > 0 )
      CODE1("\t\"%s\",\n", Reseau->T[i].Name );
    else
      CODE("\t\"\",\n" );

  i = Reseau->Nb_Trans-1;
  if ( strlen( Reseau->T[i].Name ) > 0 )
    CODE1("\t\"%s\",\n", Reseau->T[i].Name );
  else
    CODE("\t\"\"\n" );
  CODE("\n};\n\n");

  CODE("/* variables for use with the circular buffer\t\t\t*/\n");
  CODE("int Time;\t\t/* The time, modulo the buffer size */\n");
  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] )
      CODE2("int Time_M_%d;\t\t/* Time - %d */\n", i, i );
  CODELN;

  CODE("/* tables and counts for routing generation\t\t\t\t*/\n");
  for( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Routing = Reseau->P[i].Routing;

    if ( The_Routing != NULL ) {
      switch ( The_Routing->Type ) {

      case 'R':
	CODE1("int Count_%d = 0;\n", i );
	CODE2("double Proba_%d[%d] = {", i, Reseau->P[i].Outdegree );
      
	Cum_Proba = 0.0;
	for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	  Cum_Proba += The_Routing->Probas[j];
	  CODE1(" %6.4f,", Cum_Proba );
	}
	CODE(" 1.0 };\n");
	CODE2("int Dest_%d[%d] = {", i, Reseau->P[i].Outdegree );
	L_Scan = Reseau->P[i].Outputs;
	for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	  CODE1(" %d,", L_Scan->Dest );
	  L_Scan = L_Scan->Next;
	}
	CODE1(" %d };\n", L_Scan->Dest );

	break;

      case 'D':
	CODE1("int Count_%d = 0;\n", i );
	CODE1("int Phase_%d = 0;\n", i );
	CODE2("int Cycle_%d[%d] = {", i, The_Routing->Period );
	for ( j=0; j<The_Routing->Period-1 ; j++ ) {
	  CODE1(" %d,", The_Routing->Cycle[j] );
	}
	CODE1(" %d };\n",
	      The_Routing->Cycle[The_Routing->Period-1] );

	break;

      default:
	if ( Is_Compet_Pl[i] )
	  CODE1("int Count_%d = 0;\n", i );

	break;
      }
    } /* if ( Routing != NULL ) */
    else if ( Is_Compet_Pl[i] )
	  CODE1("int Count_%d = 0;\n", i );

    CODE1("int H_%d[NB_TRANS];\n", i );

    if ( ( Is_Compet_Pl[i] ) && !Already_Visited[i] ) {
      /* generate Marking code 						*/
      CODE2("/* Function to compute the marking of place %d (%s) */\n",
	    i, Reseau->P[i].Name );
      Gen_Marking_Function( Reseau, i, TRUE );
      Gen_Marking_Function( Reseau, i, FALSE );
      Already_Visited[i] = TRUE;
    }
  }

  if ( Trace ) {

    CODE("/* declarations for trace handling */\n");

    CODE("char  *gss_trans_name[NB_TRANS] = {\n");
    for( i=0; i<Reseau->Nb_Trans-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->T[i].Data );
	
    CODE1("\t\"%s\"\n", Reseau->T[Reseau->Nb_Trans-1].Data );
    CODE("};\n\n");

    CODE("char  *gss_place_name[NB_PLACES] = {\n");
    for( i=0; i<Reseau->Nb_Places-1; i++ )
      CODE1("\t\"%s\",\n", Reseau->P[i].Data );

    CODE1("\t\"%s\"\n", Reseau->P[Reseau->Nb_Places-1].Data );
    CODE("};\n\n");

/*     CODE("char  *gss_platra_link_name[NB_PLACES][NB_TRANS];\n"); */
/*     for( i=0; i<Reseau->Nb_Places; i++ ) */
/*       SCAN( L_Scan, Reseau->P[i].Outputs ) */
/* 	CODE3("gss_platra_link_name[%d][%s] = \"%s\",\n",  */
/* 	      i, L_Scan->Dest, L_Scan->Data ); */

/*     CODE("char  *gss_trapla_link_name[NB_TRANS][NB_PLACES];\n"); */
/*     for( i=0; i<Reseau->Nb_Trans; i++ ) */
/*       SCAN( L_Scan, Reseau->T[i].Outputs ) */
/* 	CODE3("gss_trapla_link_name[%d][%s] = \"%s\",\n",  */
/* 	      i, L_Scan->Dest, L_Scan->Data ); */

    CODE("/* table of initial populations\t\t\t\t\t*/\n");
    CODE("int\t\t\tInitial_Pop[NB_PLACES] = {\n");

    for( i=0; i<Reseau->Nb_Places-1; i++ ) {
      CODE1("\t%d", Reseau->P[i].Init_Pop );
      CODE(",\n");
    }
    CODE1("\t%d", Reseau->P[Reseau->Nb_Places-1].Init_Pop );
    CODE("\n};\n\n");

    CODE("int   Cur_X_Evt[NB_TRANS+1];\n");
    CODE("int   Cur_Y_Evt[NB_TRANS+1];\n");
    CODELN;
    CODE("FILE  *Tracefile;\n");
    CODE("int   More_Events;\n");
    CODE("int   Was_X_Evt, Was_Y_Evt;\n");
    CODE("int   Min_X_Val, Min_Y_Val;\n");
    CODE("int   Arg_Min_X, Arg_Min_Y;\n");
    CODE("int   Old_Date, New_Date;\n");
    CODELN;
    CODE1("int  Trace_Start = %d;\n", Trace_Start );
    CODE1("int  Trace_End = %d;\n", Trace_End );
  }
}

void gen_trace_before_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  if ( The_Trans.Indegree > 1 )
    CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
  /* first remove tokens from input places, except dummy ones */
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Inputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( Time>%d ? Y[%d][ Time - %d ] : 0 )", 
	      M_Scan->Dest, Delay, Delay );
	else
	  CODE2("  Marking += Y[%d][ Time - %d ]", M_Scan->Dest, Delay );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	CODE1("  Marking -= Y[%d][ Time ]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then move the tokens */
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_place_name[%d], ", L_Scan->Dest );
      CODE1("gss_trans_name[%d] );\n", Trans_Id );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then switch transition status */
  CODE1("  CODE(\"MODIFY %s \\\"status\\\" = 0\\n\");\n",
	The_Trans.Name );
}

void gen_trace_after_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  /* move tokens to (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {

      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_trans_name[%d],", Trans_Id );
      CODE1("gss_place_name[%d] );\n", L_Scan->Dest );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* finally change marking of (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( Time>%d ? Y[%d][ Time - %d ] : 0 )", 
	      M_Scan->Dest, Delay, Delay );
	else
	  CODE2("  Marking += Y[%d][ Time - %d ]", M_Scan->Dest, Delay );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	CODE1("  Marking -= Y[%d][ Time ]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  CODE("  CODE(\"eol+\\n\");\n");
}

void	write_trans_name( The_Trans )
Tr_Des	The_Trans;
{
  if ( NULL != The_Trans.Name )
    printf( "%s", The_Trans.Name );
  else
    printf( "_TR_%d", The_Trans.Tr_Id );
}

/* T1, T2: btw 0 and NB_TRANS-1 */
void	Write_Token_Total( Reseau, Place_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  int		Delay;

  if ( Already_Visited[ Place_Id ] ) {
    CODE1("Nb_Tokens = Count_%d;\n", Place_Id );
  }
  else {
    if ( Initial )
      CODE1("Nb_Tokens = Avail[%d][Time];\n", Place_Id );
    else
      CODE1("Nb_Tokens = %d;\n", Reseau->P[Place_Id].Init_Pop );

    SCAN( L_Scan, Reseau->P[Place_Id].Inputs ) {
      CODE( "Nb_Tokens += ");

      Delay = IRINT( Scale * 
		     ( Reseau->P[Place_Id].Pl_Law->Parameters[0]
		       + Reseau->T[L_Scan->Dest].Tr_Law->Parameters[0] ) );

      if ( Initial ) {
	CODE1( "( Time >= %d ? ", Delay );
	CODE2( "Y[%d][Time-%d] ", L_Scan->Dest, Delay );
	if ( L_Scan->Weight > 1 )
	  CODE1("* %d", L_Scan->Weight );
	CODE(" : 0 );\n");
      }
      else {
	if ( Delay == 0 )
	  CODE1( "Y[%d][ Time ]", L_Scan->Dest );
	else
	  CODE2( "Y[%d][ Time_M_%d ]", L_Scan->Dest, Delay );

	if ( L_Scan->Weight > 1 )
	  CODE1(" * %d", L_Scan->Weight );
	CODE(";\n");
      }
    }
  }
}

void	Write_Token_Routing( Reseau, Place_Id, Trans_Id, Initial )
PetriNet	*Reseau;
int		Place_Id;
int		Trans_Id;
boolean		Initial;
{
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  The_Routing = Reseau->P[Place_Id].Routing;

  if ( Reseau->P[Place_Id].Outdegree > 1 ) {
    if ( !Already_Visited[Place_Id] ) {
      CODE2("/* Computing number of tokens arrived in place %d (%s) */\n",
	    Place_Id, Reseau->P[Place_Id].Name );

      if ( The_Routing != NULL ) {
	switch( The_Routing->Type ) {

	case 'R':

	  CODE2("for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
		Place_Id, Place_Id );
	  CODE("  U = U_0_1();\n");
	  CODE1("  for( j=0; U > Proba_%d[j]; j++ );\n", Place_Id );
	  CODE2("  H_%d[Dest_%d[j]]++;\n", Place_Id, Place_Id );
	  CODE("}\n");

	  Already_Visited[ Place_Id ] = TRUE;

	  break;

	case 'D':

	  CODE2("for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
		Place_Id, Place_Id );
	  CODE3("  H_%d[ Cycle_%d[ Phase_%d ] ]++;\n",
		Place_Id, Place_Id, Place_Id );
	  CODE1("  Phase_%d++;\n", Place_Id );
	  CODE3("  Phase_%d = ( Phase_%d %% %d );\n", 
		Place_Id, Place_Id, The_Routing->Period );
	  CODE("}\n");

	  Already_Visited[ Place_Id ] = TRUE;

	  break;

	case 'P':

	  CODE("/* Start firing as long as  transition ");
	  CODE2("%d (%s) is enabled */\n", 
		Trans_Id, Reseau->T[Trans_Id].Name );
	  CODE("Is_Enabled = TRUE;\n");
	  CODE1("while( Is_Enabled && ( Count_%d < Nb_Tokens ) ) {\n", 
	       Place_Id );
	  CODE("  Is_Enabled = ");
	  L_Scan = Reseau->T[Trans_Id].Inputs;
	  /* Call to the marking function, with all Time parameters */
	  if ( Initial )
	    CODE("( I_");
	  else
	    CODE("( ");
	  CODE2("Marking_%d() >= %d )", L_Scan->Dest, L_Scan->Weight );
	  SCAN( L_Scan, Reseau->T[Trans_Id].Inputs->Next ) {
	    CODE2("\n      && ( Marking_%d () >= %d )",
		 L_Scan->Dest, L_Scan->Weight );
	  }
	  CODE(";\n");
	  CODE("  if ( Is_Enabled ) {\n");
	  CODE2("    H_%d[ %d ]++;\n", Place_Id, Trans_Id );
	  CODE1("    Count_%d++;\n", Place_Id );
	  CODE("  }\n");
	  CODE("}\n");

	  break;

	default:

	  printf("%s: error: unexpected routing policy (%c)\n",
		 TOOLNAME, The_Routing->Type );
	} /* switch(...) */
      } /* if ( Routing != NULL ) */
      else {
	  CODE1("Count_%d = Nb_Tokens;\n", Place_Id );
	  Already_Visited[ Place_Id ] = TRUE;
      }
    }

    CODE2("Nb_Tokens = H_%d[%d];\n", Place_Id, Trans_Id );
  }
}

void Write_Token_Reduction( The_Link )
Liste_Link	The_Link;
{
  if ( The_Link->Weight > 1 )
    CODE1("Nb_Tokens /= %d;\n", The_Link->Weight );
}

void	gen_code_node( Reseau, The_Node, Initial, Trace )
PetriNet	*Reseau;
int		The_Node;
boolean		Initial;
boolean		Trace;
{
  Liste_Link	L_Scan;
  Tr_Des	The_Trans;
  int		Pred_Pl_Id;
  int		Delay;

  The_Trans = Reseau->T[ The_Node ];

  CODE("/* Evolution of transition ");
  write_trans_name( Reseau->T[The_Node] );
  CODE(" */\n"); 

  switch( The_Trans.Indegree ) {

  case 0:
    /* this is a source but it should be recycled...	*/
    CODE2("Nb_Tokens = Y[%d][Time_M_%d] + 1;\n", 
	  The_Node, Fire_Time[The_Node] );
    printf("%s: warning: non recycled source (%d). ", TOOLNAME );
    printf("Assumed recycled.\n" );

    CODE1("Y[%d][Time] = Nb_Tokens;\n", The_Node );

    break;

  case 1:
    /* single input. No min */
    Pred_Pl_Id = The_Trans.Inputs->Dest;
    Write_Token_Total( Reseau, Pred_Pl_Id, Initial );
    Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial );
    Write_Token_Reduction( The_Trans.Inputs );

    CODE1("Y[%d][Time] = Nb_Tokens;\n", The_Node );

    break;

  default:

    CODE("Tmp_Y = Infinity;\n" );

    for( L_Scan=The_Trans.Inputs; L_Scan != NULL; L_Scan=L_Scan->Next ) {

      Pred_Pl_Id = L_Scan->Dest;
      Write_Token_Total( Reseau, Pred_Pl_Id, Initial );
      Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial );
      Write_Token_Reduction( L_Scan );
      
      CODE("Tmp_Y = MIN( Tmp_Y, Nb_Tokens );\n" );
    }

    CODE1("Y[%d][Time] = Tmp_Y;\n", The_Node );

  } /* switch () */

#define OLDY(x)	CODE3("( Time>%d ? Y[%d][Time-%d] : 0 )", The_Node, x, x )

  if ( Trace ) {
    Delay = Fire_Time[The_Node];
    CODE("Present_Nb = ");
    if ( Initial )
      OLDY( Delay );
    else
      CODE2("Y[%d][ Time - %d ]", The_Node, Delay );
    CODE(";\n");
    
    CODE3("Y[%d][(Time>%d?Time-%d:0)]", 
	  The_Node, Delay, Delay );
    CODE3(" = Y[%d][(Time>%d?Time-%d:0)];\n", 
	  The_Node, Delay+1, Delay+1 );
    CODE("if( ( Time <= Trace_End ) && ( Time >= Trace_Start ) )\n");
    CODE3("while( Y[%d][(Time>%d?Time-%d:0)] < Present_Nb ) {\n",
	  The_Node, Delay, Delay );
    CODELN;
    gen_trace_before_firing( Reseau, The_Node, Initial );
    gen_trace_after_firing( Reseau, The_Node, Initial );
    CODELN;
    CODE3("  Y[%d][(Time>%d?Time-%d:0)]++;\n", The_Node, Delay, Delay );
    CODE("}\n");
  }
  CODELN;
}

void	gen_preparation_code( Reseau )
PetriNet	*Reseau;
{
  int		i;
  Liste_Link	L_Scan;

  if ( Has_Competition ) {
    CODE("/* Prepare enabling count variables for transitions with ");
    CODE("conflicts */\n");

    for( i=0; i<Reseau->Nb_Trans; i++ )
      if ( Is_Compet_Tr[i] )
	CODE2("Y[%d][Time] = Y[%d][Time_M_1];\n", i, i );

    CODELN;
  }
}

void	gen_increment_code( Max_Delay )
int	Max_Delay;
{
  int	i;

  CODE("/* Increment of the time variables */\n");
  CODE("Time = ( Time + 1 ) %% BUF_SIZE;\n");
  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] )
      CODE2("Time_M_%d = ( Time + %d ) %% BUF_SIZE;\n", 
	    i, Buf_Size - i );
  CODELN;
}

void 	gen_loop_code( Reseau, The_Layer_Struct, Initial, Trace )
PetriNet	*Reseau;
Layer_Struct	The_Layer_Struct;
boolean		Initial;
boolean		Trace;
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;

  if ( Trace ) {
    CODE("CODE1(\"MODIFY TheClock \\\"ctime\\\" = %%7.3f");
    CODE("\\n\", (double)Time );\n");
  }

  for( Num_Layer = 0; Num_Layer < The_Layer_Struct->Nb_Levels; Num_Layer++)
    for ( Num_Node = 0; 
	  Num_Node < The_Layer_Struct->Level[Num_Layer]->Nb_Nodes;
	  Num_Node++ ) {
	    
      The_Node = The_Layer_Struct->Level[Num_Layer]->Node_Id[Num_Node];
      gen_code_node( Reseau, The_Node, Initial, Trace );
    }
}

boolean Is_Dummy( The_Place )
Pl_Des	The_Place;
{
  if ( NULL != The_Place.Name )
    return FALSE;
  else
    return ( 0 == strncmp( The_Place.Name, "_Dummy_Place", 12 ) );
}

int	gen_c_code(Reseau, The_Layer_Struct, Trace )
PetriNet	*Reseau;
Layer_Struct	The_Layer_Struct;
boolean		Trace;
{
  int		Max_Dependency;
  char		File_Name[512];
  int		i;

  Max_Dependency = rint( Scale * max( Max_Delay( Reseau ), 
				      Max_Lag( Reseau ) ) );
  
  /* adjusting buffer size and trace parameters		*/
  Buf_Size = min( Buf_Size, Scale*Reseau->Sim_Params->Nb_Firings + 1 );
  Buf_Size = max( Buf_Size, Max_Dependency );

  if ( Trace ) {
    Trace_Start = max( Scale*Trace_Start, 0 );
    Buf_Size = max( Buf_Size, Trace_Start );
  }

  /* initializing some data structures */
  Init_Generation( Reseau, Max_Dependency, Scale );

  /* generating global parameters of the graph */
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_graph_parameters.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )
    gen_file_header( Reseau, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_declarations.h" );
  if ( NULL != freopen( File_Name, "w", stdout ) )    
    gen_declarations( Reseau, Max_Dependency, Scale, Trace );
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the initialization loop.		*/
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_initial_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) ) {
    gen_preparation_code( Reseau );
    gen_loop_code( Reseau, The_Layer_Struct, TRUE, Trace );
    gen_increment_code( Max_Dependency );
  }
  else	
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  /* code for the main loop. 			*/
  for ( i=0; i<Reseau->Nb_Places; i++ )
    Already_Visited[i] = FALSE;
  strcpy( File_Name, Tmp_Dir );
  strcat( File_Name, "pnc_evolution.c" );
  if ( NULL != freopen( File_Name, "w", stdout ) ) {
    gen_preparation_code( Reseau );
    gen_loop_code( Reseau, The_Layer_Struct, FALSE, Trace );
    gen_increment_code( Max_Dependency );
  }
  else
    fprintf( stderr, "%s: error opening %s\n", TOOLNAME, File_Name );

  return TRUE;
}
