/************************************************************************/
/* module to compile an PN into an executable performing its equational	*/
/* simulation.								*/
/************************************************************************/
#include <unistd.h>
#include "pn.h"
#define RESOURCES
#include "resources.h"

#define TOOLNAME	"pn_sim"

/* Extern from the (Common) module transcript.o */
extern boolean	Write_PN_As_PNL();

/* Externs from the module resource.o */
extern void 	get_resources();
extern char     **Resource_Value;
extern boolean	*Is_Available;
extern int  	execute();

#define STOP    (char *)0
#define RV(x)   Resource_Value[ x ]

boolean Resources_Are_Present()
{
  return ( Is_Available[ C_COMPILER ] && Is_Available[ PN2C ] );
}

boolean	Compile_PN(
PetriNet	*Reseau,
boolean 	Checking,
char		*Tmp_Dir,
boolean 	Keep )
{
  char	Sim_Name[1024];
  char	Src_Name[1024];
  char	Lib_Name[1024];

  boolean	All_Ok;

  get_resources( Checking );

  All_Ok = Resources_Are_Present();

  if ( (!Checking) && All_Ok ) {

    /* Dumping the network for immediate re-reading... 	*/
    /* Sounds stupid, but not simple to avoid...        	*/      
    Write_PN_As_PNL( Reseau, RV( TMP_PNL_FILE ), TOOLNAME );

    if ( strlen( Tmp_Dir ) == 0 )
      strcpy( Sim_Name, RV( TMP_DIR ) );
    else 
      strcpy( Sim_Name, Tmp_Dir );
    strcat( Sim_Name, "/" );
    strcat( Sim_Name, RV( SIM_NAME ) );

    strcpy( Lib_Name, "-I" );
    strcat( Lib_Name, RV( TMP_DIR ) );

    execute( TOOLNAME,
	     RV( PN2C ), "-f", RV( TMP_PNL_FILE ), "-tmp", RV( TMP_DIR ), 
	     STOP );

    execute( TOOLNAME,
	     RV( C_COMPILER ), RV( C_SOURCE_FILE ), "-o", Sim_Name,
	     Lib_Name, "-lm", "-O2",
	     STOP );

    execute( TOOLNAME, Sim_Name, STOP );

    /* cleaning up */
    if ( !Keep ) {
      unlink( RV( TMP_PNL_FILE ) );
      unlink( Sim_Name );
      if ( strlen( Tmp_Dir ) == 0 ) {
	strcpy( Src_Name, RV( TMP_DIR ) );
	strcat( Src_Name, "/pnc_initial_evolution.c" );
	unlink( Src_Name );
	strcpy( Src_Name, RV( TMP_DIR ) );
	strcat( Src_Name, "/pnc_evolution.c" );
	unlink( Src_Name );
	strcpy( Src_Name, RV( TMP_DIR ) );
	strcat( Src_Name, "/pnc_graph_parameters.h" );
	unlink( Src_Name );
	strcpy( Src_Name, RV( TMP_DIR ) );
	strcat( Src_Name, "/pnc_declarations.h" );
	unlink( Src_Name );
      }
      else {
	strcpy( Src_Name, Tmp_Dir );
	strcat( Src_Name, "/pnc_initial_evolution.c" );
	unlink( Src_Name );
	strcpy( Src_Name, Tmp_Dir );
	strcat( Src_Name, "/pnc_evolution.c" );
	unlink( Src_Name );
	strcpy( Src_Name, Tmp_Dir );
	strcat( Src_Name, "/pnc_graph_parameters.h" );
	unlink( Src_Name );
	strcpy( Src_Name, Tmp_Dir );
	strcat( Src_Name, "/pnc_declarations.h" );
	unlink( Src_Name );
      }
    }

  }
  else if (!Checking) {
    fprintf( stderr, "%s: error: some resources are missing ", TOOLNAME );
    fprintf( stderr, "(check with -c)\n");
    return FALSE;
  }

  /* execution presumed successful */
  return TRUE;
}
