/* A program for simulating Free Choice Nets based on the evolution	*/
/* equations of Baccelli/Foss/Gaujal, and some more by Gaujal/Jean-Marie*/

#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <time.h>
#include "pn.h"

#define TOOLNAME	"pn_sim"
#define max(x,y)        ( (x>y) ? x : y )
#define IRINT(x)	(int)rint(x)

/* additionnal declaration, not in the standard "math.h" */
extern double rint(double x);

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

/* Externs from the module parse_pn.o */
extern PetriNet		*Read_Network( char *File_Name );
extern void		dumpe_network();

/* Externs from the module analyze_pn.o */
extern boolean		topological_analysis();
extern boolean		Has_All_Deterministic_Objects();
extern boolean		Has_Simulable_Places();
extern boolean		Has_Real_Sources(PetriNet *, boolean, int *);
extern void		Dump_Layer_Struct();
extern double		Max_Lag();
extern double		Max_Delay();
extern int		Find_Time_Scale(PetriNet *Reseau, boolean, 
					int, boolean, int *);

/* Externs from modules of pn_sim. */
extern void		simulate();
extern void		init_simulation();
extern void		dumpe_simulation();
extern void		write_stat_results();
extern boolean		Compile_PN(PetriNet *Reseau, boolean Checking, 
				   char *Tmp_Dir, boolean Keep );

int main(int argc, char *argv[] )
{
  static char		File_Name[128] = "";
  PetriNet 		*Reseau;
  int			i;
  int			Max_Time = 1000;
  int			Buf_Size = 1024;
  int			Max_Dependency;
  boolean		Help = FALSE;
  boolean		Compile = FALSE;
  boolean		Keep = FALSE;
  boolean		Checking = FALSE;
  boolean		Args_Ok = TRUE;
  int			Scale = 10;
  struct timeval	t_debut;
  struct timeval	t_fin;
  Layer_Struct		The_Layer_Struct;
  char			*Tmp_Dir;

  Tmp_Dir = (char*) malloc( sizeof( char ) );
  strcpy( Tmp_Dir, "" );

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name )
      else if GET_INT( "-bs", Buf_Size )
      else if GET_INT( "-scale", Scale )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-c", Checking, TRUE )
      else if SWITCH( "-compile", Compile, TRUE )
      else if SWITCH( "-keep", Keep, TRUE )
      else if GET_INT( "-tmax", Max_Time )
      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {

	printf("Usage: %s [-f <file>] [-tmax <int>] [-bs <int>]",
	       TOOLNAME );
	printf(" [-scale <int>]\n\t[-compile] [-keep] [-tmp] [-c] ");
	printf(" [-debug] [-h] [-?]\n" );
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmp", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Tmp_Dir = (char*) realloc( Tmp_Dir, 
				     2 + (unsigned)strlen( argv[i] ) * 
				     sizeof( char ) );
	  strcpy( Tmp_Dir, argv[i] );
	  if ( ( strlen( Tmp_Dir ) != 0 )
	       && (Tmp_Dir[strlen(Tmp_Dir)-1] != '/' ) )
	    strcat( Tmp_Dir, "/" );
	  i++;
	}
	else {  
	  printf("%s: directory expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

      The_Layer_Struct = (Layer_Struct) 
	malloc( sizeof( struct LAYER_STRUCT ) );

      if ( Has_All_Deterministic_Objects( Reseau, FALSE, NULL ) 
	   && Has_Simulable_Places( Reseau, FALSE, NULL )
	   && Has_Real_Sources( Reseau, FALSE, NULL )
	   && ( 0 != ( Scale 
		= Find_Time_Scale( Reseau, TRUE, Scale, FALSE, NULL ) ) ) ) {

	if ( topological_analysis( Reseau, The_Layer_Struct ) ) {

	  if (Debug)
	    Dump_Layer_Struct( The_Layer_Struct );

	  Reseau->Sim_Params->Nb_Firings = Max_Time;

	  Max_Dependency = IRINT( Scale * 
				  max( Max_Delay( Reseau ), 
				       Max_Lag( Reseau ) ) );

	  Buf_Size = max( Buf_Size, 1 + Max_Dependency );

	  if ( !Compile ) {
	    gettimeofday(&t_debut,(struct timezone *)0);
	    simulate(Reseau, The_Layer_Struct, Scale, Buf_Size);
	    gettimeofday(&t_fin,(struct timezone *)0);
	    
	    if (Debug)
	      dumpe_simulation(Reseau, Buf_Size);

	    write_stat_results(Reseau, Scale, Buf_Size, t_debut, t_fin);
	  }
	  else {
	    (void)Compile_PN( Reseau, Checking, Tmp_Dir, Keep );
	  }

	  return 0;
	}
	else {
	  printf("%s: zero time cycle encountered. No simulation.\n", 
		 TOOLNAME );
	  return -1;
	}
      }
      else {
	printf("%s: the network doesn't satisfy the assumptions. ", TOOLNAME );
	printf("No simulation.\n");
	return -1;
      }
    }
    else {
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, File_Name );
      return -2;
    }

    return 0;
  }
  else if ( !Help )
    return -1;
  else 
    return 0;
}

