/************************************************************************/
/* Utilities to control the external programs or files used by pn_sim	*/
/* The file resources.h can be generated from data file resouces.dat	*/
/************************************************************************/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "pn.h"
#include "resources.h"

#define TOOLNAME "pn_sim"

extern	char	*getenv();
char	**Resource_Value;
boolean	*Is_Available;

void	get_resources( boolean	Checking )
{
  char	*The_Val;
  int	i;
  char  *Home;
  char	*Sub_Path;
  char	*Orig_Path;
  char	*All_Path = NULL;
  char	*Res_Name;
  char	*Full_Name;

  Home = getenv( "ERSHOME" );
  Orig_Path = getenv( "PATH" );
  All_Path = strdup( Orig_Path );

  Resource_Value = (char**) calloc( NB_RESOURCES, sizeof(char*) );
  Is_Available = (boolean*) calloc( NB_RESOURCES, sizeof(boolean) );

  Res_Name = (char*) calloc( 256, sizeof(char) );
  Full_Name = (char*) calloc( 256, sizeof(char) );

  for (i=0; i<NB_RESOURCES; i++) {

    Is_Available[i] = FALSE;

    /* if specified in the environment, get it	*/
    if ( NULL != ( The_Val = getenv( Resource_Name[i] ) ) ) {
      Resource_Value[i] = (char*) calloc( (unsigned)(1+strlen(The_Val)),
					  sizeof( char ) );
      strcpy( Resource_Value[i], The_Val );
    }
    else {
      /* just use the default */
      Resource_Value[i] = (char*) 
	calloc( (unsigned)(1+strlen(Default_Value[i])), sizeof( char ) );
      strcpy( Resource_Value[i], Default_Value[i] );
    }

    if ( Resource_Value[i][0] == '@' ) {
      /* it is relative to ERSHOME. Look for it there */
      if ( NULL != Home ) {
	strcpy( Res_Name, Home );
	Resource_Value[i][0] = '/';
	Res_Name = strcat( Res_Name, Resource_Value[i] );
      }
      else {
	/* no ERSHOME found. Pretending it is "/" */
	Resource_Value[i][0] = '/';
      }
    }
    else {
      strcpy( Res_Name, Resource_Value[i] );
    }

    /* reinitialize path, possibly damaged by "strtok" */
    if ( Resource_Value[i][0] != '/' ) {
      if ( Orig_Path != NULL )
	All_Path = strdup( Orig_Path );
      Sub_Path = strtok( All_Path, ":" );
    }
    else
      Sub_Path = NULL;

    do {
      if ( Debug )
	printf( "%s: testing dir %s\n", TOOLNAME, 
		( Sub_Path == NULL ? "<NULL>" : Sub_Path ) );

      if ( NULL != Sub_Path ) {
	strcpy( Full_Name, Sub_Path );
	Full_Name = strcat( Full_Name, "/" );
	Full_Name = strcat( Full_Name, Res_Name );
      }
      else
	strcpy( Full_Name, Res_Name );

      if ( 0 == access( Full_Name, Access_Code[i] ) ) {
	if ( Checking )
	  fprintf( stderr, "%s: %s found (%s)\n",
		   TOOLNAME, Full_Name, Resource_Name[i] );
	Resource_Value[i] = realloc( Resource_Value[i], 
				     (unsigned)(1+strlen( Full_Name )) );
	strcpy( Resource_Value[i], Full_Name );
	Is_Available[i] = TRUE;
      }

      Sub_Path = strtok( (char*)NULL, ":" );
    } while ( !Is_Available[i] && ( NULL != Sub_Path) );

    if ( !Is_Available[i] && Checking && Is_Essential[i] )
      fprintf( stderr, "%s: %s not accessible (%s)\n",
	       TOOLNAME, Resource_Value[i],
	       Resource_Name[i] );
  }

  /* cleanup */
  if ( All_Path != NULL )
    free( All_Path );
}
