#include <stdio.h>
#include <malloc.h>
#include <math.h>

#ifndef BOOLEAN
#define BOOLEAN
typedef short   boolean;
#define TRUE    1
#define FALSE   0
#endif

#define	TOOLNAME	"md1"
#define VERSION 	"v1.1"

#define	MAX(x,y)	( (x>y) ? (x) : (y) )

boolean	Debug;
boolean	Queue = FALSE;
boolean Log = FALSE;
boolean Log10 = FALSE;

void	Print_Range(double,double,double,double,double,double*,double*);
void	Build_Ref_Values(int,double,double*,double*);
void	Build_Pow_Table(int,double,double*);

int main(
int	argc,
char	*argv[] )
{
	static char		Result_File_Name[256] = "";
	int			i;
	int			N_Max;
	boolean			Help = FALSE;
	boolean			Args_Ok = TRUE;
	boolean			Value = FALSE;
	double			x_Start = 0.0;
	double			x_Stop = 10.0;
	double			x_Step = 0.1;
	double			x_Val;
	double			Lambda = 0.0;
	double			Mu = 1.0;
	double			*Pow_Rho_F;
	double			*Phi; /* voir notes MD1-12 et. seq. */

	i = 1;

	if (argc>1)
	do {
	if ( 0 == strcmp( "-o", argv[i] ) )
	{	i++;
		if ( i < argc )
		{	strcpy( Result_File_Name, argv[i] );
			i++;
		}
		else {	printf("%s: filename expected\n",argv[0]);
			Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-lambda", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &Lambda ) ) )
		    i++;
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-mu", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &Mu ) ) )
		    i++;
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-xfrom", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Start ) ) )
		    i++;
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-xto", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Stop ) ) )
		    i++;
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-xval", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Val ) ) )
		{   i++;
		    Value = TRUE;
		}
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-xstep", argv[i] ) )
	{	i++;
		if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Step ) ) )
		    i++;
		else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
		    Args_Ok = FALSE;
		}
	}
	else if ( 0 == strcmp( "-queue", argv[i] ) )
	{	Queue = TRUE;
		i++;
	}
	else if ( 0 == strcmp( "-log", argv[i] ) )
	{	Log = TRUE;
		Log10 = FALSE;
		i++;
	}
	else if ( 0 == strcmp( "-log10", argv[i] ) )
	{	Log10 = TRUE;
		Log = FALSE;
		i++;
	}
	else if ( 0 == strcmp( "-debug", argv[i] ) )
	{	Debug = TRUE;
		i++;
	}
	else if ( ( 0 == strcmp( "-h", argv[i] ) )
	       || ( 0 == strcmp( "-?", argv[i] ) ) )
	{
	    printf("Usage: %s [-o <file>] [-lambda <real>] [-mu <real>] ", 
								argv[0]);
	    printf("[-xfrom <real>]\n\t[-xto <real>] [-xval <real>] ");
	    printf("[-xstep <real>] [-queue] [-log] [-log10]");
	    printf("\n\t[-debug] [-h] [-?]\n");
		Args_Ok = FALSE;
		Help = TRUE;
		i++;
	}
	else if ( 0 == strcmp( "-v", argv[i] ) )
	{	printf("%s (Version %s)\n",argv[0],VERSION);
		Args_Ok = FALSE;
		i++;
	}
	else	i++; /* skip unknown argument */
	}
	while ( i < argc );

	if (Args_Ok)
	{
	    if ( Value )
		{
		    x_Start = x_Val;
		    x_Stop = x_Val;
		}

	    N_Max = 1 + (int) floor( x_Stop * Mu );

	    /* allocate here the memory */
	    Pow_Rho_F = (double*) calloc( N_Max, sizeof( double ) );
	    Phi = (double*) calloc( N_Max, sizeof( double ) );

	    if ( N_Max > 0 )
	    {
		Build_Pow_Table( N_Max, Lambda/Mu, Pow_Rho_F );
		Build_Ref_Values( N_Max, Lambda/Mu, Pow_Rho_F, Phi );
	    }

	    Print_Range( x_Start, x_Stop, x_Step, 
				Lambda, Mu, Pow_Rho_F, Phi );
	    return 0;
	}

	else if (Help)
	    return 0;
	else
	    return 1;
}

/* procedure pour creer la table Pow_Rho_F des rho^s/s! */
void Build_Pow_Table(int	N_Max,
		     double	Rho,
		     double	*Pow_Rho_F )
{
  int	i;

  Pow_Rho_F[0] = 1.0;
  for ( i=1; i<=N_Max; i++ )
    Pow_Rho_F[i] = Pow_Rho_F[i-1] * Rho / (double)i;

}

/* procedure pour creer la table Phi */
void Build_Ref_Values(int	N_Max,
		      double	Rho,
		      double	*Pow_Rho_F,
		      double	*Phi )
{
  int		i;
  int		s;
  double	sum;
  double  	Lexp;

  Lexp = exp( Rho );
  /* first, initialize Phi */
  Phi[0] = (1-Rho)*Lexp;
  if ( Debug )
    printf("%8.4f", Phi[0] );

  /* second, build recursively the Phis */
  for ( i=1; i<=N_Max; i++ ) {
    sum = Lexp*Phi[i-1];
    for ( s=1; s<=i; s++ ) {
      sum -= Phi[i-s] * Pow_Rho_F[s];
    }
    Phi[i] = sum;
    if ( Debug )
      printf("%8.4f", Phi[i] );
  }
  if ( Debug )
    printf("\n");
}

/* procedure pour imprimer les valeurs de la distribution */
void	Print_Range(double	Start,
		    double	Stop,
		    double	Step,
		    double	Lambda,
		    double	Mu,
		    double	*Pow_Rho_F,
		    double	*Phi )
{
  int	s;
  int	p;
  int	Cur_i;	/* index of the current interval */
  int	Nb_Points;
  double  Val;
  double	Rho;
  double	prod;
  double	F;

  Rho = Lambda / Mu;

  if ( Start > Stop ) {
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    exit( 1 );
  }
  else if ( Step < 0.0 ) {
    fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    exit( 1 );
  }

  Nb_Points = ( Stop == Start ?
		1 :
		1 + floor( ( Stop - Start ) / Step ) );

  if ( Debug ) {
    printf("Waiting time distribution for the M/D/1" );
    printf(" on range [%f,%f]\n", Start, Stop );
    printf("Load factor = %f\n", Rho );
  }

  Cur_i = 0;

  for ( p=0; p<Nb_Points; p++ ) {
    Val = Start + p*Step;
    /* find the interval */

    Cur_i = (int) floor( Val * Mu );

    if ( Val < 0.0 )
      F = 0.0;
    else if ( Val == (Cur_i+1) / Mu )
      F = Phi[Cur_i];
    else {	/* j'applique la convolution de la page MD1-12 */
      F = Phi[Cur_i];
      prod = (Cur_i+1) - Mu*Val;
      for ( s=1; s<=Cur_i; s++ ) {
	F += Phi[Cur_i-s] * Pow_Rho_F[s] * prod;
	prod *= (Cur_i+1) - Mu*Val;
      }
    }
    F *= exp( Rho * ( Val*Mu - (Cur_i+1) ) );

    if ( !Queue )
      printf("%10.6f %12.8f\n", Val, F );
    else {
      if ( !Log && !Log10 )
	printf("%10.6f %14.10e\n", Val, 1.0-F );
      else if ( Log )
	printf("%10.6f %12.8f\n", Val, log(1.0-F) );
      else /* Log10 == TRUE */
	printf("%10.6f %12.8f\n", Val, log10(1.0-F) );
    }
  }
}
