#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "stats.h"

#define	TOOLNAME	"laplace"

boolean	Debug;

/* forward declaration */
void	Compute_Laplace_Range( double Start, double Stop, double Step );

int main(
int	argc,
char	*argv[] )
{
  static char		Sample_File_Name[128] = "";
  static char		Result_File_Name[128] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Range = TRUE;
  double		s_Start = 0.0;
  double		s_Stop = 10.0;
  double		s_Step = 0.1;
  double		s_Val;

  i = 1;

  if (argc>1)
    do {
      if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc )
	  {	strcpy( Sample_File_Name, argv[i] );
	  i++;
	  }
	else {	printf("%s: filename expected\n",argv[0]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc )
	  {	strcpy( Result_File_Name, argv[i] );
	  i++;
	  }
	else {	printf("%s: filename expected\n",argv[0]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-sfrom", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &s_Start ) ) )
	  i++;
	else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-sto", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &s_Stop ) ) )
	  i++;
	else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-sstep", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &s_Step ) ) )
	  i++;
	else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-sval", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &s_Val ) ) )
	  {   i++;
	  Range = FALSE;
	  }
	else {	printf("%s: real expected after %s\n",argv[0],argv[i-1]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
	Debug = TRUE;
	i++;
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) )
	{
	  printf("Usage: %s [-f <file>] [-o <file>] ", argv[0]);
	  printf("[-sval <real>] [-sfrom <real>]\n\t[-sto <real>] ");
	  printf("[-sstep <real>] [-debug] [-h] [-?]\n");
	  Args_Ok = FALSE;
	  Help = TRUE;
	  i++;
	}
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n",argv[0],VERSION);
	Args_Ok = FALSE;
	i++;
      }
      else	i++; /* skip unknown argument */
    }
    while ( i < argc );

  if (Args_Ok)
    {
      if ( strlen(Sample_File_Name) != 0 )
	freopen( Sample_File_Name, "r", stdin );
      if ( strlen(Result_File_Name) != 0 )
	freopen( Result_File_Name, "w", stdout );

      if ( Range )
	Compute_Laplace_Range( s_Start, s_Stop, s_Step );
      else
	Compute_Laplace_Range( s_Val, s_Val, 0.0 );

      return 0;
    }
  else if (Help)
    return 0;
  else
    return 1;
}

void	Compute_Laplace_Range(double	Start,
			      double	Stop,
			      double	Step )
{
  int		Nb_Samples;
  int		Nb_Points;
  double	Sample;
  double	*Sum;
  double	*Val;
  int		p;

  if ( Start > Stop ) {
	
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    exit( 1 );
  }
  else if ( Step < 0.0 ) {
    fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    exit( 1 );
  }

  Nb_Points = ( Stop == Start ?
		1 :
		1 + floor( ( Stop - Start ) / Step ) );

  Sum = (double*) calloc( Nb_Points, sizeof( double ) );
  Val = (double*) calloc( Nb_Points, sizeof( double ) );
  for ( p=0; p<Nb_Points; p++ ) {
    Sum[p] = 0.0;
    Val[p] = Start + p*Step;
  }
  Val[Nb_Points-1] = Stop;

  Nb_Samples = 0;

  while ( EOF != scanf( "%lf\n", &Sample ) ) {
	
    for ( p=0; p<Nb_Points; p++ )
      Sum[p] += exp( - Sample * Val[p] );
    Nb_Samples++;
  }

  if ( Nb_Samples == 0 ) {
	
    fprintf(stderr, "%s: no samples read\n", TOOLNAME );
    exit( 1 );
  }
  else {
    if ( Debug )
      {   printf("Samples read: %d\n", Nb_Samples );
      printf("Empirical laplace transform on range [%f,%f]\n", 
	     Start, Stop );
      }
    for ( p=0; p<Nb_Points; p++ )
      printf("%10.6f %f\n", Val[p], Sum[p]/Nb_Samples);
  }

}
