#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include "stats.h"
#include "parseargs.h"

#define	TOOLNAME	"moment"

boolean	Debug = FALSE;
boolean	Newline = TRUE;

/* forward declaration */
int	Compute_Moment( int N, int Adj );

int main(
int	argc,
char	*argv[] )
{
  static char		In_File_Name[128] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  int			Moment = 1;
  int			Adjustment_Period = -1;
  int			Result;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )

      else if GET_INT( "-n", Moment )
      else if GET_INT( "-adjustment", Adjustment_Period )

      else if IS_VERSION

      else if ( 0 == strcmp( "-debug", argv[i] ) )  {
	Debug = TRUE;
	Newline = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-nonewline", argv[i] ) ) {
	Newline = FALSE;
	Debug = FALSE;
	i++;
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", TOOLNAME);
	printf("[-n <int>] [-adjustment <int>] ");
	printf("[-nonewline]\n\t[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok ) {
    if ( strlen(In_File_Name) != 0 )
      freopen( In_File_Name, "r", stdin );

    Result = Compute_Moment( Moment, Adjustment_Period );

    return Result;
  }
  else if (Help)
    return 0;
  else
    return 1;
}

double	Pow( double	x,
	     int	n )
{
  static int	i;
  static double	Res;

  Res = 1.0;
  for ( i=0; i<n; i++ )
    Res *= x;

  return Res;
}

int	Compute_Moment( int Order,
			int Adjustment_Period )
{
  int		Nb_Samples;
  double	Sample;
  double	Sum;
  double	Cur_Estimate;

  Nb_Samples = 0;
  Sum = 0.0;
  Cur_Estimate = 0.0;

  while ( EOF != scanf( "%lf\n", &Sample ) ) {

    switch( Order ) {
    case 1:
      Sum += ( Sample - Cur_Estimate );
      break;
    case 2:
      Sum += ( Sample*Sample - Cur_Estimate );
      break;
    default:
      Sum += ( Pow( Sample, Order ) - Cur_Estimate );
    }
    Nb_Samples++;

    if ( ( Adjustment_Period > 0 ) 
	 && ( ( Nb_Samples % Adjustment_Period ) == 0 ) ){
      /* Revise current estimate to limit better the drift */
      Cur_Estimate += Sum / Nb_Samples;
      Sum = 0.0;
      if ( Debug ) {
	printf("# %9d samples. Adjusting estimate to: %f\n", 
	       Nb_Samples, Cur_Estimate );
      }
    }
  }

  if ( Nb_Samples == 0 ) {
    fprintf(stderr, "%s: no samples read\n", TOOLNAME );
    return 1;
  }
  else {
    if ( Debug ) {
      printf("# Samples read: %d\n", Nb_Samples );
      printf("# Empirical moment of order %d: ", Order );
    }
    printf("%f", Cur_Estimate + Sum / Nb_Samples);
    if ( Newline )
      printf("\n");

    return 0;
  }
}
