/* programme/module pour l'analyse topologique et la classification 	*/
/* des transitions d'un RdP 						*/

#include <stddef.h>
#include <malloc.h>

#include "taskgraph.h"

#define max(x,y)        ( (x>y) ? x : y )

void dumpe_adj(boolean	*A, int	n)
{
  int i;
  int j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++)
      printf("%1d",A[i*n+j]);
    printf("\n");
  }
}

void find_cycle(
/* finds a cycle in graph A, knowing A is strongly connected	*/
/* Skips "recycles".						*/
boolean	*A,
int	The_Size,
Graph	*Reseau )
{
  int 		i;
  int		j;
  boolean	*Visited;	/* matrice de booleens	*/

  Visited = (boolean*) calloc( The_Size, sizeof( boolean ) );

  dumpe_adj(A,The_Size);
  for (i=0; i<The_Size; i++) {
    Visited[i] = FALSE;;
  }

  i = 0;
  while ( ! Visited[i] ) {
    printf("%s(%1d) -> ",Reseau->T[i].Task_Name,i);
    Visited[i] = TRUE;
    j = 0;
    while ( (j==i) || !A[i*The_Size+j] )
      j++;
    i=j;
  }

  printf("%s(%1d)\n",Reseau->T[i].Task_Name,i);

  free( Visited );
}

int find_layer_structure(
/* extracts a layer structure from an adjacency matrix Adj, of size	*/
/* The_Size, and stores it in the structure Level.			*/ 
/* The space needed for the layer structure must have been allocated	*/
/* Returns the number of layers found, 0 if a cycle exists.		*/
Graph	*Reseau,
boolean	*Adj,
int	The_Size )
{
  int		Layer;
  int		Left_Tasks;
  int		InDegree;
  boolean	Found_One;
  boolean	*Present;	/* matrice de booleens */
  Layer_Desc	Niveau;
  int		i;
  int		j;

  Present = (boolean*) calloc( The_Size, sizeof( boolean ) );

  for (i=0; i<The_Size; i++) {
    Present[i] = TRUE;
    Reseau->Level[i] = (Layer_Desc)NULL;
  }
	
  Left_Tasks = The_Size;
  Found_One = TRUE;
  for ( Layer = 0; (Left_Tasks > 0) && Found_One ; Layer++ ) {
    Reseau->Level[Layer] = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );
    Niveau = Reseau->Level[Layer];
    Niveau->Nb_Tasks = 0;
    Niveau->Task_Id = (int*) calloc( (unsigned)Left_Tasks, sizeof (int) );

    /* look for nodes without fathers */
    Found_One = FALSE;
    for (j=0; j<The_Size; j++)
      if ( Present[j] ) {
	InDegree = 0;
	for (i=0; i<The_Size; i++)
	  if ( Present[i] && Adj[i*The_Size+j] )
	    InDegree++;
	if (InDegree == 0) {
	  /* node j has no fathers... */
	  Niveau->Task_Id[Niveau->Nb_Tasks] = j;
	  Niveau->Nb_Tasks++;
	  Found_One = TRUE;
	}
      }

    if ( Found_One ) {
      	for (i=0; i<Niveau->Nb_Tasks; i++)
	  Present[Niveau->Task_Id[i]] = FALSE;
	Left_Tasks -= Niveau->Nb_Tasks;
    }
  }

  /* cleanup */
  free( Present );

  if ( Found_One )
    return Layer;
  else
    return 0;
}

void find_prev_structure(Graph *Reseau, boolean *Adj, int The_Size)
/* extracts a layer structure from an adjacency matrix Adj, of size	*/
/* The_Size, and stores it in the structure Level.			*/ 
/* The space needed for the layer structure is allocated.		*/
{
  Layer_Desc	Prev_Ptr;
  Layer_Desc	The_Prevs;
  int		i;
  int		j;
	
  The_Prevs = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );
  The_Prevs->Task_Id = (int*) calloc( The_Size, sizeof (int) );

  for ( j=0; j<The_Size; j++) {
    Reseau->T[j].Prev = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );

    The_Prevs->Nb_Tasks = 0;
		
    for (i=0; i<The_Size; i++)
      if ( Adj[i*The_Size+j] ) {
	The_Prevs->Task_Id[The_Prevs->Nb_Tasks] = i;
	The_Prevs->Nb_Tasks++;
      }
    
    Reseau->T[j].Prev = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );
    Prev_Ptr = Reseau->T[j].Prev;
    Prev_Ptr->Nb_Tasks = The_Prevs->Nb_Tasks;
    Prev_Ptr->Task_Id = (int*) calloc( (unsigned)The_Prevs->Nb_Tasks, 
				       sizeof (int) );

    for (i=0; i<The_Prevs->Nb_Tasks; i++)
      Prev_Ptr->Task_Id[i] = The_Prevs->Task_Id[i];
  }
}

Graph *Reseau;

boolean topological_analysis(Graph *Reseau)
{
  boolean	*Adj;		/* matrice d'adjacence	*/
  boolean	*Adj0;		/* arcs de poids nul	*/
  int		Max_Depth = 0;

  int		i;
  int		j;
  Liste_Link	L_Scan;
  boolean	Result = TRUE;

  Adj = (boolean*) 
    calloc ( Reseau->Size*Reseau->Size, sizeof( boolean ) );
  Adj0 = (boolean*) 
    calloc ( Reseau->Size*Reseau->Size, sizeof( boolean ) );

  /* generate adj matrix, and matrix of zero weighted arcs */
  for (i=0; i<Reseau->Size; i++) {

    /* reset to FALSE, for safety */
    for (j=0; j<Reseau->Size; j++) {

      Adj[i*Reseau->Size+j] = FALSE;
      Adj0[i*Reseau->Size+j] = FALSE;
    }

    /* now scan liste of links */
    SCAN( L_Scan , Reseau->T[i].Neighbors )
      {	j = L_Scan->Val->Task_To;
      Adj[i*Reseau->Size+j] = TRUE;
      if ( L_Scan->Val->Link_Type == PRECEDENCE )
	Adj0[i*Reseau->Size+j] = TRUE;
      else if ( L_Scan->Val->Link_Type == ITERATION )
	Max_Depth = max( Max_Depth,
			 L_Scan->Val->Depth );
      }
  }

  Reseau->Max_Dependency = Max_Depth;

  Reseau->Level = (Layer_Desc*) calloc ( Reseau->Size, sizeof( Layer_Desc ) );

  if ( 0 != ( Reseau->Nb_Layers = 
	      find_layer_structure(Reseau,Adj0,Reseau->Size) ) ) {
      find_prev_structure(Reseau,Adj,Reseau->Size);
      Result = TRUE;
  }
  else {
    printf("Cyclic task graph!!\n");
    find_cycle(Adj0, Reseau->Size, Reseau);
    find_prev_structure(Reseau,Adj,Reseau->Size);
    /* on le fait quand meme au cas ou l'analyse 	*/
    /* est demandee.					*/
    Result = FALSE;
  }

  /* cleanup */
  free( Adj );
  free( Adj0 );

  /* end */
  return Result;
}

