%{

#include "taskgraph.h"

#define YYDEBUG 1

extern boolean		Debug;

extern Graph	 	*Parse_TG(Liste_Task, Liste_Link);
extern Liste_Task	Extract_Tasks(Liste_Object);
extern Liste_Link	Extract_Links(Liste_Object);
extern void		Free_Object_List(Liste_Object);
extern Task 		*Parse_Task(char*,char*,Law_Desc);
extern Link_Desc	Parse_Link(char*,Link_Desc);
extern Link_Desc	Parse_Data_Link(SType,BiPoint,Law_Desc);
extern Link_Desc	Parse_Program_Link(SType,BiPoint,int);
extern Law_Desc		Parse_Law(char Nom, Liste_Reel Params, 
				  Liste_Reel Params2, Liste_Loi The_SubLaws);
extern Liste_Object	Append_Object(Liste_Object,Liste_Object);
extern Liste_Object	Task2Object(Task*);
extern Liste_Object	Link2Object(Link_Desc);
extern BiPoint		Parse_Ends(char*,char*);

/* The result of the parsing is to be put into the (global!) variable	*/
/* The_Task_Graph. It is defined in "parse_tg.c".			*/
/* Before I used "yyval" but this does not work with flex/bison.	*/
#ifndef USE_YYLVAL
extern Graph	*The_Task_Graph;	
#endif

extern boolean	Parse_Debug;
%}

%start reseau

%union{
	int		ival;
	double		rval;
	char		*sval;
	BiPoint		bpval;
	Graph		*un_reseau;
	Law_Desc	loi;
	Link_Desc	link;
	Liste_Link	plist;
	Task		*task;
	Liste_Task 	tlist;
	Liste_Object	olist;
	Liste_Reel	rlist;
	Liste_Loi	Llist;
}

%token	<sval>		STRING
%token	<ival>		INTGR
%token	<ival>		SEP 			/* pas de valeur, en fait */
%token	<ival>		PARB PARE		/* pas de valeur non plus */
%token	<ival>		TASK_BEG LINK_BEG END
%token	<ival>		LINK_ORIG LINK_DEST DATA PROGRAM
%token	<ival>		LINK_SYNCH LINK_TYPE OR_LINK AND_LINK
%token	<ival>		DEPTH
%token	<ival>		SERV
%token	<rval>		REAL
%token	<loi>		EXPO ERLG BERN GEOM0 GEOM1 UNIF CST DISC ONOFF IPP HEXP
%token  <loi>		PARETO WEIBULL LOGNO

%type	<un_reseau>	reseau
%type	<task>		task
%type	<link>		link link_dcl
%type	<loi>		loi_dcl service_dcl
%type   <Llist>		law_list
%type	<rlist>		real_list
%type   <olist>         liste_objects object
%type	<ival>		synchro_dcl
%type	<ival>		depth_dcl
%type	<sval>		orig_dcl dest_dcl
%type	<bpval>		ends_dcl
%%

reseau	:	liste_objects
		{
#ifdef USE_YYLVAL
			$$ = Parse_TG( Extract_Tasks( $1 ), 
						   Extract_Links( $1 ) );
#else
			The_Task_Graph = Parse_TG( Extract_Tasks( $1 ), 
						   Extract_Links( $1 ) );
#endif
			Free_Object_List( $1 );
			if ( Debug )
			  printf("Reseau ok\n");
		}
		;

liste_objects : 
		object
		{	$$ = Append_Object( NULL, $1 );
		}
		| liste_objects object
		{	$$ = Append_Object( $1, $2 );
		}
		;

object :
		task
		{	$$ = Task2Object( $1 );
		}
		|
		link
		{	$$ = Link2Object( $1 );
		}
		;

task :	
		TASK_BEG STRING service_dcl END
		{ 	$$ = Parse_Task( $2, (char *)NULL, $3 );
			if (Parse_Debug)
				printf("Task %s recognized\n",$2);
		}
		|
		TASK_BEG STRING STRING service_dcl END
		{ 	$$ = Parse_Task( $2, $3, $4 );
			if (Parse_Debug)
				printf("Task %s recognized\n",$2);
		}
		;

link :		LINK_BEG STRING link_dcl END
		{ 	$$ = Parse_Link( $2, $3 );
			if (Parse_Debug)
				printf("Link %s recognized\n",$2);
		}
		;

link_dcl :
		LINK_TYPE DATA synchro_dcl ends_dcl service_dcl
		{	$$ = Parse_Data_Link( $3, $4, $5 );
		}
		|
		LINK_TYPE PROGRAM synchro_dcl ends_dcl depth_dcl
		{	$$ = Parse_Program_Link( $3, $4, $5 );
		}
		;

synchro_dcl :	LINK_SYNCH AND_LINK
		{	$$ = AND;
		}
		|
		LINK_SYNCH OR_LINK
		{	$$ = OR;
		}
		;

service_dcl :	SERV loi_dcl
		{	$$ = $2;
		}
		;

loi_dcl : 	EXPO PARB real_list PARE
		{	$$ = Parse_Law( 'E', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| CST PARB real_list PARE
		{	$$ = Parse_Law( 'C', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| BERN PARB real_list PARE
		{	$$ = Parse_Law( 'B', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| GEOM0 PARB real_list PARE
		{	$$ = Parse_Law( 'g', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| GEOM1 PARB real_list PARE
		{	$$ = Parse_Law( 'G', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| ERLG PARB INTGR SEP real_list PARE
		{	$$ = Parse_Law( 'R', Append_Real( $5, (double)$3 ),
					(Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| UNIF PARB real_list PARE
		{	$$ = Parse_Law( 'U', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| PARETO PARB real_list PARE
		{	$$ = Parse_Law( 'p', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| LOGNO PARB real_list PARE
		{	$$ = Parse_Law( 'L', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| WEIBULL PARB real_list PARE
		{	$$ = Parse_Law( 'W', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| DISC PARB real_list PARE PARB real_list PARE
		{	$$ = Parse_Law( 'D', $3, $6,
					(Liste_Loi)NULL );
		}
		| HEXP PARB real_list PARE
		{	$$ = Parse_Law( 'H', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| ONOFF PARB real_list PARE
		{	$$ = Parse_Law( 'O', $3, (Liste_Reel)NULL,
					(Liste_Loi)NULL );
		}
		| IPP PARB REAL SEP law_list PARE
		{	$$ = Parse_Law( 'P',
					Append_Real( (Liste_Reel)NULL, $3),
					(Liste_Reel)NULL, $5 );
		}
		;

law_list :	loi_dcl
		{	$$ = Append_Law( (Liste_Loi)NULL, $1 );
		}
		| law_list SEP loi_dcl
		{	$$ = Append_Law( $1, $3 );
		}
		;

ends_dcl :	orig_dcl dest_dcl
		{	$$ = Parse_Ends( $1, $2 );
		}
		;

orig_dcl :	LINK_ORIG STRING
		{	$$ = $2;
		}
		;

dest_dcl :	LINK_DEST STRING
		{	$$ = $2;
		}
		;

depth_dcl :	DEPTH INTGR
		{	$$ = $2;
		}
		;

real_list :	REAL
		{	$$ = Append_Real( (Liste_Reel)NULL, $1 );
		}
		| INTGR
		{	$$ = Append_Real( (Liste_Reel)NULL, (double)$1 );
		}
		| real_list SEP REAL
		{	$$ = Append_Real( $1, $3 );
		}
		;

%%
#include "tgl_lex.yy.c"
