/* Translation program from TG language to EGL language.	  */

#include <stdio.h>
#include "version.h"
#include "taskgraph.h"

#define	TOOLNAME	"tg2eg"

#define MAX(x,y)	( x>y ? x : y )

extern boolean	topological_analysis(Graph *Reseau);
extern Graph	*Read_Network(char *File_Name);
extern void	dumpe_network(Graph *Reseau);
extern char	*todays_date(void);

/* forward declaration */
void	Write_TG_As_EGL(Graph *Reseau);


boolean	Has_OR_Synchronizations(Graph	*Reseau)
{
  boolean		Has_Some = FALSE;
  Liste_Link	L_Scan;
  int		i;

  for ( i=0; (i<Reseau->Size) && !Has_Some; i++ )
    SCAN( L_Scan, Reseau->T[i].Neighbors )
      Has_Some = Has_Some ||
      L_Scan->Val->Synchro_Type == OR;

  return	Has_Some;
}

int main(int argc, char *argv[])
{	
  static char	File_Name[128] = "";
  static char	Code_File_Name[128] = "";
  Graph 	*Reseau;
  int		i;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name ) 
      else if GET_STRING( "-o", Code_File_Name ) 

      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION	      

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", argv[0]);
	printf(" [-debug] [-?] [-h]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if ( Has_OR_Synchronizations( Reseau ) ) {
	printf("%s: this task graph has OR synchronizations", TOOLNAME );
	printf(". Cannot handle.\n");
	return -2;
      }

      /* no need to perform topological analysis */
      if ( strlen(Code_File_Name) != 0 )
	freopen( Code_File_Name, "w", stdout );

      Write_TG_As_EGL(Reseau);
      return 0;
    } 
    else {
      printf("%s: error: could not read network from file %s.\n",
	     TOOLNAME, File_Name);
      return -1;
    }
  } else if (!Help)
    return -1;
  else
    return 0;
}

void Write_TG_As_EGL(Graph *Reseau)
{
  int		i;
  int		j;
  int		Link_Nb;
  Link_Desc	The_Link;
  Liste_Link	L_Scan;

  printf("# egl code generated by %s version %s\n", 
	 TOOLNAME, VERSION );
  printf("# Date: %s\n", todays_date() );
  printf("#\n");

  printf("Size <- %d\n", Reseau->Size);

  /* generate transitions */
  for ( i=0; i<Reseau->Size; i++ ) {

    printf("%d ", i+1 );
    Write_Law( stdout, Reseau->T[i].Task_Law, NORMAL_PRINT_MODE );
    printf("\n");

    /* count links. This is for avoiding \n before the final ";" */
    Link_Nb = 0;
    SCAN( L_Scan, Reseau->T[i].Neighbors ) {
      Link_Nb ++;
    }

    /* scan link and generate place */
    SCAN( L_Scan, Reseau->T[i].Neighbors ) {
      The_Link = L_Scan->Val;
      j = The_Link->Task_To;
      printf("\t, %d ", j+1 );
      if ( The_Link->Depth > 0 ) {
	printf("C 0.0 %d ", The_Link->Depth );
	for ( j=0; j<The_Link->Depth; j++ ) {
	  printf("0.0 ");
	}
      }
      else {
	Write_Law( stdout, The_Link->Link_Law, NORMAL_PRINT_MODE );
	printf(" 0");
      }
      Link_Nb --;
      if ( Link_Nb > 0 )
	printf("\n");
    }

    printf(";\n");
  }
}
