/* Translation program from TG language to QNAP language.	  */

#include <stdio.h>
#include "version.h"
#include "taskgraph.h"

#define	TOOLNAME	"tg2qnap"

#define MAX(x,y)	( x>y ? x : y )

extern boolean	topological_analysis(Graph *Reseau);
extern Graph	*Read_Network(char *File_Name);
extern void	dumpe_network(Graph *Reseau);
extern char	*todays_date(void);

/* forward declaration */
void	dump_qnap_code(Graph *Reseau);

int		Sim_Length = 1000;	/* default number of iterations */
int		*Link_Count;
boolean		*Link_Exists;
boolean		Optimize = FALSE;
int		Opt_Level;
int		Nb_Links;

boolean	Has_OR_Synchronizations(Graph	*Reseau)
{
  boolean		Has_Some = FALSE;
  Liste_Link	L_Scan;
  int		i;

  for ( i=0; (i<Reseau->Size) && !Has_Some; i++ )
    SCAN( L_Scan, Reseau->T[i].Neighbors )
      Has_Some = Has_Some ||
      L_Scan->Val->Synchro_Type == OR;

  return	Has_Some;
}

int main(int argc, char *argv[])
{	
  static char	File_Name[128] = "";
  static char	Code_File_Name[128] = "";
  Graph 		*Reseau;
  int		i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name ) 
      else if GET_STRING( "-o", Code_File_Name ) 
      else if GET_INT_SWITCH( "-O", Opt_Level, Optimize, TRUE )

      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION	      

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", argv[0]);
	printf(" [-O <level>] [-tmax <number>] [-debug] [-?] [-h]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmax", argv[i] ) ) {
	i++;
	if ( !( ( i < argc )
		&& ( 0 != sscanf( argv[i], "%d", &(Sim_Length) ) )
		&& ( Sim_Length > 0 ) ) ) {
	  Args_Ok = FALSE;
	  printf("%s: positive integer expected ", argv[0]);
	  printf("after %s\n",argv[i-1]);
	}
	else {	
	  i++;
	}
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if ( Has_OR_Synchronizations( Reseau ) ) {
	printf("%s: this task graph has OR synchronizations", TOOLNAME );
	printf(". Cannot handle.\n");
	return -2;
      }
      else if ( topological_analysis(Reseau) ) {
	if ( strlen(Code_File_Name) != 0 )
	  freopen( Code_File_Name, "w", stdout );

	dump_qnap_code(Reseau);
	return 0;
      } 
      else
	return -2;
    }
    else {
      printf("%s: error: could not read network from file %s.\n",
	     TOOLNAME, File_Name);
      return -1;
    }
  } else if (!Help)
    return -1;
  else
    return 0;
}

/* Procedure to write the law. Adapted from the same from gen_eg.c */
void fprint_QNAP_law( Law_Desc The_Law)
{
	printf("    SERVICE = ");
	switch( The_Law->Name ) {
	case 'E':
		printf("EXP( %8.4f );\n", The_Law->Parameters[0] );
		break;
	case 'C':
		printf("CST( %8.4f );\n", The_Law->Parameters[0] );
		break;
	case 'B':
		printf("IF RANDU > %8.4f THEN CST(0.0) ELSE CST(1.0);\n", 
				The_Law->Parameters[0] );
		break;
	case 'G':
		printf("CST( FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
				The_Law->Parameters[0] );
		break;
	case 'H':
		printf("HEXP( %8.3f, %8.3f );\n", The_Law->Parameters[0],
				The_Law->Parameters[1] );
		break;	   
	case 'U':
		printf("UNIFORM( %8.4f, %8.4f );\n",
					The_Law->Parameters[0],
					The_Law->Parameters[1] );
		break;
	case 'R':
		printf("ERLANG( %8.4f, %4.0f );\n",
					The_Law->Parameters[1],
					The_Law->Parameters[0] );
		break;
	}
}

/* procedure to write the initial population. 			   	*/
void fprint_init_mark(Link_Desc	The_Link, int The_Class)
{	
  if ( ( The_Link->Link_Type == ITERATION ) && (The_Link->Depth>0) )
    printf("    INIT(CL_%d) = %d;\n", The_Class, The_Link->Depth );

}

/* procedure to write what is the destination of customers exiting	*/
/* "transition" The_Orig to link The_Link which is between The_Orig	*/
/* and The_Dest.							*/
void fprint_dest(int The_Orig, int The_Dest, int The_Link)
{	
  if (Optimize && !Link_Exists[The_Link])
    printf("        TA_%d,", The_Dest );
  else 	
    printf("        LI_%d,", The_Link );

}

/****************************************************************/
/* Function 	int	Link_Numbering				*/
/* Computes the number of the link located between two given	*/
/* transitions. The numbering is according the alphabetical	*/
/* order of the couple (T1, T2).				*/
/*								*/
/* Parameters:							*/
/*	Reseau:		the network where it is computed	*/
/*	T1:		the number of the origin transition	*/
/*	T2:		the number of the destination 		*/
/*			(both numbers betw 0 and NB_TRANS-1)	*/
/*								*/
/* Return Value:						*/
/*	the number of the link					*/
/****************************************************************/
int	Link_Numbering( Reseau, T1, T2 )
Graph	*Reseau;
int		T1, T2;
{	int		Num;
	Liste_Link	Scan;

	Num = Link_Count[ T1 ];
	for( Scan = Reseau->T[T1].Neighbors; 
		( Scan != NULL ) && ( Scan->Val->Task_To != T2 );
		Scan=Scan->Next )
		Num++;

/*	fprintf(stderr,"Link_N de (%d,%d)=%d\n", T1, T2, Num );
*/
	if ( Scan == NULL )
	{
		fprintf( stderr, "%s: warning: inconsistency in link/task numbering:",
			TOOLNAME );
		fprintf( stderr, "no link btw %d and %d\n", T1, T2 );
	}

	return Num;
}

/****************************************************************/
/* Function 	int	in_link_number				*/
/* Computes the number of a T/P link, between 0 and the indegree*/
/* of T2, the output transition of P. The numbering is according*/
/* to the alphabetical order of the couple (T1, T2).		*/
/*								*/
/* Parameters:							*/
/*	Reseau:		the network where it is computed	*/
/*	T1:		the number of the origin transition	*/
/*	T2:		the number of the destination 		*/
/*			(both numbers betw 0 and NB_TRANS-1)	*/
/*								*/
/* Return Value:						*/
/*	the number of the link					*/
/****************************************************************/
int	in_link_number( Reseau, T1, T2 )
Graph	*Reseau;
int		T1, T2;
{	int	Num;

	for( Num=0; (Num<Reseau->T[T2].Prev->Nb_Tasks)
			&& (Reseau->T[T2].Prev->Task_Id[Num] != T1 ); 
								Num++ );
	if ( Num == Reseau->T[T2].Prev->Nb_Tasks )
	{	fprintf( stderr, 
			"%s: warning: inconsistency in prev structure:",
			TOOLNAME );
		fprintf( stderr, "no link btw %d and %d\n", T1, T2 );
	}

	return Num;
}

int	Nb_Neighbors( Reseau, Task_Id )
Graph   *Reseau;
int	Task_Id;
{
	int		Nb = 0;
	Liste_Link	L_Scan;

	SCAN( L_Scan, Reseau->T[Task_Id].Neighbors )
		Nb++;

	return	Nb;
}

void	init_link_count( Graph	*Reseau )
{
  int		i;
  int		j;
  Link_Desc	The_Link;
  Liste_Link	L_Scan;

  Link_Count = (int*) calloc( Reseau->Size, sizeof(int) );
  Link_Count[0] = 0;

  for ( i=1; i<Reseau->Size; i++ )
    Link_Count[i] = Link_Count[i-1] 
      + Nb_Neighbors( Reseau, i-1 );
  Nb_Links = Link_Count[Reseau->Size-1] + 
    Nb_Neighbors( Reseau, Reseau->Size-1 );
  if ( Optimize ) {
    Link_Exists = (boolean*) 
      calloc( Nb_Links, sizeof(int) );
    for ( i=0; i<Nb_Links; i++ )
      Link_Exists[i] = TRUE;

    for ( i=0; i<Reseau->Size; i++ )
      SCAN( L_Scan, Reseau->T[i].Neighbors ) {
      The_Link = L_Scan->Val;
      j = The_Link->Task_To;
      Link_Exists[Link_Numbering(Reseau,i,j)] 
	= ( ( The_Link->Link_Type == ITERATION ) 
	    || ( ( The_Link->Link_Type == PRECEDENCE )
		 && ! ( ( The_Link->Link_Law->Name == 'C' )
			&& ( The_Link->Link_Law->Parameters[0] == 0.0 ) ) )
	    );
      if (Debug && !Link_Exists[Link_Numbering(Reseau,i,j)])
	fprintf(stderr,"Link %d=(%d,%d) removed\n",
		Link_Numbering(Reseau,i,j),i,j);
    }
  }

}

void dump_qnap_code(Graph *Reseau)
{
  int		i;
  int		j;
  Task		The_Task;
  Link_Desc	The_Link;
  int		Nb_Neighb;
  int		The_Dest, The_Orig;
  int		Max_Deg;
  Liste_Link	L_Scan;
  boolean		Some_Links = TRUE;
  int		Last_Link;
  int		Nb_Real_Links;

  init_link_count( Reseau );

  printf("& QNAP2 code generated by %s version %s\n", 
	 TOOLNAME, VERSION );
  printf("& Date: %s\n", todays_date() );
  printf("&\n");

  /* bloc de controle */
  printf("& Simulation Control\n");
  printf("/CONTROL/\n");
  printf("    OPTION = NSOURCE;\n");
  printf("    TMAX = %d;\n", Sim_Length );
  printf("\n");

  /* declarations */
  printf("/DECLARE/\n");
  printf("&   tasks\n");
  printf("    QUEUE");
  for (i=0; i<Reseau->Size-1; i++)
    {	printf(" TA_%d,",i);
    if ((i+1) % 8 == 0)
      printf("\n         ");
    }
  printf(" TA_%d;\n", Reseau->Size-1 );

  /* for links, check if some exist... */
  if ( Optimize )
    {	Some_Links = FALSE;
    for (i=0; i<Nb_Links; i++)
      {	Some_Links = ( Some_Links || Link_Exists[i] );
      if ( Link_Exists[i] )
	Last_Link = i;
      }
    }
  else	Last_Link = Nb_Links - 1;
  if ( Some_Links )
    {	printf("&   links\n");
    printf("    QUEUE");
    Nb_Real_Links = 0;
    for (i=0; i<Last_Link; i++)
      {	if ( (!Optimize) || Link_Exists[i] )
	{	printf(" LI_%d,",i);
	Nb_Real_Links++;
	}
      if ((Nb_Real_Links+1) % 8 == 0)
	printf("\n         ");
      }
    printf(" LI_%d;\n", Last_Link );
    }

  printf("&   classes\n");
  /* computation of the maximum indegree */
  Max_Deg = Reseau->T[0].Prev->Nb_Tasks;
  for (i=1; i<Reseau->Size; i++)
    Max_Deg = MAX( Max_Deg, Reseau->T[i].Prev->Nb_Tasks );

  printf("    CLASS");
  for (i=0; i<Max_Deg-1; i++)
    {	printf(" CL_%d,",i);
    if ((i+1) % 8 == 0)
      printf("\n         ");
    }
  printf(" CL_%d;\n", Max_Deg-1 );
  printf("\n");

  /* generate transitions */
  printf("& Stations corresponding to tasks\n");
  for ( i=0; i<Reseau->Size; i++ )
    {
      printf("/STATION/\n");
      printf("    NAME = TA_%d;\n", i);
      The_Task = Reseau->T[i];

      switch (The_Task.Prev->Nb_Tasks) {
      case 0:	/* a source that is not recycled */
	printf("    TYPE = SOURCE;\n");
	break;
      case 1: /* only one input. Nothing to be done */
	break;
      default:	/* have to make a FUSION */
	printf("    FUSION = (\n");
	for( j=0; j<The_Task.Prev->Nb_Tasks-1; j++ )
	  {	The_Orig = The_Task.Prev->Task_Id[j];
	  printf("        CL_%d, 1,\n",
		 in_link_number( Reseau, The_Orig, i ) );
	  }
	The_Orig = The_Task.Prev->
	  Task_Id[The_Task.Prev->Nb_Tasks-1];
	printf("        CL_%d, 1 ) CL_0;\n",
	       in_link_number( Reseau, The_Orig, i ) );
      }			
      fprint_QNAP_law( The_Task.Task_Law );

      Nb_Neighb = Nb_Neighbors( Reseau, The_Task.Task_Id );
      switch (Nb_Neighb) {
      case 0:	printf("    TRANSIT = OUT;\n");
	break;
      case 1:	The_Dest = The_Task.Neighbors->Val->Task_To;
	printf("    TRANSIT = ");
	fprint_dest( i, The_Dest,
		     Link_Numbering( Reseau, i, The_Dest ) );
	printf(" CL_%d;\n", 
	       in_link_number( Reseau, i, The_Dest ) );
	break;
      default:
	L_Scan = The_Task.Neighbors;
	printf("    FISSION = (\n");
	for( j=0; j<Nb_Neighb-1; j++ )
	  {	The_Dest = L_Scan->Val->Task_To;
	  fprint_dest( i, The_Dest, 
		       Link_Numbering( Reseau, i, The_Dest ) );
	  printf(" CL_%d, 1,\n",
		 in_link_number( Reseau, i, The_Dest ) );
	  L_Scan = L_Scan->Next;
	  }
	The_Dest = L_Scan->Val->Task_To;
	fprint_dest( i, The_Dest, 
		     Link_Numbering( Reseau, i, The_Dest ) );
	printf(" CL_%d, 1 );\n",
	       in_link_number( Reseau, i, The_Dest ) );
      }
      printf("\n");
    }

  /* now generate links */
  printf("& stations corresponding to communication links\n");
  Nb_Links = 0;
  for ( i=0; i<Reseau->Size; i++ )
    {
      SCAN( L_Scan, Reseau->T[i].Neighbors )
	{	The_Link = L_Scan->Val;
	j = The_Link->Task_To;
	Nb_Links++;

			
	if (Optimize
	    && !Link_Exists[Link_Numbering( Reseau, i, j ) ] )
	  /* if initial population, put it in	*/
	  /* downstram transition.		*/
	  {	if (The_Link->Depth != 0)
	    {	printf("/STATION/\n");
	    printf("    NAME = TA_%d;\n", j );
	    fprint_init_mark( The_Link,
			      in_link_number( Reseau, i, j ) );
	    printf("\n");
	    }
	  }
	else {

	  printf("/STATION/\n");

	  printf("    NAME = LI_%d;\n",
		 Link_Numbering( Reseau, i, j ) );
	  printf("    TYPE = SERVER, INFINITE;\n");
			
	  if ( The_Link->Link_Type == ITERATION )
	    printf("    SERVICE = CST( 0.0 );");
	  else	fprint_QNAP_law( The_Link->Link_Law );
	  fprint_init_mark( The_Link, 
			    in_link_number( Reseau, i, j ) );
	  printf("    TRANSIT = TA_%d;\n", j );
	  printf("\n");

	}
	}
    }

  /* exec bloc */
  printf("& execution of the analysis\n");
  printf("/EXEC/\n");
  printf("    BEGIN\n");
  printf("        SIMUL;\n");
  for (i=0; i<Reseau->Size; i++)
    {	printf("        WRITELN(\"Cycle time of transition %d:\",", 
	       i);
    printf("TIME/SERVNB(TA_%d) );\n", i );
    }
  printf("    END;\n");
  printf("/END/\n");

  exit(0);
}
