/* programme calculant le chamin critique d'un graphe de taches		*/
/* (non repetitif pour le moment) 					*/

#include <math.h>		/* for random number generation	*/
#include <stdio.h>

#include "version.h"
#include "taskgraph.h"

#define	TOOLNAME	"tg_crit_path"

#ifndef ALPHA
extern void srandom( unsigned seed );
extern long random( void );
#endif

#define	VALUE(x)	( A_Sample ? Sample( x ) : Mean( x ) )

extern Graph		*Read_Network(char *File_Name);
extern void		dumpe_network( Graph *);
extern void		Free_Task_Graph( Graph *);
extern boolean		topological_analysis( Graph *);
extern double		Sample( Law_Desc The_Law );

#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

boolean	File_Problem = FALSE;
boolean	Silent = FALSE;
boolean	Ignore = FALSE;
boolean	A_Sample = TRUE;

double	*YY;

boolean	Has_Constant_Times( Graph *Reseau )
{
  boolean	Constant = TRUE;
  int		i;
  Liste_Link	L_Scan;

  for (i=0; (i<Reseau->Size) && Constant; i++) {
    Constant = Constant 
      && ( Reseau->T[i].Task_Law->Name == 'C' );
    SCAN( L_Scan, Reseau->T[i].Neighbors)
      Constant = Constant &&
      ( L_Scan->Val->Link_Law->Name == 'C' );
  }
  
  return	Constant;
}

Link_Desc get_link( Graph *Reseau, int Orig, int Dest )
{
  Link_Desc	Resultat;
  Liste_Link	L_Scan;

  for ( L_Scan = Reseau->T[Orig].Neighbors; L_Scan != NULL; 
	L_Scan = L_Scan->Next )
    if ( L_Scan->Val->Task_To == Dest )
      Resultat = L_Scan->Val;

  return Resultat;
}

void rewind_path(Graph	*Reseau, int Start_Task, boolean Single )
{
  int		k;
  int		Prev_Id;
  boolean	Continue = TRUE;
  Link_Desc	The_Link;

  printf( "# Task %d\n", Start_Task );
  printf( "\n\"%s\"\n_E_O_E\n", Reseau->T[Start_Task].Task_Name );
  Reseau->T[Start_Task].Is_Critical = TRUE;
				
  for( k=0; (k<Reseau->T[Start_Task].Prev->Nb_Tasks) && Continue; k++) {
    Prev_Id = Reseau->T[Start_Task].Prev->Task_Id[k];
    The_Link = get_link( Reseau, Prev_Id, Start_Task );
    if ( The_Link->Is_Critical ) {
      printf( "# Link (%d->%d)\n", Prev_Id, Start_Task );
      printf( "\n\"%s\"\n_E_O_E\n", The_Link->Link_Name );

      if (!Reseau->T[Prev_Id].Is_Critical )
	rewind_path( Reseau, Prev_Id, Single );
      if ( Single )
	Continue = FALSE;
    }
  }
}

void dump_critical_paths(Graph	*Reseau,
			 double	Exec_Time,
			 boolean	Single)
{
  int		Level;
  int		j;
  boolean	Continue = TRUE;
  int		Task_Id;

  for (Level = Reseau->Nb_Layers-1; (Level >0) && Continue; Level--)
    for( j=0; (j<Reseau->Level[Level]->Nb_Tasks) && Continue; 
	 j++) {
      Task_Id = Reseau->Level[Level]->Task_Id[j];
      if ( ( YY[Task_Id] == Exec_Time )
	   && ( !Reseau->T[Task_Id].Is_Critical ) ) {

	rewind_path( Reseau, Task_Id, Single );
	if ( Single )
	  Continue = FALSE;
      }
    }
}
	
double Critical_Length(Graph *Reseau)
{	
  int		Level;
  int		j;
  int		k;
  int		The_Size;
  double	The_Max = 0.0;
  double	x;
  double	y;
  int		Task_Id, Prev_Id;
  Link_Desc	The_Link;

  The_Size = Reseau->Size;

  YY = (double*) calloc( The_Size, sizeof(double) );

  /* set times to 0.0 for Level 0 */
  for( j=0; j<Reseau->Level[0]->Nb_Tasks; j++) {
    Task_Id = Reseau->Level[0]->Task_Id[j];
    YY[Task_Id] = VALUE( Reseau->T[Task_Id].Task_Law );
    The_Max = max( The_Max, YY[Task_Id] );
    Reseau->T[Task_Id].Is_Critical = FALSE;
  }

  /* compute level by level */
  for (Level=1; Level<Reseau->Nb_Layers; Level++)
    for( j=0; j<Reseau->Level[Level]->Nb_Tasks; j++) {
      Task_Id = Reseau->Level[Level]->Task_Id[j];

      /*  scan links to find the starting instant */
      for( k=0; k<Reseau->T[Task_Id].Prev->Nb_Tasks; k++) {
	Prev_Id = Reseau->T[Task_Id].Prev->Task_Id[k];
	The_Link = get_link( Reseau, Prev_Id, Task_Id );

	/* check if not program link */
	if ( The_Link->Link_Type == PRECEDENCE ) {
	  The_Link->Val = VALUE( The_Link->Link_Law );
	  y = YY[Prev_Id] + The_Link->Val;

	  if ( k == 0 )
	    x = y;
	  else
	    switch( The_Link->Synchro_Type ) {
	    case AND:
	      x = max( x, y );
	      break;
	    case OR:
	      x = min( x, y );
	      break;
	    }
	}
      }
      /* now scan links again to find critical ones */
      for( k=0; k<Reseau->T[Task_Id].Prev->Nb_Tasks; k++) {
	Prev_Id = Reseau->T[Task_Id].Prev->Task_Id[k];
	The_Link = get_link( Reseau, Prev_Id, Task_Id );
	if ( The_Link->Link_Type == PRECEDENCE ) {
	  The_Link->Is_Critical = ( x == YY[Prev_Id] + The_Link->Val );
	}
      }

      /* finally compute finish date */
      YY[Task_Id] = x + VALUE( Reseau->T[Task_Id].Task_Law );
      The_Max = max( The_Max, YY[Task_Id] );
      Reseau->T[Task_Id].Is_Critical = FALSE;

    }

  return	The_Max;
}

int main(int argc, char *argv[])
{	
  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";
  Graph 		*Reseau;
  boolean		Help = FALSE;
  boolean		Debug = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Single = FALSE;
  boolean		Paths = FALSE;
  double		L;
  int			i;
  int			Seed;
    
  i = 1;
    
  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-ignore", Ignore, TRUE )
      else if SWITCH( "-mean", A_Sample, FALSE )
      else if SWITCH( "-paths", Paths, TRUE )
      else if SWITCH( "-sample", A_Sample, TRUE )
      else if SWITCH( "-single", Single, TRUE )
      else if SWITCH( "-silent", Silent, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-ignore] [-silent]", TOOLNAME );
	printf(" [-mean] [-paths]\n\t[-single] [-sample]");
	printf(" [-seed <number>] [-o <file>] [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-seed", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  sscanf( argv[i], "%d", &Seed );
	  i++;
	  srandom( Seed );
	  if ( Debug )
	    printf("Using seed %d\n", Seed );
	}
	else {	
	  printf("%s: integer expected after %s\n", TOOLNAME,
		       argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    Reseau = Read_Network( In_File_Name );

    if ( Reseau != NULL ) {

      if ( topological_analysis(Reseau) ) {

	if (Debug)
	  dumpe_network(Reseau);
	    
	if ( ( Reseau->Max_Dependency != 0 )
	     && !Ignore ) {
	  printf("%s: The graph is repetitive. ", TOOLNAME );
	  printf("I can't handle it yet.\n");
	}
	else {
	  if ( ( Reseau->Max_Dependency != 0 )
	       && !Silent )
	    printf("%s: Ignoring repetitivity.\n",
		   argv[0] );
	  L = Critical_Length( Reseau );
	  if (!Silent)
	    printf("Minimal execution time = ");
	  printf("%f\n", L );
		
	  if ( Paths ) {
	    if ( strlen(Out_File_Name) != 0 )
	      freopen( Out_File_Name, "w", stdout );
	    printf("# objects of a critical path,");
	    printf(" generated by %s\n", TOOLNAME );
	    dump_critical_paths( Reseau, L, Single );
	  }
	}

	Free_Task_Graph( Reseau );

	return 0;
      }
      else
	return -2;
    }
    else {
      printf("%s: error: could not read graph from file %s.\n",
	     TOOLNAME, In_File_Name );
      return -1;
    }
  } else if (Help)
    return 0;
  else 
    return -1;
}
