#
# Procedures for processing text. Useful when displaying "help" 
# from the LaTex documentation.
#
proc ProcessText { tw } {

    # first eliminate inner "contextes"
    TagContext $tw "tt" fixed
    TagContext $tw "em" italic

    # process section headers
    TagCommand $tw "section" title
    TagCommand $tw "paragraph" blue
    TagCommand $tw "paragraph\\*" blue

    # process local font changes
    TagCommand $tw "texttt" fixed

    # misc
    TagCommand $tw "mbox" nil
    TagCommand $tw "ref" blue
    TagCommand $tw "index" remove
    TagCommand $tw "ttindex" remove
    TagCommand $tw "label" remove

    # process descriptions etc
    foreach env { description itemize } {
	TagEnv $tw $env description
    }
}

# A "context" is something like:
#	{\foo ... }
# without nested parentheses.
proc TagContext { tw name tag } {
    global TXTDEBUG

    set pat "\{\\\\$name\[^\}\]*\}"
    if $TXTDEBUG {
	puts "looking for context $pat"
    }
    set istart [$tw search -forwards -count nbm -regexp $pat 1.0]
    while { $istart != "" } {
	if $TXTDEBUG {
	    puts "Found CONTEXT $pat at $istart ($nbm chars)"
	}
	$tw tag add $tag $istart "$istart + $nbm chars"
	# first remove "\}" at the end
	$tw delete "$istart + $nbm chars - 1 chars"
	# then only remove "\{\name " (including the implicitly assumed 
	# whitespace)
	$tw delete "$istart" "$istart + [expr 3+[string length $name]] chars"
	set istart [$tw search -forwards -count nbm -regexp $pat $istart]
    }
}

# A command is something like 
#	\foo{...}
# without nested braces in the argument.
# if tag is "remove", the command and the argument are removed
proc TagCommand { tw name tag } {
    global TXTDEBUG

    set pat "\\\\$name\{\[^\}\]*\}"
    if $TXTDEBUG {    
	puts "looking for command $pat"
    }

    set istart [$tw search -forwards -count nbm -regexp $pat 1.0]
    while { $istart != "" } {
	if $TXTDEBUG {
	    puts "Found COMM $name at $istart ($nbm chars)"
	}
	if { $tag != "remove" } {
	    $tw tag add $tag $istart "$istart + $nbm chars"
	    # first remove final "\}"
	    $tw delete "$istart + $nbm chars - 1 chars"
	    # then only remove "\name\{" 
	    $tw delete "$istart" "$istart + [expr 2+[string length $name]] chars"
	} else {
	    if $TXTDEBUG {
		puts "deleting range $istart to $istart + $nbm chars"
	    }
	    $tw delete "$istart" "$istart + $nbm chars"
	}
	set istart [$tw search -forwards -count nbm -regexp $pat $istart]

    }
}

# An environment is what you expect in LaTeX...
proc TagEnv { tw name tag } {
    global TXTDEBUG

    set pat "\\\\begin{$name}.*\\\\end{$name}"
    if $TXTDEBUG {
	puts "looking for environment $pat"
    }
    set istart [$tw search -forwards -count nbm -regexp $pat 1.0]
    while { $istart != "" } {
	if $TXTDEBUG {
	    puts "Found ENV $name at $istart ($nbm chars)"
	}
	$tw tag add $tag $istart "$istart + $nbm chars"
	$tw mark set begin $istart
	$tw mark set end "$istart + $nbm chars"
	# removing the end
	$tw delete "$istart + [expr $nbm-6-[string length $name]] chars" \
	    "$istart + $nbm chars"
	# removing the beginning
	$tw delete "$istart" "$istart + [expr 8+[string length $name]] chars"
	# include newlines
	$tw insert begin "\n"
	$tw insert end "\n"

	set istart [$tw search -forwards -count nbm -regexp $pat $istart]
    }
}

proc RestoreNewLines { tw } {

    set istart [$tw search -forwards -regexp -count nbm "__PAR__" 1.0]
    while { $istart != "" } {
	$tw delete "$istart" "$istart + 7 chars"
	$tw insert $istart "\n\n"
	set istart [$tw search -forwards -regexp -count nbm \
			"__PAR__" $istart]
    }
}
