#
# Module for the management of edit/select/delete modes (and others)
# 

proc Edit_Object_Attributes { win } {

    set cur_id [$win find withtag current]

    if { $cur_id == "" } { return }

    if [Is_Type $cur_id NODE] {
	Edit_Node_Attributes $cur_id
    }
}

proc Start_Arc { x y } {
    global curlinex curliney curarc
    global node_radius trans_width trans_height
    global orig_obj

    set curobj [.graph.canvas find withtag current]
    puts "pointing over $curobj"

    if { $curobj == "" } { return }

    set from [.graph.canvas coords current]
    set objtags  [.graph.canvas gettags $curobj]

    if {  -1 != [lsearch -exact $objtags NODE] } {
	set curlinex [expr [lindex $from 0]+$node_radius]
	set curliney [expr [lindex $from 1]+$node_radius]

	.graph.canvas configure -cursor cross
	if [info exists cur_arc] { unset cur_arc }
	set orig_obj $curobj
	bind .graph.canvas <1> { Continue_Arc [%W canvasx %x] [%W canvasy %y] }
    }
}

proc Continue_Arc { x y } {
    global curlinex curliney cur_arc 
    global node_radius arc_width

    # decide if it is the end ot not.
    if [Stop_Arc $x $y] {
	return 
    }

    if ![info exists cur_arc] {
	# have to start the link
	set end [clip $x $y $curlinex $curliney $node_radius]
	set cur_arc [.graph.canvas create line \
			 [lindex $end 0] [lindex $end 1] $x $y \
			 -tag ARC -width $arc_width ]
    } else {
	# have to continue the existing link
	set ptlist [.graph.canvas coords $cur_arc ]
	set ptlist [concat $ptlist [list $x $y]]
	eval {.graph.canvas coords $cur_arc} $ptlist
    }

    set curlinex $x
    set curliney $y
}

proc Abort_Arc { } {
    global cur_arc

    if [info exists cur_arc] {
	.graph.canvas delete $cur_arc
	unset cur_arc
    }
    bind .graph.canvas <1> { Start_Arc %x %y }
}

proc Stop_Arc { x y } {
    global curlinex curliney cur_arc arc_lag arc_time
    global node_radius arc_width
    global orig_obj arc_list state
    
    set cur_obj [.graph.canvas find withtag current]
    puts "Current object = $cur_obj"

    if { $cur_obj == "" } { return 0 }

    set objtags  [.graph.canvas gettags $cur_obj]
    puts " -> Has tags $objtags"
    set goal [.graph.canvas coords current]

    set end [clip $curlinex $curliney \
		 [expr [lindex $goal 0]+$node_radius] \
		 [expr [lindex $goal 1]+$node_radius] \
		 $node_radius ]
    # perhaps clipping of origin?
    if ![info exists cur_arc] {
	set start [clip $x $y $curlinex $curliney $node_radius]
	set curlinex [lindex $start 0]
	set curliney [lindex $start 1]
    }

    if ![info exists cur_arc] {
	set ptlist [list $curlinex $curliney \
			[lindex $end 0] [lindex $end 1] ]
	set cur_arc [eval {.graph.canvas create line} $ptlist \
			 -tags ARC \
			 -width $arc_width]
    } else {
	set ptlist [.graph.canvas coords $cur_arc ]
	set ptlist [concat $ptlist $end]
	eval {.graph.canvas coords $cur_arc} $ptlist
	eval {.graph.canvas itemconfigure $cur_arc} -tags ARC
    }

    # update of arc list
    lappend arc_list [list $cur_arc $orig_obj $cur_obj]
    
    .graph.canvas configure -cursor ""
    bind .graph.canvas <1> { Start_Arc %x %y }
    bind .graph.canvas <2> { Abort_Arc }
    bind .graph.canvas <3> { Exit_Arc_Mode }

    unset cur_arc
    set state(.graph.canvas) modified

    return 1
}

proc Exit_Current_Mode { } {
    global cur_mode

    if [info exists cur_mode] {
	switch $cur_mode {
	    node { 
		.edit.node configure -relief raised
	    }
	    arc { 
		.edit.arc configure -relief raised
	    }
	    info {
		.edit.info configure -relief raised
	    }
	    del { 
		.edit.del configure -relief raised
	    }
	    sel { 
		.edit.sel configure -relief raised
	    }
	}
	unset cur_mode
	.graph.canvas configure -cursor ""	
    }
}

proc Exit_Arc_Mode { } {
    global cur_arc

    if [info exists cur_arc] {
	.graph.canvas delete $cur_arc
    }

    Exit_Current_Mode
    Reset_Standard_Bindings .graph.canvas
}

proc Enter_Arc_Mode { } {
    global cur_mode

    Exit_Current_Mode

    .edit.arc configure -relief sunken
    .graph.canvas configure -cursor ""
    bind .graph.canvas <1> { Start_Arc %x %y }
    bind .graph.canvas <2> { Abort_Arc }
    bind .graph.canvas <3> { Exit_Arc_Mode }

    set cur_mode arc
}

proc Exit_Info_Mode { } {

    Exit_Current_Mode
    Reset_Standard_Bindings .graph.canvas

}

proc Enter_Info_Mode { } {
    global cur_mode

    Exit_Current_Mode

    .edit.info configure -relief sunken
    .graph.canvas configure -cursor question_arrow
    bind .graph.canvas <1> { Edit_Object_Attributes %W }
    bind .graph.canvas <2> { Exit_Info_Mode }
    bind .graph.canvas <3> { Exit_Info_Mode }

    set cur_mode info
}

#######################################################################
# Delete_Object -
# Removes an object from the canvas. Removes also the "attached" objects
# such as text, links.
# Handles the redirection of the attention to the "owner" of the object
# in case the object is a text (apparently not yet!).
#
# Parameters:
# - win		the canvas where this is happening
# - obj_id	the id of the object to remove with its "environment"
#
proc Delete_Object { win obj_id } {
    global node_list arc_list state
    global NUMBERS DEBUG

    if $DEBUG {
	puts -nonewline \
	    "Deleting object $obj_id, tags = [$win gettags $obj_id]"
    }

    set newnodes {}
    set newarcs {}

    if [Is_Type $obj_id NODE] {
	# find connected arcs and remove them
	set newarcs {}
	foreach arc $arc_list {
	    set arc_id [lindex $arc 0]
	    set nodeid [lindex $arc 1]
	    if { $nodeid == $obj_id } {
		Delete_Object $win $arc_id
	    } else {
		set nodeid [lindex $arc 2]
		if { $nodeid == $obj_id } {
		    Delete_Object $win $arc_id
		} else {
		    set newarcs [lappend newarcs $arc]
		}
	    }
	}
	# remove node itself, including from the list of nodes
	foreach node $node_list {
	    if { $obj_id != [lindex $node 0] } {
		set newnodes [lappend newnodes $node]
	    }
	}
	set arc_list $newarcs
	set node_list $newnodes

    } elseif [Is_Type $obj_id ARC] {
	foreach arc $arc_list {
	    if { $obj_id != [lindex $arc 0] } {
		set newarcs [lappend newarcs $arc]
	    }
	}
	set arc_list $newarcs
    }

    # remove the object from the canvas
    $win delete $obj_id
    # and its text too, if any
    set textid [$win find withtag "Tx$obj_id"]
    $win delete $textid

    if $DEBUG {
	puts " ...done."
    }

    # if graph empty, restore state
    if [expr [llength $node_list] == 0 ] {
	set state($win) saved
    } else {
	set state($win) modified
    }
}

#######################################################################
# Delete_Current_Object -
# Removes the "current" object, usually the one pointed at. 
#
# Parameters:
# - win		the canvas where this is happening
#
proc Delete_Current_Object { win } {
    global DEBUG 

    set cur_obj [$win find withtag current]
    if { [llength $cur_obj] > 1 } { 
	puts "Warning! More than one current object!"
    }
    if { [llength $cur_obj] > 0 } { 
	Delete_Object $win [lindex $cur_obj 0]
    }
}

proc Exit_Delete_Mode { } {

    Exit_Current_Mode
    Reset_Standard_Bindings .graph.canvas

}

proc Enter_Delete_Mode { } {
    global cur_mode

    Exit_Current_Mode

    .edit.del configure -relief sunken
    .graph.canvas configure -cursor pirate
    bind .graph.canvas <1> { Delete_Current_Object %W }
    bind .graph.canvas <2> { Exit_Delete_Mode }
    bind .graph.canvas <3> { Exit_Delete_Mode }

    set cur_mode del
}

proc Exit_Select_Mode { } {

    Exit_Current_Mode
    Reset_Standard_Bindings .graph.canvas

}

proc Enter_Select_Mode { } {
    global cur_mode

    Exit_Current_Mode

    .edit.sel configure -relief sunken
    .graph.canvas configure -cursor target
    bind .graph.canvas <1> { Select_Current_Object %W }
    bind .graph.canvas <2> { Exit_Select_Mode }
    bind .graph.canvas <3> { Exit_Select_Mode }

    set cur_mode sel
}

proc Exit_Node_Mode { } {

    Exit_Current_Mode
    Reset_Standard_Bindings .graph.canvas
}

proc Enter_Node_Mode { } {
    global cur_mode img_dir

    Exit_Current_Mode

    .edit.node configure -relief sunken    
    .graph.canvas configure -cursor "@$img_dir/node.xbm black"
    bind .graph.canvas <1> { make_new_node %W [%W canvasx %x] [%W canvasy %y] }
    bind .graph.canvas <3> { Exit_Node_Mode }

    set cur_mode node
}
