#      Init the global variable "trl"
#       which correspond of translations from latex to text
#       ex : \EG -> Event Graph
proc  InitTrl {} {
    global trl translation where_is_Mtool env

    set filetrl "$where_is_Mtool/$translation"
    if { ![file exists "~/$translation"] } {
	puts -nonewline "Copying translations file ... " 
	if { [catch "exec cp $filetrl $env(HOME)"] ==0 } {
	    puts "ok"
	}
    }
    set trl_id [open "~/$translation" r]

    while { [gets $trl_id line] != -1 } {
	set indice "\\[lindex $line 0]"
	set trl($indice) "[lrange $line 1 end]"
    }
    close $trl_id
}


#      Edit the help file of the `name` program in latex format
proc  EditHelp { name } {
    global AMAN
    global whence
    
    if { $AMAN } {
	catch {exec $whence(aman) $name} result
	LogResult "Aman $name" "$result\n"
    } else {
	EditHelpMaison $name
    }
}

proc  EditHelpMaison { name } {
    global man
    
    set win_name ".[ToLower $name]"
    set top "${win_name}.help"
    if [catch {toplevel $top}] { return }
    wm title $top "Help for $name"
    set name_icon "Help $name"
    wm iconname $top $name_icon
    # uncomment the line below to iconifie child with parent
    wm transient $top .
    set geom_parent [split [wm geometry $win_name] x+]
    set geom "+[lindex $geom_parent 2]+[expr [lindex $geom_parent 3]+[lindex $geom_parent 1]+40]"
    wm geometry $top $geom

    set man_id "[open $man($name) r]"
    
    button $top.ok -text OK -command "destroy $top"
    text $top.t -relief raised -bd 2 -yscrollcommand "$top.s set" \
	-width 80 -height 35 -wrap word -font "-adobe-new century schoolbook-medium-r-normal--*-120-*"
    scrollbar $top.s -relief flat -command "$top.t yview"
    pack $top.ok -side bottom -fill x
    pack $top.s -side right -fill y
    pack $top.t -expand yes -fill both

    $top.t tag configure bold -font "-adobe-new century schoolbook-bold-r-normal--*-120-*" 
    $top.t tag configure h_bold -font "-Adobe-Helvetica-Bold-R-Normal--*-120-*"
    $top.t tag configure bold_big -font "-adobe-new century schoolbook-bold-r-normal--*-140-*"
    $top.t tag configure it -font "-Adobe-Helvetica-Medium-O-Normal--*-120-*"
    $top.t tag configure small -font "-Adobe-Helvetica-Medium-R-Normal--*-100-*"

    set para(text) ""
    set para(name) ""
    set flag_para 0
    while {[gets $man_id line] !=  -1} {
	switch -regexp $line {
	    {^\\name\{} {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "name"
		set para(text) "$line"
	    }
	    {^\\synopsis\{}   {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "synopsis"
		set para(text) "$line"
	    }
	    {^\\descript\{}   {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "descript"
		set para(text) "$line"
	    }
	    {^\\options\{}    {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "options"
		set para(text) "$line"
	    }
	    {^\\bugs\{}       {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "bugs"
		set para(text) "$line"
	    }
	    {^\\desirables\{} {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "desirables"
		set para(text) "$line"
	    }
	    {^\\notes\{}      {
		if $flag_para {
		    TreatParagraph para $top.t
		} else {
		    set flag_para 1
		}
		set para(name) "notes"
		set para(text) "$line"
	    }
	    default { 
		if [string length "$line"] {
		    if [string compare "[string index $line 0]" "%"] {
			append para(text) " $line"
		    }
		} else {
		    append para(text) "\n"
		}
	    }
	}
    }
    if $flag_para {
	TreatParagraph para $top.t
    }
    
    bind $top.t <Any-Button-1> { }
    bind $top.t <Any-B1-Motion> { }
    bind $top.t <Any-Double-Button-1> { }
    bind $top.t <Any-Triple-Button-1> { }
    $top.t config -state disabled

    close $man_id
}


#      Call the good procedure for name, synopsys ... etc
#       dependent  of `para` parameter
proc  TreatParagraph { para w } {
    upvar $para p

    case $p(name) {
	{name} {
	    HelpNAME $w $p(text)
	}
	{synopsis} {
	    HelpSYNOPSIS $w $p(text)
	}
	{descript} {
	    HelpDESCRIPT $w $p(text)
	}
	{options} {
	    HelpOPTIONS $w $p(text)
	}
	{bugs} {
	    HelpBUGS $w $p(text)
	}
	{desirables} {
	    HelpDESIRABLES $w $p(text)
	}
	{notes} {
	    HelpNOTES $w $p(text)
	}
    }
}


#       Return a list fill with parts between braces in text
#        ex : this is the text { part one } for exemple { part two }
#             return {part one} {part two}
#       text : text with barces to be split in list
proc  FindPart { text } {

    set l ""
    set str ""
    set nb_brace 0
    
    for {set i 0}  {$i < [string length "$text"]} {incr i} {
	set c "[string index $text $i]"
	if ![string compare "$c" "\{"] {
	    if { $nb_brace==0 } {
		set first "[expr $i + 1]"
	    }
	    incr nb_brace
	} elseif ![string compare "$c" "\}"] {
	    incr nb_brace -1
	    if { $nb_brace==0 } {
		lappend l "[string range $text $first [expr $i - 1]]"
	    }
	}
    }
    return $l
}


#      Insert tabulations in string to have a good form
#       and print text in window w
#      w : window to be print in
#      text : string to be treat with tab
#      long : number of maximum char by line
proc  InsertTabInItem { w text long } {
    set i 0
    set itemstr_withtabs "     o"
    set itemstr_length [string length $text]
    while { $i < $itemstr_length } {
	append itemstr_withtabs "\t[string range $text $i [expr $i + $long-1]]\n"
	set i [expr $i+$long]
    }
    TranslateAndTagText $w $itemstr_withtabs
}

    
#      Do the translation of each "\a.string" of LaTex format
#      Manage it with is signification and is correspondant tag
#       w : window be print in 
#       text : text to be treat and tag
proc  TranslateAndTagText { w text } {
    global trl ref
    
    set t "$text"
    set long "[expr [lindex [$w configure -width] 4] - 10]"
    set item 0
    while { 1 } {
	if [regexp -indices -nocase {(\\begin\{[a-z ]+\})|(\\end\{[a-z ]+\})|(\\index\{)|(\\cite\{[a-z0-9]+\})|(\\([a-z]+|_|\#))|(\{\\[a-z]+ [^{}]+\})|(\$)} $t ind] {
	    set first "[lindex $ind 0]"
	    set last "[lindex $ind 1]"
	    set expression "[string range $t $first $last]"
	    set str "[string range $t 0 [expr $first - 1]]"
	    if { !$item } {
		InsertWithTags $w insert $str
		set t "[string range $t [expr $last +1] end]"
		switch -regexp "$expression" {
		    {^\{\\tt} {
			set str "[string range $expression 5 [expr [string length $expression] -2]]"
			InsertWithTags $w insert [TranslateText "$str"] h_bold
		    }
		    
		    {^\{\\em} {
			set str "[string range $expression 5 [expr [string length $expression] -2]]"
			InsertWithTags $w insert [TranslateText "$str"] it
		    }
		    
		    {^\\index} {
			# Procedure Supprimant L'Intervalle 1ere Derniere Accolade 
			# dans un index
			set nb_brace 1
			set i 0
			while { $nb_brace!=0 } {
			    set c "[string index $t $i]"
			    if ![string compare "$c" "\{" ] {
				incr nb_brace
			    } elseif ![string compare "$c" "\}"] {
				decr nb_brace
			    }
			    incr i
			}
			set t [string range $t $i end]
		    }
			
		    {^\\bigskip} -
		    {^\\medskip} -
		    {^\\smallskip} {
			InsertWithTags $w insert "\n" small
		    }
		    
		    {^\\cite} {
			set who_is_cite [string range $expression 6 [expr [string length $expression] - 2]]
			if [info exists ref] {
			    TranslateAndTagText $w [TranslateText "\"$ref($who_is_cite)\"" ]
			} else {
			    InsertWithTags $w insert "($who_is_cite)" small
			}
		    }

		    {^\\item} {
			set item 1
		    }
		    
		    {^\\begin\{itemize} {
			InsertWithTags $w insert "\n"
		    }
		    
		    default {
			InsertWithTags $w insert [TranslateText "$expression"]
		    }
		}
	    } else {
		switch -regexp "$expression" {
		    {^\\item} {
			append itemstr $str
			InsertTabInItem $w $itemstr $long
			set itemstr ""
		    }

		    {^\\end\{itemize} {
			append itemstr $str
			InsertTabInItem $w $itemstr $long
			set itemstr ""
			set item 0
		    }
		    
		    default {
			append itemstr [string range $t 0 $last]
		    }
		}
		set t "[string range $t [expr $last +1] end]"
	    }
	    update idletasks
	} else {
	    InsertWithTags $w insert "$t"
	    break
	}
    }
}


#      Translate string like "\___" in another string to be read
#      ex : "\#" in latex format become "#"
#      text : string to be treat
proc  TranslateText { text } {
    global trl
    
    set t $text
    while { 1 } {
	if [regexp -nocase -indices {(\\([a-z]+|_|\#|['`^][ea]))|(\$)} $t ind] {
	    set first "[lindex $ind 0]"
	    set last "[lindex $ind 1]"
	    set expression "[string range $t $first $last]"
	    switch $expression {
		{\_} {
		    regsub {\\_} $text "_" text
		}
		
		{\#} {
		    regsub {\\\#} $text "#" text
		}
		
		{\max} {
		    regsub {\\max} $text "max" text
		}
		
		{\times} {
		    regsub {\\times} $text "" text
		}
		
		{\'e} {
		    regsub {\\'e} $text "" text
		}
		
		{\`e} {
		    regsub {\\`e} $text "" text
		}
		
		{\^e} {
		    regsub {\\^e} $text "" text
		}
		
		{\`a} {
		    regsub {\\`a} $text "" text
		}
		
		{$} {
		    regsub {\$} $text {} text
		}
		
		default {
		    if [info exists trl($expression)] {
			regsub "\\$expression" $text "$trl($expression)" text
		    }
		}
	    }
	    set t [string range $t [expr $last + 1] end]
	} else {
	    break
	}
    }
    return $text
}


#      Display NAME subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpNAME { w text } {
    InsertWithTags $w insert "NAME\n" bold_big
    set L [FindPart $text]
    InsertWithTags $w insert "   [TranslateText [lindex $L 0]]" h_bold
    TranslateAndTagText $w "  -   [lindex $L 2]\n\n"
}

#      Display SYNOPSIS subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpSYNOPSIS { w text } {
    InsertWithTags $w insert "SYNOPSIS\n" bold_big
    set L "[FindPart $text]"
    InsertWithTags $w insert "  [TranslateText [lindex $L 0]]" bold
    set str "[TranslateText [lindex $L 1]]"
    set nb_char "[string length $str]"
    set max_char "[expr [lindex [$w configure -width] 4] - 20]"
    if { [lindex [split [$w index insert] .] 1] <= 7 } {
	InsertWithTags $w insert "\t"
    }
    if { $nb_char < $max_char } {
	InsertWithTags $w insert "\t$str\n\n"
    } else {
	set begin "0"
	set m "$max_char"
	while { $m<$nb_char } {
	    #   Cut the end-of-line with a "]"
	    set end [expr [string last "\]" [string range $str $begin $m]] + $begin]
	    InsertWithTags $w insert "\t[string range $str $begin $end]\n\t"
	    #   Begin line with a "["
	    set begin [expr [string first "\[" [string range $str $end end]] + $end]
	    set m "[expr $begin+$max_char]"
	}
	InsertWithTags $w insert \
	    "\t[string range $str $begin end]\n\n"
    }
}

#      Display DESCRIPT subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpDESCRIPT { w text } {
    InsertWithTags $w insert "DESCRIPTION\n" bold_big
    set text "\t[string range $text 10 [expr [string last "\}" $text] - 1]]"
    TranslateAndTagText $w "$text\n\n"
}

#      Display OPTIONS subpart of help in including "\item"
#       w : window to be print in
#       text : string to treat
proc  HelpOPTIONS { w text } {
    InsertWithTags $w insert "OPTIONS\n" bold_big
    set text "[string range $text 12 [expr [string last "\}" $text] - 1]]"
    regsub -all "\\option\{" $text "\item\{\\tt " text
    append t "\\begin\{itemize\}" $text "\\end\{itemize\}"
    TranslateAndTagText $w "$t\n\n"
}

#      Display BUGS subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpBUGS { w text } {
    InsertWithTags $w insert "BUGS\n" bold_big
    set text "\t[string range $text 6 [expr [string last "\}" $text] - 1]]"
    TranslateAndTagText $w "$text\n\n"
}

#      Display DESIRABLES subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpDESIRABLES { w text } {
    InsertWithTags $w insert "DESIRABLE DEVELOPMENT\n" bold_big
    set text "\t[string range $text 12 [expr [string last "\}" $text] - 1]]"
    TranslateAndTagText $w "$text\n\n"
}

#      Display NOTES subpart of help
#       w : window to be print in
#       text : string to treat
proc  HelpNOTES { w text } {
    InsertWithTags $w insert "NOTES\n" bold_big
    set text "\t[string range $text 7 [expr [string last "\}" $text] - 1]]"
    TranslateAndTagText $w "$text\n\n"
}
