/* Skeleton of a generic simulation code for a Petri Net		*/

#ifndef BOOLEAN
#define BOOLEAN
typedef short	boolean;
#define	TRUE	1
#define FALSE	0
#endif

#define MIN(x,y)	( x<y?x:y )

/* include principal parameters: nb of transitions, of places,	*/
/* buffer size and so on...					*/
#include "pnc_graph_parameters.h"

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#include <sys/time.h>
#include <time.h>

/* table YY of initiated firings 				*/
int	Y[BUF_SIZE][NB_TRANS];
int    	Nb_Tokens;

/* Control flag for debugging/tracing 				*/
boolean Debug = FALSE;

#ifdef Trace
/* Some variables to handle the case of simulataneous events	*/
boolean Taboo_Trans[NB_TRANS] = { FALSE };
boolean Some_Taboo = FALSE;
#endif

/* Declarations for random number generation (for routing and sources)	*/
#define MAXINT          0x7FFFFFFF

double U_0_1()
{
  return (double)random()/MAXINT;
}
double	U;

/* problem dependent variables					*/
#include "pnc_declarations.h"

/* Forward declarations */
void 	write_stat_results();
void 	dump_simulation();
boolean	Network_Is_Still_Live();
int	Death_Time();

main(argc, argv)
int	argc;
char *argv[];
{
    int   	R_Time = 0;		/* "real" time (event numbers)	*/
    boolean	Continue = TRUE;
    boolean	Print_Final_Marking = FALSE;
    int		E_Degree;		/* enabl. degree. for compet.	*/
    boolean	Is_Enabled;		/* has become obsolete.		*/
    int   	Num;
    struct timeval	t_debut, t_fin;
    double	Dt;
    int		Nb_Evt;
    int		i, j;
    int		Tmp_Y;
    int		NumArg = 1;
#ifdef Trace
    int		Marking;
    int		Present_Nb;
#endif

    while( NumArg < argc ) {
	if ( !strcmp( argv[NumArg], "-tmax" ) ) {
	    NumArg++;
	    if ( NumArg < argc ) {
		sscanf( argv[NumArg], "%d", &TMAX );
		TMAX *= SCALE;
		NumArg++;
	    }
	}
	else if ( !strcmp( argv[NumArg], "-finalmarking" ) ) {
	    Print_Final_Marking = TRUE;
	    NumArg++;
	}
	else if ( !strcmp( argv[NumArg], "-h" ) ) {
	    printf("This is a simulator compiled from a Petri net");
	    if ( strlen( PNL_INPUT ) > 0 )
	      printf(" (%s)", PNL_INPUT );
	    printf(" with pn2c\nUsage: %s [-tmax <int>] ", argv[0] );
	    printf("[-finalmarking] [-debug]\n");
	    exit(-1);
	}
	else if ( !strcmp( argv[NumArg], "-debug" ) ) {
	  Debug = TRUE;
	  NumArg++;
	}
    }
	    
#ifdef Trace
    Tracefile = fopen("trace.results","w");
    CODE("# Trace file generated by equational simulation\n#\n");
    CODE("# Get rid of the current clock and create a new one\n");
    CODE("KILL_OBJECT TheClock\n" );
    CODE("MAKE_NODE TheClock OF_CLASS clock" );
    CODE(" AT_POSITION 720 750\n" );

    CODE("# Reset the status of all transitions to idle\n");
    for ( Num=0 ; Num < NB_TRANS ; Num++ )
	CODE1("MODIFY %s \"status\" = 0\n", gss_trans_name[Num] );

    CODE("# Reset the status of all places to their initial pop\n");
    for ( Num=0 ; Num < NB_PLACES ; Num++ )
	CODE2("MODIFY %s \"token\" = %d\n", 
	      gss_place_name[Num], Initial_Pop[Num] );
    CODE("eol+\n");

    CODE("#\n# Start animation\n#\n");
#endif

    gettimeofday(&t_debut,(struct timezone *)0);

    Time = 0;
    /* compute the first events */
    for ( R_Time = 0 ; ( R_Time < MAX_DELAY ) && ( R_Time <= TMAX );
		       R_Time++ ) {

#include "pnc_initial_evolution.c"

    }

    /* compute the rest... */
    for ( ; ( R_Time <= TMAX ) && Continue; R_Time++) {

#include "pnc_evolution.c"

      /*       if ( Time == 0 ) { */
      /* 	  Continue = Network_Is_Still_Live(); */
      /*       } */
    }

    gettimeofday(&t_fin,(struct timezone *)0);

/*     Continue = Network_Is_Still_Live(); */

    /* end of simulation... dump statistics */
    if ( Continue ) {
      Write_Source_Stats( TMAX, BUF_SIZE );
      write_stat_results( TMAX, BUF_SIZE );
      if ( Print_Final_Marking )
	Final_Marking( TMAX );
    }
    else {
      Write_Source_Stats( R_Time-1, BUF_SIZE );
      write_stat_results( R_Time-1, BUF_SIZE );
      printf("The network died...\n" );
    }

    Dt = (double)((t_fin.tv_sec - t_debut.tv_sec) * 1000000.0
		  + t_fin.tv_usec - t_debut.tv_usec)/1000000.0;

    printf("Total execution time: %f", Dt );

    Nb_Evt = Nb_Arrived_Tokens( TMAX );

    for ( i=0; i < NB_TRANS; i++ )
      if ( !Is_Source[i] )
	Nb_Evt += Y[ TMAX % BUF_SIZE ][i];

    printf("\t\t(%f evt/sec)\n", (double)Nb_Evt / Dt );

    if ( Debug )
      dump_simulation( 100 );

    return 0;
}

/****************************************************************/
/* Procedure:		write_stat_results			*/
/* Dumps the results to the standard output.			*/
/* Parameters:							*/
/*	Sim_Length:	the final simulation "time"		*/
/*	Buf_Size:	the size of result tables		*/
/****************************************************************/
void write_stat_results( Sim_Length, Buf_Size )
int	Sim_Length;
int	Buf_Size;
{
  int	i;
  int	Nb_Enabl;

  for (i=0; i<NB_TRANS; i++) {
    if ( !Is_Source[i] ) {
      Nb_Enabl = Y[Sim_Length % Buf_Size][i];
      printf("Y_%1d[%1d] = %6d    Cycle time: ", 
	     i, Sim_Length / SCALE, Nb_Enabl );
      if ( Nb_Enabl > 0 )
	printf("%10.4f", (double)Sim_Length / (double)Nb_Enabl / SCALE );
      else
	printf("  Infinity");

      if ( strlen( Trans_Name[i] ) > 0 )
	printf("  (%s)\n", Trans_Name[i] );
      else
	printf("\n");
    }
  }
}

/****************************************************************/
/* Procedure:		dump_simulation				*/
/* Dumps the counter Y to the standard output.			*/
/* Parameters:							*/
/*	Dump_Length:	the size to be dumped			*/
/****************************************************************/
void dump_simulation( Dump_Length )
int	Dump_Length;
{
  int	i,j;

  printf("Y----------------\n");
  for (i=0; i<Dump_Length; i++) {
    for( j=0; j<NB_TRANS; j++ )
      printf("%4d ", Y[i][j]);
    printf("\n");
  }
}

/****************************************************************/
/* Function:		Network_Is_Still_Live			*/
/* Checks that the network is still evolving			*/
/* Parameters:							*/
/*	None							*/
/* Returns:							*/
/*	TRUE	if network is still living			*/
/*	FALSE	otherwise					*/
/****************************************************************/
boolean Network_Is_Still_Live()
{
  boolean	Is_Dead = TRUE;
  int		i;

  for ( i=0; i < NB_TRANS; i++ )
    Is_Dead = Is_Dead 
      && ( Y[ BUF_SIZE - 1 ][i] == Y[ BUF_SIZE - 1 - Fire_Time[i] ][i] );

  return !Is_Dead;
}

/****************************************************************/
/* Function:		Death_Time				*/
/* Returns the time at which the network died, assuming it did	*/
/* Parameters:							*/
/*	None							*/
/* Returns:							*/
/*	an integer, the time of last activity.			*/
/****************************************************************/
int Death_Time( Upper_Bound )
int	Upper_Bound;
{
  int		i;
  int		T;
  boolean	Continue = TRUE;

  for ( T = Upper_Bound - MAX_DELAY; Continue && ( T >= 0 ); T-- )
    for ( i=0; i < NB_TRANS; i++ )
      Continue = Continue 
	&& ( Y[ T ][i] == Y[ T + Fire_Time[i] ][i] );

  return T+1;
}
